/////////////////////////////////////////////////////////////////////////////
/// @file EuclideanPersistentMain
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-07-01
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#include <capd/auxil/ApplicationDesc.h>
#include <capd/auxil/Logger.h>

#include <boost/program_options.hpp>


#include <string>
#include <iostream>
#include <capd/persistence/AbstractPointCloud.h>
#include <boost/scoped_ptr.hpp>
#include <fstream>
#include <boost/tokenizer.hpp>
#include <capd/persistence/PointCloudTypes.h>
#include <capd/persistence/PointCloudUtils.h>
#include <capd/persistence/PointCloudSimplicialCellSComplex.h>
#include <capd/persistence/ComputePersistentHomology.h>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>

using namespace capd;

namespace {
  const capd::auxil::ApplicationDesc applicationDesc("capdRedHom::example::EuclideanPersistentMain",
                                                     "Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>",
                                                     "2014-07-01",
                                                     "");
}


int main(int argc, char *argv[])
{

  try {
    INIT_CAPD_LOGGER;
    CAPD_INFO("Entering application:\n" << applicationDesc);

    namespace po = boost::program_options;

    po::options_description desc("Allowed options");
    po::positional_options_description positionalDesc;

    desc.add_options()
      ("help,h", "produce help message");


    std::string filename;
    short dimension = 0;
    short maxComplexDimension = 0;
    size_t numberOfEgdesLimit = 0;
    size_t degreeLimit = 0;
    double epsilonLimi = 0;
    std::string diagramFilenamePrefix;

    desc.add_options()
      // ("f,files", po::value<std::vector<std::string> >(&filenames)->multitoken(), "list of files");
      ("outputDiagram,o", po::value<std::string>(&diagramFilenamePrefix)->required(), "output diagram file name prefix")
      ("dimension,d", po::value<short>(&dimension)->default_value(0), "points dimension")
      ("maxComplexDimension,D", po::value<short>(&maxComplexDimension)->required(), "maximal complex dimension")
      ("numberOfEgdesLimit", po::value<size_t>(&numberOfEgdesLimit), "number of edges limit")
      ("degreeLimit", po::value<size_t>(&degreeLimit), "degree of an vertex in graph")
      ("epsilonLimi,e", po::value<double>(&epsilonLimi)->default_value(0), "distance limit in VR complex");

    positionalDesc.add("filename", 1);
    desc.add_options()("filename", po::value<std::string>(&filename), "input file name");

    po::variables_map vm;
    po::store(po::command_line_parser(argc, argv).options(desc).positional(positionalDesc).run(), vm);

    if (vm.count("help") > 0) {
      std::cout << desc << std::endl;
      return 0;
    }

    po::notify(vm);

    std::istream* input;
    boost::scoped_ptr<std::fstream> inputFile;

    if (filename.empty()) {
      CAPD_INFO("Reading from stdin");

      if (dimension == 0) {
        throw std::runtime_error("Cannot guess points dimension when reading from stdin");
      }

      input = &std::cin;
    } else {
      if (dimension == 0) {
        std::ifstream dummy(filename.c_str());
        std::string line;
        std::getline(dummy, line);
        dummy.close();

        std::stringstream lineStream(line);
        dimension = std::vector<double>(std::istream_iterator<double>(lineStream),
                                        std::istream_iterator<double>()).size();
        CAPD_INFO("Guess dimension: " << dimension);
      }
      inputFile.reset(new std::fstream(filename.c_str()));
      input = inputFile.get();
    }

    using namespace capd::persistence;
    typedef EuclideanPointCloudType<0> Traits;
    typedef Traits::PointCloudType PointCloud;
    typedef PointCloudInputStreamIterator<Traits>::type IStreamIterator;
    boost::scoped_ptr<PointCloud> pc;

    if (degreeLimit > 0) {
      CAPD_INFO("Building Point Cloud - degree limit: " << degreeLimit);
      pc.reset(makeDegreeLimitedPointCloud<Traits>(IStreamIterator(*input), IStreamIterator(),
                                         degreeLimit, epsilonLimi));
    } else {
      CAPD_INFO("Building Point Cloud - number of edges limit: " << numberOfEgdesLimit);
      pc.reset(makeSizeLimitedPointCloud<Traits>(IStreamIterator(*input), IStreamIterator(),
                                         numberOfEgdesLimit, epsilonLimi));
    }

    CAPD_INFO("Point Cloud with: " << pc->size() << " points and " << pc->edges() << " edges constructed");
    CAPD_INFO("Building simplicial complex with maximu dimension: " << maxComplexDimension);

    typedef PointCloudSimplicialCellSComplex PCComplex;
    PCComplex pccomplex(*pc, maxComplexDimension);
    pccomplex();

    CAPD_INFO("Complex constructed, its size: " << pccomplex.size());
    CAPD_INFO("Computing persistence homology");

    typedef capd::persistence::ComputePersistentHomology<PointCloudSimplicialCellSComplex> ComputePersistentHomology;
    ComputePersistentHomology persistence(pccomplex);
    persistence();

    CAPD_INFO("TOTAL Persistent pairs: " << persistence.pairs().size() << ", cycles: " << persistence.cycles().size());

    for (PCComplex::Complex::Dim dim = 0, maxDim = pccomplex.dim(); dim <= maxDim; ++dim) {
      std::string diagramFilename = diagramFilenamePrefix + "." + boost::lexical_cast<std::string>((short)dim) + ".diagram";
      ComputePersistentHomology::Diagram diagram = persistence.diagram(dim);
      ComputePersistentHomology::Pairs pairs = persistence.pairs(dim);
      ComputePersistentHomology::Cycles cycles = persistence.cycles(dim);
      std::ofstream outputDiagram(diagramFilename.c_str());

      CAPD_INFO("Diagram size in dim: " << (int)dim << " diagram: " << diagram.size() << ", pairs: " << pairs.size() << ", cycles: " << cycles.size());
      CAPD_INFO("Writing diagram to file: " << diagramFilename);

      typedef ComputePersistentHomology::Diagram::value_type Pair;
      BOOST_FOREACH(Pair pair, diagram) {
        outputDiagram << pair.first << " " << pair.second << "\n";
      }
    }

    return 0;

  } catch (std::exception& ex) {
    std::cerr << "Error: " << ex.what() << std::endl;
    return 1;
  } catch (const char* ex) {
    std::cerr << "Error: " << ex << std::endl;
    return 1;
  } catch (const std::string& ex) {
    std::cerr << "Error: " << ex << std::endl;
    return 1;
  } catch(...) {
    std::cerr << "Unknown exception" << std::endl;
    return 1;
  }

  return 1;
}
