/// @addtogroup bitSet
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file SComplexFiltrT.h
///
/// @author Marian Mrozek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) Marian Mrozek 2005-2011
//
// This file constitutes a part of the CAPD library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.uj.edu.pl/ for details.





#if !defined(_SCOMPLEX_FILTR_H_)
#define _SCOMPLEX_FILTR_H_


#include <vector>
#include <map>
#include <deque>
#include <algorithm>
#include <boost/bind.hpp>
#include <boost/ref.hpp>
#include "capd/homAlgebra/ChainContainer.h"

template<class StrategyComplexT,class Scalar, class CellChain, class ReductionPair>
struct restoreCoefStrategy;

template<class Scalar, class CellChain, class ReductionPair>
struct restoreCoefStrategy<SimplexSComplex,Scalar,CellChain,ReductionPair>;


template <typename ComplexT, typename ScalarT=int, typename IntT = int>
class SComplexFiltrT{
public:
  typedef ComplexT Complex;
  typedef ScalarT Scalar;
  typedef typename Complex::Cell Cell;
  typedef FreeModule<Cell, capd::vectalg::Matrix<Scalar,0,0>, IntT> CellFreeModule;
  typedef ChainT<ChainContainer<Cell, Scalar> > CellChain;
  typedef std::vector<CellChain> CellChains;
  typedef std::vector<CellChains> CellChainsByDim;

  typedef std::pair<Cell, Cell> ReductionPair;
  typedef std::deque<ReductionPair> ReducedCells;
  typedef std::deque<Cell> ExtractedCells;

public:

  SComplexFiltrT(Complex& complex, const ExtractedCells& extractedCells, const ReducedCells& reducedCells, const ReducedCells& coreducedCells):
    _complex(complex), _extractedCells(extractedCells)
  {
    BOOST_FOREACH(const ReductionPair& pair, reducedCells)
			addReducedPair(pair);
    BOOST_FOREACH(const ReductionPair& pair, coreducedCells)
			addReducedPair(pair);
    BOOST_FOREACH(const ReductionPair& pair, reducedCells)
    	addReducedPairDim1(pair);
    BOOST_REVERSE_FOREACH(const ReductionPair& pair, coreducedCells)
    	addReducedPairDim1(pair);
  }

  explicit SComplexFiltrT(Complex& complex):
    _complex(complex)
  {
  }

  void operator()(bool rfcReduce = true, bool restoreBaseHomologyChains = true)
  {
    ReducibleFreeChainComplexOverZFromSComplexAlgorithm<Complex, Scalar> rfcBuilder(_complex, true);
    _rfcComplexCR = rfcBuilder.template build< ReducibleFreeChainComplex<CellFreeModule> >();

    if (rfcReduce) {
			Stopwatch swr;
      int t = _rfcComplexCR().reduce();
			std::cout<<"Reduce "<< t <<" took: "<<swr<<"\n";
    }

		std::vector<std::vector<Cell> > simpleHomologyGenerators;
  	bool fullyReduced=_rfcComplexCR().removeSimpleHomologyGenerators(simpleHomologyGenerators);
  	const int topDimSimple=simpleHomologyGenerators.size()-1;

    CRef<FreeChainComplex<CellFreeModule> > fcComplexCR(new FreeChainComplex<CellFreeModule>(_rfcComplexCR()));

    _homologyGradedModule = HomAlgFunctors<CellFreeModule>::HomologyViaSmith(fcComplexCR);

    const int topDimFromSmith =_homologyGradedModule().TopDim();
    int topDim = std::max(topDimFromSmith,topDimSimple);

    _baseHomologyChains.resize(topDim + 1);
    _baseHomologyChainOrders.resize(topDim + 1);

    for(int dim=0; dim <= topDim; ++dim){
      if (dim <= topDimFromSmith){
        // This code should be optimized, note that the chains in baseHomologyChain[q]
        // are brought to its final form in the next for loop!!
        _homologyGradedModule().component(dim).exportBaseChains(_baseHomologyChains[dim], _baseHomologyChainOrders[dim]);
      }
    	if(dim<=topDimSimple){
      	for(unsigned int j=0;j<simpleHomologyGenerators[dim].size();++j){
					//std::cout<<"Adding simple homology generator in dim"<<dim<<" = "<<CellChain(simpleHomologyGenerators[dim][j]) <<"\n";
        	_baseHomologyChains[dim].push_back(CellChain(simpleHomologyGenerators[dim][j]));
        	_baseHomologyChainOrders[dim].push_back(0);
      	}
			}
    }

    BOOST_FOREACH(const Cell& extracted, _extractedCells) {
      if (_baseHomologyChains.size() < extracted.getDim() + 1) {
        _baseHomologyChains.resize(extracted.getDim() + 1);
        _baseHomologyChainOrders.resize(extracted.getDim() + 1);
      }
      _baseHomologyChains[extracted.getDim()].push_back(CellChain(extracted));
      _baseHomologyChainOrders[extracted.getDim()].push_back(0);
    }

    _baseReducedHomologyChains = _baseHomologyChains;


		if(restoreBaseHomologyChains)
		{
			Stopwatch swr;
			int c = 0;
			for (int dim=0; dim <= topDim; ++dim) {
				for(typename std::vector<CellChain>::iterator it = _baseHomologyChains[dim].begin();
						it != _baseHomologyChains[dim].end(); ++it){
					restore(*it);
					c++;
				}
			}
			std::cout<<"Restore "<<c<<" took "<<swr<<"\n";
		}

  }

  const QuotientGradedModule<CellFreeModule>& filteredHomModule() const{
    return _homologyGradedModule();
  }

  const CellChainsByDim& baseHomologyChains() const{
    return _baseHomologyChains;
  }

  const CellChainsByDim& baseReducedHomologyChains() const{
    return _baseReducedHomologyChains;
  }

  int topDim() const {
    return _baseHomologyChains.size()-1;
  }


//GJ build chain from coefficients in vector v in dimension dim
  template<typename vectorT>
  CellChain buildChain(const int & dim, const vectorT &v)
  {
    CellChain chain;

    for(int i = 0; i < v.dimension(); ++i)
      {
        chain+= (v[i] * _baseHomologyChains[dim][i]);
      }
    return chain;
  }

  int maxCycleSize(int q);
  void cycleSizeSpectrum(int q,std::map<int,int>& A_spect);


  friend std::ostream& operator<<(std::ostream& out, const SComplexFiltrT& A_filtr){
    out << " --- Homology generators --- \n\n";
    for(unsigned int q=0;q<A_filtr._baseHomologyChains.size();++q){
      out << "Dimension " << q << "\n";
      for(unsigned int j=0;j<A_filtr._baseHomologyChains[q].size();++j){
        out << "* Generator " << j << " of order " << A_filtr._baseHomologyChainOrders[q][j] << "\n";
        out << A_filtr._baseHomologyChains[q][j] << "\n";
				out << "* reduced \n";
        out << A_filtr._baseReducedHomologyChains[q][j] << "\n";
      }
    }
    return out;
  }

  void addReducedPair(const ReductionPair& reducedPair)
  {
    BOOST_ASSERT(reducedPair.first.getDim() > reducedPair.second.getDim());

		if(reducedPair.first.getDim() >= _reducedPairsByDim.size())
	    _reducedPairsByDim.resize(reducedPair.first.getDim() + 1);
    
		_reducedPairsByDim[reducedPair.first.getDim()].push_back(reducedPair);
		_reducedPairs.push_back(reducedPair);
		
  }

	void addReducedPairDim1(const ReductionPair& reducedPair)
  {
    BOOST_ASSERT(reducedPair.first.getDim() > reducedPair.second.getDim());
	
		if(reducedPair.first.getDim()==1)
			_reducedPairsDim1.push_back(reducedPair);

  }


	void restore(CellChain& cellChain) const
	{
		if (cellChain.size() == 0) {
			return;
		}

		_rfcComplexCR().restore(cellChain);

		const int chainDim = cellChain.ownDim();

		if (chainDim >= _reducedPairsByDim.size() || chainDim <= 0) { // less than 0 means empty chain, for 0 we don't need to restore
			return;
		}

		BOOST_REVERSE_FOREACH(const ReductionPair& reductionPair, _reducedPairsByDim[chainDim]) {
			BOOST_ASSERT(reductionPair.first.getDim() == chainDim);

			const Scalar lambda = _complex.coincidenceIndex(reductionPair.first, reductionPair.second);

			Scalar chainMulCbd = restoreCoefStrategy<Complex,Scalar,CellChain,ReductionPair>::compute(reductionPair, cellChain, _complex);
			
			if (chainMulCbd != Scalar(0)) {
			  cellChain.accessAt(reductionPair.first) -= chainMulCbd / lambda;
			}
		}   
	}

	void reduce(CellChain& cellChain) const
	{
		const int chainDim = cellChain.ownDim();

		if (chainDim < 0) { // less than 0 means empty chain
			return;
		}
		if(chainDim == 0 && _reducedPairsByDim.size() > 1){
			BOOST_FOREACH(const ReductionPair& reductionPair, _reducedPairsDim1) {

				BOOST_ASSERT(reductionPair.first.getDim() == chainDim + 1);
				const Scalar chainAt = cellChain.at(reductionPair.second);

				if (chainAt == Scalar(0))
				  continue;

				const Scalar lambda = chainAt / Scalar(_complex.coincidenceIndex(reductionPair.first, reductionPair.second));

				typedef typename Complex::Iterators::BdCells BdCells;
				BdCells bdCells = _complex.iterators().bdCells(reductionPair.first);

				BOOST_FOREACH(const typename BdCells::iterator::value_type& bd, bdCells) {

					Scalar coincidenceIndex = _complex.coincidenceIndex(reductionPair.first, bd);

					if (coincidenceIndex == Scalar(0))
					  continue;

					cellChain.accessAt(bd) -= lambda * coincidenceIndex;
					cellChain.normalize(bd);
				}
			}
		}
	if(chainDim > 0){
		BOOST_FOREACH(const ReductionPair& reductionPair, _reducedPairs)
		{

			if(reductionPair.first.getDim() == chainDim + 1){
				const Scalar chainAt = cellChain.at(reductionPair.second);

				if (chainAt != Scalar(0)){

					const Scalar lambda = chainAt / Scalar(_complex.coincidenceIndex(reductionPair.first, reductionPair.second));

					typedef typename Complex::Iterators::BdCells BdCells;
					BdCells bdCells = _complex.iterators().bdCells(reductionPair.first);

					BOOST_FOREACH(const typename BdCells::iterator::value_type& bd, bdCells) {

						Scalar coincidenceIndex = _complex.coincidenceIndex(reductionPair.first, bd);

						if(coincidenceIndex == Scalar(0))
						  continue;

						cellChain.accessAt(bd) -= lambda * coincidenceIndex;
						cellChain.normalize(bd);
					}
				}
			}

			if(reductionPair.first.getDim() == chainDim){
				const Scalar chainAt = cellChain.at(reductionPair.first);

				if (chainAt != Scalar(0)) {
				  cellChain.accessAt(reductionPair.first) = 0;
				  cellChain.normalize(reductionPair.first); // TODO do it in one instruction
				}
			}
		}
	}

	_rfcComplexCR().reduce(cellChain);
}

private:

  Complex& _complex;

  // CRef to QuotientGradedModule remaining after removing simple generators from RFCC
  // This should be changed as soon as the the direct sum operation is available
  CRef<QuotientGradedModule<CellFreeModule> > _homologyGradedModule;
  CRef<ReducibleFreeChainComplex<CellFreeModule> > _rfcComplexCR;

  // vector (grade indexed) of vectors of homology base chains
  std::vector<std::vector<CellChain> > _baseHomologyChains;
  std::vector<std::vector<CellChain> > _baseReducedHomologyChains;
  std::vector<std::vector<IntT> > _baseHomologyChainOrders;
  std::vector<std::deque<ReductionPair> > _reducedPairsByDim;
  std::deque<ReductionPair> _reducedPairsDim1;
  std::deque<ReductionPair> _reducedPairs;
  const ExtractedCells _extractedCells;

};

// -------------------------------------------------------------------------------------- //

template<class StrategyComplexT,class Scalar, class CellChain, class ReductionPair>
struct restoreCoefStrategy
{
	static Scalar compute(const ReductionPair& reductionPair,CellChain& cellChain, StrategyComplexT& _complex)
	{
		Scalar chainMulCbd = Scalar(0);

		typedef typename StrategyComplexT::Iterators::CbdCells CbdCells;

		CbdCells cbdCells = _complex.iterators().cbdCells(reductionPair.second);

		BOOST_FOREACH(const typename CbdCells::iterator::value_type& cbd, cbdCells) {
			Scalar chainAt = cellChain.at(cbd);

			if (chainAt != Scalar(0))
			  continue;
			Scalar coincindenceIndex = _complex.coincidenceIndex(cbd,reductionPair.second);

			chainMulCbd += coincindenceIndex * chainAt;
		}
		return chainMulCbd;
	}
};


template<class Scalar, class CellChain, class ReductionPair>
struct restoreCoefStrategy<SimplexSComplex,Scalar,CellChain,ReductionPair>
{
	static Scalar compute(const ReductionPair& reductionPair,CellChain& cellChain, SimplexSComplex& _complex)
	{
		typedef typename SimplexSComplex::Iterators::CbdCells CbdCells;

		Scalar chainMulCbd = Scalar(0);

		int chainSize = cellChain.size();
		int cbdSize = _complex.getSimplexById(reductionPair.second.getId())->coborder.size();

		if(chainSize > cbdSize)
		{
			CbdCells cbdCells = _complex.iterators().cbdCells(reductionPair.second);

			BOOST_FOREACH(const typename CbdCells::iterator::value_type& cbd, cbdCells) 
			{
				Scalar chainAt = cellChain.at(cbd);

				if (chainAt == Scalar(0))
				  continue;

				Scalar coincindenceIndex = _complex.coincidenceIndex(cbd,reductionPair.second);

				chainMulCbd += coincindenceIndex * chainAt;
			}
		}
		else
		{
			BOOST_FOREACH(typename CellChain::value_type& it, cellChain)
			{
				Scalar coincindenceIndex = _complex.coincidenceIndex(it.first,reductionPair.second);

				if (coincindenceIndex == Scalar(0))
				  continue;

				chainMulCbd += coincindenceIndex * it.second;
			}
		}
		return chainMulCbd;
	}
};



#endif // _SCOMPLEX_FILTR_H_
/// @}

