/////////////////////////////////////////////////////////////////////////////
/// @file Algorithms.h
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2013-12-03
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2013 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.

#ifndef CAPD_FILE_ALGORITHMS_H
#define CAPD_FILE_ALGORITHMS_H

#include <vector>
#include <stdint.h>
#include <cstddef>
#include <string>
#include <memory>

namespace capd
{

  namespace persistence
  {
    class DistanceMatrixPersistentHomology;
    class PersistentSignature;
  }

  namespace apiRedHom
  {
    class Complex;

    struct EulerCharacteristic
    {
      Complex& _complex;

      explicit EulerCharacteristic(Complex& complex): _complex(complex) {}
      long operator()();
    };

    struct BettiNumbersOverZ
    {
      Complex& complex;

      explicit BettiNumbersOverZ(Complex& complex): complex(complex) {}
      std::vector<int> operator()();
    };

    struct BettiNumbersOverZp
    {
      Complex& complex;
      int p;

      explicit BettiNumbersOverZp(Complex& complex, int p): complex(complex), p(p) {}
      std::vector<int> operator()();
    };

    struct BettiNumbersOverZ2
    {
      Complex& complex;

      explicit BettiNumbersOverZ2(Complex& complex): complex(complex) {}
      std::vector<int> operator()();
    };

    std::vector<int> ComputeBettiNumbersOverZ(Complex& complex);
    std::vector<int> ComputeBettiNumbersOverZp(Complex& complex, int p);
    std::vector<int> ComputeBettiNumbersOverZ2(Complex& complex);
    std::vector<int> ComputeBettiNumbers(Complex& complex, int modulo = 0);

    struct Signature
    {
      std::vector<std::pair<int, std::vector<int> > > raw;
      std::string str;
    };

    struct HomologyOverZ
    {
      Complex& complex;

      explicit HomologyOverZ(Complex& complex): complex(complex) {}
      Signature operator()();
    };

    struct HomologyOverZp
    {
      Complex& complex;
      int p;

      explicit HomologyOverZp(Complex& complex, int p): complex(complex), p(p) {}
      Signature operator()(); // result Signature uses int, but you can interprete it in Zp
    };

    struct HomologyOverZ2
    {
      Complex& complex;

      explicit HomologyOverZ2(Complex& complex): complex(complex) {}
      Signature operator()();  // result Signature uses int, but you can interprete it in Zp
    };

    Signature ComputeHomologyOverZ(Complex& complex);
    Signature ComputeHomologyOverZp(Complex& complex, int p);
    Signature ComputeHomologyOverZ2(Complex& complex);


    struct HomologyGeneratorsOverZ
    {
      Complex& complex;
      void operator()();
    };

    struct Coreduction
    {
      Complex& complex;
      size_t operator()();
    };

    struct RedHomOrchestratorOverZ
    {
      typedef int Scalar;

      explicit RedHomOrchestratorOverZ(Complex& complex);
      std::vector<Scalar> operator()();

      void reductions(bool v) { _reductions = v; }
      void coreductions(bool v) { _coreductions = v; }
      void morse(bool v) { _morse = v; }

    private:
      Complex& _complex;
      bool _reductions;
      bool _coreductions;
      bool _morse;
    };

    // struct FundGroup
    // {
    //   Complex& complex;

    //   explicit FundGroup(Complex& complex): complex(complex) {}

    //   std::vector<int> operator()();
    // };

    struct FundamentalGroup
    {
      Complex& complex;
      int _vectorFieldType;

      explicit FundamentalGroup(Complex& complex, int vectorFieldType):
        complex(complex), _vectorFieldType(vectorFieldType)
      {}

      std::vector<int> operator()();
    };


    struct RelativeBettiNumbers
    {
      Complex& _complex;
      Complex& _subcomplex;
      bool _differenceClosed;
      bool _reductions;
      bool _coreductions;
      bool _morse;

      RelativeBettiNumbers(Complex& complex, Complex& subcomplex, bool differenceClosed):
        _complex(complex), _subcomplex(subcomplex), _differenceClosed(differenceClosed),
        _reductions(true),
        _coreductions(true),
        _morse(true)
      {}

      void reductions(bool v) { _reductions = v; }
      void coreductions(bool v) { _coreductions = v; }
      void morse(bool v) { _morse = v; }

    };

    struct RelativeBettiNumbersOverZ: RelativeBettiNumbers
    {

      RelativeBettiNumbersOverZ(Complex& complex, Complex& subcomplex, bool differenceClosed):
        RelativeBettiNumbers(complex, subcomplex, differenceClosed)
      {}

      std::vector<int> operator()();
    };

    struct RelativeBettiNumbersOverZp: RelativeBettiNumbers
    {
      int p;

      RelativeBettiNumbersOverZp(Complex& complex, Complex& subcomplex, bool differenceClosed, int p):
        RelativeBettiNumbers(complex, subcomplex, differenceClosed), p(p)
      {}

      std::vector<int> operator()();
    };

    struct RelativeBettiNumbersOverZ2: RelativeBettiNumbers
    {
      RelativeBettiNumbersOverZ2(Complex& complex, Complex& subcomplex, bool differenceClosed):
        RelativeBettiNumbers(complex, subcomplex, differenceClosed)
      {}

      std::vector<int> operator()();
    };

    std::vector<int> ComputeRelativeBettiNumbersOverZ(Complex& complex, Complex& subcomplex, bool differenceClosed);
    std::vector<int> ComputeRelativeBettiNumbersOverZp(Complex& complex, Complex& subcomplex, bool differenceClosed, int p);
    std::vector<int> ComputeRelativeBettiNumbersOverZ2(Complex& complex, Complex& subcomplex, bool differenceClosed);


    struct BettiNumbersViaSmithFormOverZ
    {
      typedef int Scalar;
      Complex& _complex;

      explicit BettiNumbersViaSmithFormOverZ(Complex& complex): _complex(complex) {}

      std::vector<Scalar> operator()();
    };


    struct ImagePersistentHomology
    {
      typedef std::vector<std::vector<double> > Image;
      typedef std::vector<std::pair<double, double> > Diagram;

      explicit ImagePersistentHomology(const Image& image):
        _image(image)
      {}

      const Diagram& operator()();
      const Diagram& diagram(size_t dim) const;
      const Diagram& diagram() const;

    private:
      const Image& _image;
      std::auto_ptr<persistence::PersistentSignature> _signature;
    };

    struct DistanceMatrixPersistentHomology
    {
      typedef std::vector<std::vector<double> > DistanceMatrix;
      typedef std::vector<std::pair<double, double> > Diagram;
      typedef std::vector<size_t> Cycles;
      typedef std::vector<std::pair<size_t, size_t> > Pairs;

      DistanceMatrixPersistentHomology(const DistanceMatrix& distance, size_t dim, size_t degreeLimit):
        _distance(distance), _dim(dim), _degreeLimit(degreeLimit), _impl(0)
      {}

      ~DistanceMatrixPersistentHomology();

      void operator()();
      Pairs pairs(int dim = -1) const;
      Cycles cycles(int dim = -1) const;
      Diagram diagram(int dim = -1) const;

    private:
      DistanceMatrixPersistentHomology(const DistanceMatrixPersistentHomology&);

      const DistanceMatrix& _distance;
      const size_t _dim;
      const size_t _degreeLimit;
      persistence::DistanceMatrixPersistentHomology* _impl;
    };

    struct ZigZag
    {

      ZigZag(Complex& complex, size_t dim):
        _complex(complex), _dim(dim)
      {}

      void operator()();

    private:
      Complex& _complex;
      const size_t _dim;
    };

    struct CountCriticalCells
    {

      explicit CountCriticalCells(Complex& complex, int method):
        _complex(complex), _method(method)
      {}

      size_t operator()();

    private:
      Complex& _complex;
      int _method;
    };

  }
}

#endif // CAPD_FILE_ALGORITHMS_H
