/////////////////////////////////////////////////////////////////////////////
/// @file Complex.h
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2013-11-28
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2013 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.

#ifndef CAPD_FILE_COMPLEX_H
#define CAPD_FILE_COMPLEX_H

#include <capd/apiRedHom/AnyWrapper.h>

#include <stdint.h>
#include <boost/tuple/tuple.hpp>
#include <boost/utility.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/range/concepts.hpp> // there is a bug in any_range.hpp, we need to include it
#include <boost/range/any_range.hpp>
#include <boost/function.hpp>
#include <boost/optional.hpp>

#include <vector>
#include <iosfwd>
#include <capd/auxil/Logger.h>

namespace capd
{

  namespace apiRedHom
  {
    struct DefaultComplexTag {};

    template<typename ComplexTag>
    struct ComplexTraits {};

    template<>
    struct ComplexTraits<DefaultComplexTag>
    {
      typedef int Dim; // TODO we cannot set it for a different type (int8_t), because conversions from CubSComplex::DIM doesn't work
      typedef int Size;
      typedef Size Id;
    };


    class CellsCollection
    {

    public:
      class ComplexCell
      {
      public:
        typedef ComplexTraits<DefaultComplexTag> Traits;
        typedef Traits::Dim Dim;
        typedef Traits::Id Id;

      private:
        struct IComplexCell
        {
          virtual Dim dim() const = 0;
          virtual Id id() const = 0;
        };

        template<typename Wrapped>
        class WrappedCell: public IComplexCell
        {
        public:
          WrappedCell(const Wrapped& object):
            _wrapped(object)
          {}

          Wrapped& rawCell() { return _wrapped; }
          const Wrapped& rawCell() const { return _wrapped; }

          Dim dim() const { return _wrapped.getDim(); }
          Id id() const { return _wrapped.getId(); }

        private:
          Wrapped _wrapped;
        };
      public:

        template<typename T>
        ComplexCell(const T& object):
          _wrapped(new WrappedCell<T>(object))
        {}

        ComplexCell() {}

        template<typename T>
        T& rawCell() { return dynamic_cast<WrappedCell<T>&>(*_wrapped).rawCell(); }

        Dim dim() const { return _wrapped->dim(); }
        Dim id() const { return _wrapped->id(); }

      private:
        boost::shared_ptr<IComplexCell> _wrapped;
      };

    public:

      typedef boost::any_range<ComplexCell, boost::forward_traversal_tag, ComplexCell, std::ptrdiff_t> AnyRange;

      typedef AnyRange::iterator iterator;
      typedef AnyRange::const_iterator const_iterator;

      CellsCollection() {}

      template<typename Collection>
      CellsCollection(const Collection& collection):
        _collection(collection, -1),
        _range(_collection.get<Collection>().begin(), _collection.get<Collection>().end())
      {

      }

      iterator begin() { return _range.begin(); }
      iterator end() { return _range.end(); }
      const_iterator begin() const { return _range.begin(); }
      const_iterator end() const { return _range.end(); }

    private:
      AnyWrapper _collection; // keep a copy of a collection
      AnyRange _range;
    };


    class ComplexIterators
    {

    public:
      typedef CellsCollection AllCells;
      typedef CellsCollection DimCells;
      typedef CellsCollection CbdCells;
      typedef CellsCollection BdCells;


      ComplexIterators(const AnyWrapper& w = AnyWrapper()):
        _iterators(w) {}

      AllCells allCells();

    private:
      AnyWrapper _iterators;
    };

    template<typename ComplexT>
    struct SetFunctions;

    class Complex
    {
    public:
      typedef boost::any InternalComplex;
      typedef ComplexTraits<DefaultComplexTag> Traits;
      typedef Traits::Size Size;
      typedef Traits::Dim Dim;

      typedef CellsCollection::ComplexCell Cell;

      AnyWrapper& anyComplex() { return _complex; }

      template<typename T>
      T& rawComplex() { return *_complex.template get<boost::shared_ptr<T> >(); }

      template<typename T>
      boost::shared_ptr<T> rawComplexPtr() { return _complex.template get<boost::shared_ptr<T> >(); }

      ComplexIterators iterators();

      Dim dim() const { return _dim(); }
      Size cardinality() const { return _cardinality(); }

      virtual ~Complex();

    protected:
      Complex(){}

      explicit Complex(const AnyWrapper& complex):
        _complex(complex)
      {}

      template<typename ThisT>
      void set_complex(const AnyWrapper& complex);

      AnyWrapper _complex;
      boost::function<Dim ()> _dim;
      boost::function<Size ()> _cardinality;
      CAPD_CLASS_LOGGER;
    };


    class CubicalComplex: public Complex
    {

    public:
      typedef size_t Coordinate;

      typedef std::pair<Coordinate, bool> Interval;
      typedef std::vector<Interval> CellCode;
      typedef boost::any_range<CellCode, boost::forward_traversal_tag, CellCode&, std::ptrdiff_t> CellCodeRange;
      typedef std::vector<Coordinate> FullCube;

      static const int TYPE_DEFAULT;
      static const int TYPE_OLD;
      static const int TYPE_DENSE;
      static const int TYPE_SPARSE;

      CubicalComplex(const std::vector<size_t>& size,
                     const CellCodeRange& cubes = CellCodeRange(), int type = TYPE_DEFAULT);


      void insert(const CellCode& cube) { _insert(cube); }
      void insert(const FullCube& cube) { _insertFullCube(cube); }
      Cell operator[](const CellCode& cube) { return _operator_array(cube); }
      boost::optional<Cell> find(const CellCode& cube) { return _find(cube); }

      void fillWithBoundaries() { _fillWithBoundaries(); }

    protected:
      friend struct SetCubicalComplex;
      friend struct SetFunctions<CubicalComplex>;

      boost::function<void (const CellCode& cube)> _insert;
      boost::function<void (const FullCube& cube)> _insertFullCube;
      boost::function<Cell (const CellCode& cube)> _operator_array;

      boost::function<boost::optional<Cell> (const CellCode& cube)> _find;
      boost::function<void ()> _fillWithBoundaries;
    };


    class SimplicialComplex: public Complex
    {

    public:
      typedef int Point;
      typedef std::vector<Point> CellCode;
      typedef boost::any_range<CellCode, boost::forward_traversal_tag, CellCode&, std::ptrdiff_t> CellCodeRange;

      explicit SimplicialComplex(const CellCodeRange& simplices = CellCodeRange());

      void insert(const CellCode& simplex) { _insert(simplex); }
      Cell operator[](const CellCode& simplex) { return _operator_array(simplex); }
      boost::optional<Cell> find(const CellCode& simplex) { return _find(simplex); }

    protected:
      friend struct CreateSimplicialComplex;
      friend struct SetFunctions<SimplicialComplex>;

      boost::function<void (const CellCode& simplex)> _insert;
      boost::function<Cell (const CellCode& simplex)> _operator_array;
      boost::function<boost::optional<Cell> (const CellCode& simplex)> _find;
    };


    class GeneralizedSComplex: public Complex
    {

    public:
      typedef size_t Id;
      typedef size_t CellCode;
      typedef boost::tuple<Id, Id, int> KappaMapEntry;
      typedef boost::any_range<KappaMapEntry, boost::forward_traversal_tag, const KappaMapEntry&, std::ptrdiff_t> KappaMapEntries;
      typedef std::vector<Dim> Dims;

      GeneralizedSComplex(const Dims& dims, const KappaMapEntries& kappaMapEntries);

    private:
      friend struct CreateSComplex;
      friend struct SetFunctions<GeneralizedSComplex>;

      boost::function<void (const CellCode& cellCode)> _insert;
      boost::function<Cell (const CellCode& cellCode)> _operator_array;
      boost::function<boost::optional<Cell> (const CellCode& cellCode)> _find;
    };

  }
}




#endif // CAPD_FILE_COMPLEX_H
