#ifndef ALGS_AKQ_REDUCE_STRATEGY_HPP_
#define ALGS_AKQ_REDUCE_STRATEGY_HPP_

#include "CellDimIndexer.hpp"

#include <boost/optional.hpp>
#include <boost/foreach.hpp>
#include <boost/iterator/filter_iterator.hpp>
#include <boost/tuple/tuple.hpp>

#include <utility>
#include <stack>
#include <map>
#include <vector>
#include <algorithm>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/typeof/typeof.hpp>
#include <capd/basicalg/doubleFun.h>
#include <unordered_map>
#include <boost/functional/hash.hpp>
#include <capd/auxil/Logger.h>


namespace capd
{
  namespace complex
  {

    template<typename SComplexT, typename OutputComplexT, typename Scalar>
    class AKQReduceStrategy : public DefaultReduceStrategy<SComplexT>
    {
    protected:
      using DefaultReduceStrategyBase<SComplexT>::complex;

    public:
      enum AKQType {UNSET, KING, QUEEN, ACE};

      typedef std::unordered_map<int, Scalar> PathsInfo;
      typedef std::unordered_map<std::pair<int,int>, Scalar, boost::hash<std::pair<int,int> > >  CoeffsMap;

      typedef OutputComplexT OutputComplexType;
      typedef SComplexT SComplex;
      typedef DefaultReduceStrategyTraits<SComplex> Traits;
      typedef typename SComplex::Cell Cell;
      typedef typename SComplex::Cell::Id Id;

      typedef CellDimIndexer<Cell> DimIndexerType;
      DimIndexerType dimIndexer;
      typename DimIndexerType::iterator extractIt;
      typename DimIndexerType::iterator extractEnd;


      AKQReduceStrategy(SComplex& _complex): DefaultReduceStrategy<SComplex>(_complex)
      {
        Stopwatch total;
        maxExtractDim = _complex.getDim();
        extractDim = 0;

        const int csize = _complex.size();

        BOOST_AUTO(allCells, _complex.iterators(1).allCells());
        const int card = std::distance(allCells.begin(), allCells.end());

        morse.resize(csize, 0);
        followMemoTable.resize(csize, 0);
        akq.resize(csize);
        kerKing.resize(csize, 0);

        infosUsed = 0;
        {
          PathsInfo pi;
          pathsInfoPool.resize(card, pi);
        }

        dimIndexer = DimIndexerType(this->complex.iterators(1).allCells().begin(),
                                    this->complex.iterators(1).allCells().end(),
                                    maxExtractDim);

        extractIt = dimIndexer.begin();
        extractEnd = dimIndexer.end();
      }

      SComplex& getComplex() const
      {
        return complex;
      }

      template<typename CellT1, typename CellT2>
      void coreduce(CellT1& a, CellT2& b)
      {
        BOOST_ASSERT(a.getDim() != b.getDim());

        if (a.getDim() > b.getDim()) {
          kingGetsMarried(a, b);
        } else {
          kingGetsMarried(b, a);
        }

        a.setColor(2);
        b.setColor(2);
      }

      typename Traits::Extract::result_type extract()
      {
        while (extractIt != extractEnd && extractIt->getColor() != 1)
          ++extractIt;

        if (extractIt != extractEnd)
          {
            BOOST_ASSERT(extractIt->getColor() == 1);

            int v = (extractIt->getDim() == 0) ? 0 : calcMorseValue(*extractIt);

            morse[extractIt->getId()] = v;
            aces.push_back(extractIt->getId());
            akq[extractIt->getId()] = ACE;

            return typename Traits::Extract::result_type::value_type(*extractIt);
          }


        reportPaths();
        return typename Traits::Extract::result_type();
      }

      OutputComplexType* getOutputComplex()
      {
        return outputComplex;
      }

    protected:

      template<typename CellT>
      int calcMorseValue(const CellT& cell)
      {
        int val = 0;
        BOOST_FOREACH(typename SComplex::Iterators::BdCells::iterator::value_type el, complex.iterators().bdCells(cell)) {
          val = std::max(val, morse[el.getId()]);
        }

        return val + 1;
      }

      template<typename CellT1, typename CellT2>
      void kingGetsMarried(const CellT1& king, const CellT2& queen)
      {
        akq[king.getId()] = KING;
        akq[queen.getId()] = QUEEN;
        const int v = calcMorseValue(king);
        morse[king.getId()] = v;
        morse[queen.getId()] = v;
        kerKing[queen.getId()] = king.getId();
      }

      template<typename CellT1, typename CellT2>
      Scalar toAceCoeff(const CellT1& x, const CellT2& y)
      {
        BOOST_ASSERT(akq[y.getId()] == ACE);
        BOOST_ASSERT(akq[x.getId()] != QUEEN);

        return complex.coincidenceIndex(x, y);
      }

      template<typename Cell1, typename Cell2, typename Cell3>
      Scalar toKingCoeff(const Cell1& x, const Cell2& y, const Cell3& y_star)
      {
        BOOST_ASSERT(akq[x.getId()] != QUEEN);
        BOOST_ASSERT(akq[y_star.getId()] == QUEEN);
        BOOST_ASSERT(akq[y.getId()] == KING);

        return Scalar(-1) * Scalar(complex.coincidenceIndex(x, y_star) / complex.coincidenceIndex(y, y_star));
      }

      template<typename CellT1, typename CellT2>
      const PathsInfo& followPathMemo(const CellT1& curr, const CellT2& from)
      {
        PathsInfo *&coeffs = followMemoTable[curr.getId()];

        if (coeffs != 0)
          return *coeffs;

        if (akq[curr.getId()] == UNSET)
          return emptyPathsInfo;

        BOOST_ASSERT(akq[curr.getId()] != ACE || curr.getId() == from.getId());
        coeffs = &pathsInfoPool.at(infosUsed++);
        const int ourMorseValue = morse[curr.getId()];

        // case 1: to ace
        BOOST_FOREACH(typename SComplex::Iterators::BdCells::iterator::value_type toCell,
                      this->complex.iterators().bdCells(curr)) {
          const Id to = toCell.getId();

          if (akq[to] == ACE && morse[to] < ourMorseValue) {
            const Scalar newCoeff = toAceCoeff(curr, toCell);
            (*coeffs)[to] += newCoeff;
          }
        }
        //compactCoeffs();

        // case 2: to king
        BOOST_FOREACH(typename SComplex::Iterators::BdCells::iterator::value_type bd, this->complex.iterators().bdCells(curr)) {
          if (akq[bd.getId()] != QUEEN)
            continue;

          Cell to = getComplex()[kerKing[bd.getId()]];

          if (akq[to.getId()] != KING) {
            continue;
          }

          BOOST_ASSERT(akq[to.getId()] == KING);

          if (morse[to.getId()] < ourMorseValue) {
            const Scalar newCoeff = toKingCoeff(curr, to, bd);
            const PathsInfo &fromHere = followPathMemo(to, from);

            for (typename PathsInfo::const_iterator it = fromHere.begin(); it != fromHere.end(); ++it) {
              (*coeffs)[it->first] += it->second*newCoeff;
            }
          }
        }

        for (typename PathsInfo::iterator it = coeffs->begin(), end = coeffs->end(); it != end;) {
            if (it->second == Scalar(0)) {
              it = coeffs->erase(it++);
            } else {
              ++it;
            }
        }


        return *coeffs;
      }

      void reportPaths()
      {
        CAPD_DEBUG_TMPL("Building output complex...");

        Stopwatch total;
        BOOST_FOREACH(Id ace, aces) {
          Cell cell = getComplex()[ace];
          const PathsInfo &cf = followPathMemo(cell, cell);
          for (typename PathsInfo::const_iterator it = cf.begin(); it != cf.end(); ++it) {
            coeffs[std::make_pair(ace, it->first)] += it->second;
          }
        }

        typedef std::pair<std::pair<int,int>,Scalar> Triple;
        std::vector<typename OutputComplexType::Dim> dims(aces.size());

        std::unordered_map<size_t, size_t> from0;

        BOOST_FOREACH(Id ace, aces) {
          size_t next = from0.size();
          Cell cell = getComplex()[ace];
          dims[next] = cell.getDim();
          from0[ace] = next;
        }

        typename OutputComplexType::KappaMap kap;

        BOOST_FOREACH(Triple p, coeffs) {
          long coef = convertToLong(p.second);
          if (coef != 0) {
            kap.push_back(boost::make_tuple(from0[p.first.first], from0[p.first.second],  coef));
          }
        }

        outputComplex = new OutputComplexType(3, dims, kap, 1);
        CAPD_DEBUG_TMPL("...output complex ready, critical cells: " << dims.size());
      }

      // temporarily public for fundamental group computations
    public:

      std::vector<int> morse;
      std::vector<AKQType> akq;
      std::vector<Id> kerKing;
      std::vector<Id> aces;

      int infosUsed;
      std::vector<PathsInfo> pathsInfoPool;

      std::vector<PathsInfo*> followMemoTable;

      int extractDim;
      int maxExtractDim;
      CoeffsMap coeffs;

      PathsInfo emptyPathsInfo;
      OutputComplexType *outputComplex;
      CAPD_CLASS_LOGGER;
    };

  }
}

#endif
