/////////////////////////////////////////////////////////////////////////////
/// @file AcyclicNeighbourhood
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-03-18
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_ACYCLICNEIGHBOURHOOD_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_ACYCLICNEIGHBOURHOOD_H

// #include
#include "TheCubicalAcyclicity.h"
#include <capd/complex/ComplexTraits.h>
#include <capd/complex/BdNeighbours.h>
#include <boost/range/adaptors.hpp>
#include <boost/utility/enable_if.hpp>
#include <boost/multi_array.hpp>
#include <boost/bind.hpp>

namespace capd
{
  namespace complex
  {

    template<typename Complex, size_t DIM, typename Implemented=void>
    class AcyclicNeighbourhood
    {
      typedef typename ComplexTraits<Complex>::Cell Cell;
    public:
      AcyclicNeighbourhood(Complex&) {}

      bool operator()(const Cell&) const
      {
        BOOST_ASSERT(false);
        return false;
      }
    };


    template<typename Complex, size_t DIM>
    class AcyclicNeighbourhood<Complex, DIM,
                               typename boost::enable_if<IsRegularCubical<Complex> >::type >
    {
    public:
      typedef typename ComplexTraits<Complex>::Cell Cell;
      typedef typename Complex::Color Color;

      AcyclicNeighbourhood(Complex& complex, Color color=Color(1)):
        _complex(complex), _color(color) {}

      bool operator()(const Cell& cell)
      {
        BdNeighbours<Complex, DIM> _neighbours(_complex, cell, DIM);
        auto neighbours = _neighbours | boost::adaptors::filtered(boost::bind(&Cell::getColor, _1) == _color);
        auto config = computeConfiguration(cell, neighbours);
        return TheCubicalAcyclicity<DIM>::getInstance()(config);
      }

    private:

      template<typename NeighboursRange>
      boost::multi_array<size_t, DIM> computeConfiguration(const Cell& cell, const NeighboursRange& neighbours) const
      {
        boost::array<size_t, DIM> shape;
        std::fill(shape.begin(), shape.end(), 3);

        boost::multi_array<size_t, DIM> config(shape);
        std::fill(config.origin(), config.origin() + config.num_elements(), 2);

        for (auto neighbour : neighbours) {
          auto coord = computeCoord(cell, neighbour);
          config(coord) = 1;
        }

        return config;
      }

      boost::array<size_t, DIM> computeCoord(const Cell& cell, const Cell& neighbour) const
      {
        auto cellCode = cell.code();
        auto neighbourCode = neighbour.code();
        boost::array<size_t, DIM> coord;

        for (size_t dim = 0; dim < DIM; ++dim) {
          BOOST_ASSERT(cellCode[dim].second);
          BOOST_ASSERT(neighbourCode[dim].second);

          coord[dim] = (1 + cellCode[dim].first) - neighbourCode[dim].first;
          BOOST_ASSERT(0 <= coord[dim] && coord[dim] <= 2);
        }
        //        CAPD_DEBUG("coord: " << coord);
        return coord;
      }

      Complex& _complex;
      Color _color;
    };

    template<size_t DIM, typename Complex>
    bool IsAcyclicNeighbourhood(Complex& complex, const typename ComplexTraits<Complex>::Cell& cell)
    {
      AcyclicNeighbourhood<Complex, DIM> algo(complex);
      return algo(cell);
    }

  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_ACYCLICNEIGHBOURHOOD_H
