/////////////////////////////////////////////////////////////////////////////
/// @file AssignIds
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-06-30
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_ASSIGNIDS_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_ASSIGNIDS_H

// #include
#include <vector>
#include <boost/iterator/counting_iterator.hpp>
#include <boost/algorithm/string/join.hpp>
#include <capd/auxil/ignoreUnused.h>

namespace capd
{
  namespace complex
  {

    // TODO reimplement it using boost::flat_map
    template <typename Iterator, typename Id, typename ObjectComparator=std::less<typename Iterator::value_type> >
    class AssignIds
    {
    public:
      typedef typename Iterator::value_type Object;

      static const Id npos = -1;

    private:

      struct ComparatorWithDummy
      {
        explicit ComparatorWithDummy(Iterator begin, const Object& dummy):
          _begin(begin),
          _dummy(dummy), _objectComparator()
        {}

        bool operator()(const Id& a, const Id&) const
        {
          const Object& aObj = *(_begin + a);
          const Object& bObj = _dummy; // according to specification of lower_bound, the second argument is always equal to searched value;
          return _objectComparator(aObj, bObj);
        }

      private:
        Iterator _begin;
        const Object& _dummy;
        const ObjectComparator _objectComparator;
      };

      struct Comparator
      {
        explicit Comparator(Iterator begin):
          _begin(begin), _objectComparator()
        {}

        bool operator()(const Id& a, const Id& b) const
        {
          const Object& aObj = *(_begin + a);
          const Object& bObj = *(_begin + b);
          return _objectComparator(aObj, bObj);
        }

      private:
        Iterator _begin;
        const ObjectComparator _objectComparator;
      };

    public:

      AssignIds(Iterator begin, Iterator end, size_t size):
        _begin(begin),
        _size(size),
        _ids(_size),
        _rankToPosition(boost::counting_iterator<int>(0),
                        boost::counting_iterator<int>(_size))
      {
        capd::auxil::ignoreUnused(end);
      }

      Id reverse(size_t pos) const { return _ids[pos]; }
      const std::vector<Id>& reversed() const { return _ids; }

      Id operator[](const Object& obj) const
      {
        ComparatorWithDummy compareWithDummy(_begin, obj);
        typename std::vector<Id>::const_iterator pos = std::lower_bound(_rankToPosition.begin(), _rankToPosition.end(),
                                                                        _size + 1,
                                                                        compareWithDummy);
        if (pos != _rankToPosition.end() && ! compareWithDummy(*pos, _size + 1)) {
          return _ids[*pos];
        }

        return npos;
      }

      virtual ~AssignIds(){}

      void operator()()
      {
        std::sort(_rankToPosition.begin(), _rankToPosition.end(),
                  Comparator(_begin));

        std::vector<Id> groups(_size);

        for (size_t idx = 0; idx < _size; ++idx) {
          groups[idx] = (idx > 0 ?  groups[idx - 1] : 0);
          groups[idx] += (idx > 0 && *(_begin + _rankToPosition[idx]) != *(_begin +  _rankToPosition[idx - 1]));
        }

        for (size_t idx = 0; idx < _size; ++idx) {
          _ids[_rankToPosition[idx]] = groups[idx];
        }

      }

    private:
      Iterator _begin;
      const size_t _size;
      std::vector<Id> _ids;
      std::vector<Id> _rankToPosition;
    };

    template <typename Id, typename Rng, typename ObjectComparator=std::less<typename Rng::value_type> >
    AssignIds<typename Rng::const_iterator, Id, ObjectComparator>
    MakeAssignIds(const Rng& rng)
    {
      return AssignIds<typename Rng::const_iterator, Id, ObjectComparator>(rng.begin(), rng.end(),
                                                                           boost::size(rng));
    }
  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_ASSIGNIDS_H
