/////////////////////////////////////////////////////////////////////////////
/// @file BuildGreedyDiscreteVectorField
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-04-02
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_BUILDGREEDYDISCRETEVECTORFIELD_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_BUILDGREEDYDISCRETEVECTORFIELD_H

// #include
#include <capd/complex/CellsByDim.h>
#include "DiscreteVectorField.h"
#include <capd/complex/ComplexFunctors.h>
#include <capd/complex/LoggerHelpers.h>

#include <deque>

#include <boost/optional.hpp>

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/topological_sort.hpp>
#include <boost/iterator/counting_iterator.hpp>
#include <algorithm>
#include <boost/range/adaptors.hpp>
#include <boost/container/flat_map.hpp>
#include <boost/range.hpp>
#include <boost/range/adaptors.hpp>
#include <boost/lambda/bind.hpp>
#include <functional>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/construct.hpp>
#include <utility>

namespace capd
{
  namespace complex
  {

    template<typename Complex>
    struct BuildGreedyDiscreteVectorFieldReduction
    {
      typedef capd::complex::CellsByDimReversed<Complex> CellsByDim;
      typedef capd::complex::CbdCells<Complex> GradientCells;
      typedef capd::complex::BdCells<Complex> CogradientCells;
    };

    template<typename Complex>
    struct BuildGreedyDiscreteVectorFieldCoreduction
    {
      typedef capd::complex::CellsByDim<Complex> CellsByDim;
      typedef capd::complex::BdCells<Complex> GradientCells;
      typedef capd::complex::CbdCells<Complex> CogradientCells;
    };

    template<typename ComplexT, typename ScalarT, typename Strategy=BuildGreedyDiscreteVectorFieldCoreduction<ComplexT> >
    class BuildGreedyDiscreteVectorField
    {

    public:
      typedef ComplexT Complex;
      typedef typename Complex::Cell Cell;
      typedef typename Complex::Id Id;
      typedef typename Complex::Dim Dim;
      typedef typename Complex::Color Color;

      typedef capd::complex::DiscreteVectorField<Complex, ScalarT> DiscreteVectorField;

      // typedef boost::adjacency_list<boost::vecS, boost::vecS, boost::directedS,
      //                               boost::property<boost::capd_complex_cell_dimension_t, Dim>,
      //                               boost::property<boost::capd_edge_coincidence_index_t, Scalar> >
      // MorseGraph;


      explicit BuildGreedyDiscreteVectorField(Complex& complex):
        _complex(complex),
        _discreteVectorField(_complex),
        _nextColor(2),
        _cellsByDim(_complex),
        _cellsByDimIt(_cellsByDim.begin()),
        _gradientCells(_complex),
        _cogradientCells(_complex),
        _waitsForProcessign(_complex.size())
      {
      }

      const DiscreteVectorField& operator()()
      {
        CAPD_DEBUG_TMPL("Starting...");
        while (extract()) {
          while (!_toProcess.empty()) {
            Id nextId = _toProcess.front();
            _toProcess.pop_front();

            Cell next = _complex[nextId];
            CAPD_TRACE_TMPL("Processing: " << " dim: " << (int) next.getDim() << " code: " << next.code() << " color: " << (size_t) next.getColor());

            if (next.getColor() == 1) {
              std::pair<int, Cell> match = uniqueMatching(next);
              if (match.first == 1) {
                processNeighbours(match.second); // before changing color, so we can use its actual color
                match.second.setColor(_nextColor);
                next.setColor(_nextColor);

                _discreteVectorField.match(next, match.second);
              } else if (match.first == 0) {
                processNeighbours(next);
              }
            }
            _waitsForProcessign[nextId] = false;
          }
        }

        CAPD_DEBUG_TMPL("... finished");
        return _discreteVectorField;
      }

    private:
      bool extract()
      {
        while (_cellsByDimIt != _cellsByDim.end() && _cellsByDimIt->getColor() != 1) {
          ++_cellsByDimIt;
        }

        if (_cellsByDimIt != _cellsByDim.end()) {
          Cell extracted = *_cellsByDimIt;
          CAPD_TRACE_TMPL("Extracting: "  << " dim: " << (int) extracted.getDim() << " code: " << extracted.code() << " color: " << (size_t) extracted.getColor());
          processNeighbours(extracted); // before changing color, so we can use its actual color
          markCritical(extracted);
          return true;
        }

        return false;
      }

      void markCritical(Cell& cell)
      {
        BOOST_ASSERT(cell.getColor() == 1);
        cell.setColor(_nextColor);
        _discreteVectorField.markCritical(cell);
      }

      std::pair<int, Cell> uniqueMatching(const Cell& cell)
      {
        std::pair<int, Cell> result(0, Cell());

        for (auto gradient : _gradientCells(cell)) {
          if (gradient.getColor() == cell.getColor()) {
            if (result.first == 1) {
              result.first = 2;
              break;
            } else {
              result.first = 1;
              result.second = gradient;
            }
          }
        }
        return result;
      }

      void processNeighbours(const Cell& cell)
      {
        for (auto next : _cogradientCells(cell)) {
          if (!_waitsForProcessign[next.getId()] && next.getColor() == cell.getColor()) {
            _toProcess.push_back(next.getId());
            _waitsForProcessign[next.getId()] = true;
            CAPD_TRACE_TMPL("Adding: " <<  " dim: " << (int) next.getDim() << " code: " << next.code() << " color: " << (size_t) next.getColor());
          }
        }
      }


      BuildGreedyDiscreteVectorField(const BuildGreedyDiscreteVectorField&);
      BuildGreedyDiscreteVectorField& operator=(const BuildGreedyDiscreteVectorField&);

      Complex& _complex;
      DiscreteVectorField _discreteVectorField;

      Color _nextColor;
      typename Strategy::CellsByDim _cellsByDim;
      typename Strategy::CellsByDim::iterator _cellsByDimIt;
      typename Strategy::GradientCells _gradientCells;
      typename Strategy::CogradientCells _cogradientCells;

      std::deque<Id> _toProcess;
      std::vector<bool> _waitsForProcessign;

      CAPD_CLASS_LOGGER;
    };


    template<typename Complex, typename Scalar>
    struct BuildGreedyDiscreteVectorFieldViaCoreductions {
      typedef BuildGreedyDiscreteVectorField<Complex, Scalar, BuildGreedyDiscreteVectorFieldCoreduction<Complex> > type;

    };

    template<typename Complex, typename Scalar>
    struct BuildGreedyDiscreteVectorFieldViaReductions {
      typedef BuildGreedyDiscreteVectorField<Complex, Scalar, BuildGreedyDiscreteVectorFieldReduction<Complex> > type;
    };

    template<typename Scalar, typename Complex>
    size_t CountCriticalCellsViaCoreduction(Complex& complex)
    {
      typename BuildGreedyDiscreteVectorFieldViaCoreductions<Complex, Scalar>::type algo(complex);
      return algo().criticalCells();
    }

    template<typename Scalar, typename Complex>
    size_t CountCriticalCellsViaReduction(Complex& complex)
    {
      typename BuildGreedyDiscreteVectorFieldViaReductions<Complex, Scalar>::type algo(complex);
      return algo().criticalCells();
    }


  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_BUILDGREEDYDISCRETEVECTORFIELD_H
