/////////////////////////////////////////////////////////////////////////////
/// @file BuildMorseGraph
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-04-02
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_BUILDMORSEGRAPH_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_BUILDMORSEGRAPH_H

// #include
#include "HasseDiagram.h"
#include <capd/homAux/MergeMaps.h>
#include <boost/shared_ptr.hpp>
#include <boost/container/flat_map.hpp>
#include <boost/lambda/construct.hpp>
#include <boost/range/adaptors.hpp>
#include <capd/auxil/Logger.h>
#include <capd/homAux/PositionInSequence.h>

namespace capd
{
  namespace complex
  {



    template<typename DiscreteVectorField>
    class BuildMorseGraph
    {

      typedef typename DiscreteVectorField::Scalar Scalar;
      typedef typename DiscreteVectorField::Dim Dim;
      typedef typename DiscreteVectorField::Graph DVFGraph;
      typedef typename DiscreteVectorField::Graph::Id Id;
      typedef boost::container::flat_map<Id, Scalar> Boundary;

      struct BoundaryClosure
      {
        Boundary upDim;
        Boundary sameDim;
        Boundary downDim;
      };

    public:
      typedef HasseDiagram<Dim, Scalar> Graph;

      BuildMorseGraph(const DiscreteVectorField& discreteVectorField):
        _discreteVectorField(discreteVectorField),
        _boundaries(_discreteVectorField.graph().size()),
        _graph(_discreteVectorField.criticalCells())
      {}

      void operator()()
      {
        closeCriticalBoundary();
        buildGraph();
      }

      const Graph& graph() const { return _graph; }
      Graph& graph() { return _graph; } // temporary, we should implement std::move for the graph

      const std::vector<Id>& originalIds() const { return _discreteVectorField.allCriticalCells(); }

    private:
      BuildMorseGraph(const BuildMorseGraph&);
      BuildMorseGraph& operator=(const BuildMorseGraph&);

      void buildGraph()
      {
        namespace bl = boost::lambda;
        namespace ba = boost::adaptors;

        auto criticalCellSeqNo = capd::homAux::ComputePositionInSequence(this->originalIds());

        for (auto criticalCell : this->originalIds()) {
          _graph.set(criticalCellSeqNo[criticalCell], _discreteVectorField.complex()[criticalCell].getDim());
        }

        for (auto criticalCell : this->originalIds()) {
          Id ccU = criticalCellSeqNo[criticalCell];
          for (auto neighbour : _boundaries[criticalCell].downDim) {
            if (neighbour.second != Scalar(0) && // TODO we can skip 0's during MergeMaps, it should speed up
                _discreteVectorField.isCritical(neighbour.first)) {
              Id ccV = criticalCellSeqNo[neighbour.first];
              _graph.connect(ccU, ccV, neighbour.second);
            }
          }
        }
      }

      void closeCriticalBoundary()
      {
        const DVFGraph& graph = _discreteVectorField.graph();

        for (Id vertex : graph.vertices()) {
          if (_discreteVectorField.isCritical(vertex)) {
            _boundaries[vertex].sameDim[vertex] = Scalar(1);
          }
        }

        for (Id source : graph.orderedVertices()) {
          BoundaryClosure& sourceBd = _boundaries[source];

          for (auto neighbour : graph.neighbours(source)) {
            const Id target = boost::get<0>(neighbour);
            const Scalar ci = boost::get<1>(neighbour);
            const bool goUp = boost::get<2>(neighbour);
            BoundaryClosure& targetBd = _boundaries[target];

            if (goUp) {
              sourceBd.upDim = closeCriticalBoundary(ci, sourceBd.upDim, targetBd.sameDim);
              sourceBd.sameDim = closeCriticalBoundary(ci, sourceBd.sameDim, targetBd.downDim);
            } else {
              sourceBd.sameDim = closeCriticalBoundary(ci, sourceBd.sameDim, targetBd.upDim);
              sourceBd.downDim = closeCriticalBoundary(ci, sourceBd.downDim, targetBd.sameDim);
            }
          }
        }
      }

      Boundary closeCriticalBoundary(const Scalar& ci, const Boundary& source, const Boundary& target)
      {
        namespace bl = boost::lambda;
        typedef typename Boundary::value_type BoundaryValue;

        auto multiply = bl::bind(std::multiplies<Scalar>(), ci,
                                 bl::bind(&BoundaryValue::second, bl::_1)
                                 );
        auto multiplyPair = bl::bind(bl::constructor<BoundaryValue>(),
                                     bl::bind(&BoundaryValue::first, bl::_1),
                                     multiply);
        auto targetMultiplied = target | boost::adaptors::transformed(multiplyPair);

        return homAux::DoMergeMaps(source, targetMultiplied, std::plus<Scalar>());
      }

      const DiscreteVectorField& _discreteVectorField;
      std::vector<BoundaryClosure> _boundaries;
      Graph _graph;
    };


  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_BUILDMORSEGRAPH_H
