/////////////////////////////////////////////////////////////////////////////
/// @file CellComplex
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-09-13
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_CELLCOMPLEX_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_CELLCOMPLEX_H

// #include
#include "ColoredIteratorFamily.h"
#include "CellComplex_IteratorFamily.h"
#include <capd/complex/SComplexBase.h>
#include <boost/optional.hpp>
#include <vector>
#include <algorithm>
#include <iosfwd>
#include <capd/homAux/UniversalVariable.h>

namespace capd
{
  namespace complex
  {


    template<typename MemoryModel>
    class CellComplex: public SComplexBase
    {

    public:
      typedef typename MemoryModel::Id Id;
      typedef typename MemoryModel::Color Color;
      typedef typename MemoryModel::Dim Dim;

      typedef typename MemoryModel::EnumerateCellCodes EnumerateCellCodes;
      typedef typename EnumerateCellCodes::CellCode CellCode;

      typedef typename MemoryModel:: template ValueById<bool> AvailableCellsType;
      typedef typename MemoryModel:: template ValueById<Color> ColorsType;
      typedef typename AvailableCellsType::type AvailableCells;
      typedef typename ColorsType::type Colors;

      template<typename ComplexT>
      class CellImpl
      {
      public:
        typedef ComplexT Complex;
        typedef typename Complex::Color Color;
        typedef typename Complex::Dim Dim;
        typedef typename Complex::Id Id;

        CellImpl(): _id(-1), _complex(NULL) {}

        explicit CellImpl(Id id, Complex* complex): _id(id), _complex(complex)
        {
          BOOST_ASSERT(id < complex->_availableCells.size());
        }

        operator CellImpl<const ComplexT>() const
        {
          return CellImpl<const ComplexT>(this->_id, this->_complex);
        }

        Id id() const { return _id; }
        Id getId() const { return id(); }

        Dim dim() const { return _complex->dim(_id); }
        Dim getDim() const { return dim(); }

        Color color() const { return _complex->_colors[_id]; }
        Color getColor() const { return color(); } //TODO refactor, remove getColor

        void setColor(Color color) { _complex->_colors[_id] = color; }

        CellCode code() const
        {
          return (*_complex->_enumerateCellCodes)[_id];
        }

        bool operator<(const CellImpl& other) const
        {
          return this->_id < other._id;
        }
      private:
        Id _id;
        Complex* _complex;
      };

      typedef CellImpl<CellComplex> Cell;
      typedef CellImpl<const CellComplex> ConstCell;

      typedef CellComplex_IteratorFamily<CellComplex, Cell, ConstCell> Iterators;
      typedef CellComplex_IteratorFamily<const CellComplex, ConstCell, ConstCell> ConstIterators;
      typedef ColoredIteratorFamily<Iterators> ColoredIterators;
      typedef ColoredIteratorFamily<ConstIterators> ColoredConstIterators;


      CellComplex(const CellComplex&);

      CellComplex(EnumerateCellCodes* enumerateCellCodes, bool reference=false):
        _enumerateCellCodes(enumerateCellCodes, reference),
        _availableCells(AvailableCellsType::Create(_enumerateCellCodes->size())),
        _colors(ColorsType::Create(_enumerateCellCodes->size()))
      {
      }

      explicit CellComplex(const EnumerateCellCodes& enumerateCellCodes):
        _enumerateCellCodes(enumerateCellCodes),
        _availableCells(AvailableCellsType::Create(_enumerateCellCodes->size())),
        _colors(ColorsType::Create(_enumerateCellCodes->size()))
      {
      }


      virtual ~CellComplex()
      {
      }

      size_t size() const { return _enumerateCellCodes->size(); }

      size_t cardinality() const {
        typedef typename ColoredConstIterators::AllCells AllCells;
        AllCells allCells = this->iterators(1).allCells();
        return std::distance(allCells.begin(), allCells.end());
      }


      CellCode cellCode() const { return _enumerateCellCodes->cellCode(); }

      Dim dim() const { return this->_enumerateCellCodes->dim(); }
      Dim getDim() const { return dim(); }

      Iterators iterators()
      {
        return Iterators(*this);
      }

      ConstIterators iterators() const
      {
        return ConstIterators(*this);
      }

      ColoredIterators iterators(Color color)
      {
        return ColoredIterators(iterators(), color);
      }

      ColoredConstIterators iterators(Color color) const
      {
        return ColoredConstIterators(iterators(), color);
      }

      ConstCell operator[](const CellCode& cellCode) const
      {
        Id id = (*_enumerateCellCodes)[cellCode];
        BOOST_ASSERT(exists(id));
        return ConstCell(id, this);
      }

      ConstCell operator[](Id id) const
      {
        BOOST_ASSERT(exists(id));
        return ConstCell(id, this);
      }

      Cell operator[](const CellCode& cellCode)
      {
        Id id = (*_enumerateCellCodes)[cellCode];
        BOOST_ASSERT(exists(id));
        return Cell(id, this);
      }

      Cell operator[](Id id)
      {
        BOOST_ASSERT(exists(id));
        return Cell(id, this);
      }

      boost::optional<Cell> find(const CellCode& cellCode)
      {
        const Id id = (*_enumerateCellCodes)[cellCode];
        if (!exists(id)) {
          return boost::optional<Cell>();
        }

        return boost::optional<Cell>(Cell(id, this));
      }

      boost::optional<Cell> find(Id id)
      {
        if (!exists(id)) {
          return boost::optional<Cell>();
        }

        return boost::optional<Cell>(Cell(id, this));
      }

      bool exists(Id id) const
      {
        return id < _availableCells.size() && _availableCells[id];
      }

      Cell insert(const CellCode& cellCode)
      {
        const Id id = (*_enumerateCellCodes)[cellCode];
        _availableCells[id] = true;
        _colors[id] = 1;
        return Cell(id, this);
      }

      void remove(const CellCode& cellCode)
      {
        const Id id = (*_enumerateCellCodes)[cellCode];
        _availableCells[id] = false;
      }

      void color(const Cell& cell, Color color)
      {
        _colors[cell.id()] = color;
      }

      void fillWithBoundaries()
      {
        typedef typename Iterators::AllCells AllCells;
        AllCells all = this->iterators().allCells();
        for (typename AllCells::iterator it = all.begin(), end = all.end();
             it != end; ++it) {
          fillWithBoundary(*it, it->color(), false);
        }
      }

      void fillWithBoundary(const Cell& cell, Color color=Color(), bool keepLower=true)
      {
        if (!_availableCells[cell.id()]) {
          _availableCells[cell.id()] = true;
          _colors[cell.id()] = color;
        } else if (!keepLower || _colors[cell.id()] > color) {
            _colors[cell.id()] = color;
        }


        typename EnumerateCellCodes::Bd bd = _enumerateCellCodes->bd(cell.id());
        for (typename EnumerateCellCodes::Bd::const_iterator it = bd.begin(), end = bd.end();
             it != end; ++it) {
          Cell bdCell(*it, this);
          fillWithBoundary(bdCell, color, keepLower);
        }
      }

      void resetColors(Color color)
      {
        std::fill(_colors.begin(), _colors.end(), color);
      }

    protected:

      EnumerateCellCodes& enumerateCellCodes()
      {
        return *_enumerateCellCodes;
      }

      const EnumerateCellCodes& enumerateCellCodes() const
      {
        return *_enumerateCellCodes;
      }

    private:
      template<typename ComplexT, typename CellT, typename ConstCellT>
      friend class CellComplex_IteratorFamily;

      Dim dim(Id id) const
      {
        return _enumerateCellCodes->dim(id);
      }


      capd::homAux::UniversalVariable<EnumerateCellCodes> _enumerateCellCodes;
      AvailableCells _availableCells;
      Colors _colors;
    };


  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_CELLCOMPLEX_H
