/////////////////////////////////////////////////////////////////////////////
/// @file CellComplex_IteratorFamily
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-02-24
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_CELLCOMPLEX_ITERATORFAMILY_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_CELLCOMPLEX_ITERATORFAMILY_H

// #include
#include <boost/iterator/transform_iterator.hpp>
#include <boost/range/iterator_range.hpp>
#include <boost/iterator/filter_iterator.hpp>

namespace capd
{
  namespace complex
  {

    template<typename ComplexT, typename CellT, typename ConstCellT>
    class CellComplex_IteratorFamily
    {

    public:
      typedef ComplexT Complex;
      typedef CellT Cell;
      typedef ConstCellT ConstCell;
      typedef typename Complex::Dim Dim;

      CellComplex_IteratorFamily(): _complex(0)
      {}

      explicit CellComplex_IteratorFamily(Complex& complex):
        _complex(&complex)
      {}

      virtual ~CellComplex_IteratorFamily() {}

    private:

      struct IdToCell
      {
        typedef Cell result_type;

        IdToCell(Complex* cellComplex):
          _cellComplex(cellComplex)
        {}

        IdToCell():
          _cellComplex(0)
        {}

        Cell operator()(typename Complex::Id id) const
        {
          return (*_cellComplex)[id];
        }

        Complex* _cellComplex;
      };

      struct AvailableIdPred
      {
        AvailableIdPred(Complex* complex):
          _complex(complex)
        {}

        AvailableIdPred():
          _complex(0)
        {}

        bool operator()(typename Complex::Id id)
        {
          return (*_complex)._availableCells[id];
        }

        Complex* _complex;
      };


      struct DimPred
      {
        explicit DimPred(Dim dim): _dim(dim) {}
        bool operator()(const Cell& cell) const
        {
          return cell.getDim() == _dim;
        }
        Dim _dim;
      };


    public:
      typedef boost::filter_iterator<AvailableIdPred, typename Complex::EnumerateCellCodes::All::iterator> AllAvailableCells;
      typedef boost::transform_iterator<IdToCell, AllAvailableCells> TransforAllCells;
      typedef boost::iterator_range<TransforAllCells> AllCells;

      typedef boost::filter_iterator<DimPred, typename AllCells::iterator> DimCellsIterator;
      typedef boost::iterator_range<DimCellsIterator> DimCells;

      typedef boost::filter_iterator<AvailableIdPred, typename Complex::EnumerateCellCodes::Bd::iterator> BdAvailableCells;
      typedef boost::transform_iterator<IdToCell, BdAvailableCells> TransforBdCells;
      typedef boost::iterator_range<TransforBdCells> BdCells;

      typedef boost::filter_iterator<AvailableIdPred, typename Complex::EnumerateCellCodes::Cbd::iterator> CbdAvailableCells;
      typedef boost::transform_iterator<IdToCell, CbdAvailableCells> TransforCbdCells;
      typedef boost::iterator_range<TransforCbdCells> CbdCells;


      AllCells allCells()
      {
        typename Complex::EnumerateCellCodes::All all = _complex->_enumerateCellCodes->all();

        return AllCells(TransforAllCells(AllAvailableCells(AvailableIdPred(_complex),
                                                        all.begin(), all.end()),
                                         IdToCell(_complex)),
                       TransforAllCells(AllAvailableCells(AvailableIdPred(_complex),
                                                        all.end(), all.end()),
                                       IdToCell(_complex)));
      }

      BdCells bdCells(const ConstCell& cell)
      {
        typename Complex::EnumerateCellCodes::Bd bd = _complex->_enumerateCellCodes->bd(cell.id());

        return BdCells(TransforBdCells(BdAvailableCells(AvailableIdPred(_complex),
                                                        bd.begin(), bd.end()),
                                       IdToCell(_complex)),
                       TransforBdCells(BdAvailableCells(AvailableIdPred(_complex),
                                                        bd.end(), bd.end()),
                                       IdToCell(_complex)));
      }

      CbdCells cbdCells(const ConstCell& cell)
      {
        typename Complex::EnumerateCellCodes::Cbd cbd = _complex->_enumerateCellCodes->cbd(cell.id());
        return CbdCells(TransforCbdCells(CbdAvailableCells(AvailableIdPred(_complex),
                                                           cbd.begin(), cbd.end()),
                                         IdToCell(_complex)),
                        TransforCbdCells(CbdAvailableCells(AvailableIdPred(_complex),
                                                           cbd.end(), cbd.end()),
                                         IdToCell(_complex)));
      }



      DimCells dimCells(Dim dim)
      {
        AllCells allCells = this->allCells();
        return DimCells(DimCellsIterator(DimPred(dim), allCells.begin(), allCells.end()),
                        DimCellsIterator(DimPred(dim), allCells.end(), allCells.end()));
      }

    private:
      Complex* _complex;

    };


  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_CELLCOMPLEX_ITERATORFAMILY_H
