/////////////////////////////////////////////////////////////////////////////
/// @file CellSComplex
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-06-27
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_CELLSCOMPLEX_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_CELLSCOMPLEX_H

#include "AssignIds.h"

#include <algorithm>
#include <boost/typeof/typeof.hpp>
#include <boost/assert.hpp>
#include <utility>
#include <boost/function.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/iterator/counting_iterator.hpp>
#include <boost/iterator/function_input_iterator.hpp>
#include <boost/assign.hpp>
#include <boost/iterator.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/iterator/counting_iterator.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/shared_ptr.hpp>



namespace capd
{
  namespace complex
  {

    template<typename Object, typename Id, typename ObjectStorage, typename ObjectComparator >
    class IdentifiableObjectContainer
    {

    public:

      static const Id npos = -1;

      template <typename Iterator>
      IdentifiableObjectContainer(Iterator begin, Iterator end, size_t size):
        _ownStorage(new ObjectStorage(begin, end)),
        _storage(*_ownStorage),
        _assignIds(_storage.begin(), _storage.end(), size)
      {
        _assignIds();
      }

      explicit IdentifiableObjectContainer(const ObjectStorage& storage):
        _ownStorage(),
        _storage(storage),
        _assignIds(_storage.begin(), _storage.end(), _storage.size())
      {
        _assignIds();
      }

      Id operator[](const Object& obj) const
      {
        return _assignIds[obj];
      }

    private:
      boost::shared_ptr<ObjectStorage> _ownStorage;
      const ObjectStorage& _storage;
      AssignIds<typename ObjectStorage::const_iterator, Id, ObjectComparator> _assignIds;
    };



    template<typename CellCodeT, typename Traits, typename CellCodeStorage>
    class CellSComplex
    {

    public:
      typedef CellCodeT CellCode;
      typedef typename Traits::Id Id;
      typedef typename Traits::Dim Dim;
      typedef typename Traits::Color Color;

      typedef boost::iterator_range<typename std::vector<Id>::const_iterator> Boundary;


      struct Cell
      {
        explicit Cell(Id id, const CellSComplex& complex): _id(id), _complex(complex) {}

        Dim dim() const { return _complex._dims[_id]; }
        Color color() const { return _complex._colors[_id]; }
        Id id() const { return _id; }

      protected:
        const Id _id;
        const CellSComplex& _complex;
      };

      friend struct Cell;

      template<typename BoundaryGenerator, typename Dimension,
               typename PositionToIdCallback>
      CellSComplex(const CellCodeStorage& storage,
                   BoundaryGenerator boundaryGenerator,
                   Dimension dimension,
                   PositionToIdCallback positionToIdCallback):
        _boundaryRange(storage.size()),
        _dims(storage.size()),
        _colors(storage.size()),
        _objects(storage)
      {
        init(storage.begin(), storage.end(), storage.size(),
             boundaryGenerator, dimension,
             positionToIdCallback);
      }

      template<typename BoundaryGenerator, typename Dimension>
      CellSComplex(const CellCodeStorage& storage,
                   BoundaryGenerator boundaryGenerator,
                   Dimension dimension):
        _boundaryRange(storage.size()),
        _dims(storage.size()),
        _colors(storage.size()),
        _objects(storage),
        _dim(0)
      {
        init(storage.begin(), storage.end(), storage.size(),
             boundaryGenerator, dimension);
      }

      template<typename Iterator, typename BoundaryGenerator, typename Dimension>
      CellSComplex(Iterator begin, Iterator end, size_t size,
                   BoundaryGenerator boundaryGenerator,
                   Dimension dimension):
        _boundaryRange(size),
        _dims(size),
        _colors(size),
        _objects(begin, end, size),
        _dim(0)
      {
        init(begin, end, size, boundaryGenerator, dimension);
      }

      virtual ~CellSComplex() {}

      Id size() const { return _dims.size(); }
      Dim dim() const { return _dim; }
      Cell operator[](const CellCode& cellCode) const { return Cell(_objects[cellCode], *this); }
      Cell operator[](Id cellId) const { return Cell(cellId, *this); }

      void setColor(Id id, const Color& color) { _colors[id] = color; }


      Boundary boundary(const Cell& cell) const {
        return Boundary(_boundaries.begin() + _boundaryRange[cell.id()].first,
                        _boundaries.begin() + _boundaryRange[cell.id()].second);
      }

    private:

      struct EmptyCallback { void operator()(Id,Id) const {} };

      template<typename Iterator, typename BoundaryGenerator,
               typename Dimension>
      void init(Iterator begin, Iterator end, size_t size,
                BoundaryGenerator boundaryGenerator,
                Dimension dimension)
      {
        init(begin, end, size, boundaryGenerator, dimension, EmptyCallback());
      }

      template<typename Iterator, typename BoundaryGenerator,
               typename Dimension, typename PositionToIdCallback>
      void init(Iterator begin, Iterator end, size_t size,
                BoundaryGenerator boundaryGenerator,
                Dimension dimension, PositionToIdCallback positionToIdCallback)
      {

        size_t currPosition = 0;

        for (Iterator curr = begin; curr != end; ++curr) {
          const Id currId = _objects[*curr];
          BOOST_ASSERT(currId != _objects.npos);

          typedef typename BoundaryGenerator::result_type Boundary;
          Boundary boundary = boundaryGenerator(*curr);
          _boundaryRange[currId].second = boundary.size();

          _dims[currId] = dimension(*curr);
          positionToIdCallback(currPosition++, currId);
        }

        for (size_t idx = 1; idx < size; ++idx) {
          _boundaryRange[idx].first = _boundaryRange[idx - 1].second;
          _boundaryRange[idx].second += _boundaryRange[idx - 1].second;
        }

        _boundaries.resize(_boundaryRange[size - 1].second);

        for (Iterator curr = begin; curr != end; ++curr) {
          const Id currId = _objects[*curr];
          BOOST_ASSERT(currId != _objects.npos);

          typedef typename BoundaryGenerator::result_type Boundary;
          Boundary boundary = boundaryGenerator(*curr);

          size_t bdPos = _boundaryRange[currId].first;

          for (typename Boundary::const_iterator bdIt = boundary.begin(),
                 bdEnd = boundary.end(); bdIt != bdEnd; ++bdIt) {
            const Id bdId = _objects[*bdIt];
            BOOST_ASSERT(bdId != _objects.npos);
            BOOST_ASSERT(bdId < currId); // so we have good order for a filtration

            _boundaries[bdPos++] = bdId;
          }

        }

        if (!_dims.empty()) {
          _dim = *std::max_element(_dims.begin(), _dims.end());
        }
      }


      std::vector<std::pair<Id, Id> > _boundaryRange;
      std::vector<Id> _boundaries;
      std::vector<Dim> _dims;
      std::vector<Color> _colors;
      IdentifiableObjectContainer<CellCode, Id, CellCodeStorage, typename Traits::CellCodeComparator> _objects;
      Dim _dim;
    };


    template<typename Traits, typename CellCodeStorage = std::vector<typename Traits::CellCode> >
    class SimplicialCellSComplex: public CellSComplex<typename Traits::CellCode, Traits, CellCodeStorage>
    {
    public:
      typedef CellSComplex<typename Traits::CellCode, Traits, CellCodeStorage> BaseComplex;
      using typename BaseComplex::CellCode;
      using typename BaseComplex::Dim;

    private:
      struct BoundaryGenerator
      {
        struct CellBoundaries
        {
          typedef CellCode result_type;
          const CellCode& _cell;

          explicit CellBoundaries(const CellCode& cell): _cell(cell) {}

          CellCode operator()(Dim _skipDim) const
          {
            BOOST_ASSERT(_cell.size() > 0);
            BOOST_ASSERT((size_t)_skipDim < _cell.size());
            CellCode result;
            result.reserve(_cell.size() - 1);

            std::for_each(_cell.begin(), _cell.begin() + _skipDim, boost::assign::push_back(result));
            std::for_each(_cell.begin() + _skipDim + 1, _cell.end(), boost::assign::push_back(result));

            return result;
          }
        };

        typedef boost::counting_iterator<Dim> DimIterator;
        typedef boost::transform_iterator<CellBoundaries, DimIterator> CellBoundariesIterator;
        typedef boost::iterator_range<CellBoundariesIterator> result_type;

        result_type operator()(const CellCode& cell) const
        {
          const size_t bds = (cell.size() == 1 ? 0 : cell.size());
          return result_type(CellBoundariesIterator(DimIterator(0), CellBoundaries(cell)),
                             CellBoundariesIterator(DimIterator(bds), CellBoundaries(cell)));
        }
      };

      struct Dimension
      {
        Dim operator()(const CellCode& cell) const { return cell.size() - 1; }
      };

    public:

      template<typename Iterator>
      SimplicialCellSComplex(Iterator begin, Iterator end, size_t size):
        BaseComplex(begin, end, size, BoundaryGenerator(), Dimension())
      {}

      template<typename PositionToIdCallback>
      explicit SimplicialCellSComplex(const CellCodeStorage& storage,
                                      PositionToIdCallback positionToIdCallback):
        BaseComplex(storage, BoundaryGenerator(), Dimension(), positionToIdCallback)
      {}

      explicit SimplicialCellSComplex(const CellCodeStorage& storage):
        BaseComplex(storage, BoundaryGenerator(), Dimension())
      {}

      virtual ~SimplicialCellSComplex() {}
    };

    template<typename T>
    struct CellSComplexTraitsT
    {
      typedef T Color;
      typedef T Dim;
      typedef T Id;
    };

    template<typename T>
    struct SimplicialCellSComplexTraitsT: public CellSComplexTraitsT<T>
    {
      typedef std::vector<int> CellCode;

      struct CellCodeComparator
      {
        bool operator()(const CellCode& a, const CellCode& b) const
        {
          if (a.size() < b.size()) {
            return true;
          } else if (a.size() > b.size()) {
            return false;
          } else {
            return a < b;
          }
        }
      };
    };

  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_CELLSCOMPLEX_H
