/////////////////////////////////////////////////////////////////////////////
/// @file Chain.h
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2013-12-13
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2013 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.

#ifndef CAPD_FILE_CHAIN_H
#define CAPD_FILE_CHAIN_H

#include <map>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/assert.hpp>
#include <algorithm>


namespace capd
{
namespace complex
{

  template<typename ComplexT, typename ScalarT>
  class Chain
  {
  public:
    typedef ComplexT Complex;
    typedef ScalarT Scalar;

    typedef typename Complex::Id Id;
    typedef typename Complex::Cell Cell;
    typedef typename Complex::ConstCell ConstCell;
    typedef std::map<Id, Scalar> Map; // TODO use google hash(?)

    typedef typename Map::const_iterator const_iterator;
    typedef typename Map::const_iterator iterator;
    typedef typename Map::value_type value_type;

    explicit Chain(){}

    explicit Chain(const Cell& cell, Scalar coef = Scalar(1))
    {
      _map[cell.getId()] = coef;
    }

    explicit Chain(const Id cell, Scalar coef = Scalar(1))
    {
      _map[cell] = coef;
    }

    template<typename It>
    Chain(It b, It e)
    {
      while(b != e) {
	_map[b->getId()] += Scalar(1);
	++b;
      }

      compact();
    }

    const Scalar operator[](const Cell& cell) const
    {
      return (*this)[cell.getId()];
    }

    const Scalar operator[](const Id id) const
    {
      const_iterator it = _map.find(id);
      return (it == _map.end() ? Scalar(0) : it->second);
    }

    Chain& operator+=(const value_type& v)
    {
      typename Map::iterator it = _map.find(v.first);
      if (it != _map.end()) {
	it->second += v.second;

	if (it->second == Scalar(0)) {
	  _map.erase(it);
	}
      } else if (v.second != Scalar(0)){
	_map.insert(v);
      }

      return *this;
    }

    Chain& operator+=(const Chain& other)
    {
      // TODO can we can speed it up?
      BOOST_FOREACH(const value_type& chainElem, other) {
	(*this) += chainElem;
      }

      return *this;
    }

    Chain& insert(const value_type& v)
    {
      return insert(v.first, v.second);
    }

    Chain& insert(const Cell& cell, const Scalar& v)
    {
      return insert(cell.getId(), v);
    }

    Chain& insert(Id id, const Scalar& v)
    {
      if (v == Scalar(0)) {
        typename Map::iterator it = _map.find(id);
        if (it != _map.end()) {
          _map.erase(it);
        }
      } else {
        _map[id] = v;
      }
      return *this;
    }

    Chain& operator*=(const Scalar& v)
    {
      namespace bl = boost::lambda;
      std::for_each(_map.begin(), _map.end(), bl::bind(&value_type::second, bl::_1) *= v);
      compact();
      return *this;
    }

    const_iterator begin() const { return _map.begin(); }
    const_iterator end() const { return _map.end(); }

    size_t size() const { return _map.size(); }
    bool empty() const { return size() == 0; }

    bool isZero() const
    {
      namespace bl = boost::lambda;

      if (_map.empty()) {
        return true;
      }
      return std::all_of(_map.begin(), _map.end(),
                                      bl::bind(&value_type::second, bl::_1) == 0);
    }

    void compact()
    {
      for (typename Map::iterator it = _map.begin(), end = _map.end(); it != end;) {
	if (it->second == Scalar(0)) {
	  _map.erase(it++);
	} else {
	  ++it;
	}
      }
    }

    Chain boundary(const Complex& complex) const
    {
      Chain result;

      BOOST_FOREACH(const value_type& chainElem, *this) {
	ConstCell cell = complex[chainElem.first];

	// TODO const_cast because of non-implemented const methods in SimplexSComplex
	typename Complex::ColoredIterators::BdCells bdCells = const_cast<Complex&>(complex).iterators(cell.getColor()).bdCells(cell);

	BOOST_FOREACH(const Cell bdCell, bdCells) {
	  result += value_type(bdCell.getId(), chainElem.second * complex.coincidenceIndex(cell, bdCell));
	}
      }

      result.compact();

      return result;
    }

  private:
    Map _map;
  };

}
}

#endif // CAPD_FILE_CHAIN_H
