/////////////////////////////////////////////////////////////////////////////
/// @file ComplexFunctors
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-03-20
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_COMPLEXFUNCTORS_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_COMPLEXFUNCTORS_H

// #include
#include <capd/complex/ComplexTraits.h>
#include <boost/lambda/bind.hpp>


namespace capd
{
  namespace complex
  {

    template<typename Cell>
    struct CellColor
    {
      typedef typename Cell::Color result_type;

      result_type
      operator()(const Cell& cell) const
      {
        return cell.getColor();
      }
    };

    template<typename Complex>
    CellColor<typename ComplexTraits<Complex>::Cell>
    MakeCellColor()
    {
      return CellColor<typename ComplexTraits<Complex>::Cell>();
    }

    template<typename Cell>
    struct CellId
    {
      typedef typename Cell::Id result_type;

      result_type
      operator()(const Cell& cell) const
      {
        return cell.getId();
      }
    };

    template<typename Complex>
    CellId<typename ComplexTraits<Complex>::Cell>
    MakeCellId()
    {
      return CellId<typename ComplexTraits<Complex>::Cell>();
    }

    template<typename Cell>
    struct CellDim
    {
      typedef typename Cell::Dim result_type;

      result_type
      operator()(const Cell& cell) const
      {
        return cell.getDim();
      }
    };

    template<typename Complex>
    CellDim<typename ComplexTraits<Complex>::Cell>
    MakeCellDim()
    {
      return CellDim<typename ComplexTraits<Complex>::Cell>();
    }

    template<typename Complex>
    struct DimCells
    {
      explicit DimCells(Complex& complex):
        _complex(complex)
      {}

      typedef typename ComplexTraits<Complex>::Iterators::DimCells result_type;
      result_type operator()(typename Complex::Dim dim) const
      {
        return _complex.iterators().dimCells(dim);
      }

      Complex& _complex;
    };

    template<typename Complex>
    struct DimCellsReversed // We cannot iterate over a range[0..dim + 1) in reversed direction (boost::reverse_iterator) using DimCells when typename Complex::Dim == size_t, because it requires a range [dim, -1) and size_t is signed. So we need to use this functor
    {
      explicit DimCellsReversed(Complex& complex):
        _complex(complex)
      {}

      typedef typename ComplexTraits<Complex>::Iterators::DimCells result_type;
      result_type operator()(typename Complex::Dim dim) const
      {
        return _complex.iterators().dimCells(_complex.getDim() - dim);
      }

      Complex& _complex;
    };


    template<typename Complex>
    struct BdCells: std::unary_function<const typename ComplexTraits<Complex>::Cell&,
                                        typename ComplexTraits<Complex>::Iterators::BdCells>
    {
      explicit BdCells(Complex& complex): _complex(&complex){}

      BdCells(): _complex(0) {}

      typename ComplexTraits<Complex>::Iterators::BdCells
      operator()(const typename ComplexTraits<Complex>::Cell& cell) const
      {
        return _complex->iterators().bdCells(cell);
      }

    private:
      Complex* _complex;
    };

    template<typename Complex>
    struct CbdCells: std::unary_function<const typename ComplexTraits<Complex>::Cell&,
                                        typename ComplexTraits<Complex>::Iterators::CbdCells>
    {
      explicit CbdCells(Complex& complex): _complex(&complex){}

      CbdCells(): _complex(0) {}

      typename ComplexTraits<Complex>::Iterators::CbdCells
      operator()(const typename ComplexTraits<Complex>::Cell& cell) const
      {
        return _complex->iterators().cbdCells(cell);
      }

    private:
      Complex* _complex;
    };

    template<typename Complex>
    struct IdToCell: std::unary_function<const typename Complex::Id&,
                                         typename ComplexTraits<Complex>::Cell>
    {
      explicit IdToCell(Complex& complex): _complex(&complex){}

      IdToCell(): _complex(0) {}

      typename ComplexTraits<Complex>::Cell
      operator()(const typename Complex::Id& id) const
      {
        return (*_complex)[id];
      }

    private:
      Complex* _complex;
    };

    template<typename Complex>
    IdToCell<Complex> MakeIdToCell(Complex& complex)
    {
      return IdToCell<Complex>(complex);
    }

  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_COMPLEXFUNCTORS_H
