/////////////////////////////////////////////////////////////////////////////
/// @file ComplexSelector
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-02-26
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_COMPLEXSELECTOR_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_COMPLEXSELECTOR_H

// #include
#include "Tags.h"
#include "CubCellComplex.h"
#include "CubSComplex.h"
#include "SimplexSComplex.hpp"
#include "SComplex.h"
#include "SComplexDefaultTraits.h"

#include <boost/mpl/back_inserter.hpp>
#include <boost/mpl/push_back.hpp>
#include <boost/mpl/fold.hpp>
#include <boost/mpl/range_c.hpp>
#include <boost/mpl/vector.hpp>
#include <boost/mpl/joint_view.hpp>
#include <boost/mpl/copy.hpp>
#include <boost/mpl/front.hpp>


#include <stdexcept>
#include <boost/mpl/deref.hpp>
#include <boost/type_traits/is_same.hpp>

namespace capd
{
  namespace complex
  {

    template<typename TagT, unsigned _DIM, typename ColorT>
    struct ComplexSelectorBase
    {
      typedef TagT Tag;
      static const unsigned DIM = _DIM;
      static const unsigned TYPE_ID = Tag::ID * 1000 + DIM;
    };

    template<typename Tag, unsigned DIM, typename Color>
    struct ComplexSelector;


    template<unsigned DIM, typename Color>
    struct ComplexSelector<DenseCubicalTag, DIM, Color>: ComplexSelectorBase<DenseCubicalTag, DIM, Color>
    {
      typedef CubCellComplex<capd::complex::DenseCubMemoryModel<DIM, Color> > type;
    };


    template<unsigned DIM>
    struct ComplexSelector<OldCubicalTag, DIM, size_t>: ComplexSelectorBase<OldCubicalTag, DIM, size_t>
    {
      typedef CubSComplex<DIM> type;
    };


//     template<unsigned DIM, typename Color>
//     struct ComplexSelector<SparseCubicalTag, DIM, Color>: ComplexSelectorBase<SparseCubicalTag, DIM, Color>
//     {
// //      typedef CubCellComplex<capd::complex::DenseCubMemoryModel<DIM, Color> > type;
//     };

    template<unsigned DIM, typename Color>
    struct ComplexSelector<SimplicialTag, DIM, Color>: ComplexSelectorBase<SimplicialTag, DIM, Color>
    {
      typedef SimplexSComplex type;
    };

    template<unsigned DIM, typename Color>
    struct ComplexSelector<GeneralizedTag, DIM, Color>: ComplexSelectorBase<GeneralizedTag, DIM, Color>
    {
      typedef SComplex<capd::complex::SComplexDefaultTraits> type;
    };


    struct SimplicialComplexes:
      boost::mpl::vector<
      ComplexSelector<SimplicialTag, 0, size_t>
      >
    {
      static size_t id()
      {
        return boost::mpl::front<type>::type::TYPE_ID;
      }
    };

    struct GeneralizedComplexes:
      boost::mpl::vector<
      ComplexSelector<GeneralizedTag, 0, size_t>
      >
    {
      static size_t id()
      {
        return boost::mpl::front<type>::type::TYPE_ID;
      }
    };


    template<typename Tag, typename DIM, typename Color>
    struct ApplyComplexSelector
    {
      typedef ComplexSelector<Tag, DIM::value, Color> type;
    };

    template<int MIN_DIM, int MAX_DIM>
    struct _DenseCubicalComplexes:
      boost::mpl::fold<boost::mpl::range_c<int, MIN_DIM, MAX_DIM>,
                       boost::mpl::vector<>,
                       boost::mpl::push_back<boost::mpl::_1,
                                             ApplyComplexSelector<DenseCubicalTag, boost::mpl::_2, size_t> >
                       >::type
    {};

    template<int MIN_DIM, int MAX_DIM>
    struct _OldCubicalComplexes:
      boost::mpl::fold<boost::mpl::range_c<int, MIN_DIM, MAX_DIM>,
                       boost::mpl::vector<>,
                       boost::mpl::push_back<boost::mpl::_1,
                                             ApplyComplexSelector<OldCubicalTag, boost::mpl::_2, size_t> >
                       >::type
    {};

    typedef _DenseCubicalComplexes<2, 4> DenseCubicalComplexesLow;
    typedef _DenseCubicalComplexes<4, 7> DenseCubicalComplexesMedium;
    typedef _DenseCubicalComplexes<7, 11> DenseCubicalComplexesHigh;

    typedef _OldCubicalComplexes<2, 4> OldCubicalComplexesLow;
    typedef _OldCubicalComplexes<4, 7> OldCubicalComplexesMedium;
    typedef _OldCubicalComplexes<7, 11> OldCubicalComplexesHigh;

    template<typename Seq>
    struct FlatMerge:
      boost::mpl::fold<Seq,
                       boost::mpl::vector<>,
                       boost::mpl::copy<boost::mpl::_2, boost::mpl::back_inserter<boost::mpl::_1> >
                       >::type
    {};

    typedef FlatMerge<
      boost::mpl::vector<DenseCubicalComplexesLow, OldCubicalComplexesLow>
      > CubicalComplexesLow;

    typedef FlatMerge<
      boost::mpl::vector<DenseCubicalComplexesMedium, OldCubicalComplexesMedium>
      > CubicalComplexesMedium;

    typedef FlatMerge<
      boost::mpl::vector<DenseCubicalComplexesHigh, OldCubicalComplexesHigh>
      > CubicalComplexesHigh;


    struct DenseCubicalComplexes: FlatMerge<
      boost::mpl::vector<DenseCubicalComplexesLow, DenseCubicalComplexesMedium, DenseCubicalComplexesHigh>
      >
    {
      static size_t id(size_t dim)
      {
        BOOST_ASSERT(dim >= 2);
        return boost::mpl::front<type>::type::TYPE_ID + dim - 2;
      }
    };


    struct OldCubicalComplexes: FlatMerge<
      boost::mpl::vector<OldCubicalComplexesLow, OldCubicalComplexesMedium, OldCubicalComplexesHigh>
      >
    {
      static size_t id(size_t dim)
      {
        BOOST_ASSERT(dim >= 2);
        return boost::mpl::front<type>::type::TYPE_ID + dim - 2;
      }
    };


    struct CubicalComplexes: FlatMerge<boost::mpl::vector<
                                         DenseCubicalComplexes,
                                         OldCubicalComplexes
                                         > >
    {
      typedef int Type;
      static const Type OLD = OldCubicalTag::ID;
      static const Type DENSE = DenseCubicalTag::ID;
      static const Type SPARSE = SparseCubicalTag::ID;

      static size_t id(size_t dim, Type type)
      {
        if (type == OLD) {
          return OldCubicalComplexes::id(dim);
        } else if (type == DENSE) {
          return DenseCubicalComplexes::id(dim);
        } else {
          BOOST_ASSERT_MSG(false, "Wrong cubical complex type");
        }
        return 0;
      }

    };

    struct Complexes: FlatMerge<boost::mpl::vector<CubicalComplexes, SimplicialComplexes, GeneralizedComplexes> >
    {};


    template<typename Complex>
    struct SelectorForComplex:
      boost::mpl::deref<
      typename boost::mpl::find_if<Complexes, boost::is_same<
                                                Complex,
                                                boost::mpl::deref<boost::mpl::_1>
                                                >
                                   >::type>

    {};

  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_COMPLEXSELECTOR_H
