/////////////////////////////////////////////////////////////////////////////
/// @file ComputeHomologyOrchestrator
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-07-23
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////


#include <capd/complex/ComputeHomologyOrchestrator.h>


#if !defined(CAPD_FILE_CAPDREDHOM_COMPLEX_COMPUTEHOMOLOGYORCHESTRATOR_HPP) && !defined(USE_PRECOMPILED_CAPDREDHOM_COMPLEX_COMPUTEHOMOLOGYORCHESTRATOR_HPP)
#define CAPD_FILE_CAPDREDHOM_COMPLEX_COMPUTEHOMOLOGYORCHESTRATOR_HPP

#include "Coreduction.h"
#include "Shave.h"
#include "SComplex.h"
#include "SComplexDefaultTraits.h"

#include "AKQStrategy.hpp"

#include <capd/auxil/Logger.h>
#include <capd/complex/ComputeHomologySignature.hpp>
#include <capd/complex/SqueezedComplex.h>

namespace capd
{
  namespace complex
  {

    //ComputeHomologyOrchestrator::ComputeHomologyOrchestrator() {}

    //ComputeHomologyOrchestrator::~ComputeHomologyOrchestrator() {}

    template<typename ComplexT, typename ScalarT, typename IntT>
      std::vector<IntT> ComputeHomologyOrchestrator<ComplexT, ScalarT, IntT>::operator()()
    {
      Complex& complex = _complex;
      size_t coreductionsExtracted = 0;

      if (_reductions) {
        CAPD_INFO_TMPL("Starting shave algorithm");
        boost::shared_ptr<typename ShaveAlgorithmFactory<Complex>::DefaultAlgorithm> algo =
          ShaveAlgorithmFactory<Complex>::createDefault(complex);
        const size_t reduced = (*algo)();
        CAPD_INFO_TMPL("Shaved: " << reduced);
      }

      if (_coreductions) {
        CAPD_INFO_TMPL("Starting coreduction algorithm");
        boost::shared_ptr<typename CoreductionFactory<Complex, int, int>::DefaultAlgorithm> algo =
          CoreductionFactory<Complex, int, int>::createDefault(complex);
        const size_t reduced = (*algo)();
        coreductionsExtracted += algo->getExtractedSignature().bettiNumber(0);

        CAPD_INFO_TMPL("Coreduced: " << reduced);
      }

      if (_morse) {
        CAPD_INFO_TMPL("Starting morse algorithm");
        const size_t size = complex.size();
        const size_t cardinality = complex.cardinality();
        const double density = (.0 + cardinality)/size;
        CAPD_INFO_TMPL("Complex size: " << size << " cardinality: " << cardinality << " density: " << density << " squeezeFactor: " << _squeezeFactor);

        typedef complex::SComplex<complex::SComplexDefaultTraits> OutputComplex;

        if (density <= _squeezeFactor) {
          CAPD_INFO_TMPL("Squeezing complex");
          typedef SqueezedComplex<Complex> Squeezed;;
          Squeezed squeezed(complex);

          CAPD_INFO_TMPL("Running morse on squeezed complex, size: " << squeezed.size());
          typedef AKQReduceStrategy<Squeezed, OutputComplex, Scalar> AKQ;
          capd::complex::Coreduction<AKQ, Scalar, Int> algorithm(new AKQ(squeezed));
          algorithm();

          ComputeHomologySignature<OutputComplex, Scalar, Int> computeSignature(*algorithm.getStrategy()->getOutputComplex(), _useAlgebraicReductions);
          _signature = computeSignature();

        } else {
          typedef AKQReduceStrategy<Complex, OutputComplex, Scalar> AKQ;
          capd::complex::Coreduction<AKQ, Scalar, Int> algorithm(new AKQ(complex));
          algorithm();

          ComputeHomologySignature<OutputComplex, Scalar, Int> computeSignature(*algorithm.getStrategy()->getOutputComplex(), _useAlgebraicReductions);
          _signature = computeSignature();
        }
      } else {
        ComputeHomologySignature<Complex, Scalar, Int> computeSignature(complex, _useAlgebraicReductions);
        _signature = computeSignature();
      }

      if (_coreductions) {
        if (coreductionsExtracted > 0) {
          std::vector<IntT> extractedBetti(1);
          extractedBetti[0] = coreductionsExtracted;
          _signature += HomologySignature<ScalarT, IntT>(extractedBetti);
        }
      }

      _signature.shrink();

      return _signature.ranks();
    }

  }
}


#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_COMPUTEHOMOLOGYORCHESTRATOR_HPP
