/////////////////////////////////////////////////////////////////////////////
/// @file ComputeHomologySignature
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-09-20
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#include "ComputeHomologySignature.h"

#if !defined(CAPD_FILE_CAPDREDHOM_COMPLEX_COMPUTEHOMOLOGYSIGNATURE_HPP) && !defined(USE_PRECOMPILED_CAPDREDHOM_COMPLEX_COMPUTEHOMOLOGYSIGNATURE_HPP)

#define CAPD_FILE_CAPDREDHOM_COMPLEX_COMPUTEHOMOLOGYSIGNATURE_HPP

#include "SComplexToFreeChainComplex.hpp"
#include "BuildFiltration.hpp"

#include <capd/vectalg/Matrix.hpp>
#include <capd/homAlgebra/homAlgFunctors.hpp>
#include <capd/auxil/CRef.h>
#include <capd/auxil/Logger.h>
#include <capd/rings/Z2.h>
#include <capd/persistence/PHATInterface.h>

#include <boost/mpl/bool.hpp>
namespace capd
{
  namespace complex
  {

    //ComputeHomologySignature::ComputeHomologySignature() {}

    //ComputeHomologySignature::~ComputeHomologySignature() {}

    namespace
    {
      template<typename Scalar, typename Int, typename Complex>
      HomologySignature<Scalar, Int> ComputeHomologySignatureViaFreeChainComplex(Complex& _complex, bool useAlgebraicReductions)
      {
        if (useAlgebraicReductions) {
          typedef SComplexToFreeChainComplex<Complex, Scalar, Int, boost::mpl::true_> RFCBuilder;
          RFCBuilder fccBuilder(_complex);
          CRef<typename RFCBuilder::FreeChainComplexT> fccCR = fccBuilder();
          CAPD_INFO_TMPL("Compute homology signature for: " << fccCR().size() << " cells using algebraic reductions and Smith Normal Form");
          CRef<HomologySignature<Scalar, Int> > signature = HomAlgFunctors<typename RFCBuilder::FreeModuleT>::homSignViaAR_Random(fccCR);
          (*signature).shrink();

          return signature();
        } else {
          typedef SComplexToFreeChainComplex<Complex, Scalar, Int, boost::mpl::false_> RFCBuilder;
          RFCBuilder fccBuilder(_complex);
          CRef<typename RFCBuilder::FreeChainComplexT> fccCR = fccBuilder();
          CAPD_INFO("Compute homology signature  using Smith Normal Form (withou algebraic reductions!)");
          CRef<HomologySignature<Scalar, Int> > signature=HomAlgFunctors<typename RFCBuilder::FreeModuleT>::homSignViaSmith(fccCR);
          (*signature).shrink();

          return signature();
        }
      }

    }

    template<typename ComplexT, typename ScalarT, typename IntT>
    HomologySignature<ScalarT, IntT> ComputeHomologySignature<ComplexT, ScalarT, IntT>::operator()()
    {
      CAPD_INFO_TMPL("Computing homology signature, complex size: " << _complex.size() << " cardinality: " << _complex.cardinality());
      return ComputeHomologySignatureViaFreeChainComplex<Scalar, Int>(_complex, _useAlgebraicReductions);
    }

    template<typename ComplexT, typename IntT>
    HomologySignature<Z2, IntT> ComputeHomologySignature<ComplexT, Z2, IntT>::operator()()
    {
      CAPD_INFO_TMPL("Computing homology signature, complex size: " << _complex.size() << " cardinality: " << _complex.cardinality());

      using capd::persistence::PHATInterface;

      if (PHATInterface::enabled()) {
        CAPD_DEBUG_TMPL("Homology signature over Z2 with PHAT");
        CAPD_INFO_TMPL("Building PHAT filtration");
        BuildFiltration<Complex> buildFiltration(_complex);
        buildFiltration();
        typename BuildFiltration<Complex>::Filtration filtration = buildFiltration.filtration();

        PHATInterface phat;
        PHATInterface::Pairs pairs;
        PHATInterface::Cycles cycles;
        std::vector<PHATInterface::FiltrationElement> phatFiltration(filtration.size());

        //TODO convert on the fly with transform iterator ?
        for (size_t idx = 0; idx < filtration.size(); ++idx) {
          phatFiltration[idx].first = filtration[idx].first;
          phatFiltration[idx].second.assign(filtration[idx].second.begin(), filtration[idx].second.end());
        }

        CAPD_INFO_TMPL("Computing homology using PHAT...");
        phat(PHATInterface::Filtration(phatFiltration.begin(), phatFiltration.end()), phatFiltration.size(), pairs, cycles);
        CAPD_INFO_TMPL("...done");

        std::vector<IntT> betti(_complex.getDim() + 1);
        for (size_t idx = 0; idx < cycles.size(); ++idx) {
          const typename Complex::Dim dim = buildFiltration[cycles[idx]].getDim();
          ++betti[dim];
        }

        HomologySignature<Scalar, IntT> signature(betti);
        signature.shrink();
        return signature;
      } else {
        CAPD_WARN_TMPL("Computing homology signature over Z2 with standard Smith Normal Form. Reconfigure with '--with-phat=<paht to PHAT>' to use faster implementation");

        return ComputeHomologySignatureViaFreeChainComplex<Scalar, Int>(_complex, _useAlgebraicReductions);
      }
    }

  }
}


#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_COMPUTEHOMOLOGYSIGNATURE_HPP
