/////////////////////////////////////////////////////////////////////////////
/// @file Coreduction.h
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2013-12-04
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2013 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.

#ifndef _CAPD_REDHOM_COMPLEX_FILE_COREDUCTION_H
#define _CAPD_REDHOM_COMPLEX_FILE_COREDUCTION_H

#include "SReductionPair.h"
#include "DefaultReduceStrategy.hpp"
#include <capd/homAlgebra/HomologySignature.h>
#include <capd/complex/LoggerHelpers.h>

#include <deque>

#include <boost/shared_ptr.hpp>
#include <boost/type_traits.hpp>
#include <boost/utility/enable_if.hpp>
#include <boost/type_traits/is_same.hpp>
#include <boost/assign/list_inserter.hpp>
#include <capd/auxil/Logger.h>

namespace capd
{
  namespace complex
  {

    template<typename StrategyT, typename ScalarT, typename IntT, typename ReductionCallback = void*>
    class Coreduction {

    public:
      typedef StrategyT Strategy;
      typedef ScalarT Scalar;

      typedef typename Strategy::SComplex SComplex;
      typedef typename Strategy::Cell Cell;

      explicit Coreduction(Strategy* _strategy, ReductionCallback reductionCallback = ReductionCallback()):
        strategy(_strategy), _reductionCallback(reductionCallback)
      {}

      ~Coreduction() {
        if (strategy)
          delete strategy;
      }

      size_t operator()()
      {
        CAPD_DEBUG_TMPL("Starting coreductions...");
        cellIdsToProcess.resize(strategy->getComplex().size());

        size_t reduced=0;

        for(;;){
          if (coreduceNextPair()) {
            reduced += 2;
          } else {
            //CAPD_DEBUG_TMPL("Extracting ...");
            typename StrategyT::Traits::Extract::result_type sourceFace = strategy->extract();

            if(sourceFace){
              CAPD_TRACE_TMPL("Extracting: "  << " dim: " << (int) sourceFace->getDim() << " code: " << sourceFace->code() << " color: " << (size_t) sourceFace->getColor());

              if (extractedSignature.size() < (size_t) sourceFace->getDim() + 1) {
                extractedSignature.resize(sourceFace->getDim() + 1);
              }
              extractedSignature[sourceFace->getDim()] += IntT(1);

              storeGenerator(*sourceFace);
              addCellsToProcess(*sourceFace);

              strategy->reduce(*sourceFace);
              ++reduced;
              //TODO we should mark the complex as not closed, so second corredoction won't fail.
              // Still smth is wrong :/ MJ
              //strategy->getComplex().setClosed(false);
            }else{
              //CAPD_DEBUG_TMPL("Nothing to extract");
              break; // no base face left: quit any further processing
            }
          }
        }

        return reduced; // the number of cells removed
      }

      Strategy* getStrategy() { return strategy; }

      const HomologySignature<Scalar, IntT>& getExtractedSignature() const {
        return extractedSignature;
      }

    private:

      template<typename ImplT1, typename ImplT2, typename Callback, typename Enable=void>
      struct CallReductionCallback
      {
CallReductionCallback(Strategy& /*startegy*/, Callback& /*callback*/)
        {}

void operator()(const ImplT1& /*a*/)
        {}

void operator()(const ImplT1& /*a*/, const ImplT2& /*b*/)
        {}
      };

      template<typename ImplT1, typename ImplT2, typename Callback>
      struct CallReductionCallback<ImplT1, ImplT2, Callback, typename boost::disable_if<boost::is_same<Callback, void*> >::type>
      {
        CallReductionCallback(Strategy& strategy, Callback& callback):
          _callback(callback),
          _strategy(strategy)
        {}


        void operator()(const ImplT1& a)
        {
          _callback(makeSReductionPair(_strategy.getComplex(), a));
        }

        void operator()(const ImplT1& a, const ImplT2& b)
        {
          _callback(makeSReductionPair(_strategy.getComplex(), a, b));
        }

      private:
        Callback& _callback;
        Strategy& _strategy;
      };

      template<typename ImplT>
      void storeGenerator(const ImplT& a)
      {
        CallReductionCallback<ImplT, ImplT, ReductionCallback> call(*strategy, _reductionCallback);
        call(a);
      }

      template<typename ImplT1, typename ImplT2>
      void storeCoreductionPair(const ImplT1& a, const ImplT2& b)
      {
        CallReductionCallback<ImplT1, ImplT2, ReductionCallback> call(*strategy, _reductionCallback);
        call(a, b);
      }


      bool coreduceNextPair()
      {
        while(true) {
          while (! cellsToProcess.empty() ) {
            Cell* cell = &cellsToProcess.front();
            CAPD_TRACE_TMPL("Processing: " << " dim: " << (int) cell->getDim() << " code: " << cell->code() << " color: " << (size_t) cell->getColor());

            if (! strategy->reduced(*cell)) {
              typename StrategyT::Traits::template GetCoreductionPair<Cell>::result_type coreductionPair = strategy->getCoreductionPair(*cell);

              if (coreductionPair) {
                //CAPD_TRACE_TMPL("Found pair: " << " " << coreductionPair->code() << " " << (int)coreductionPair->getColor());
                doCoreduction(*coreductionPair, *cell);
                cellIdsToProcess[cell->getId()] = false;
                cellsToProcess.pop_front();
                return true;
              } else {
                addCellsToProcess(*cell);
                //CAPD_TRACE_TMPL("No pair");
              }
            }
            cellIdsToProcess[cell->getId()] = false;
            cellsToProcess.pop_front();
          }

          const bool force = strategy->forceCoreductionPair(boost::assign::push_back(cellsToProcess));
          if (!force) {
            return false;
          }
        }
      }

      template<typename ImplT>
      void addCellsToProcess(const ImplT& sourceFace) {
        typename SComplex::ColoredIterators::CbdCells cbdCells = strategy->getComplex().iterators(1).cbdCells(sourceFace);
        for (typename SComplex::ColoredIterators::CbdCells::iterator cbdn = cbdCells.begin(),
               end = cbdCells.end(); cbdn != end; ++cbdn) {
          typename SComplex::ColoredIterators::CbdCells::iterator::value_type v = *cbdn;

          if (!cellIdsToProcess[v.getId()]) {
            cellsToProcess.push_back(v);
            cellIdsToProcess[v.getId()] = true;
            CAPD_TRACE_TMPL("Adding: " <<  " dim: " << (int) v.getDim() << " code: " << v.code() << " color: " << (size_t) v.getColor());
          }
        }
      }


      template<typename ImplT1, typename ImplT2>
      void doCoreduction(ImplT1& a, ImplT2& b) {
        storeCoreductionPair(a, b);
        strategy->coreduce(a, b);
        addCellsToProcess(a);
      }

      Strategy* strategy;
      bool storeReducedCells;
      std::deque<Cell> cellsToProcess;
      std::vector<bool> cellIdsToProcess;
      HomologySignature<Scalar, IntT> extractedSignature;
      ReductionCallback _reductionCallback;
      CAPD_CLASS_LOGGER;
    };

    template<typename Scalar, typename IntT, typename Complex, typename ReductionCallback>
    Coreduction<DefaultReduceStrategy<Complex>, Scalar, IntT, ReductionCallback>
    makeCoreductionDefault(Complex& complex, ReductionCallback callback)
    {
      return Coreduction<DefaultReduceStrategy<Complex>, Scalar, IntT, ReductionCallback>
        (new DefaultReduceStrategy<Complex>(complex), callback);
    }

    template<typename Scalar, typename IntT, typename Complex>
    Coreduction<DefaultReduceStrategy<Complex>, Scalar, IntT>
    makeCoreductionDefault(Complex& complex)
    {
      return Coreduction<DefaultReduceStrategy<Complex>, Scalar, IntT>
        (new DefaultReduceStrategy<Complex>(complex));
    }

    template<typename Complex, typename Scalar, typename IntT>
    class CoreductionFactory {

    public:
      typedef Coreduction<DefaultReduceStrategy<Complex>, Scalar, IntT> DefaultAlgorithm;

      static boost::shared_ptr<DefaultAlgorithm>  createDefault(Complex& s) {
        return boost::shared_ptr<DefaultAlgorithm>(new DefaultAlgorithm(new DefaultReduceStrategy<Complex>(s)));
      }

    };
  }
}


#endif // CAPD_FILE_COREDUCTION_H
