/////////////////////////////////////////////////////////////////////////////
/// @file CoreductionPairInclusionChainModel.h
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2013-12-31
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2013 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.

#ifndef CAPD_FILE_COREDUCTIONPAIRINCLUSIONCHAINMODEL_H
#define CAPD_FILE_COREDUCTIONPAIRINCLUSIONCHAINMODEL_H

#include "Chain.h"
#include "SReductionPair.h"

#include <map>
#include <utility>

namespace capd
{
namespace complex
{

  template<typename Complex, typename Scalar, typename CoreductionPairs>
  class CoreductionPairInclusionChainModel
  {

  public:
    typedef void result_type;
    typedef typename Complex::Id Id;
    typedef typename Complex::Cell Cell;
    typedef typename Complex::ConstCell ConstCell;
    typedef capd::complex::Chain<Complex, Scalar> Chain;
    typedef SReductionPair<Complex> ReductionPair;
    typedef std::map<ReductionPair, Scalar> CoefInvs;

    explicit CoreductionPairInclusionChainModel(const Complex& complex,
						const CoreductionPairs& pairs):
      _complex(complex), _pairs(pairs)
    {}

    void operator()(Chain& chain)
    {
      Chain chainBd = boundary(chain);

      typedef typename Chain::value_type ChV;
      BOOST_REVERSE_FOREACH(const ReductionPair& pair, _pairs) {
	const Scalar coefInChainBd = chainBd[pair.face];

	if (coefInChainBd != Scalar(0)) {
	  const Scalar lambda = coefInv(pair);
	  chain += typename Chain::value_type(pair.coface, (- coefInChainBd) * lambda);

	  Chain cofaceBd = boundary(Chain(pair.coface, pair.dim));
	  cofaceBd *= (- coefInChainBd * lambda);
	  chainBd += cofaceBd;
	}
      }

      chain.compact();
    }

  private:
    const Scalar& coefInv(const ReductionPair& pair)
    {
      typename CoefInvs::iterator it = _coefInvs.find(pair);
      if (it == _coefInvs.end()) {
        it = _coefInvs.insert(typename CoefInvs::value_type(pair, _complex.coincidenceIndex(_complex[pair.coface], _complex[pair.face]))).first;
      }

      BOOST_ASSERT(it->second == Scalar(1) || it->second == Scalar(-1));
      return it->second;
    }

    Chain boundary(const Chain& chain) const
    {
      Chain result;
      typedef typename Chain::value_type ChV;
      BOOST_FOREACH(const ChV& chainElem, chain) {
	ConstCell cell = _complex[chainElem.first];
	typename Complex::Iterators::BdCells bdCells = const_cast<Complex&>(_complex).iterators().bdCells(cell);

	BOOST_FOREACH(const Cell bdCell, bdCells) {
	  result += ChV(bdCell.getId(), chainElem.second * _complex.coincidenceIndex(cell, bdCell));
	}
      }

      result.compact();
      return result;
    }

    const Complex& _complex;
    const CoreductionPairs& _pairs;
    CoefInvs _coefInvs;
  };

  template<typename Scalar, typename Complex, typename CoreductionPairs>
  CoreductionPairInclusionChainModel<Complex, Scalar, CoreductionPairs>
  makeCoreductionPairInclusionChainModel(const Complex& complex, const CoreductionPairs& pairs)
  {
    return CoreductionPairInclusionChainModel<Complex, Scalar, CoreductionPairs>(complex, pairs);
  }

}
}

#endif // CAPD_FILE_COREDUCTIONPAIRINCLUSIONCHAINMODEL_H
