/////////////////////////////////////////////////////////////////////////////
/// @file CubCellComplex.h
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-09-16
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_CUBCELLCOMPLEX_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_CUBCELLCOMPLEX_H

// #include
#include "Tags.h"
#include "CellComplex.h"
#include "CubicTessellation.h"

#include <vector>

#include <boost/assign.hpp>
#include <boost/mpl/vector.hpp>

namespace capd
{
  namespace complex
  {



    template<uint8_t DIM, typename ColorT>
    struct DenseCubMemoryModel
    {
      typedef CubicTessellation<DIM> EnumerateCellCodes;
      typedef DenseCubicalTag Tag;

      typedef size_t Id;
      typedef ColorT Color;
      typedef uint8_t Dim;

      typedef typename EnumerateCellCodes::CellCode CellCode;

      template<typename Value>
      struct ValueById
      {
        typedef std::vector<Value> type;

        static type Create(size_t size)
        {
          return type(size);
        }
      };

    };

    template<uint8_t DIM>
    struct FilteredDenseCubMemoryModel: DenseCubMemoryModel<DIM, float>
    {

    };


    template<typename MemoryModel>
    class CubCellComplex: public CellComplex<MemoryModel>
    {

    public:
      typedef boost::mpl::vector<typename MemoryModel::Tag, CubicalTag, RegularCubicalTag> Tags;

      typedef CellComplex<MemoryModel> BaseCellComplex;
      typedef typename BaseCellComplex::Cell Cell;
      typedef typename BaseCellComplex::ConstCell ConstCell;
      typedef typename BaseCellComplex::CellCode CellCode;
      typedef typename BaseCellComplex::EnumerateCellCodes::FullCube FullCube;
      typedef typename BaseCellComplex::Dim Dim;

      explicit CubCellComplex(const std::vector<size_t>& lengths):
        BaseCellComplex(new typename MemoryModel::EnumerateCellCodes(lengths))
      {}

      virtual ~CubCellComplex() {}


      FullCube fullCube()
      {
        return this->enumerateCellCodes().fullCube();
      }

      CellCode operator()(const FullCube& fullCube)
      {
        BOOST_ASSERT(this->dim() == fullCube.size());
        CellCode cellCode = this->cellCode();

        for (typename CellCode::size_type idx = 0; idx < this->dim(); ++idx) {
          cellCode[idx].first = fullCube[idx];
          cellCode[idx].second = true;
        }

        return cellCode;
      }

      CellCode operator()(size_t x, size_t y)
      {
        std::vector<size_t> cube = boost::assign::list_of(x)(y);
        return (*this)(cube);
      }

      CellCode operator()(size_t x, size_t y, size_t z)
      {
        std::vector<size_t> cube = boost::assign::list_of(x)(y)(z);
        return (*this)(cube);
      }

      int coincidenceIndex(const ConstCell& cellA, const ConstCell& cellB) const {
        const CellCode& a = cellA.code();
        const CellCode& b = cellB.code();

        int res = 0;
        int sgn = 1;

        for (Dim i = 0, end = this->dim(); i < end; ++i) {
          if (! (a[i].first == b[i].first || a[i].first + (a[i].second) == b[i].first)) {
            return 0; // b[i] left side doesn't intersect a[i] interval
          }

          if (a[i].second == 1 && b[i].second == 1) {
            sgn = -sgn; //both nondegenerated, go to next
          } else if (a[i].second == 0 && b[i].second == 0) {
            // both degenerated
          } else if (b[i].second == 1) {
            return 0; // a[i] is inside b[i]
          } else { // b[i] is inside a[i]
            if (res != 0) {
              return 0; // second time, so not proper face
            }
            res = sgn;
            if (a[i].first != b[i].first) { // b[i] is the right face of a[i]
              res = -res;
            }
          }
        }
        return res;
      }

    private:

    };

    template<uint8_t DIM, typename ColorT>
    struct DenseCubCellComplex
    {
      typedef CubCellComplex<DenseCubMemoryModel<DIM, ColorT> > type;
    };

  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_CUBCELLCOMPLEX_H
