/////////////////////////////////////////////////////////////////////////////
/// @file CubSComplex
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-01-17
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_CUBSCOMPLEX_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_CUBSCOMPLEX_H

#include "Tags.h"

#include <boost/assert.hpp>
// there is a bug in any_range.hpp, we need to include it
#include <boost/range/concepts.hpp>
#include <boost/utility.hpp>

#include <boost/range/any_range.hpp>
#include <boost/bind.hpp>
#include <boost/optional.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/mpl/vector.hpp>
#include <utility>



#include "SComplexBase.h"

#include <capd/auxil/Stopwatch.h>
#include <capd/auxil/embeddingDim.h>
#include <capd/auxil/CRef.h>

#include <capd/repSet/RepSet.h>
#include <capd/repSet/ElementaryCube.h>
#include <capd/bitSet/BitSetT.h>
#include <capd/bitSet/EuclBitSetT.h>
#include <capd/cubSet/CubCellSetT.h>



namespace capd
{
    namespace complex
    {

      class CubSComplexBase: public SComplexBase
      {
      public:
        explicit CubSComplexBase(int dim): _dim(dim)
        {}

        virtual ~CubSComplexBase(){}
        int dim() const { return _dim; } //not virtual for inlining

      private:
        int _dim;
      };


template<int DIM>
class CubSComplex: public CubSComplexBase
{
  class CellImpl;

  typedef BitSetT<BitmapT<unsigned long int> > BitSet;
  typedef EuclBitSetT<BitSet, DIM> EuclBitSet;

public:
typedef boost::mpl::vector<CubicalTag, RegularCubicalTag, OldCubicalTag> Tags;

  typedef CubCellSetT<EuclBitSet> BCubCellSet;

  class BitCoordCellImpl;
  class DynamicBitCoordCellImpl;
  class BitCoordPtrCellImpl;

  template<typename, typename Enable=void>
  class CellProxy;

  typedef CellProxy<BitCoordCellImpl> Cell;
  typedef CellProxy<BitCoordCellImpl> ConstCell;
  typedef CellProxy<BitCoordPtrCellImpl> CellRef;
  typedef CellProxy<DynamicBitCoordCellImpl> DynamicCell;

  typedef int Dim;
  typedef int Color;
  static const int ColorValues = 1;
  //typedef std::pair<const BCubCellSet::BitCoordIterator::WordIterator, int> Id;
  typedef long Id;

private:

  template<typename NumeratorT, bool isConst>
  class IteratorProvider;


  class BasicNumerator;
  class CellNumerator;
  class CellDimNumerator;
  class CbdNumerator;
  class BdNumerator;

  template<Color color>
  struct FixedColorChecker {
    bool operator()(bool b) const {
      return (color == 1 && b) || (color == 2 && !b);
    }
  };

  struct ColorChecker {
    Color color;

    ColorChecker(Color _color): color(_color) {}

    bool operator()(bool b) const {
      return (color == 1 && b) || (color == 2 && !b);
    }
  };

  struct NoBitChecker {
    bool operator()(bool /*b*/) const {
      return true;
    }
  };

  template<bool isConst, typename BitChecker>
  class IteratorsImpl;

  template<bool isConst, typename BitChecker>
  friend class IteratorsImpl;

public:
  typedef size_t Coordinate;
  typedef std::pair<Coordinate, bool> Interval;
  typedef std::vector<Interval> CellCode;
  typedef std::vector<Coordinate> FullCube;

  typedef boost::any_range<CellCode, boost::forward_traversal_tag, CellCode&, std::ptrdiff_t> CellCodeRange;

  typedef IteratorsImpl<false, NoBitChecker> Iterators;
  typedef IteratorsImpl<true, NoBitChecker> ConstIterators;


  typedef IteratorsImpl<false, ColorChecker> ColoredIterators;
  typedef IteratorsImpl<true, ColorChecker> ColoredConstIterators;

  explicit CubSComplex(const std::vector<size_t>& size,
                       const CellCodeRange& cubes = CellCodeRange());

  explicit CubSComplex(RepSet<ElementaryCube>& repSet, bool addEmptyCollar = true);
  explicit CubSComplex(const CRef<BCubCellSet>& __bCubCellSet, bool addEmptyCollar = true);


  size_t size() const { return const_cast<BCubCellSet&>(bCubCellSet).getBmpSizeInBits(); }

  size_t cardinality() const {
    typedef typename ColoredConstIterators::AllCells AllCells;
    AllCells allCells = this->iterators(1).allCells();
    return std::distance(allCells.begin(), allCells.end());
  }


  Dim getDim() const { return DIM; }

  void insert(const CellCode& cube);

  CellCode operator()(const FullCube& fullCube)
  {
    BOOST_ASSERT(this->getDim() == (Dim) fullCube.size());
    CellCode cellCode(this->getDim());

    for (Dim idx = 0; idx < this->getDim(); ++idx) {
      cellCode[idx].first = fullCube[idx];
      cellCode[idx].second = true;
    }

    return cellCode;
  }

  void fillWithBoundaries()
  {
    for (Dim i = 0; i < DIM; ++i) {
      bCubCellSet.fillWithBoundaries();
    }
  }

  ConstIterators iterators() const;
  Iterators iterators();

  ColoredConstIterators iterators(const Color& color) const;
  ColoredIterators iterators(const Color& color);

  template<typename ImplT>
  bool getUniqueFace(const CellProxy<ImplT>& cell, DynamicCell& coface) const;

  template<typename ImplT>
  bool getUniqueCoFace(const CellProxy<ImplT>& cell, DynamicCell& coface) const;

  template<typename ImplT1, typename ImplT2>
  int coincidenceIndex(const CellProxy<ImplT1> &a, const CellProxy<ImplT2> &b) const;

  ConstCell operator[](const CellCode& cube) const
  {
    int realCoord[DIM];
    bitmapCoord(cube, realCoord);

    typename BCubCellSet::BitCoordIterator bitIt(bCubCellSet, realCoord);
    BOOST_ASSERT(bitIt.inSpace());

    return BitCoordCellImpl(bitIt);
  }

  Cell operator[](const CellCode& cube)
  {
    return ((const CubSComplex*)this)->operator[](cube);
  }

  ConstCell operator[](Id id) const
  {
    typename BCubCellSet::BitCoordIterator bitIt(bCubCellSet, id);
    typename BCubCellSet::BitCoordIterator bitCoordIt(bitIt);

    BOOST_ASSERT(bitCoordIt.inSpace());

    return BitCoordCellImpl(bitCoordIt);
  }

  const Cell operator[](Id id)
  {
    return ((const CubSComplex*)this)->operator[](id);
  }

  boost::optional<Cell> find(const CellCode& cube)
  {
    int realCoord[DIM];
    bitmapCoord(cube, realCoord);

    typename BCubCellSet::BitCoordIterator bitIt(bCubCellSet, realCoord);
    if (!bitIt.inSpace() || bitIt.getBit() == 0) {
      return boost::optional<Cell>();
    }

    return boost::optional<Cell>(BitCoordCellImpl(bitIt));
  }

  boost::optional<Cell> find(Id id)
  {
    typename BCubCellSet::BitCoordIterator bitIt(bCubCellSet, id);
    typename BCubCellSet::BitCoordIterator bitCoordIt(bitIt);

    if (!bitCoordIt.inSpace() || bitCoordIt.getBit() == 0) {
      return boost::optional<Cell>();
    }

    return boost::optional<Cell>(BitCoordCellImpl(bitCoordIt));
  }

  CRef<BCubCellSet> getCubCellSet()
  {
      return _bCubCellSet;
  }

protected:

  static std::vector<int> computeBitmapSize(const std::vector<size_t>& v)
  {
    using boost::lambda::_1;
    std::vector<int> r(v.begin(), v.end());
    std::for_each(r.begin(), r.end(), _1 *= 2 + 1);
    return r;
  }


  static void bitmapCoord(const CubSComplex<DIM>::CellCode& cube, int* realCoord);
  static void cubeCode(const int* realCoord, CubSComplex<DIM>::CellCode& cube);


private:
  //  CubSComplex(const CubSComplex& other);

  //  boost::shared_ptr<RepSet<ElementaryCube> > repSet;
  CRef<BCubCellSet> _bCubCellSet;
  BCubCellSet& bCubCellSet;

};

  }
}

#include "CubSComplex_Cell.hpp"
#include "CubSComplex_IteratorProvider.hpp"
#include "CubSComplex_Iterators.hpp"
#include "CubSComplex_ColoredIterators.hpp"
#include "CubSComplex_Numerators.hpp"

namespace capd
{
    namespace complex
    {

template<int DIM>
inline CubSComplex<DIM>::CubSComplex(const std::vector<size_t>& size,
                                     const CubSComplex<DIM>::CellCodeRange& cubes):
  CubSComplexBase(DIM),
  _bCubCellSet(new BCubCellSet(&computeBitmapSize(size)[0], true)),
  bCubCellSet(_bCubCellSet())
{
  bCubCellSet.addEmptyCollar();

  std::for_each(cubes.begin(), cubes.end(), boost::bind(&CubSComplex::insert, this, _1));
}


template<int DIM>
inline CubSComplex<DIM>::CubSComplex(RepSet<ElementaryCube>& _repSet,
                                     bool addEmptyCollar /* = true */):
  CubSComplexBase(DIM),
  _bCubCellSet(new BCubCellSet(_repSet)),
  bCubCellSet(_bCubCellSet())
{

  if (addEmptyCollar)
  {
    bCubCellSet.addEmptyCollar();
  }
}

template<int DIM>
inline CubSComplex<DIM>::CubSComplex(const CRef<BCubCellSet>& __bCubCellSet,
                                     bool addEmptyCollar /* = true */):
  CubSComplexBase(DIM),
  _bCubCellSet(__bCubCellSet),
  bCubCellSet(_bCubCellSet())
{

  if (addEmptyCollar)
  {
    bCubCellSet.addEmptyCollar();
  }
}

template<int DIM>
inline void CubSComplex<DIM>::bitmapCoord(const CubSComplex<DIM>::CellCode& cube, int* realCoord)
{
  for (size_t idx = 0; idx < DIM; ++idx) {
    realCoord[idx] = 2*cube[idx].first + cube[idx].second + 2; // +2 because of new coordinates in empty collar
  }
}

template<int DIM>
inline void CubSComplex<DIM>::cubeCode(const int* realCoord, CubSComplex<DIM>::CellCode& cube)
{
  cube.resize(DIM);
  for (size_t idx = 0; idx < DIM; ++idx) {
    cube[idx].second = (realCoord[idx] % 2);
    cube[idx].first = (realCoord[idx] - 2) / 2;
  }
}


template<int DIM>
inline void CubSComplex<DIM>::insert(const CubSComplex<DIM>::CellCode& cube)
{
  BOOST_ASSERT((Dim) cube.size() == getDim());
  int realCoord[DIM];
  bitmapCoord(cube, realCoord);
  bCubCellSet.insert(realCoord);
}

template<int DIM>
inline typename CubSComplex<DIM>::Iterators CubSComplex<DIM>::iterators() {
  return Iterators(*this);
}

template<int DIM>
inline typename CubSComplex<DIM>::ConstIterators CubSComplex<DIM>::iterators() const {
  return ConstIterators(*this);
}

template<int DIM>
inline typename CubSComplex<DIM>::ColoredConstIterators CubSComplex<DIM>::iterators(const Color& color) const {
  BOOST_ASSERT(color == 1 || color == 2);
  return ColoredConstIterators(*this, ColorChecker(color));
}

template<int DIM>
inline typename CubSComplex<DIM>::ColoredIterators CubSComplex<DIM>::iterators(const Color& color) {
  BOOST_ASSERT(color == 1 || color == 2);
  return ColoredIterators(*this, ColorChecker(color));
}

template<int DIM>
template<typename ImplT>
inline bool CubSComplex<DIM>::getUniqueCoFace(const CellProxy<ImplT>& cell, DynamicCell& coface) const {
  if (bCubCellSet.isFreeFace(const_cast<typename CubSComplex::BCubCellSet::BitCoordIterator&>(cell.getBitCoordIt()), coface.getBitCoordIt())) {
  	 return true;
  } else {
  	 return false;
  }
}

template<int DIM>
template<typename ImplT>
inline bool CubSComplex<DIM>::getUniqueFace(const CellProxy<ImplT>& cell, DynamicCell& coface) const {
  if (bCubCellSet.isFreeCoFace(const_cast<typename CubSComplex::BCubCellSet::BitCoordIterator&>(cell.getBitCoordIt()), coface.getBitCoordIt())) {
  	 return true;
  } else {
  	 return false;
  }
}

template<int DIM>
template<typename ImplT1, typename ImplT2>
inline int CubSComplex<DIM>::coincidenceIndex(const CellProxy<ImplT1> &_a, const CellProxy<ImplT2> &_b) const {
  int res = 0;
  int sgn = 1;

  const typename CubSComplex::BCubCellSet::BitCoordIterator& a = _a.getBitCoordIt();
  const typename CubSComplex::BCubCellSet::BitCoordIterator& b = _b.getBitCoordIt();

  for (Dim i = 0, end = DIM; i < end; ++i) {
  	 if (! (a[i]/2 == b[i]/2 || a[i]/2 + (a[i]%2) == b[i]/2)) {
         return 0; // b[i] left side doesn't intersect a[i] interval
       }

       if (a[i] % 2 == 1 && b[i] % 2 == 1) {
         sgn = -sgn; //both nondegenerated, go to next
       } else if (a[i] % 2 == 0 && b[i] % 2 == 0) {
         // both degenerated
       } else if (b[i] % 2 == 1) {
         return 0; // a[i] is inside b[i]
       } else { // b[i] is inside a[i]
         if (res != 0) {
           return 0; // second time, so not proper face
         }
         res = sgn;
         if (a[i] / 2 != b[i]/2) { // b[i] is the right face of a[i]
           res = -res;
         }
       }
  }
  return res;
}

    }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_CUBSCOMPLEX_H
