/////////////////////////////////////////////////////////////////////////////
/// @file CubicTessellation
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-09-11
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_CUBICTESSELLATION_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_CUBICTESSELLATION_H

// #include
#include <capd/complex/EnumerateCellCodes.h>
#include <capd/complex/StaticArray.h>
#include <stdint.h>
#include <vector>
#include <utility>
#include <boost/assert.hpp>
#include <boost/iterator/iterator_facade.hpp>
#include <boost/range/iterator_range.hpp>
#include <algorithm>




namespace capd
{
  namespace complex
  {


    template<typename Value, size_t SIZE>
    struct SelectVector
    {
      typedef StaticArray<Value, SIZE> type;

      static type create(size_t /*size*/)
      {
        return type();
      }
    };


    template<typename Value>
    struct SelectVector<Value, 0>
    {
      typedef std::vector<Value> type;

      static type create(size_t size)
      {
        return type(size);
      }
    };

    template<uint8_t _DIM>
    struct DimVariable
    {
      DimVariable(uint8_t /*dim*/)
      {
      }
      uint8_t operator()() const { return _DIM; }
    };

    template<>
    struct DimVariable<0>
    {
      DimVariable(uint8_t dim) : _dim(dim) {}
      uint8_t operator()() const { return _dim; }

      const uint8_t _dim;
    };



    template<uint8_t _DIM>
    class CubicTessellation: public EnumerateCellCodes<
      typename SelectVector<std::pair<size_t, bool>, _DIM>::type,
      size_t>
    {

    public:
      typedef SelectVector<std::pair<size_t, bool>, _DIM>  CellCodeType;
      typedef SelectVector<size_t, _DIM>  FullCubeType;
      typedef typename FullCubeType::type FullCube;
      typedef EnumerateCellCodes<typename CellCodeType::type, size_t> Enumerator;
      using typename Enumerator::CellCode;
      using typename Enumerator::Index;

      typedef uint8_t Dim;
      typedef SelectVector<size_t, _DIM> LengthsType;
      typedef typename LengthsType::type Lengths;

      template<typename Derive>
      class IteratorBase: public boost::iterator_facade<Derive, const Index&, boost::bidirectional_traversal_tag>
      {
      private:
        friend class boost::iterator_core_access;

        bool equal(const IteratorBase& other) const
        {
          return this->_index == other._index;
        }

        const Index& dereference() const { return _index; }

      protected:

        IteratorBase(const CubicTessellation& cubicTessellation, Index index):
          _cubicTessellation(&cubicTessellation), _index(index)
        {}

        IteratorBase():
          _cubicTessellation(0), _index(-1)
        {}

        CubicTessellation const* _cubicTessellation;
        Index _index;
      };


      class AllIterator: public IteratorBase<AllIterator>
      {

      public:
        AllIterator(){}

        AllIterator(const CubicTessellation& cubicTessellation, Index index):
          IteratorBase<AllIterator>(cubicTessellation, index)
        {}

      private:
        friend class boost::iterator_core_access;

        void increment()
        {
          ++this->_index;
        }
      };

      class BdIterator: public IteratorBase<BdIterator>
      {

      public:
        BdIterator(const CubicTessellation& cubicTessellation, Index index):
          IteratorBase<BdIterator>(cubicTessellation, 0),
          _cellCode(cubicTessellation[index]),
          _dimIdx(0)
        {
          increment();
        }

        explicit BdIterator(const CubicTessellation& cubicTessellation):
          IteratorBase<BdIterator>(cubicTessellation, -1),
          _dimIdx(0)
        {}

        explicit BdIterator():
          _dimIdx(0)
        {}

      private:
        friend class boost::iterator_core_access;

        void increment()
        {
          if (this->_index == Index(-1)) {
            return;
          }

          while (_dimIdx/2 < this->_cubicTessellation->dim() && ! _cellCode[_dimIdx/2].second) {
            ++_dimIdx;
          }

          if (_dimIdx/2 >= this->_cubicTessellation->dim()) {
            this->_index = -1;
          } else {
            _cellCode[_dimIdx/2].second = false;
            _cellCode[_dimIdx/2].first += _dimIdx % 2;
            this->_index = (*this->_cubicTessellation)[_cellCode]; // TODO do it with precomputed offset. It depends only on _cellCode.second configuration
            _cellCode[_dimIdx/2].first -= _dimIdx % 2;
            _cellCode[_dimIdx/2].second = true;
            ++_dimIdx;
          }
        }

        CellCode _cellCode;
        Dim _dimIdx;
      };

      class CbdIterator: public IteratorBase<CbdIterator>
      {

      public:
        CbdIterator(const CubicTessellation& cubicTessellation, Index index):
          IteratorBase<CbdIterator>(cubicTessellation, 0),
          _cellCode(cubicTessellation[index]),
          _dimIdx(0)
        {
          increment();
        }

        explicit CbdIterator(const CubicTessellation& cubicTessellation):
          IteratorBase<CbdIterator>(cubicTessellation, -1),
          _dimIdx(0)
        {}

        explicit CbdIterator():
          _dimIdx(0)
        {}

      private:
        friend class boost::iterator_core_access;

        void increment()
        {
          if (this->_index == Index(-1)) {
            return;
          }

          do {
            while (_dimIdx/2 < this->_cubicTessellation->dim() && _cellCode[_dimIdx/2].second) {
              ++_dimIdx;
            }

            if (_dimIdx/2 >= this->_cubicTessellation->dim()) {
              break;
            }

            if ((_cellCode[_dimIdx/2].first == 0 && _dimIdx % 2 == 1)
                ||
                 (_cellCode[_dimIdx/2].first == this->_cubicTessellation->_gridLengths[_dimIdx/2] && _dimIdx % 2 == 0)) {
              ++_dimIdx;
            } else {
              break;
            }
          } while(true);

          if (_dimIdx/2 >= this->_cubicTessellation->dim()) {
            this->_index = -1;
          } else {
            _cellCode[_dimIdx/2].second = true;
            _cellCode[_dimIdx/2].first -= _dimIdx % 2;
            this->_index = (*this->_cubicTessellation)[_cellCode]; // TODO do it with precomputed offset. It depends only on _cellCode.second configuration
            _cellCode[_dimIdx/2].first += _dimIdx % 2;
            _cellCode[_dimIdx/2].second = false;
            ++_dimIdx;
          }
        }

        CellCode _cellCode;
        Dim _dimIdx;
      };

    public:

      typedef boost::iterator_range<AllIterator> All;
      typedef boost::iterator_range<BdIterator> Bd;
      typedef boost::iterator_range<CbdIterator> Cbd;

      explicit CubicTessellation(const std::vector<size_t>& lengths): // in terms of full cubes
        _lengths(computeLengths(lengths)), _size(computeSize(lengths)), _dim(_lengths.size()),
        _gridLengths(lengths)
      {}

      All all() const { return All(AllIterator(*this, 0), AllIterator(*this, size())); }
      Bd bd(Index index) const { return Bd(BdIterator(*this, index), BdIterator(*this)); }
      Cbd cbd(Index index) const { return Cbd(CbdIterator(*this, index), CbdIterator(*this)); }

      size_t size() const { return _size; }
      Dim dim() const { return _dim(); }

      CellCode cellCode() const { return CellCodeType::create(_dim()); }
      FullCube fullCube() const { return FullCubeType::create(_dim()); }

      Index operator[](const CellCode& cellCode) const
      {
        BOOST_ASSERT(cellCode.size() == _dim());

        Index id = 0;
        size_t subSize = _size;

        for (Dim idx = 0; idx < _dim(); ++idx) {
          BOOST_ASSERT(cellCode[idx].first < _lengths[idx]);
          const size_t realCellCode = 2*cellCode[idx].first + cellCode[idx].second;
          subSize /= _lengths[idx];
          id += realCellCode * subSize;
        }

        BOOST_ASSERT(id < _size);
        return id;
      }

      CellCode operator[](Index index) const
      {
        BOOST_ASSERT(index < _size);

        CellCode cellCode = this->cellCode();
        Index id = index;
        size_t subSize = _size;

        for (Dim idx = 0; idx < _dim(); ++idx) {
          subSize /= _lengths[idx];
          const size_t realCellCode = id / subSize;
          id %= subSize;
          BOOST_ASSERT(realCellCode / 2 < _lengths[idx]);
          cellCode[idx] = std::make_pair(realCellCode / 2, realCellCode % 2);
        }

        BOOST_ASSERT(id == 0);
        return cellCode;
      }

      Dim dim(Index index) const
      {
        BOOST_ASSERT(index < _size);

        Dim dim = 0;
        Index id = index;
        size_t subSize = _size;

        for (Dim idx = 0; idx < _dim(); ++idx) {
          subSize /= _lengths[idx];
          const size_t realCellCode = id / subSize;
          id %= subSize;
          dim += realCellCode % 2;
        }

        return dim;
      }

      virtual ~CubicTessellation() {}

    private:
      static Lengths computeLengths(const std::vector<size_t>& lengths)
      {
        const Dim dim = lengths.size();
        Lengths realLengths = LengthsType::create(dim);

        for (Dim idx = 0; idx < dim; ++idx) {
          realLengths[idx] = (lengths[idx] + 1) * 2 - 1;
        }

        return realLengths;
      }

      static size_t computeSize(const std::vector<size_t>& lengths)
      {
        size_t size = lengths.size() > 0 ? 1 : 0;
        const Dim dim = lengths.size();

        for (Dim idx = 0; idx < dim; ++idx) {
          size *= (lengths[idx] + 1) * 2 - 1;
        }

        return size;
      }


      const Lengths _lengths;
      const size_t _size;
      const DimVariable<_DIM> _dim;
      const Lengths _gridLengths;
    };


  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_CUBICTESSELLATION_H
