/////////////////////////////////////////////////////////////////////////////
/// @file DiscreteVectorField
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-03-31
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_DISCRETEVECTORFIELD_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_DISCRETEVECTORFIELD_H

// #include
#include <boost/assert.hpp>
#include "DiscreteVectorFieldGraph.h"
#include <boost/lambda/bind.hpp>
#include <boost/lambda/lambda.hpp>


namespace capd
{
  namespace complex
  {

    template<typename ComplexT, typename ScalarT>
    class DiscreteVectorField
    {

      struct ValidPred
      {
        ValidPred(): _dvf(0) {}

        explicit ValidPred(const DiscreteVectorField& dvf):
          _dvf(&dvf)
        {}

        template<typename Vertex>
        bool operator()(const Vertex& v) const
        {
          return _dvf->isInField(v);
        }
        const DiscreteVectorField* _dvf;
      };

    public:
      typedef ComplexT Complex;
      typedef ScalarT Scalar;
      typedef DiscreteVectorFieldGraph<Complex, Scalar, ValidPred> Graph;
      typedef typename Complex::Dim Dim;
      typedef typename Complex::Id Id;


      explicit DiscreteVectorField(Complex& complex):
        _complex(complex),
        _maxCriticalCells(_complex.size()),
        _graph(_complex, ValidPred(*this)),
        _match(_maxCriticalCells),
        _isKing(_maxCriticalCells, false),
        _available(_maxCriticalCells, false)
      {
        std::copy(boost::make_counting_iterator((size_t)0),
                  boost::make_counting_iterator((size_t)_maxCriticalCells),
                  _match.begin());

        namespace bl = boost::lambda;
        boost::for_each(_complex.iterators(1).allCells(),
                        bl::var(_available)[bl::bind(MakeCellId<Complex>(), bl::_1)] = true);
      }

      Complex& complex() const { return _complex; }

      size_t criticalCells() const { return _criticalCells.size(); }
      const std::vector<Id>& allCriticalCells() const { return _criticalCells; }
      Id criticalCellId(size_t number) const { return _criticalCells[number]; }

      bool isInField(Id id) const
      {
        return _available[id];
      }

      bool isProcessed(Id id) const
      {
        return isCritical(id) || isKing(id) || isQueen(id);
      }

      bool isCritical(Id id) const
      {
        return _isKing[id] && _match[id] == id;
      }

      bool isKing(Id id) const
      {
        return _isKing[id] && _match[id] != id;
      }

      bool isQueen(Id id) const
      {
        return (!_isKing[id]) && _match[id] != id;
      }

      template<typename Cell>
      void markCritical(const Cell& cell)
      {
        _isKing[cell.getId()] = true;
        updateGraph(cell);
        _criticalCells.push_back(cell.getId());
      }

      template<typename CellU, typename CellV>
      void match(const CellU& cellU, const CellV& cellV)
      {
        const Dim dimU = cellU.getDim();
        const Dim dimV = cellV.getDim();
        const Id idU = cellU.getId();
        const Id idV = cellV.getId();

        BOOST_ASSERT(idU != idV);
        BOOST_ASSERT((dimU > dimV && dimU - dimV == 1)
                     ||
                     (dimV > dimU && dimV - dimU == 1));
        BOOST_ASSERT(!isProcessed(idU));
        BOOST_ASSERT(!isProcessed(idV));

        _match[idU] = idV;
        _match[idV] = idU;
        if (dimU > dimV) {
          _isKing[idU] = true;
        } else {
          _isKing[idV] = true;
        }

        BOOST_ASSERT(isInField(idU));
        BOOST_ASSERT(!isCritical(idU));
        BOOST_ASSERT(isInField(idV));
        BOOST_ASSERT(!isCritical(idV));
        BOOST_ASSERT(isKing(idU) || isQueen(idU));
        BOOST_ASSERT(isKing(idV) || isQueen(idV));
        updateGraph(cellU, cellV);
      }

      Id matched(Id id) const
      {
        BOOST_ASSERT(isQueen(id) || isKing(id));
        return _match[id];
      }

      const Graph& graph() const { return _graph; }

    private:

      template<typename CellU>
      void updateGraph(const CellU& cell)
      {
        BOOST_ASSERT(isCritical(cell.getId()));
        for (auto bdCell : _complex.iterators().bdCells(cell)) {
          _graph.addBdEdge(cell, bdCell);
        }
      }

      template<typename CellU, typename CellV>
      void updateGraph(const CellU& cellU, const CellV& cellV)
      {
        BOOST_ASSERT((isKing(cellU.getId()) && isQueen(cellV.getId()))
                     ||
                     (isKing(cellV.getId()) && isQueen(cellU.getId())));
        if (! (isKing(cellU.getId()) && isQueen(cellV.getId()))) {
          updateGraph(cellV, cellU);
          return;
        }

        BOOST_ASSERT(isKing(cellU.getId()) && isQueen(cellV.getId()));

        _graph.match(cellU, cellV);

        for (auto bdCell : _complex.iterators().bdCells(cellU)) {
          if (!_available[bdCell.getId()]) {
            continue;
          }

          if (bdCell.getId() != cellV.getId()) {
            _graph.addBdEdge(cellU, bdCell);
          }
        }

        for (auto bdCell : _complex.iterators().bdCells(cellV)) {
          if (!_available[bdCell.getId()]) {
            continue;
          }

          if (bdCell.getId() != cellU.getId()) {
            _graph.addBdEdge(cellV, bdCell);
          }
        }

      }


      DiscreteVectorField(const DiscreteVectorField&);
      DiscreteVectorField& operator=(const DiscreteVectorField&);

      Complex& _complex;
      std::vector<Id> _criticalCells;
      size_t _maxCriticalCells;
      Graph _graph;
      std::vector<Id> _match;
      std::vector<bool> _isKing;
      std::vector<bool> _available; // only because CubSComplex has 0-1 colors, so we do not know which cells were in the original complex. Otherwaise cell color is enough.
    };


  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_DISCRETEVECTORFIELD_H
