/////////////////////////////////////////////////////////////////////////////
/// @file DiscreteVectorFieldGraph
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-03-31
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_DISCRETEVECTORFIELDGRAPH_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_DISCRETEVECTORFIELDGRAPH_H

// #include
#include <boost/graph/adjacency_list.hpp>
#include <capd/complex/ComplexFunctors.h>
#include <boost/range/algorithm/for_each.hpp>
#include <boost/graph/filtered_graph.hpp>
#include <boost/graph/topological_sort.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/range/iterator_range.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/graph/properties.hpp>


namespace boost {


}

namespace capd
{
  namespace complex
  {

    // TODO MJ:
    // This clas should be implementated as adaptor Complex -> Boost.Graph. Now it takes additional memory.
    template<typename Complex, typename Scalar, typename VertexPredicate>
    struct DiscreteVectorFieldGraph
    {
    public:
      typedef typename Complex::Id Id;

    private:
      struct edge_coincidence_index_t
      {
        typedef boost::edge_property_tag kind;
      };

      struct edge_in_matching_t
      {
        typedef boost::edge_property_tag kind;
      };

      typedef boost::no_property VertexProps;
      typedef boost::property<edge_coincidence_index_t, Scalar,
                              boost::property<edge_in_matching_t, bool> >
      EdgeProps;


      typedef boost::adjacency_list<boost::vecS, boost::vecS, boost::directedS,
                                    VertexProps, EdgeProps>
      FullGraph;

      struct EdgePredicate
      {
        typedef bool result_type;
        EdgePredicate(): _graph(0) {}

        EdgePredicate(const FullGraph& graph, VertexPredicate p):
          _graph(&graph),
          _predicate(p)
        {}

        result_type operator()(const typename boost::graph_traits<FullGraph>::edge_descriptor& e) const
        {
          return _predicate(boost::source(e, *_graph))
            &&
            _predicate(boost::target(e, *_graph));
        }

        const FullGraph* _graph;
        VertexPredicate _predicate;
      };

      typedef boost::filtered_graph<FullGraph, EdgePredicate, VertexPredicate> Graph;
      typedef typename boost::graph_traits<Graph>::edge_descriptor Edge;
      typedef typename boost::graph_traits<Graph>::out_edge_iterator EdgeIterator;
      typedef typename boost::graph_traits<Graph>::vertex_descriptor Vertex;
      typedef typename boost::graph_traits<Graph>::vertex_iterator VertexIterator;

      struct VertexSuccessor
      {
        typedef boost::tuple<Id, Scalar, bool> result_type;
        VertexSuccessor(): _dvfg(0) {}

        explicit VertexSuccessor(const DiscreteVectorFieldGraph& dvfg):
          _dvfg(&dvfg)
        {}

        result_type operator()(const Edge& e) const
        {
          return boost::make_tuple(boost::target(e, _dvfg->graph()), _dvfg->_coincidence[e], _dvfg->_inMatching[e]);
        }

        const DiscreteVectorFieldGraph* _dvfg;
      };

    public:
      typedef boost::iterator_range<boost::transform_iterator<VertexSuccessor, EdgeIterator> > Neighbours;
      typedef boost::iterator_range<boost::transform_iterator<Id (*)(const Vertex&), VertexIterator> > Vertices;

      DiscreteVectorFieldGraph(Complex& complex, VertexPredicate predicate):
        _complex(complex),
        _fullGraph(_complex.size()),
        _graph(_fullGraph, EdgePredicate(_fullGraph, predicate), predicate),
        _coincidence(boost::get(edge_coincidence_index_t(), _fullGraph)),
        _inMatching(boost::get(edge_in_matching_t(), _fullGraph))
      {
      }

      size_t size() const { return boost::num_vertices(graph()); }

      std::vector<Id> orderedVertices() const
      {
        std::vector<Id> orderedVertices;
        orderedVertices.reserve(size());
        boost::topological_sort(graph(), std::back_inserter(orderedVertices));

        return orderedVertices;
      }

      Vertices vertices() const
      {
        auto vertices = boost::vertices(graph());
        return Vertices(boost::make_transform_iterator(vertices.first, boost::lexical_cast<Id, Vertex>),
                        boost::make_transform_iterator(vertices.second, boost::lexical_cast<Id, Vertex>));
      }

      Neighbours neighbours(Id vertex) const
      {
        auto edges = boost::out_edges(vertex, graph());
        return Neighbours(boost::make_transform_iterator(edges.first, VertexSuccessor(*this)),
                          boost::make_transform_iterator(edges.second, VertexSuccessor(*this)));
      }

      template<typename CellU, typename CellV>
      void addBdEdge(const CellU& u, const CellV& v)
      {
        BOOST_ASSERT(u.getDim() > v.getDim() && u.getDim() - v.getDim() == 1);
        Edge edge = boost::add_edge(u.getId(), v.getId(), _fullGraph).first;
        _coincidence[edge] = _complex.coincidenceIndex(u, v);
        _inMatching[edge] = false;
      }

      template<typename CellU, typename CellV>
      void match(const CellU& u, const CellV& v)
      {
        if (u.getDim() < v.getDim()) {
          match(v, u);
          return;
        }
        BOOST_ASSERT(u.getDim() > v.getDim() && u.getDim() - v.getDim() == 1);
        Edge edge = boost::add_edge(v.getId(), u.getId(), _fullGraph).first;
        _coincidence[edge] = Scalar(-1) / _complex.coincidenceIndex(u, v);
        _inMatching[edge] = true;
      }


    private:
      // DiscreteVectorFieldGraph(const DiscreteVectorFieldGraph&);
      // DiscreteVectorFieldGraph& operator=(const DiscreteVectorFieldGraph&);

      Graph& graph() { return _graph; }
      const Graph& graph() const { return _graph; }

      Complex& _complex;
      FullGraph _fullGraph;
      Graph _graph;

      typename boost::property_map<FullGraph, edge_coincidence_index_t>::type _coincidence;
      typename boost::property_map<FullGraph, edge_in_matching_t>::type _inMatching;
    };

  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_DISCRETEVECTORFIELDGRAPH_H
