/////////////////////////////////////////////////////////////////////////////
/// @file FilteredComplex
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-03-13
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_FILTEREDCOMPLEX_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_FILTEREDCOMPLEX_H

// #include
#include "FilteredIteratorFamily.h"
#include <boost/optional.hpp>
#include <boost/range/distance.hpp>

namespace capd
{
  namespace complex
  {


    template<typename ComplexT, typename FilterT>
    class FilteredComplex
    {

    public:
      typedef ComplexT Complex;
      typedef typename Complex::Tags Tags;
      typedef typename Complex::Cell OrgCell;
      typedef typename Complex::ConstCell OrgConstCell;
      typedef typename Complex::CellCode OrgCellCode;
      typedef typename Complex::Id OrgId;
      typedef typename Complex::Dim OrgDim;
      typedef OrgId Id;
      typedef OrgCellCode CellCode;
      typedef OrgDim Dim;
      typedef typename OrgCell::Color Color;

      typedef FilterT Filter;
      typedef OrgCell Cell;
      typedef OrgConstCell ConstCell;

      typedef FilteredIteratorFamily<typename Complex::Iterators, Filter> Iterators;
      typedef FilteredIteratorFamily<typename Complex::ConstIterators, Filter> ConstIterators;
      typedef FilteredIteratorFamily<typename Complex::ColoredIterators, Filter> ColoredIterators;
      typedef FilteredIteratorFamily<typename Complex::ColoredConstIterators, Filter> ColoredConstIterators;

      explicit FilteredComplex(Complex& complex, Filter filter=Filter()):
        _complex(complex), _filter(filter)
      {}

      Dim getDim() const { return _complex.getDim(); }

      Iterators iterators()
      {
        return Iterators(this->complex().iterators(), _filter);
      }

      ConstIterators iterators() const
      {
        return ConstIterators(this->complex().iterators(), _filter);
      }

      ColoredIterators iterators(Color color)
      {
        return ColoredIterators(this->complex().iterators(color), _filter);
      }

      ColoredConstIterators iterators(Color color) const
      {
        return ColoredConstIterators(this->complex().iterators(color), _filter);
      }

      ConstCell operator[](const CellCode& cellCode) const
      {
        ConstCell res = _complex[cellCode];
        BOOST_ASSERT(_filter(res));
        return res;
      }

      ConstCell operator[](Id id) const
      {
        ConstCell res = _complex[id];
        BOOST_ASSERT(_filter(res));
        return res;
      }

      Cell operator[](const CellCode& cellCode)
      {
        Cell res = _complex[cellCode];
        BOOST_ASSERT(_filter(res));
        return res;
      }

      Cell operator[](Id id)
      {
        Cell res = _complex[id];
        BOOST_ASSERT(_filter(res));
        return res;
      }

      boost::optional<Cell> find(const CellCode& cellCode)
      {
        boost::optional<Cell> res = this->complex().find(cellCode);

        if (! _filter(res)) {
          return boost::optional<Cell>();
        }

        return res;
      }

      boost::optional<Cell> find(Id id)
      {
        boost::optional<Cell> res = this->complex().find(id);

        if (! _filter(res)) {
          return boost::optional<Cell>();
        }

        return res;
      }

      size_t size() const
      {
        return _complex.size();
      }

      size_t cardinality() const
      {
        return boost::distance(iterators(1).allCells());
      }

      virtual ~FilteredComplex() {}

    private:
      Complex& complex()
      {
        return _complex;
      }

      const Complex& complex() const
      {
        return _complex;
      }

      Complex& _complex;
      Filter _filter;
    };


    template<typename Complex, typename Filter>
    FilteredComplex<Complex, Filter> makeFilteredComplex(Complex& complex, Filter filter)
    {
      return FilteredComplex<Complex, Filter>(complex, filter);
    }
  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_FILTEREDCOMPLEX_H
