/////////////////////////////////////////////////////////////////////////////
/// @file FundamentalGroup
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-04-12
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////



#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_FUNDAMENTALGROUP_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_FUNDAMENTALGROUP_H

// #include
#include "BuildMorseComplex.h"
#include "FundamentalGroupPaths.h"

#include <capd/complex/LoggerHelpers.h>
#include <map>
#include <vector>
#include <set>
#include <list>
#include <boost/assert.hpp>

namespace capd
{
  namespace complex
  {

    // This is a refactorization of FundGroup written by Piotr Brendel
    template<typename DiscreteVectorField>
    class FundamentalGroup
    {

      typedef typename DiscreteVectorField::Id Id;
      typedef typename DiscreteVectorField::Scalar Scalar;
      typedef typename BuildMorseComplex<DiscreteVectorField, bool>::MorseComplex MorseComplex;
      typedef capd::complex::FundamentalGroupPaths<DiscreteVectorField> FundamentalGroupPaths;

      typedef std::set<Id> Cells;
      typedef std::vector<Cells> CellsByDim;
      typedef std::vector<std::pair<Id, Scalar> >    Chain;

      typedef std::pair<Id, int>                          RelatorComponent;
      typedef std::vector<RelatorComponent>               Relator;
      typedef std::vector<Relator>                        Relators;

    public:
      explicit FundamentalGroup(const DiscreteVectorField& dvf):
        _dvf(dvf),
        _morseComplex(ConvertToMorseComplex<bool>(dvf))
      {}

      void operator()()
      {
        CAPD_DEBUG_TMPL("getting cells data");
        GetCells(_cellsByDim, _2Boundaries);

        if (_cellsByDim[0].size() > 1) {
          CAPD_DEBUG_TMPL("creating spanning tree");
          CreateSpanningTree();
        }

        CAPD_DEBUG_TMPL("computing relators");
        ComputeRelators();
        SimplifyRelators();
      }

      std::vector<int> HapInterfaceVector() const
      {
        std::vector<int> ret;

        std::map<Id, int> symbols;
        {
          const Cells &_1cells = _cellsByDim[1];
          ret.push_back(_1cells.size());
          int c = 1;
          typename Cells::const_iterator it = _1cells.begin();
          typename Cells::const_iterator itEnd = _1cells.end();
          for ( ; it != itEnd; ++it)
            {
              symbols[*it] = c++;
            }
        }

        {
          ret.push_back(_relators.size());
          typename Relators::const_iterator it = _relators.begin();
          typename Relators::const_iterator itEnd = _relators.end();
          for ( ; it != itEnd; ++it)
            {
              Relator r = *it;
              ret.push_back(r.size());
              typename Relator::iterator jt = r.begin();
              typename Relator::iterator jtEnd = r.end();
              for ( ; jt != jtEnd; ++jt)
                {
                  ret.push_back(symbols[jt->first]);
                  ret.push_back(jt->second);
                }
            }
        }

        return ret;
      }

    private:
      FundamentalGroup(const FundamentalGroup&);
      FundamentalGroup& operator=(const FundamentalGroup&);



      void CreateSpanningTree()
      {
        std::list<Id> _1cells(_cellsByDim[1].begin(), _cellsByDim[1].end());
        _cellsByDim[1].clear();
        Cells verts;
        Cells &edges = _spanningTreeEdges;
        verts.insert(*_cellsByDim[0].begin());

        bool found = true;
        while (found)
          {
            found = false;
            typename std::list<Id>::iterator it = _1cells.begin();
            typename std::list<Id>::iterator itEnd = _1cells.end();
            while (it != itEnd) {

              BOOST_ASSERT(_morseComplex[*it].getDim() == 1);
                Chain boundary = GetBoundary(*it);
                if (boundary.size() == 2)
                  {
                    typename Chain::iterator cit = boundary.begin();
                    Id v0 = cit->first;
                    cit++;
                    Id v1 = cit->first;
                    BOOST_ASSERT(v0 != v1);

                    bool v0Found = verts.find(v0) != verts.end();
                    bool v1Found = verts.find(v1) != verts.end();
                    if (v0Found && !v1Found)
                      {
                        verts.insert(v1);
                        edges.insert(*it);
                        it = _1cells.erase(it);
                        found = true;
                      }
                    else if (!v0Found && v1Found)
                      {
                        verts.insert(v0);
                        edges.insert(*it);
                        it = _1cells.erase(it);
                        found = true;
                      }
                    else if (v0Found && v1Found)
                      {
                        _cellsByDim[1].insert(*it);
                        it = _1cells.erase(it);
                      }
                    else
                      {
                        ++it;
                      }
                  }
                else
                  {
                    CAPD_WARN_TMPL("boundary.size =  " << boundary.size() << " is it ok?");
                    ++it;
                    //BOOST_ASSERT(false);
                  }
              }
          }

        CAPD_DEBUG_TMPL(verts.size() << " " << _cellsByDim[0].size());
        BOOST_ASSERT(verts.size() == _cellsByDim[0].size());

        _cellsByDim[1].insert(_1cells.begin(), _1cells.end());

        // TODO!
        // if we assume that space may be disconnected
        // we need to remove "unused" cells
      }

      void ComputeRelators()
      {
        typename std::map<Id, Chain>::iterator it = _2Boundaries.begin();
        typename std::map<Id, Chain>::iterator itEnd = _2Boundaries.end();

        for ( ; it != itEnd; ++it)
          {
            Cells &edges = _spanningTreeEdges;
            Relator r;
            Chain boundary = it->second;
            typename Chain::iterator jt = boundary.begin();
            typename Chain::iterator jtEnd = boundary.end();

            for ( ; jt != jtEnd; ++jt)
              {
                // if edge is not contained in the spanning tree
                // add it as relator with proper sign
                // TODO:
                // if we assume that space may be disconnected then
                // we need to check if 2-boundary is in current connected component
                // (for example by checking if a vertex is contained in spanning tree)
                if (edges.find(jt->first) == edges.end())
                  {
                    r.push_back(RelatorComponent(jt->first, jt->second));
                  }
              }
            if (r.size() > 0)
              {
                _relators.push_back(r);
              }
          }
      }

      void SimplifyRelators()
      {
        std::set<Id>& generators = _cellsByDim[1];
        size_t relatorsCount = _relators.size();
        std::vector<bool> unusedRelators(relatorsCount);
        bool reduced = true;
        while (reduced)
          {
            reduced = false;
            for (size_t i = 0; i < relatorsCount; i++)
              {
                if (unusedRelators[i])
                  {
                    continue;
                  }
                Relator r = _relators[i];
                size_t reducedCount = 0;
                typename Relator::iterator generatorToReduce;
                for (typename Relator::iterator it = r.begin(); it != r.end(); ++it)
                  {
                    if (generators.find(it->first) == generators.end())
                      {
                        reducedCount++;
                      }
                    else
                      {
                        generatorToReduce = it;
                      }
                  }
                if (reducedCount >= r.size() - 1)
                  {
                    if (reducedCount == r.size() - 1)
                      {
                        generators.erase(generatorToReduce->first);
                      }
                    unusedRelators[i] = true;
                    reduced = true;
                  }
              }
            CAPD_DEBUG_TMPL("generators left: "<<generators.size());
          }
        Relators newRelators;
        for (size_t i = 0; i < relatorsCount; i++)
          {
            if (unusedRelators[i])
              {
                continue;
              }
            Relator r;
            for (typename Relator::iterator it = _relators[i].begin();
                 it != _relators[i].end();
                 ++it)
              {
                if (generators.find(it->first) != generators.end())
                  {
                    r.push_back(*it);
                  }
              }
            BOOST_ASSERT(r.size() > 1);
            newRelators.push_back(r);
          }
        _relators.assign(newRelators.begin(), newRelators.end());
      }

      void GetCells(CellsByDim& cellsByDim,
                    std::map<Id, Chain>& _2Boundaries)
      {
        size_t maxDim = _dvf.complex().getDim();
        cellsByDim.resize(std::max((size_t)3, maxDim + 1)); // always has dim 0, 1, 2

        for (size_t dim = 0; dim <= maxDim; dim++) {
          for (auto cell : _morseComplex.iterators(1).dimCells(dim)) {
            cellsByDim[dim].insert(cell.getId());
          }
        }

        // if there are some 2-cells, take its boundaries
        FundamentalGroupPaths homotopicPaths(_dvf);
        _2Boundaries.clear();
        if (cellsByDim.size() > 2)
          {
            Cells _2cells = cellsByDim[2];
            typename Cells::iterator it = _2cells.begin();
            typename Cells::iterator itEnd = _2cells.end();
            for ( ; it != itEnd; ++it) {
              BOOST_ASSERT(_morseComplex[*it].getDim() == 2);
              _2Boundaries[*it] = homotopicPaths.GetHomotopicBoundary(*it);
            }
          }
      }


      Chain GetBoundary(const Id& cellId)
      {
        Chain boundary;
        auto cell = _morseComplex[cellId];

        for(auto bdCell : _morseComplex.iterators(1).bdCells(cell)) {
          int ci = _morseComplex.coincidenceIndex(cell, bdCell);
          BOOST_ASSERT(ci != 0);
          boundary.push_back(std::pair<Id, int>(bdCell.getId(), ci));
        }
        return boundary;
      }


      const DiscreteVectorField& _dvf;
      MorseComplex _morseComplex;

      CellsByDim              _cellsByDim;
      std::map<Id, Chain>     _2Boundaries;
      Cells                   _spanningTreeEdges;
      Relators                _relators;

      CAPD_CLASS_LOGGER;
    };


    template<typename DiscreteVectorField>
    std::vector<int> FundamentalGroupHapInterfaceVector(const DiscreteVectorField& dvf)
    {
      FundamentalGroup<DiscreteVectorField> algo(dvf);
      algo();
      return algo.HapInterfaceVector();
    }

  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_FUNDAMENTALGROUP_H
