/////////////////////////////////////////////////////////////////////////////
/// @file FundamentalGroupPaths
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-04-13
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_FUNDAMENTALGROUPPATHS_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_FUNDAMENTALGROUPPATHS_H

// #include
#include "FundamentalGroupOrderedBd.h"
#include <boost/range/irange.hpp>
#include <boost/bimap.hpp>
#include <capd/auxil/Logger.h>

namespace capd
{
  namespace complex
  {

    // This is a refactorization of AKQHomotopicPaths written by Piotr Brendel
    template<typename DiscreteVectorField>
    class FundamentalGroupPaths
    {

    public:

      typedef typename DiscreteVectorField::Id Id;
      typedef typename DiscreteVectorField::Complex Complex;
      typedef std::vector<std::pair<Id, int> >   Chain;
      typedef std::pair<Id, int>                 PathCell;
      typedef std::list<PathCell>                         Path;
      typedef std::map<Id, Path>                 PathsMap;
      typedef typename PathsMap::iterator                 PathsMapIterator;
      typedef boost::bimap<Id, Id>     AcesMap;


      explicit FundamentalGroupPaths(const DiscreteVectorField& dvf):
        _dvf(dvf)
      {
        ComputeAcesMap();
      }


      Chain GetHomotopicBoundary(const Id& cellId)
      {
        // cell needs to be an ace
        assert(_acesMap.right.find(cellId) != _acesMap.right.end());
        // we get an original cell
        Id originalCellId = _acesMap.right.at(cellId);
        assert(_dvf.isCritical(originalCellId));

        // we take its boundary cells and create a homotopic path
        Path bdPath = GetOrdered2Boundary(originalCellId);
        Path homotopicPath;
        for (typename Path::iterator jt = bdPath.begin(); jt != bdPath.end(); ++jt)
          {
            Path tmpPath;
            GetHomotopicPath(*jt, tmpPath);
            homotopicPath.insert(homotopicPath.end(), tmpPath.begin(), tmpPath.end());
          }

        // finally we need to map original complex cells' ids to the output complex cells' ids
        Chain boundary;
        for (typename Path::iterator jt = homotopicPath.begin(); jt != homotopicPath.end(); ++jt)
          {
            // only ACES can be present in the final homotopic path
            assert(_dvf.isCritical(jt->first));
            assert(_acesMap.left.find(jt->first) != _acesMap.left.end());
            Id ace = _acesMap.left.at(jt->first);
            int ci = jt->second;//_outputComplex->coincidenceIndex(cell, ace);
            assert(ci != 0);
            boundary.push_back(std::pair<Id, int>(ace, ci));
          }
        return boundary;
      }

      void GetHomotopicPath(const Path& path, Path& outPath)
      {
        outPath.clear();
        typename Path::const_iterator it = path.begin();
        typename Path::const_iterator itEnd = path.end();
        for ( ; it != itEnd; it++)
          {
            Path tmpPath;
            GetHomotopicPath(*it, tmpPath);

            // reducing words
            bool reduced = false;
            while (!reduced && outPath.size() > 0 && tmpPath.size() > 0)
              {
                if (outPath.back().first == tmpPath.front().first)
                  {
                    int sum = outPath.back().second + tmpPath.front().second;
                    if (sum == 0)
                      {
                        // removing last element from current list
                        typename Path::iterator last = outPath.end();
                        last--;
                        outPath.erase(last);
                        tmpPath.erase(tmpPath.begin());
                      }
                    else
                      {
                        outPath.back().second = sum;
                        tmpPath.erase(tmpPath.begin());
                      }
                  }
                else
                  {
                    reduced = true;
                  }
              }

            // if there's still something left just add it at the end
            outPath.insert(outPath.end(), tmpPath.begin(), tmpPath.end());
          }
      }

      void GetHomotopicPath(const PathCell& cell, Path& outPath)
      {
        if (_dvf.isKing(cell.first))
          {
            // no path
            outPath.clear();
          }
        else if (_dvf.isCritical(cell.first))
          {
            // critical cell
            outPath.clear();
            outPath.push_back(cell);
          }
        else
          {
            assert(_dvf.isQueen(cell.first));
            PathsMapIterator it = GetQueenHomotopicPath(cell.first);
            assert(it != _queenPaths.end());
            if (cell.second > 0)
              {
                outPath.assign(it->second.begin(), it->second.end());
              }
            else
              {
                ReverseNegate(it->second, outPath);
              }
          }
      }

      PathsMapIterator GetQueenHomotopicPath(Id cellId)
      {
        // we can only compute homotopic path for QUEEN cell
        assert(_dvf.isQueen(cellId));

        // if the path has been already computed, we return it
        PathsMapIterator hpIt = _queenPaths.find(cellId);
        if (hpIt != _queenPaths.end()) {
          return hpIt;
        }

        Path prevCells;
        Path nextCells;
        Path* currentCells = &prevCells;
        int orientation = 0;

        // then, we take a boundary of KING cell and compute
        // alternative path for traversing given QUEEN cell
        Path boundary = GetOrdered2Boundary(_dvf.matched(cellId));
        typename Path::iterator it = boundary.begin();
        typename Path::iterator itEnd = boundary.end();
        for ( ; it != itEnd; ++it)
          {
            int ci = it->second;
            assert(ci != 0);
            if (it->first == cellId)
              {
                orientation = ci;
                currentCells = &nextCells;
              }
            else
              {
                currentCells->push_back(*it);
              }
          }
        assert(orientation != 0);

        // based on orientation of QUEEN cell we join two paths ("from" it and "to")
        Path bdPath;
        if (orientation > 0)
          {
            Path tmp;
            ReverseNegate(prevCells, tmp);
            bdPath.insert(bdPath.end(), tmp.begin(), tmp.end());
            ReverseNegate(nextCells, tmp);
            bdPath.insert(bdPath.end(), tmp.begin(), tmp.end());
          }
        else
          {
            bdPath.insert(bdPath.end(), nextCells.begin(), nextCells.end());
            bdPath.insert(bdPath.end(), prevCells.begin(), prevCells.end());
          }

        // finally we need to compute homotopic paths for every cell on path
        Path homotopicPath;
        GetHomotopicPath(bdPath, homotopicPath);

        return _queenPaths.insert(std::make_pair(cellId, homotopicPath)).first;
      }

      std::list<std::pair<Id, int> >
      GetOrdered2Boundary(const Id& cellId)
      {
        std::list<std::pair<Id, int> > boundary;
        auto cell = _dvf.complex()[cellId];
        BOOST_ASSERT(cell.getDim() == 2);

        for(auto bdCell : _dvf.complex().iterators().bdCells(cell)) {
          int ci = _dvf.complex().coincidenceIndex(cell, bdCell);
          assert(ci != 0);
          boundary.push_back(std::pair<Id, int>(bdCell.getId(), ci));
        }

        FundamentalGroupOrderedBd<Complex>(boundary);
        return boundary;
      }

      void ReverseNegate(const Path& path, Path& outPath)
      {
        outPath.clear();
        typename Path::const_reverse_iterator it = path.rbegin();
        typename Path::const_reverse_iterator itEnd = path.rend();
        for ( ; it != itEnd; ++it)
          {
            outPath.push_back(std::make_pair(it->first, -it->second));
          }
      }

      void Negate(Path& path)
      {
        typename Path::iterator it = path.begin();
        typename Path::iterator itEnd = path.end();
        for ( ; it != itEnd; ++it) {
            it->second = -it->second;
          }
      }

    private:
      FundamentalGroupPaths(const FundamentalGroupPaths&);
      FundamentalGroupPaths& operator=(const FundamentalGroupPaths&);


      void ComputeAcesMap()
      {
        for (size_t critical : boost::irange((size_t)0, _dvf.criticalCells())) {
          _acesMap.insert(typename AcesMap::value_type(_dvf.criticalCellId(critical), critical));
        }
      }


      const DiscreteVectorField& _dvf;
      PathsMap        _queenPaths;
      AcesMap         _acesMap;
      CAPD_CLASS_LOGGER;
    };


  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_FUNDAMENTALGROUPPATHS_H
