/////////////////////////////////////////////////////////////////////////////
/// @file HasseDiagram
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-04-03
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_HASSEDIAGRAM_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_HASSEDIAGRAM_H

// #include
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/properties.hpp>
#include <boost/range/iterator_range.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/lexical_cast.hpp>

namespace capd
{
  namespace complex
  {

    struct edge_coincidence_index_t
    {
      typedef boost::edge_property_tag kind;
    };

    struct vertex_dimension_t
    {
      typedef boost::vertex_property_tag kind;
    };


    template<typename DimT, typename ScalarT>
    class HasseDiagram
    {

    public:

      typedef DimT Dim;
      typedef ScalarT Scalar;

      typedef boost::adjacency_list<boost::vecS, boost::vecS, boost::bidirectionalS,
                                    boost::property<vertex_dimension_t, Dim>,
                                    boost::property<edge_coincidence_index_t, Scalar>
                                    > Graph;

      typedef typename boost::graph_traits<Graph>::edge_descriptor Edge;
      typedef typename boost::graph_traits<Graph>::vertex_descriptor Vertex;
      typedef typename boost::graph_traits<Graph>::vertex_iterator VertexIterator;
      typedef Vertex Index;

      struct CellCode
      {
        Index vertex;
        explicit CellCode(Index v): vertex(v)
        {}

        bool operator<(const CellCode& rhs) const { return vertex < rhs.vertex; }
        bool operator==(const CellCode& rhs) const { return vertex == rhs.vertex; }

        friend std::ostream& operator<<(std::ostream& out, const CellCode& cellCode)
        {
          out << cellCode.vertex;
          return out;
        }
      };

      struct EdgeTargeIndex
      {
        typedef Index result_type;
        const Graph& graph;

        EdgeTargeIndex(const Graph& graph): graph(graph) {}

        Index operator()(const Edge& edge) const
        {
          return boost::target(edge, graph);
        }
      };

      struct EdgeSourceIndex
      {
        typedef Index result_type;
        const Graph& graph;

        EdgeSourceIndex(const Graph& graph): graph(graph) {}

        Index operator()(const Edge& edge) const
        {
          return boost::source(edge, graph);
        }
      };

      typedef boost::iterator_range<boost::transform_iterator<Index (*)(const Vertex&), VertexIterator> > All;
      typedef boost::iterator_range<boost::transform_iterator<EdgeTargeIndex,
                                                              typename boost::graph_traits<Graph>::out_edge_iterator> > Bd;
      typedef boost::iterator_range<boost::transform_iterator<EdgeSourceIndex,
                                                              typename boost::graph_traits<Graph>::in_edge_iterator> > Cbd;


      explicit HasseDiagram(size_t size):
        _graph(size),
        _coincidence(boost::get(edge_coincidence_index_t(), _graph)),
        _dimension(boost::get(vertex_dimension_t(), _graph)),
        _maxDim(0)
      {}

      size_t size() const { return boost::num_vertices(_graph); }

      void set(const Vertex& v, Dim dim)
      {
        _dimension[v] = dim;
        if (dim > _maxDim) {
          _maxDim = dim;
        }
      }

      void connect(const Vertex& u, const Vertex& v, Scalar ci)
      {
        BOOST_ASSERT(_dimension[u] > _dimension[v]);
        Edge edge = boost::add_edge(u, v, _graph).first;
        _coincidence[edge] = ci;
      }

      Scalar coincidence(Index u, Index v) const
      {
        std::pair<Edge, bool> edge = boost::edge(u, v, _graph);
        if (edge.second) {
          auto m = boost::get(edge_coincidence_index_t(), (const Graph&)_graph);
          return boost::get(m, edge.first);
        }
        return Scalar(0);
      }

      Dim dim(Index index) const
      {
        auto m = boost::get(vertex_dimension_t(), _graph);
        return m[index];
      }
      Dim dim() const { return _maxDim; }

      CellCode operator[](Index vertex) const { return CellCode(vertex); }
      Index operator[](CellCode cellCode) const { return cellCode.vertex; }

      All all() const
      {
        auto vertices = boost::vertices(_graph);
        return All(boost::make_transform_iterator(vertices.first, boost::lexical_cast<Index, Vertex>),
                   boost::make_transform_iterator(vertices.second, boost::lexical_cast<Index, Vertex>));
      }

      Bd bd(Index v) const
      {
        auto edges = boost::out_edges(v, _graph);
        return Bd(boost::make_transform_iterator(edges.first, EdgeTargeIndex(_graph)),
                  boost::make_transform_iterator(edges.second, EdgeTargeIndex(_graph)));
      }

      Cbd cbd(Index v) const
      {
        auto edges = boost::in_edges(v, _graph);
        return Cbd(boost::make_transform_iterator(edges.first, EdgeSourceIndex(_graph)),
                   boost::make_transform_iterator(edges.second, EdgeSourceIndex(_graph)));
      }

    private:
      //      HasseDiagram(const HasseDiagram&);
      HasseDiagram& operator=(const HasseDiagram&);

      Graph _graph;
      typename boost::property_map<Graph, edge_coincidence_index_t>::type _coincidence;
      typename boost::property_map<Graph, vertex_dimension_t>::type _dimension;
      Dim _maxDim;
    };

  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_HASSEDIAGRAM_H
