/////////////////////////////////////////////////////////////////////////////
/// @file HomologyModel.h
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-01-01
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2014 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.

#ifndef CAPD_FILE_HOMOLOGYMODEL_H
#define CAPD_FILE_HOMOLOGYMODEL_H

#include "SComplexToFreeChainComplex.h"
#include "Chain.h"
#include "Coreduction.h"
#include "CoreductionPairInclusionChainModel.h"
#include "CoreductionPairProjectionChainModel.h"

#include <capd/homAlgebra/HomologySignature.h>
#include <capd/homAlgebra/homAlgFunctors.hpp>
#include <capd/redAlg/ReducibleFreeChainComplex.hpp>

#include <boost/typeof/typeof.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/bind.hpp>
#include <boost/assign.hpp>
#include <boost/function.hpp>
#include <boost/range/iterator_range.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/mpl/bool.hpp>

#include <deque>

namespace capd
{
namespace complex
{


  template<typename ComplexT, typename ScalarT, typename Int>
  class HomologyModel
  {
    struct ConvertToDomain;

  public:
    typedef ComplexT Complex;
    typedef ScalarT Scalar;
    typedef capd::complex::Chain<Complex, Scalar> Chain;
    typedef boost::iterator_range<typename std::vector<Chain>::const_iterator> ModelChains;

    typedef boost::transform_iterator<ConvertToDomain, typename ModelChains::const_iterator> ConvertToDomainIterator;
    //typedef boost::iterator_range<typename std::vector<Chain>::const_iterator> DomainChains;
    typedef boost::iterator_range<ConvertToDomainIterator> DomainChains;


    explicit HomologyModel(Complex& complex):
      _complex(complex), _homologyDimBound(0), _runCoreduction(true)
    {}

    void runCoreduction(bool v) { _runCoreduction = v; }

    Complex& complex() { return _complex; }
    const Complex& complex() const { return _complex; }
    size_t homologyDimBound() const { return _homologyDimBound; }

    size_t operator()()
    {
      if (_runCoreduction) {
	coreduction();
      }

      buildRFC();

      computeHomologyBase();
      return _homologyDimBound;
    }

    ModelChains modelChains(int dim) const
    {
      return ModelChains(_baseHomologyChains[dim].begin(), _baseHomologyChains[dim].end());
    }

    DomainChains domainChains(int dim) const
    {
      return DomainChains(ConvertToDomainIterator(_baseHomologyChains[dim].begin(),
						  ConvertToDomain(*this)),
			  ConvertToDomainIterator(_baseHomologyChains[dim].end(),
						  ConvertToDomain(*this)));
    }

    void operator>>(Chain& chain) const
    {
      namespace bl = boost::lambda;

      std::for_each(_chainInclusionMaps.begin(), _chainInclusionMaps.end(),
		    bl::bind(bl::_1, bl::var(chain)));
    }

    void operator<<(Chain& chain) const
    {
      namespace bl = boost::lambda;

      std::for_each(_chainProjectionMaps.begin(), _chainProjectionMaps.end(),
		    bl::bind(bl::_1, bl::var(chain)));
    }

    template<typename Function>
    void evaluate(Function function) const
    {
      for (size_t dim = 0; dim < _homologyDimBound; ++dim) {
	BOOST_FOREACH(const Chain& chain, modelChains(dim)) {
	  function(dim, chain);
	}
      }
    }

  private:
    typedef typename Complex::Cell Cell;
    typedef typename Complex::Id Id;
    typedef FreeModule<Id, capd::vectalg::Matrix<Scalar,0,0>, Int> FreeModuleType;
    typedef ReducibleFreeChainComplex<FreeModuleType> ReducibleFreeChainComplexType;
    typedef FreeChainComplex<FreeModuleType> FreeChainComplexType;

    typedef boost::function<void (Chain&)> ChainInclusionMap;
    typedef boost::function<void (Chain&)> ChainProjectionMap;

    struct ConvertToDomain
    {
      explicit ConvertToDomain(const HomologyModel& model):
	_model(model)
      {}

      typedef Chain result_type;
      Chain operator()(const Chain& chain) const
      {
	Chain result = chain;
	_model >> result;
	return result;
      }

      const HomologyModel& _model;
    };

    void coreduction()
    {
      namespace bl = boost::lambda;
      namespace b = boost;

      BOOST_AUTO(alg, (complex::makeCoreductionDefault<Scalar, Int>(_complex, b::assign::push_back(_coreductionPairs))));
      alg();

      BOOST_AUTO(inclModel, makeCoreductionPairInclusionChainModel<int>(_complex, _coreductionPairs));
      _chainInclusionMaps.push_back(b::bind(inclModel, _1));

      BOOST_AUTO(projModel, makeCoreductionPairProjectionChainModel<int>(_complex, _coreductionPairs));
      _chainProjectionMaps.push_back(b::bind(projModel, _1));
    }

    void buildRFC()
    {
      SComplexToFreeChainComplex<Complex, Scalar, Int, boost::mpl::true_> fccBuilder(const_cast<Complex&>(_complex)); //TODO ConvertToRFC requires non-const
      RFCComplexCR = fccBuilder();
    }

    void computeHomologyBase()
    {
      CRef<FreeChainComplexType> FCComplexCR(new FreeChainComplexType(RFCComplexCR()));
      CRef<QuotientGradedModule<FreeModuleType> > homologyGradedModule=HomAlgFunctors<FreeModuleType>::HomologyViaSmith(FCComplexCR);

      const int topDim = homologyGradedModule().TopDim();
      std::vector<std::vector<Int> > baseHomologyChainOrders(std::max(1, topDim + 1));
      _baseHomologyChains.resize(std::max(1, topDim + 1));

      for(int dim=0; dim <= topDim; ++dim) {
        homologyGradedModule().component(dim).exportBaseChains(_baseHomologyChains[dim], baseHomologyChainOrders[dim]);
      }

      BOOST_FOREACH(const SReductionPair<Complex>& pair, _coreductionPairs) {
	if (pair.dim == 0) {
	  _baseHomologyChains[0].push_back(Chain(_complex[pair.coface]));
	  baseHomologyChainOrders[0].push_back(0);
	}
      }

      _homologyDimBound = _baseHomologyChains.size();
      while (_homologyDimBound > 0 && _baseHomologyChains[_homologyDimBound - 1].size() == 0) _homologyDimBound--;
    }

    Complex& _complex;
    size_t _homologyDimBound;
    bool _runCoreduction;

    CRef<ReducibleFreeChainComplexType> RFCComplexCR;
    std::vector<std::vector<Chain> > _baseHomologyChains;
    std::deque<SReductionPair<Complex> > _coreductionPairs;

    std::vector<ChainInclusionMap> _chainInclusionMaps;
    std::vector<ChainProjectionMap> _chainProjectionMaps;
  };


}
}

#endif // CAPD_FILE_HOMOLOGYMODEL_H
