/////////////////////////////////////////////////////////////////////////////
/// @file SComplexKappaBuilderFromCubes.h
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-02-13
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2014 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.

#ifndef CAPD_FILE_SCOMPLEXKAPPABUILDERFROMCUBES_H
#define CAPD_FILE_SCOMPLEXKAPPABUILDERFROMCUBES_H

#include "capd/repSet/ElementaryCube.h"

#include <boost/tuple/tuple.hpp>

#include <vector>
#include <set>


namespace capd
{
namespace complex
{

  template<typename Id, typename Dim>
  class SComplexKappaBuilderFromCubes
  {
  public:
    typedef boost::tuple<Id, Id, int> KappaMapEntry;
    typedef std::vector<KappaMapEntry> KappaMap;
    typedef std::vector<Dim> Dims;

    SComplexKappaBuilderFromCubes(int dim):
      _dim(dim),
      _fullCubes(true)
    {}

    const KappaMap& kappaMap() const { return _kappaMap; }
    const Dims& dims() const { return _dims; }

    void addCell(const std::vector<int>& coords)
    {
      if (_fullCubes) {
	getId(ElementaryCube(coords));
      } else {
	int data[_dim];
	bool parity[_dim];

	for (int i=0; i < _dim; ++i) {
	  data[i] = (coords[i]/2);
	  parity[i] = (coords[i] % 2);
	}
	getId(ElementaryCube(data, parity, _dim));
      }
    }

    void operator()()
    {
      _kappaMap.clear();
      computedBoundaries.clear();

      CubeIds tmpCubeIds = cubeIds;
      for (CubeIds::iterator it = tmpCubeIds.begin(), end = tmpCubeIds.end(); it != end; ++it) {
	computeKappa(it->first, it->second);
      }

      const int size = cubeIds.size();
      _dims.resize(size);

      for (CubeIds::iterator it = cubeIds.begin(), end = cubeIds.end(); it != end; ++it) {
	_dims[it->second] = it->first.dimension();
      }
    }

  private:

    void computeKappa(const ElementaryCube& cube, int id)
    {
      if (computedBoundaries.find(id) != computedBoundaries.end())
	{
	  return;
	}

      computedBoundaries.insert(id);

      typedef std::map<ElementaryCube, int> Boundary;
      Boundary boundary;
      cube.boundary(boundary);

      for (Boundary::iterator bIt = boundary.begin(), bEnd = boundary.end(); bIt != bEnd; ++bIt)
	{
	  const int bId = getId(bIt->first);
	  _kappaMap.push_back(boost::make_tuple(id, bId, bIt->second));
	  computeKappa(bIt->first, bId);
	}

    }

    int getId(ElementaryCube cube)
    {
      CubeIds::iterator it = cubeIds.find(cube);
      if (it == cubeIds.end())
	{
	  it = cubeIds.insert(std::make_pair(cube, cubeIds.size())).first;
	}
      return it->second;
    }

    typedef std::map<ElementaryCube, int> CubeIds;
    std::set<int> computedBoundaries;

    Dim _dim;
    bool _fullCubes;
    CubeIds cubeIds;
    KappaMap _kappaMap;
    Dims _dims;

  };


}
}

#endif // CAPD_FILE_SCOMPLEXKAPPABUILDERFROMCUBES_H
