/////////////////////////////////////////////////////////////////////////////
/// @file SqueezedComplex
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-01-20
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_SQUEEZEDCOMPLEX_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_SQUEEZEDCOMPLEX_H

// #include
#include <capd/complex/BasicCellProxy.hpp>
#include "ColoredIteratorFamily.h"
#include <boost/optional.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/typeof/typeof.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <capd/complex/SqueezeIds.h>
#include <boost/iterator/iterator_adaptor.hpp>
#include <boost/foreach.hpp>
#include <boost/iterator/filter_iterator.hpp>
#include <iterator>
#include <capd/auxil/Logger.h>
#include <capd/complex/ComplexFunctors.h>

namespace capd
{
  namespace complex
  {

    template<typename ComplexT, typename OrgComplexT, typename OrgCellT, typename OrgConstCellT>
    class SqueezedCellImpl
    {
    public:
      typedef typename OrgCellT::Color Color;
      typedef typename ComplexT::Id Id;
      typedef typename ComplexT::Dim Dim;
      typedef typename ComplexT::CellCode CellCode;

      SqueezedCellImpl(): _complex(NULL), _orgComplex(NULL) {}
      explicit SqueezedCellImpl(ComplexT* complex, OrgComplexT* orgComplex,
                                OrgCellT orgCell,
                                Id id): _complex(complex), _orgComplex(orgComplex),
                                        _orgCell(orgCell),
                                        _id(id)
      {}

      operator SqueezedCellImpl<const ComplexT, const OrgComplexT, OrgConstCellT, OrgConstCellT>() const
      {
        return SqueezedCellImpl<const ComplexT, const OrgComplexT, OrgConstCellT, OrgConstCellT>(_complex, _orgComplex,
                                                                                                 _orgCell, _id);
      }

      const OrgCellT& getOrg() const { return _orgCell; }
      OrgCellT& getOrg() { return _orgCell; }
      const Id& getId() const { return _id; }
      const Color getColor() const { return _orgCell.getColor(); }
      void setColor(Color color) { return _orgCell.setColor(color); }

      CellCode code() const { return _orgCell.code(); }
      const Dim getDim() const { return _orgCell.getDim(); }

    private:
      ComplexT* _complex;
      OrgComplexT* _orgComplex;
      OrgCellT _orgCell;
      Id _id;
    };

    template<typename ComplexT, typename CellT, typename IteratorT>
    class SqueezedComplexIterator: public boost::iterator_adaptor<
      SqueezedComplexIterator<ComplexT, CellT, IteratorT>,
      IteratorT,
      CellT,
      boost::use_default,
      CellT>
    {
    public:
      explicit SqueezedComplexIterator():
        _complex(NULL)
      {}

      explicit SqueezedComplexIterator(ComplexT* complex, IteratorT iterator):
        SqueezedComplexIterator::iterator_adaptor_(iterator),
        _complex(complex)
      {}

    private:
      friend class boost::iterator_core_access;

      typename SqueezedComplexIterator::iterator_adaptor_::reference dereference() const
      {
        return _complex->cellFromSqueezeIt(this->base_reference());
      }

      ComplexT* _complex;
    };

    template<typename ComplexT, typename CellT, typename IteratorT>
    class SqueezedComplexIteratorFromOrg: public boost::iterator_adaptor<
      SqueezedComplexIteratorFromOrg<ComplexT, CellT, IteratorT>,
      IteratorT,
      CellT,
      boost::use_default,
      CellT>
    {
    public:
      explicit SqueezedComplexIteratorFromOrg():
        _complex(NULL)
      {}

      explicit SqueezedComplexIteratorFromOrg(ComplexT* complex, IteratorT iterator):
        SqueezedComplexIteratorFromOrg::iterator_adaptor_(iterator),
        _complex(complex)
      {}

    private:
      friend class boost::iterator_core_access;

      typename SqueezedComplexIteratorFromOrg::iterator_adaptor_::reference dereference() const
      {
        return _complex->cellFromOrgId(this->base_reference()->getId());
      }

      ComplexT* _complex;
    };


    template<typename ComplexT, typename CellT, typename ConstCellT, typename Iterator, typename OrgIteratorFamily>
    class SqueezedComplexIteratorFamily
    {
      struct DimPred
      {
        explicit DimPred(typename ComplexT::Dim dim): _dim(dim) {}
        bool operator()(const CellT& cell) const
        {
          return cell.getDim() == _dim;
        }
        typename ComplexT::Dim _dim;
      };


      struct NotSqueezedPred
      {
        explicit NotSqueezedPred(ComplexT* complex): _complex(complex) {}

        template<typename OrgCell>
        bool operator()(const OrgCell& cell) const
        {
          return _complex->notSqueezed(cell.getId());
        }
        ComplexT* _complex;
      };

      typedef Iterator AllIterator;
      typedef boost::filter_iterator<DimPred, AllIterator> DimIterator;
      typedef boost::filter_iterator<NotSqueezedPred, typename OrgIteratorFamily::BdCells::iterator> OrgBdIterator;
      typedef SqueezedComplexIteratorFromOrg<ComplexT, CellT, OrgBdIterator> BdIterator;
      typedef boost::filter_iterator<NotSqueezedPred, typename OrgIteratorFamily::CbdCells::iterator> OrgCbdIterator;
      typedef SqueezedComplexIteratorFromOrg<ComplexT, CellT, OrgCbdIterator> CbdIterator;


    public:
      typedef ComplexT Complex;
      typedef CellT Cell;
      typedef ConstCellT ConstCell;

      typedef boost::iterator_range<AllIterator> AllCells;
      typedef boost::iterator_range<DimIterator> DimCells;
      typedef boost::iterator_range<BdIterator> BdCells;
      typedef boost::iterator_range<CbdIterator> CbdCells;

      explicit SqueezedComplexIteratorFamily(ComplexT* complex, OrgIteratorFamily orgIteratorFamily):
        _complex(complex), _orgIteratorFamily(orgIteratorFamily)
      {}


      AllCells allCells()
      {
        return AllCells(_complex->begin(), _complex->end());
      }

      DimCells dimCells(const typename ComplexT::Dim& dim)
      {
        return DimCells(DimIterator(DimPred(dim), _complex->begin(), _complex->end()),
                        DimIterator(DimPred(dim), _complex->end(), _complex->end()));
      }

      BdCells bdCells(const ConstCell& cell)
      {
        BOOST_AUTO(bdCells, _orgIteratorFamily.bdCells(cell.getOrg()));
        return BdCells(BdIterator(_complex, OrgBdIterator(NotSqueezedPred(_complex), bdCells.begin(), bdCells.end())),
                       BdIterator(_complex, OrgBdIterator(NotSqueezedPred(_complex), bdCells.end(), bdCells.end())));
      }

      CbdCells cbdCells(const ConstCell& cell)
      {
        BOOST_AUTO(cbdCells, _orgIteratorFamily.cbdCells(cell.getOrg()));
        return CbdCells(CbdIterator(_complex, OrgCbdIterator(NotSqueezedPred(_complex), cbdCells.begin(), cbdCells.end())),
                       CbdIterator(_complex, OrgCbdIterator(NotSqueezedPred(_complex), cbdCells.end(), cbdCells.end())));
      }

    private:
      ComplexT* _complex;
      OrgIteratorFamily _orgIteratorFamily;
    };


    template<typename OrgComplex>
    class SqueezedComplex
    {

    public:
      typedef typename OrgComplex::Cell OrgCell;
      typedef typename OrgComplex::ConstCell OrgConstCell;
      typedef typename OrgComplex::CellCode OrgCellCode;
      typedef typename OrgComplex::Id OrgId;
      typedef typename OrgComplex::Dim OrgDim;
      typedef OrgId Id;
      typedef OrgCellCode CellCode;
      typedef OrgDim Dim;
      typedef typename OrgCell::Color Color;

      typedef SqueezedCellImpl<const SqueezedComplex, const OrgComplex, OrgConstCell, OrgConstCell> ConstCell;
      typedef SqueezedCellImpl<SqueezedComplex, OrgComplex, OrgCell, OrgConstCell> Cell;
      typedef SqueezedComplexIterator<SqueezedComplex, Cell, typename SqueezeIds<Id>::iterator> iterator;
      typedef SqueezedComplexIterator<const SqueezedComplex, ConstCell, typename SqueezeIds<Id>::const_iterator> const_iterator;


      typedef SqueezedComplexIteratorFamily<SqueezedComplex, Cell, ConstCell, iterator, typename OrgComplex::Iterators> Iterators;
      //typedef SqueezedComplexIteratorFamily<const SqueezedComplex, ConstCell, typename SqueezeIds<Id>::const_iterator, typename OrgComplex::ConstIterators> ConstIterators;
      typedef ColoredIteratorFamily<Iterators> ColoredIterators;

      explicit SqueezedComplex(OrgComplex& orgComplex):
        _orgComplex(orgComplex)
      {
        BOOST_AUTO(allCells, orgComplex.iterators(1).allCells());
        _squeezeIds = SqueezeIds<Id>(boost::make_transform_iterator(allCells.begin(), CellId<OrgCell>()),
                                     boost::make_transform_iterator(allCells.end(), CellId<OrgCell>()));
      }

      size_t size() const { return _squeezeIds.size(); }
      size_t cardinality()
      {
        BOOST_AUTO(allCells, iterators(1).allCells());
        return std::distance(allCells.begin(), allCells.end());
      }

      bool isClosed() const { return _orgComplex.isClosed(); }
      void setClosed(bool v) { _orgComplex.setClosed(v); }

      Dim getDim() const { return _orgComplex.getDim(); }

      iterator begin()
      {
        return iterator(this, _squeezeIds.begin());
      }

      iterator end()
      {
        return iterator(this, _squeezeIds.end());
      }

      const_iterator begin() const
      {
        return const_iterator(this, _squeezeIds.begin());
      }

      const_iterator end() const
      {
        return const_iterator(this, _squeezeIds.end());
      }

      Iterators iterators()
      {
        return Iterators(this, _orgComplex.iterators());
      }

      ColoredIterators iterators(Color color)
      {
        return ColoredIterators(iterators(), color);
      }

      Cell cellFromSqueezeIt(const typename SqueezeIds<Id>::iterator& it)
      {
        return Cell(this, &_orgComplex, _orgComplex[it->first], it->second);
      }

      ConstCell cellFromSqueezeIt(const typename SqueezeIds<Id>::const_iterator& it) const
      {
        return ConstCell(this, &_orgComplex, _orgComplex[it->first], it->second);
      }

      Cell cellFromOrgId(const OrgId& id)
      {
        return Cell(this, &_orgComplex, _orgComplex[id], _squeezeIds[id]);
      }

      ConstCell cellFromOrgId(const OrgId& id) const
      {
        return ConstCell(this, &_orgComplex, _orgComplex[id], _squeezeIds[id]);
      }

      bool notSqueezed(const OrgId& id) const
      {
        return _squeezeIds.contains(id);
      }

      Cell operator[](const Id& id) {
        return Cell(this, &_orgComplex, _orgComplex[_squeezeIds.reverse(id)], id);
      }

      ConstCell operator[](const Id& id) const {
        return ConstCell(this, &_orgComplex, _orgComplex[_squeezeIds.reverse(id)], id);
      }

      int coincidenceIndex(const Cell& cell1, const Cell& cell2) const
      {
        return _orgComplex.coincidenceIndex(cell1.getOrg(), cell2.getOrg());
      }

      SqueezeIds<Id>& squeezeIds() { return _squeezeIds; }
    private:
      SqueezedComplex(const SqueezedComplex&);
      SqueezedComplex& operator=(const SqueezedComplex&);

      OrgComplex& _orgComplex;
      SqueezeIds<Id> _squeezeIds;
      CAPD_CLASS_LOGGER;
    };


  }
}

namespace boost
{
    template<typename OrgComplex>
    struct range_mutable_iterator<capd::complex::SqueezedComplex<OrgComplex> >
    {
      typedef typename capd::complex::SqueezedComplex<OrgComplex>::iterator type;
    };

    template<typename OrgComplex>
    struct range_const_iterator<capd::complex::SqueezedComplex<OrgComplex> >
    {
      typedef typename capd::complex::SqueezedComplex<OrgComplex>::const_iterator type;
    };

}
#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_SQUEEZEDCOMPLEX_H
