/////////////////////////////////////////////////////////////////////////////
/// @file ZigZag
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-03-06
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_COMPLEX_ZIGZAG_H
#define CAPD_FILE_CAPDREDHOM_COMPLEX_ZIGZAG_H

// #include
#include "AcyclicOvergrowth.h"
#include "AcyclicShaving.h"
#include "CubicalComplexFullCubesBounds.h"
#include "FilteredComplex.h"
#include <boost/ref.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/lambda/lambda.hpp>
#include <capd/auxil/Logger.h>
#include <boost/assert.hpp>
#include <stdexcept>
#include <boost/utility/enable_if.hpp>
#include <capd/complex/ComplexTraits.h>


namespace capd
{
  namespace complex
  {

    template<typename Complex, size_t DIM, typename Enable=void>
    class ZigZag
    {

    public:
      ZigZag(Complex&, size_t)
      {}

      void operator()()
      {
        CAPD_WARN_TMPL("Computing ZigZag for this type of complex not implemented");
      }
    };


    template<typename Complex, size_t DIM>
    class ZigZag<Complex, DIM, typename boost::enable_if<IsRegularCubical<Complex> >::type>
    {

    public:
      explicit ZigZag(Complex& complex, size_t dim=DIM):
        _complex(complex), _dim(dim)
      {
        BOOST_ASSERT(DIM == _dim);
      }

      void operator()()
      {
        CAPD_TRACE_TMPL("Computing ZigZag in (static) dimension: " << DIM);

        auto bounds = makeCubicalComplexFullCubesBounds(_complex);
        namespace bl = boost::lambda;
        auto boundedComplex = makeFilteredComplex(_complex, bl::bind(bl::var(bounds),
                                                                     bl::_1));
        bool changed = true;

        while (changed) {
          const size_t added = DoAcyclicOvergrowth<DIM>(boundedComplex);
          const size_t reduced = DoAcyclicShaving<DIM>(boundedComplex);
          changed = bounds.adjust(boundedComplex);
          CAPD_TRACE_TMPL("reduced: " << reduced << " added: " << added << " bounds changed:" << changed);
        }

        for (auto cell : _complex.iterators(1).allCells()) {
          if (cell.getDim() != DIM) {
            cell.setColor(2);
          }
        }

        _complex.fillWithBoundaries();
      }

    private:
      Complex& _complex;
      size_t _dim;
      CAPD_CLASS_LOGGER;
    };


    template<size_t DIM, typename Complex>
    void DoZigZag(Complex& complex)
    {
      ZigZag<Complex, DIM> algo(complex);
      algo();
    }

    template<typename Complex>
    void DoZigZag(Complex& complex, size_t dim)
    {
      if (dim == 2) {
        ZigZag<Complex, 2> nested(complex, dim);
        nested();
      } else if (dim == 3) {
        ZigZag<Complex, 3> nested(complex, dim);
        nested();
      } else {
        throw std::runtime_error("Not implemented. For now an implementation only for dimension 2 and 3");
      }
    }

  }
}

#endif // CAPD_FILE_CAPDREDHOM_COMPLEX_ZIGZAG_H
