/// @addtogroup homologicalAlgebra
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file HomologySignature.h
///
/// @author Marian Mrozek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2006 by the CAPD Group.
//
// This file constitutes a part of the CAPD library,
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details.

#if !defined(_HOMOLOGYSIGNATURE_H_)
#define _HOMOLOGYSIGNATURE_H_
#include <vector>
#include <sstream>
#include <string>

#include "capd/homAlgebra/FGAGrpSignature.h"

template<typename ScalarType, typename IntType = int>
class HomologySignature : public std::vector<FGAGrpSignature<ScalarType, IntType> >{
  public:

  typedef std::vector<std::pair<IntType, std::vector<IntType> > > Signature;

  HomologySignature(){}
  HomologySignature(std::vector<FGAGrpSignature<ScalarType, IntType> >& A_FGAGrSignVect){
    *(std::vector<FGAGrpSignature<ScalarType, IntType> >*)(this)=A_FGAGrSignVect;
  }
  HomologySignature(const std::vector<IntType>& A_Betti){
    for(int i=0;i<(int)A_Betti.size();++i) {
      this->push_back(FGAGrpSignature<ScalarType, IntType> (A_Betti[i]));
    }
    shrink();
  }
  int topDim() const { return this->size()-1;}

  HomologySignature& operator+=(const HomologySignature& hs2){
    unsigned int i=0;
    for(i=0;i<this->size();++i){
      if(i<hs2.size()) (*this)[i]+=hs2[i];
    }
    for(;i<hs2.size();++i){
      this->push_back(hs2[i]);
    }
    return *this;
  }

  friend std::ostream& operator<<(std::ostream& out, const HomologySignature& fga){
    if(fga.topDim()<0){
      out << "  0" << std::endl;
    }else{
      for(int i=0;i<(int)fga.size();++i){
        out << "  H_" << i << " = " << fga[i] << std::endl;
      }
    }
    return out;
  }

  void shrink(){
    if(this->empty()) return;
    typename std::vector<FGAGrpSignature<ScalarType, IntType> >::iterator it=this->end();
    while((--it)>=this->begin() and it->isTrivial()) this->erase(it);
  }

  ScalarType bettiNumber(int i) const {
    if(i>=0 and i<=topDim()) return (*this)[i].rank();
    else return 0;
  }

  std::string bettiVector() const{
    std::ostringstream o;
    for(unsigned int i=0;i<this->size();++i){
      if(i) o << ",";
      o << (*this)[i].rank();
    }
    return o.str();
  }

  std::vector<IntType> ranks() const{
    std::vector<IntType> result(this->size());
    size_t nonZero = 0;

    for(size_t i=0;i<this->size();++i){
      result[i] = (*this)[i].rank();
      if (result[i] != 0) {
        nonZero = i + 1;
      }
    }

    result.resize(nonZero);
    return result;
  }

  Signature raw() const
  {
    Signature signature(this->size());

    for(unsigned int i=0;i<this->size();++i){
      signature[i] = (*this)[i].rawSignature();
    }

    return signature;
  }

};
#endif //_HOMOLOGYSIGNATURE_H_
/// @}
