/////////////////////////////////////////////////////////////////////////////
/// @file IteratedMultiValuedFunctor
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-03-10
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_HOMAUX_ITERATEDMULTIVALUEDFUNCTOR_H
#define CAPD_FILE_CAPDREDHOM_HOMAUX_ITERATEDMULTIVALUEDFUNCTOR_H

// #include

#include "MultiValuedTransformIterator.h"

#include <boost/mpl/for_each.hpp>
#include <boost/mpl/range_c.hpp>
#include <stdexcept>
#include <boost/range/any_range.hpp>
#include <iterator>
#include <vector>
#include <capd/auxil/ignoreUnused.h>
#include <capd/auxil/Logger.h>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>

namespace capd
{
  namespace homAux
  {

    template<size_t DEPTH, typename Func>
    class IteratedMultiValuedFunctor
    {

    public:
      typedef typename Func::argument_type argument_type;
      typedef MultiValuedTransformRange<IteratedMultiValuedFunctor<DEPTH - 1, Func>,
                                        typename Func::result_type::const_iterator> result_type;

      IteratedMultiValuedFunctor(Func func = Func()): _func(func)
      {}

      IteratedMultiValuedFunctor(size_t depth, Func func = Func()): _func(func)
      {
        auxil::ignoreUnused(depth);
        BOOST_ASSERT(depth == DEPTH); // we nned depth for signature compatibility
      }

      result_type operator()(const argument_type& arg)
      {
        auto x = _func(arg);
        return result_type(x.begin(), x.end(), IteratedMultiValuedFunctor<DEPTH - 1, Func>(_func));
      }

    private:
      Func _func;
    };

    template<typename Func>
    class IteratedMultiValuedFunctor<1, Func>
    {

    public:
      typedef typename Func::argument_type argument_type;
      typedef typename Func::result_type result_type;

      IteratedMultiValuedFunctor(Func func = Func()): _func(func)
      {}

      IteratedMultiValuedFunctor(size_t depth, Func func = Func()): _func(func)
      {
        auxil::ignoreUnused(depth);
        BOOST_ASSERT(depth == 1); // we nned depth for signature compatibility
      }

      result_type operator()(const argument_type& arg)
      {
        return _func(arg);
      }

    private:
      Func _func;
    };

    template<typename Func>
    class IteratedMultiValuedFunctor<0, Func>
    {

    public:
      typedef typename Func::argument_type argument_type;
      typedef std::vector<typename Func::result_type::value_type> result_type;

      static const size_t MAX_DEPTH = 10;

      IteratedMultiValuedFunctor(size_t depth=0, Func func = Func()):
        _depth(depth), _func(func)
      {
        if (_depth > MAX_DEPTH) {
          throw std::runtime_error("Too big recursive depth call");
        }
      }

      result_type operator()(const argument_type& arg)
      {
        result_type result;
        bool called = false;
        CallStatic callStatic = {_func, arg, result, _depth, called};

        namespace bl=boost::lambda;

        typedef boost::mpl::range_c<size_t, 1, MAX_DEPTH> Range;
        boost::mpl::for_each<Range>(callStatic);
        BOOST_ASSERT(called);

        return result;
      }

    private:

      struct CallStatic
      {
        template<typename Depth>
        void doit(Depth)
        { // do it in a separeted function, so no stack allocation. It is a problem, because mpl::for_each is recursive
          BOOST_ASSERT(!_called);
          IteratedMultiValuedFunctor<Depth::value, Func> nested(_func);
          auto result = nested(_arg);
          _result.insert(_result.end(), result.begin(), result.end());
          _called = true;
        }

        template<typename Depth>
        void operator()(Depth d)
        {
          if (_depth == Depth::value) {
            doit(d);
          }
        }

        Func& _func;
        const argument_type& _arg;
        result_type& _result;
        size_t _depth;
        bool& _called;
      };

      size_t _depth;
      Func _func;
    };


  }
}


#endif // CAPD_FILE_CAPDREDHOM_HOMAUX_ITERATEDMULTIVALUEDFUNCTOR_H
