/////////////////////////////////////////////////////////////////////////////
/// @file MergeMaps
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-03-30
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_HOMAUX_MERGEMAPS_H
#define CAPD_FILE_CAPDREDHOM_HOMAUX_MERGEMAPS_H

// #include
#include <boost/function_output_iterator.hpp>
#include <algorithm>
#include <boost/lambda/bind.hpp>
#include <deque>
#include <utility>
#include <boost/container/flat_map.hpp>
#include <boost/container/container_fwd.hpp>
#include <boost/type_traits/remove_const.hpp>

namespace capd
{
  namespace homAux
  {

    template<typename Range1, typename Range2, typename MergeOp>
    class MergeMaps
    {
    public:
      typedef typename boost::remove_const<typename Range1::value_type::first_type>::type Key;
      typedef typename boost::remove_const<typename Range1::value_type::second_type>::type Value;

      typedef boost::container::flat_map<Key, Value> result_type;
      typedef std::deque<std::pair<Key, Value> > Container;

      struct MergeWithLast
      {
        MergeWithLast(Container& tmp, MergeOp& mergeOp):
          _tmp(tmp), _mergeOp(mergeOp)
        {}

        void operator()(const std::pair<Key, Value>& v) const
        {
          if (_tmp.empty()) {
            _tmp.push_back(v);
          } else {
            BOOST_ASSERT(_tmp.back().first <= v.first);
            if (_tmp.back().first == v.first) {
              _tmp.back().second = _mergeOp(_tmp.back().second, v.second);
            } else {
              _tmp.push_back(v);
            }
          }
        }
        Container& _tmp;
        MergeOp& _mergeOp;
      };

      struct Compare
      {
        template<typename P1, typename P2>
        bool operator()(const P1& x, const P2& y) const
        {
          return x.first < y.first;
        }
      };

      MergeMaps(const Range1& rng1, const Range2& rng2, MergeOp mergeOp=MergeOp())
        :_rng1(rng1), _rng2(rng2), _mergeOp(mergeOp)
      {}

      result_type operator()()
      {
        namespace bl = boost::lambda;
        Container tmp;
        typedef typename Range1::value_type Rng1V;
        typedef typename Range2::value_type Rng2V;

        std::merge(_rng1.begin(), _rng1.end(),
                   _rng2.begin(), _rng2.end(),
                   boost::make_function_output_iterator(MergeWithLast(tmp, _mergeOp)),
                   Compare()
                   );


        return result_type(boost::container::ordered_unique_range,
                           tmp.begin(), tmp.end());
      }

    private:
      const Range1& _rng1;
      const Range2& _rng2;
      MergeOp _mergeOp;
    };

    template<typename Range1, typename Range2, typename MergeOp>
    typename MergeMaps<Range1, Range2, MergeOp>::result_type
    DoMergeMaps(const Range1& rng1, const Range2& rng2, MergeOp mergeOp=MergeOp())
    {
      MergeMaps<Range1, Range2, MergeOp> algo(rng1, rng2, mergeOp);
      return algo();
    }

  }
}

#endif // CAPD_FILE_CAPDREDHOM_HOMAUX_MERGEMAPS_H
