/////////////////////////////////////////////////////////////////////////////
/// @file MultiValuedTransformIterator
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-03-10
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#ifndef CAPD_FILE_CAPDREDHOM_HOMAUX_MULTIVALUEDTRANSFORMITERATOR_H
#define CAPD_FILE_CAPDREDHOM_HOMAUX_MULTIVALUEDTRANSFORMITERATOR_H

// #include
#include <boost/range/iterator_range_core.hpp>
#include <boost/utility/result_of.hpp>
#include <boost/optional.hpp>

#undef BOOST_CONCEPT_ASSERT
#define BOOST_CONCEPT_ASSERT(Model)

namespace capd
{
  namespace homAux
  {

    template<typename MultiValuedFunction, typename Iterator>
    class MultiValuedTransformIterator:
      public boost::iterator_facade<
      MultiValuedTransformIterator<MultiValuedFunction, Iterator>,
      typename boost::result_of<MultiValuedFunction(typename std::iterator_traits<Iterator>::reference)>::type::value_type,
      boost::forward_traversal_tag,
      typename boost::result_of<MultiValuedFunction(typename std::iterator_traits<Iterator>::reference)>::type::value_type
      >
    {

      typedef typename boost::result_of<MultiValuedFunction(typename std::iterator_traits<Iterator>::reference)>::type
      MultiValue;
      typedef typename MultiValue::iterator MultiValueIterator;

    public:
      typedef MultiValuedFunction Function;

      MultiValuedTransformIterator(): _isEnd(true), _distance(0)
      {
      }

      MultiValuedTransformIterator(const Iterator& it, const Iterator& itEnd, MultiValuedFunction func = MultiValuedFunction()):
        _it(it), _itEnd(itEnd),
        _isEnd(_it == _itEnd),
        _func(func),
        _multiValue(),
        _multiValueIt(),
        _distance(0)
      {
        if (!_isEnd) {
          _multiValue = _func(*_it);
          _multiValueIt = _multiValue->begin();
          makeSure();
        }
      }

      MultiValuedTransformIterator(const MultiValuedTransformIterator& other):
        _it(other._it), _itEnd(other._itEnd),
        _isEnd(other._isEnd),
        _func(other._func),
        _multiValue(other._multiValue),
        _multiValueIt(),
        _distance(other._distance)
      {
        if (_multiValue) {
          _multiValueIt = _multiValue->begin();
        }
        if (!_isEnd) {
          std::advance(*_multiValueIt, _distance);
        }
      }

      MultiValuedTransformIterator& operator=(const MultiValuedTransformIterator& other)
      {
        _it = other._it;
        _itEnd = other._itEnd;
        _isEnd = other._isEnd;
        _func = other._func;
        _multiValue = other._multiValue;

        if (_multiValue) {
          _multiValueIt = _multiValue->begin();
        }

        _distance = other._distance;
        if (!_isEnd) {
          std::advance(*_multiValueIt, _distance);
        }
        return *this;
      }

    private:
      friend class boost::iterator_core_access;

      bool equal(MultiValuedTransformIterator const& other) const
      {
        return (true == _isEnd && true == other._isEnd) ||
          (false == _isEnd && false == other._isEnd &&
           other._it == _it
           && other._distance == _distance);
      }

      void makeSure()
      {
        while (_it != _itEnd && *_multiValueIt == _multiValue->end()) {
          ++_it;

          if (_it != _itEnd) {
            _multiValue = _func(*_it);
            _multiValueIt = _multiValue->begin();
            _distance = 0;
          } else {
            break;
          }
        }

        _isEnd = (_it == _itEnd);
      }

      void increment()
      {
        if (*_multiValueIt != _multiValue->end()) {
          ++(*_multiValueIt);
          ++_distance;
        }

        makeSure();
      }

      typename MultiValue::value_type dereference() const
      {
        return **_multiValueIt;
      }

      Iterator _it;
      Iterator _itEnd;
      bool _isEnd;
      MultiValuedFunction _func;
      boost::optional<MultiValue> _multiValue;
      boost::optional<MultiValueIterator> _multiValueIt;
      size_t _distance;
    };

    template<typename MultiValuedFunction, typename Iterator>
    class MultiValuedTransformRange: public boost::iterator_range<MultiValuedTransformIterator<MultiValuedFunction, Iterator> >
    {
      typedef MultiValuedTransformIterator<MultiValuedFunction, Iterator> MultiIt;
      typedef boost::iterator_range<MultiIt> Base;

    public:
      typedef MultiValuedFunction Function;

      MultiValuedTransformRange() {}

      MultiValuedTransformRange(const Iterator& it, const Iterator& itEnd, MultiValuedFunction func = MultiValuedFunction()):
        Base(MultiIt(it, itEnd, func), MultiIt(itEnd, itEnd, func))
      {}

    };

    template<typename MultiValuedFunction, typename Iterator>
    MultiValuedTransformRange<MultiValuedFunction, Iterator>
    makeMultiValuedTransformRange(Iterator begin, Iterator end, MultiValuedFunction func=MultiValuedFunction())
    {
      return MultiValuedTransformRange<MultiValuedFunction, Iterator>(begin, end, func);
    }

  }
}

#endif // CAPD_FILE_CAPDREDHOM_HOMAUX_MULTIVALUEDTRANSFORMITERATOR_H
