/// @addtogroup multiEngHom
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file CrbHomMD.h
///
/// @author Marian Mrozek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) Marian Mrozek 2005-2010
//
// This file constitutes a part of the CAPD library,
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details.

#if !defined(_CRB_HOM_T_H_)
#define _CRB_HOM_T_H_

#include <iostream>
#include <fstream>
#include <sstream>
#include <exception>
#include <new>
using namespace std;

#include "capd/vectalg/MatrixSlice.h"
#include "capd/matrixAlgorithms/intMatrixAlgorithms.hpp"

//#include "capd/auxil/embeddingDim.h"
#include "capd/auxil/ofstreamcout.h"
#include "capd/auxil/Stopwatch.h"
#include "capd/auxil/memSize.h"
#include "capd/auxil/CRef.h"
#include "capd/auxil/stringOstream.h"

using namespace capd;
using namespace matrixAlgorithms;


#include "capd/homAlgebra/homologicalAlgebra.hpp"
#include "capd/cubSet/readCubCellSet.hpp"
#include "capd/homAlgebra/homAlgFunctors.hpp"
#include "capd/cubSet/cubSetFunctors.hpp"
#include "capd/auxil/EmbDimException.h"



extern ofstreamcout fcout;


#include "capd/cubSet/acyclicConfigs.hpp"

#include "capd/RedHom/ReductionPairT.h"
#include "capd/cubSet/CubFaceIndex.h"
#include "capd/RedHom/CubCellSetFiltrT.hpp"

#include "capd/homAlgebra/FreeChainComplex.hpp"
#include "capd/redAlg/ReducibleFreeChainComplex.hpp"
#include "capd/repSet/ECellMDCodeT.h"



template <int dim>
struct CrbHom{

  typedef ECellMDCodeT<cluster,dim> CubFaceIndexType;
  typedef ECellMDCodeT<cluster,dim> ElementaryCellType;

  typedef EuclBitSetT<BitSetT<BitmapT<cluster> >,dim> EuclBitSetType;
  typedef CubSetT<EuclBitSetType> SCubSet;
  typedef CubCellSetT<EuclBitSetType,ReductionPairT<CubFaceIndexType> > SCubCelSet;

  typedef capd::vectalg::Matrix<int,0,0> MatrixType;
  typedef MatrixType::ColumnVectorType ColumnVectorType;
  typedef MatrixType::ScalarType ScalarType;
  typedef FreeModule<ElementaryCellType,MatrixType> FreeModuleType;
  typedef FreeChainComplex<FreeModule<ElementaryCellType,MatrixType> > FreeChainComplexType;
  typedef ChainT<std::map<ElementaryCellType,int> > ChainType;

  static CrbHom selector;

  // This method runs the particular homology engine for the cubical set in the prescribed file
  // in the prescribed dimension given as the dim template parameter
  void operator()(const char* A_fileName, int A_verbose){
    if(A_verbose>=3) fcout.turnOn();
    else fcout.turnOff();

    CRef<SCubCelSet> cubCellSetCR;
    try{
      cubCellSetCR=readCubCellSet<SCubSet,SCubCelSet>(A_fileName);
    }catch(EmbDimException){
      throw EmbDimException("Incorrect embDim");
    }
    if(cubCellSetCR().embDim()!=dim){
      throw EmbDimException("Incorrect embDim");
    }

    Stopwatch swTot;
    CubCellSetFiltrT<SCubCelSet,FreeChainComplexType> homFiltr(cubCellSetCR);
    std::cout << "Computed homology is: \n\n" << homFiltr  << std::endl;
    if(A_verbose) fcout << "Computation completed in " << swTot  << std::endl;
  }


}; // CrbHom<dim>

template<int dim>
CrbHom<dim> CrbHom<dim>::selector;

template <>
struct CrbHom<0>{
  CrbHom<2>& ch2;
  CrbHom<3>& ch3;
  CrbHom<4>& ch4;
  CrbHom<5>& ch5;
  CrbHom<6>& ch6;
  CrbHom<7>& ch7;
  CrbHom<8>& ch8;
  CrbHom<9>& ch9;
  CrbHom<10>& ch10;

  CrbHom():
    ch2(CrbHom<2>::selector),
    ch3(CrbHom<3>::selector),
    ch4(CrbHom<4>::selector),
    ch5(CrbHom<5>::selector),
    ch6(CrbHom<6>::selector),
    ch7(CrbHom<7>::selector),
    ch8(CrbHom<8>::selector),
    ch9(CrbHom<9>::selector),
    ch10(CrbHom<10>::selector)
  {
  }

  void failure(){
    string s;
    s << " CRb is not available for this dimension.\n";
    throw EmbDimException(s);
  }

  // This method runs the particular homology engine for the cubical set in the prescribed file
  // trying various dimensions
  void operator()(const char* A_fileName, int A_verbose){
    for(int i=2;i<=12;++i){
      try{
        switch(i){
          case 2: ch2(A_fileName,A_verbose);
                  return;
          case 3: ch3(A_fileName,A_verbose);
                  return;
          case 4: ch4(A_fileName,A_verbose);
                  return;
          case 5: ch5(A_fileName,A_verbose);
                  return;
          case 6: ch6(A_fileName,A_verbose);
                  return;
          case 7: ch7(A_fileName,A_verbose);
                  return;
          case 8: ch8(A_fileName,A_verbose);
                  return;
          case 9: ch9(A_fileName,A_verbose);
                  return;
          case 10: ch10(A_fileName,A_verbose);
                  return;
          default:;
        }
      }catch(EmbDimException& e){
        if(e.what() != std::string("Incorrect embDim")) throw;
      };
    }
    failure();
  }


};// CrbHom<0>

#endif //_CRB_HOM_T_H_
/// @}
