/////////////////////////////////////////////////////////////////////////////
/// @file BuildVietorisRips.h
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-06-25
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2014 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.

#ifndef CAPD_FILE_BUILDVIETORISRIPS_H
#define CAPD_FILE_BUILDVIETORISRIPS_H

#include "AbstractPointCloud.h"

#include <boost/function.hpp>

namespace capd
{
  namespace persistence
  {

    class BuildVietorisRips
    {

    public:
      typedef AbstractPointCloud::Id Id;
      typedef std::vector<Id> Simplex;

      explicit BuildVietorisRips(const AbstractPointCloud& cloud, size_t maxDim):
        _cloud(cloud),
        _maxDim(maxDim),
        _neighbours(cloud.size())
      {
        for (Id idx = 0, size = _cloud.size(); idx < size; ++idx) {
          _cloud.neighbours(idx, _neighbours[idx]);
          std::sort(_neighbours[idx].begin(), _neighbours[idx].end());
        }
      }

      virtual ~BuildVietorisRips()
      {}

      // Callback signature (face, bdface, weight of face)
      template<typename Callback>
      void operator()(Callback callback)
      {
        Simplex simplex(1);
        Simplex neighb;
        const Simplex empty;
        for (Id idx = 0, size = _cloud.size(); idx < size; ++idx) {
          lowerNeighbours(idx, neighb);
          simplex[0] = idx;
          callback(simplex, empty, (double)0);
          addCofaces(simplex, neighb, callback, (double)0);
        }

      }

    private:
      void lowerNeighbours(Id vertex, Simplex& result) const
      {
        Simplex::const_iterator end = std::lower_bound(_neighbours[vertex].begin(), _neighbours[vertex].end(), vertex);
        result.assign(_neighbours[vertex].begin(), end);
      }

      template<typename Callback>
      void addCofaces(const Simplex& simplex, const Simplex& neighb, Callback& callback, double weight)
      {
        BOOST_ASSERT(simplex.size() >= 1);
        if (_maxDim > 0 && simplex.size() > _maxDim) {
          return;
        }

        std::vector<Id> newNeighb;
        std::vector<Id> neighbIntersection;
        Simplex newSimplex;

        newSimplex.reserve(simplex.size() + 1);
        newSimplex.assign(simplex.begin(), simplex.end());
        newSimplex.push_back(0);

        for (Id idx = 0, size = neighb.size(); idx < size; ++idx) {
          const Id v = neighb[idx];
          lowerNeighbours(v, newNeighb);
          newSimplex.back() = v;

          double newWeight = weight;
          for (size_t idu = 0; idu < newSimplex.size(); ++idu) {
            newWeight = std::max(newWeight, _cloud.distance(newSimplex[idu], v));
          }

          neighbIntersection.resize(std::min(newNeighb.size(), neighb.size()));
          std::vector<Id>::const_iterator intersectionEnd = std::set_intersection(neighb.begin(), neighb.end(),
                                                                                  newNeighb.begin(), newNeighb.end(),
                                                                                  neighbIntersection.begin());
          neighbIntersection.resize(intersectionEnd - neighbIntersection.begin());

          callback(newSimplex, simplex, newWeight);
          addCofaces(newSimplex, neighbIntersection, callback, newWeight);
        }

      }

      const AbstractPointCloud& _cloud;
      const size_t _maxDim;
      std::vector<Simplex> _neighbours;
    };


  }
}

#endif // CAPD_FILE_BUILDVIETORISRIPS_H
