/////////////////////////////////////////////////////////////////////////////
/// @file ComputePersistentHomology
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-09-17
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////


#ifndef CAPD_FILE_CAPDREDHOM_PERSISTENCE_COMPUTEPERSISTENCEHOMOLOGY_HPP
#define CAPD_FILE_CAPDREDHOM_PERSISTENCE_COMPUTEPERSISTENCEHOMOLOGY_HPP

#include <capd/persistence/ComputePersistentHomology.h>

#include <capd/persistence/PHATInterface.h>
#include <capd/persistence/BuildFiltration.h>
#include <capd/auxil/Logger.h>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/typeof/typeof.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/lambda/construct.hpp>

namespace capd
{
  namespace persistence
  {

    namespace b = boost;
    namespace bl = boost::lambda;

    template<typename FilteredComplexT>
    bool ComputePersistentHomology<FilteredComplexT>::enabled()
    {
      return PHATInterface::enabled();
    }

    template<typename FilteredComplexT>
    void ComputePersistentHomology<FilteredComplexT>::operator()()
    {
      if (!PHATInterface::enabled()) {
        throw std::logic_error("Cannot compute Persistent Homology without PHAT. Please configure and build with --with-phat");
      }

      typedef capd::persistence::BuildFiltration<FilteredComplex> BuildFiltration;
      BuildFiltration buildFiltration(_filteredComplex);
      buildFiltration();

      std::vector<PHATInterface::FiltrationElement> filtration(buildFiltration.size());

      for (size_t idx = 0, size = filtration.size(); idx != size; ++idx) {
        filtration[idx].first =  buildFiltration.cellDim(idx);
        std::vector<size_t> bd = buildFiltration[idx];
        filtration[idx].second.assign(bd.begin(), bd.end());
      }

      PHATInterface::Pairs pairs;
      PHATInterface::Cycles cycles;

      PHATInterface phat;
      phat(PHATInterface::Filtration(filtration.begin(), filtration.end()),
           filtration.size(), pairs, cycles);

      typedef PHATInterface::Pairs::value_type Pair;

      BOOST_AUTO(cellId, bl::bind(&BuildFiltration::cellId, bl::var(buildFiltration), bl::_1));
      BOOST_AUTO(cellColor, bl::bind(&FilteredComplex::Cell::color,
                                     bl::bind(//&FilteredComplex::operator[],
                                              static_cast<typename FilteredComplex::Cell(FilteredComplex::*)(typename FilteredComplex::Id)>(&FilteredComplex::operator[]),
                                              bl::var(_filteredComplex), bl::_1)));

      PHATInterface::Pairs::iterator pairsNewEnd = std::remove_if(pairs.begin(), pairs.end(),
                                                                  bl::bind(bl::var(cellColor),
                                                                           bl::bind(bl::var(cellId), bl::bind(&Pair::second, bl::_1)))
                                                                  ==
                                                                  bl::bind(bl::var(cellColor),
                                                                           bl::bind(bl::var(cellId), bl::bind(&Pair::first, bl::_1))));

      BOOST_AUTO(cellIdPair, bl::bind(bl::constructor<std::pair<size_t, size_t> >(),
                                      bl::bind(bl::var(cellId), bl::bind(&Pair::first, bl::_1)),
                                      bl::bind(bl::var(cellId), bl::bind(&Pair::second, bl::_1))
                                      ));

      _pairs.assign(b::make_transform_iterator(pairs.begin(), cellIdPair),
                    b::make_transform_iterator(pairsNewEnd, cellIdPair)
                    );

      _cycles.assign(b::make_transform_iterator(cycles.begin(), cellId),
                     b::make_transform_iterator(cycles.end(), cellId));
    }

    template<typename FilteredComplexT>
    typename ComputePersistentHomology<FilteredComplexT>::Pairs ComputePersistentHomology<FilteredComplexT>::pairs(typename FilteredComplex::Dim dim) const
    {
      Pairs pairs(_pairs.begin(), _pairs.end());
      const typename FilteredComplex::Dim dim_1 = -1;;

      if (dim != dim_1) {
        BOOST_AUTO(cellDim, bl::bind(&FilteredComplex::Cell::dim,
                                     bl::bind(//&FilteredComplex::operator[],
                                              static_cast<typename FilteredComplex::Cell(FilteredComplex::*)(typename FilteredComplex::Id)>(&FilteredComplex::operator[]),
                                              bl::var(_filteredComplex), bl::_1)));
        BOOST_AUTO(pairDim, bl::bind(bl::var(cellDim), bl::bind(&Pairs::value_type::first, bl::_1)));

        return Pairs(pairs.begin(),
                     std::remove_if(pairs.begin(), pairs.end(),
                                    pairDim != bl::constant(dim)));
      } else {
        return pairs;
      }
    }

    template<typename FilteredComplexT>
    typename ComputePersistentHomology<FilteredComplexT>::Cycles ComputePersistentHomology<FilteredComplexT>::cycles(typename FilteredComplex::Dim dim) const
    {
      Cycles cycles(_cycles.begin(), _cycles.end());
      const typename FilteredComplex::Dim dim_1 = -1;;

      if (dim != dim_1) {
        BOOST_AUTO(cellDim, bl::bind(&FilteredComplex::Cell::dim,
                                     bl::bind(//&FilteredComplex::operator[],
                                              static_cast<typename FilteredComplex::Cell(FilteredComplex::*)(typename FilteredComplex::Id)>(&FilteredComplex::operator[]),
                                              bl::var(_filteredComplex), bl::_1)));

        return Cycles(cycles.begin(),
                      std::remove_if(cycles.begin(), cycles.end(),
                                     cellDim != bl::constant(dim)));
      } else {
        return cycles;
      }
    }

    template<typename FilteredComplexT>
    typename ComputePersistentHomology<FilteredComplexT>::Diagram ComputePersistentHomology<FilteredComplexT>::diagram(typename FilteredComplex::Dim dim) const
    {
      using namespace boost::lambda;
      const Pairs pairs = this->pairs(dim);
      const Cycles cycles = this->cycles(dim);

      Diagram diagram;
      diagram.reserve(pairs.size() + cycles.size());

      BOOST_AUTO(cellColor, bl::bind(&FilteredComplex::Cell::color,
                                     bl::bind(//&FilteredComplex::operator[],
                                              static_cast<typename FilteredComplex::Cell(FilteredComplex::*)(typename FilteredComplex::Id)>(&FilteredComplex::operator[]),
                                              bl::var(_filteredComplex), bl::_1)));
      BOOST_AUTO(pairColor, bl::bind(bl::constructor<Diagram::value_type>(),
                                      bl::bind(bl::var(cellColor), bl::bind(&Pairs::value_type::first, bl::_1)),
                                      bl::bind(bl::var(cellColor), bl::bind(&Pairs::value_type::second, bl::_1))
                                      ));


      diagram.assign(b::make_transform_iterator(pairs.begin(), pairColor),
                     b::make_transform_iterator(pairs.end(), pairColor));

      BOOST_AUTO(makeInf, bl::bind(bl::constructor<Diagram::value_type>(),
                                   bl::bind(bl::var(cellColor), bl::_1),
                                   bl::constant(std::numeric_limits<double>::infinity())));

      diagram.insert(diagram.end(),
                     boost::make_transform_iterator(cycles.begin(), makeInf),
                     boost::make_transform_iterator(cycles.end(), makeInf));

      std::sort(diagram.begin(), diagram.end(), DiagramIntervalComparator());
      return diagram;
    }

    template<typename FilteredComplexT>
    PersistentSignature ComputePersistentHomology<FilteredComplexT>::signature() const
    {
      return PersistentSignature(*this);
    }

  }
}


#endif // CAPD_FILE_CAPDREDHOM_PERSISTENCE_COMPUTEPERSISTENCEHOMOLOGY_HPP
