/////////////////////////////////////////////////////////////////////////////
/// @file PointCloudSimplicialCellSComplex
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-06-27
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.

#ifndef CAPD_FILE_CAPDREDHOM_PERSISTENCE_POINTCLOUDSIMPLICIALCELLSCOMPLEX_H
#define CAPD_FILE_CAPDREDHOM_PERSISTENCE_POINTCLOUDSIMPLICIALCELLSCOMPLEX_H

#include <capd/complex/CellSComplex.h>
#include <deque>
#include <boost/scoped_ptr.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/range/iterator_range.hpp>




namespace capd
{
  namespace persistence
  {
    class AbstractPointCloud;

    class PointCloudSimplicialCellSComplex
    {

    public:
      struct Traits
      {
        typedef size_t Color;
        typedef char Dim;
        typedef size_t Id;
        typedef std::vector<size_t> CellCode;

        struct CellCodeComparator
        {
          bool operator()(const CellCode& a, const CellCode& b) const
          {
            if (a.size() < b.size()) {
              return true;
            } else if (a.size() > b.size()) {
              return false;
            } else {
              return a < b;
            }
          }
        };

      };

      typedef std::deque<Traits::CellCode> Storage;
      typedef capd::complex::SimplicialCellSComplex<Traits, Storage> Complex;
      typedef Complex::Dim Dim;
      typedef Complex::Id Id;

      struct Cell: Complex::Cell
      {
        Cell(Complex::Id id, const PointCloudSimplicialCellSComplex& pccomplex):
          Complex::Cell(id, pccomplex.complex()),
          _pccomplex(pccomplex)
        {}

        double color() const { return _pccomplex._weights[id()]; }

      protected:
        const PointCloudSimplicialCellSComplex& _pccomplex;
      };

      typedef Cell ConstCell;
      PointCloudSimplicialCellSComplex(const AbstractPointCloud& cloud, Complex::Dim maxDim);

      virtual ~PointCloudSimplicialCellSComplex() {}

      bool exists(Id id) const
      {
        return id < _complex->size();
      }

      size_t size() const { return _complex->size(); }
      Dim dim() const { return _complex->dim(); }
      void operator()();

      ConstCell operator[](Complex::Id id) const { return ConstCell(id, *this); }
      Cell operator[](Complex::Id id) { return Cell(id, *this); }

      struct Iterators
      {
        Iterators(const PointCloudSimplicialCellSComplex& complex):
          _complex(complex)
        {}

        struct IdToCell
        {
          typedef Cell result_type;

          IdToCell(const PointCloudSimplicialCellSComplex& complex):
            _complex(complex)
          {}

          Cell operator()(Complex::Id id) const
          {
            return _complex[id];
          }

          const PointCloudSimplicialCellSComplex& _complex;
        };

        typedef boost::transform_iterator<IdToCell, Complex::Boundary::iterator> BdCellsIterator;
        typedef boost::iterator_range<BdCellsIterator> BdCells;

        BdCells bdCells(const Cell& cell)
        {
          Complex::Boundary bd = _complex._complex->boundary(cell);
          return BdCells(BdCellsIterator(bd.begin(), IdToCell(_complex)),
                         BdCellsIterator(bd.end(), IdToCell(_complex)));
        }

        const PointCloudSimplicialCellSComplex& _complex;
      };


      Iterators iterators() const
      {
        return Iterators(*this);
      }

    private:

      Complex& complex() { return *_complex; }
      const Complex& complex() const { return *_complex; }

      const AbstractPointCloud& _cloud;
      const Complex::Dim _maxDim;

      Storage _storage;
      boost::scoped_ptr<Complex> _complex;
      std::vector<double> _weights;
    };


  }
}

#endif // CAPD_FILE_CAPDREDHOM_PERSISTENCE_POINTCLOUDSIMPLICIALCELLSCOMPLEX_H
