#if !defined(_DIRECTMAP_H_)
#define _DIRECTMAP_H_

namespace data_structures {

// DirectMap class is a fast key-value map.
// All operations (insert/find/erase) are performed in worst-case constant time.
// Keys have to be perfectly hashable (without collisions) onto [0..capacity-1] range.
// The user is responsible for calling resize function if he wants to store bigger keys.
template <typename Key, typename Value, typename Hasher>
class DirectMap {
 public:
  typedef int IntType;
  typedef std::pair<Key, Value> KeyValuePair;
    
 private:
  struct Node {
    Node() : next(-1), is_filled(false), is_removed(false) {}
    Node(const Key& _key) : next(-1), is_filled(false), is_removed(false) { item.first = _key; }
    Node(const Key& _key, const Value& _value) : item(_key, _value), next(-1), is_filled(false), is_removed(false) {}
    Node(const KeyValuePair& _item) : item(_item), next(-1), is_filled(false), is_removed(false) {}
    
    KeyValuePair item;
    IntType next;
    bool is_filled;
    bool is_removed;
  };
  
 public:
  // Iterator class used to iterate over DirectMap.
  // Garbage collects removed nodes.
  class const_iterator {
   public:
    const_iterator(Node* map, IntType i) : map_(map), i_(i) {}
    bool operator==(const const_iterator& other) const { return i_ == other.i_; }
    bool operator!=(const const_iterator& other) const { return i_ != other.i_; }
    void operator++() {
      while (map_[i_].next != -1 && map_[map_[i_].next].is_removed)
        map_[i_].next = map_[map_[i_].next].next;
      i_ = map_[i_].next;
    }
    const KeyValuePair& operator*() const { return map_[i_].item; }
    const KeyValuePair* operator->() const { return &(map_[i_].item); }
    
   private:
    Node* map_;
    IntType i_;
  };

  // Default constructor.
  DirectMap() : capacity_(0), map_(NULL), list_(-1) {}

  DirectMap(int capacity) : capacity_(capacity), map_(NULL), list_(-1) {
    map_ = new Node[capacity_];
    std::fill(map_, map_ + capacity_, Node());
  } 

  // Copy constructor.
  DirectMap(const DirectMap& other) : capacity_(other.capacity_), map_(NULL), list_(other.list_) {
    map_ = new Node[capacity_];
    std::copy(other.map_, other.map_ + capacity_, map_);
  }

  const DirectMap& operator=(const DirectMap& other) {
    delete[] map_;
    capacity_ = other.capacity_;
    map_ = new Node[capacity_];
    std::copy(other.map_, other.map_ + capacity_, map_);    
    list_ = other.list_;
    return *this;
  }

  ~DirectMap() {
    delete[] map_;
  }

  void resize(int capacity) {
    delete[] map_;
    capacity_ = capacity;
    map_ = new Node[capacity_];
    std::fill(map_, map_ + capacity_, Node());
    list_ = -1;
  }
  
  const_iterator begin() const { return const_iterator(map_, list_); }
  const_iterator end() const { return const_iterator(map_, -1); }
  bool empty() const { return list_ == -1; }
  int capacity() const { return capacity_; }

  // Find key-value pair for a given key.
  const_iterator find(const Key& key) const {
    IntType i = hash(key);
    return map_[i].is_filled ? const_iterator(map_, i) : end();
  }

  // Check if the given key is present in the map.
  bool contains(const Key& key) const {
    return map_[hash(key)].is_filled;
  }
  
  // Return a value for a given key.
  // If the key is not present, add it.
  Value& operator[](const Key& key) {
    IntType i = hash(key);
    if (!map_[i].is_filled) {
      map_[i].is_filled = true;
      map_[i].item.first = key;
      map_[i].next = list_;
      list_ = i;
    }
    return map_[i].item.second;
  }
  
  // Erase key-value pair for a given key.
  void erase(const Key& key) {
    IntType i = hash(key);
    if (map_[i].is_filled) {
      map_[i].is_filled = false;
      map_[i].is_removed = true;
      while (list_ != -1 && map_[list_].is_removed)
        list_ = map_[list_].next;
    }
  }

  // Swap in constant time.
  void swap(DirectMap& other) {
    std::swap(capacity_, other.capacity_);
    std::swap(map_, other.map_);
    std::swap(list_, other.list_);
  }
  
 private:
  IntType hash(const Key& key) const { return Hasher::hash(key); }
  
  int capacity_;
  Node* map_;
  IntType list_;
};

}

#endif // _DIRECTMAP_H_
