#if !defined(_HASHMAP_H_)
#define _HASHMAP_H_

#include <algorithm>

namespace data_structures {

// HashMap class is a fast key-value map.
// All operations (insert/find/erase) are performed in average-case constant time.
// Keys have to be hashable onto integers.
template <typename Key, typename Value, typename Hasher>
class HashMap {
 public:
  typedef int IntType;
  typedef std::pair<Key, Value> KeyValuePair;

 private:
  struct Node {
    Node(const Key& _key) : next(NULL), is_head(false), is_removed(false) { item.first = _key; }
    Node(const Key& _key, const Value& _value) : item(_key, _value), next(NULL), is_head(false), is_removed(false) {}
    Node(const KeyValuePair& _item) : item(_item), next(NULL), is_head(false), is_removed(false) {}

    KeyValuePair item;
    Node* next;
    bool is_head;
    bool is_removed;
  };

 public:
  // Iterator class used to iterate over HashMap.
  // Garbage collects removed nodes.
  class const_iterator {
   public:
    const_iterator(Node* node) : node_(node) {}
    const_iterator(const const_iterator& other) : node_(other.node_) {}
    bool operator==(const const_iterator& other) const { return node_ == other.node_; }
    bool operator!=(const const_iterator& other) const { return node_ != other.node_; }
    void operator++() {
      while (node_->next != NULL && node_->next->is_removed) {
        Node* ptr = node_->next;
        node_->next = ptr->next;
        delete ptr;
      }
      node_ = node_->next;
    }
    const KeyValuePair& operator*() const { return node_->item; }
    const KeyValuePair* operator->() const { return &(node_->item); }

   private:
    Node* node_;
  };

  // Default constructor.
  HashMap() : prime_index_(0), size_(0), list_(NULL) {
    map_ = new Node*[PRIMES[prime_index_]];
    std::fill(map_, map_ + PRIMES[prime_index_], (Node*)NULL);
  }

  HashMap(int capacity) : prime_index_(0), size_(0), list_(NULL) {
    while (PRIMES[prime_index_] < capacity)
      prime_index_++;
    map_ = new Node*[PRIMES[prime_index_]];
    std::fill(map_, map_ + PRIMES[prime_index_], (Node*)NULL);    
  }

  // Copy constructor.
  HashMap(const HashMap& other) : prime_index_(other.prime_index_), size_(other.size_), list_(NULL) {
    if (!other.empty()) {
      list_ = new Node(other.list_->item);
      Node* tail = list_;
      for (Node* other_ptr = other.list_->next; other_ptr != NULL; other_ptr = other_ptr->next) {
        tail->next = new Node(other_ptr->item);
        tail = tail->next;
      }
    }
    map_ = new Node*[PRIMES[prime_index_]];
    std::copy(other.map_, other.map_ + PRIMES[prime_index_], map_);
  }

  const HashMap& operator=(const HashMap& other) {
    delete[] map_;
    while (list_ != NULL) {
      Node* node = list_;
      list_ = list_->next;
      delete node;
    }
    prime_index_ = other.prime_index_;
    size_ = other.size_;
    if (!other.empty()) {
      list_ = new Node(other.list_->item);
      Node* tail = list_;
      for (Node* other_ptr = other.list_->next; other_ptr != NULL; other_ptr = other_ptr->next) {
        tail->next = new Node(other_ptr->item);
        tail = tail->next;
      }
    }
    map_ = new Node*[PRIMES[prime_index_]];
    std::copy(other.map_, other.map_ + PRIMES[prime_index_], map_);
    return *this;
  }

  ~HashMap() {
    delete[] map_;
    while (list_ != NULL) {
      Node* node = list_;
      list_ = list_->next;
      delete node;
    }
  }

  const_iterator begin() const { return const_iterator(list_); }
  const_iterator end() const { return const_iterator(NULL); }
  bool empty() const { return list_ == NULL; }
  int size() const { return size_; }

  // Find key-value pair for a given key.
  const_iterator find(const Key& key) const {
    return const_iterator(find(hash(key), key));
  }
  
  // Check if the given key is present in the map.
  bool contains(const Key& key) const {
    return find(key) != end();
  }

  void insert_non_existing(const Key& key, const Value& value) {
    if (size_ > PRIMES[prime_index_ + 1]) {
      prime_index_++;
      rehash();
    }
    size_++;
    insert(hash(key), new Node(key, value));
  }

  // Return a value for a given key.
  // If the key is not present, add it. 
  Value& operator[](const Key& key) {
    if (size_ > PRIMES[prime_index_ + 1]) {
      prime_index_++;
      rehash();
    }
    IntType i = hash(key);
    Node* node = find(i, key);
    if (node == NULL) {
      size_++;
      node = new Node(key);
      insert(i, node);
    }
    return node->item.second;
  }

  // Erase key-value pair for a given key.
  void erase(const Key& key) {
    if (prime_index_ > 0 && size_ < PRIMES[prime_index_ - 1]) {
      prime_index_--;
      rehash();
    }
    IntType i = hash(key);
    Node* prev;
    Node* node = find(i, key, prev);
    // Key not found
    if (node == NULL)
      return;
    size_--;
    // Node is not sublist's head, cut it out
    if (prev != NULL) {
      prev->next = node->next;
      delete node;
      return;
    }
    // It was the last node in the sublist, mark sublist as empty
    if (node->next == NULL || node->next->is_head) {
      map_[i] = NULL;
    // It was not the last node, 
    } else {
      map_[i] = node->next;
      map_[i]->is_head = true;
    }
    // We don't have access to node's predecessor, we just mark it as removed
    node->is_removed = true;
    // Node found is the list's head, erase it
    while (list_ != NULL && list_->is_removed) {
      node = list_;
      list_ = list_->next;
      delete node;
    }
  }

  // Swap in constant time.
  void swap(HashMap& other) {
    std::swap(prime_index_, other.prime_index_);
    std::swap(size_, other.size_);
    std::swap(map_, other.map_);
    std::swap(list_, other.list_);
  }

 private:
  IntType hash(const Key& key) const { return Hasher::hash(key) % PRIMES[prime_index_]; }

  void rehash() {
    delete[] map_;
    map_ = new Node*[PRIMES[prime_index_]];
    std::fill(map_, map_ + PRIMES[prime_index_], (Node*)NULL);
    Node* old_list = list_;
    list_ = NULL;
    while (old_list != NULL) {
      Node* node = old_list;
      old_list = old_list->next;
      if (node->is_removed) {
        delete node;
      } else {
        node->is_head = false;
        insert(hash(node->item.first), node);
      }
    }
  }
  
  void insert(IntType i, Node* node) {
    if (map_[i] == NULL) {
      node->is_head = true;
      node->next = list_;
      map_[i] = list_ = node;
    } else {
      node->next = map_[i]->next;
      map_[i]->next = node;
    }  
  }

  Node* find(IntType i, const Key& key) const {
    Node* ptr = map_[i];
    if (ptr == NULL || key == ptr->item.first)
      return ptr;
    ptr = ptr->next;
    while (ptr != NULL && !ptr->is_head && !(key == ptr->item.first)) {
      ptr = ptr->next;
    }
    return ptr == NULL || ptr->is_head ? NULL : ptr;
  }

  Node* find(IntType i, const Key& key, Node*& prev) const {
    prev = NULL;
    Node* ptr = map_[i];
    if (ptr == NULL || key == ptr->item.first)
      return ptr;
    prev = ptr;
    ptr = ptr->next;
    while (ptr != NULL && !ptr->is_head && !(key == ptr->item.first)) {
      prev = ptr;
      ptr = ptr->next;
    }
    return ptr == NULL || ptr->is_head ? NULL : ptr;
  }

  static const IntType PRIMES[];
  int prime_index_;
  int size_;
  Node* list_;
  Node** map_;
};

template <typename Key, typename Value, typename Hasher>
const typename HashMap<Key, Value, Hasher>::IntType HashMap<Key, Value, Hasher>::PRIMES[] = {
  3, 7, 13, 23, 53, 97, 193, 389, 769, 1543,
  3079, 6151, 12289, 24593, 49157, 98317, 196613, 393241, 786433, 1572869,
  3145739, 6291469, 12582917, 25165843, 50331653, 100663319, 201326611, 402653189, 805306457, 1610612741
};

}

#endif // _HASHMAP_H_
