#if !defined(_SPARSEMATRIX_H_)
#define _SPARSEMATRIX_H_

#include <map>
#include <set>
#include <vector>

#include "capd/redAlg/data_structures/direct_map.h"
#include "capd/redAlg/data_structures/hash_map.h"
#include "capd/redAlg/data_structures/hasher.h"

namespace data_structures {

// SparseMatrix class is an implementation of sparse matrix.
template <typename IndexType, typename ScalarType>
class SparseMatrix {
 public:
  typedef HashMap<IndexType, ScalarType, Hasher<IndexType> > Row;
  typedef HashMap<IndexType, ScalarType, Hasher<IndexType> > Column;
  typedef DirectMap<IndexType, Row, Hasher<IndexType> > Rows;
  typedef DirectMap<IndexType, Column, Hasher<IndexType> > Columns;
  typedef typename Rows::const_iterator RowsIterator;
  typedef typename Row::const_iterator RowIterator;
  typedef typename Columns::const_iterator ColumnsIterator;
  typedef typename Column::const_iterator ColumnIterator;
  
  // Default constructor.
  SparseMatrix() {}

  // Copy constructor.
  SparseMatrix(const SparseMatrix& other) : rows_(other.rows_), cols_(other.cols_) {}

  // Reserves some initial space for rows and columns.
  // Makes sense only in some implementations of Rows and Columns types (DirectMap).
  SparseMatrix(int row_count, int col_count) : rows_(row_count), cols_(col_count) {}

  // Reserves space for rows and columns.
  // Makes sense only in some implementations of Rows and Columns types (DirectMap).
  void resize(int r, int c) {
    rows_.resize(r);
    cols_.resize(c);
  }

  // Return a value assigned to (r, c) entry.
  ScalarType get(const IndexType& r, const IndexType& c) const {
    ColumnsIterator cols_it = cols_.find(c);
    if (cols_it == cols_.end())
      return ScalarType(0);
    const Column& col = cols_it->second;
    ColumnIterator col_it = col.find(r);
    return (col_it == col.end()) ? ScalarType(0) : col_it->second;
  }

  // Give read-only access to rows collection.
  const Rows& rows() const { return rows_; }

  // Give read-only access to columns collection.
  const Columns& cols() const { return cols_; }

  // Give read-only access to a row indexed by "r".
  // A row is created if does not exist yet.
  const Row& row(const IndexType& r) { return rows_[r]; }

  // Give read-only access to a column indexed by "c".
  // A column is created if does not exist yet.
  const Column& col(const IndexType& c) { return cols_[c]; }

  // Assign (r, c) entry a value.
  // Zeros are not stored in this implementation.
  void set(const IndexType& r, const IndexType& c, const ScalarType& value) {
    if (value == ScalarType(0)) {
      rows_[r].erase(c);
      cols_[c].erase(r);
    } else {
      rows_[r][c] = value; 
      cols_[c][r] = value;
    }
  }

  // Equivalent to "set" but faster.
  // The caller has to ensure that (r, c) is not yet present in the matrix.
  void set_non_existing(const IndexType& r, const IndexType& c, const ScalarType& value) {
    if (!(value == ScalarType(0))) {
      rows_[r].insert_non_existing(c, value);
      cols_[c].insert_non_existing(r, value);
    }
  }

  // Delete a row indexed by "r".
  void delete_row(const IndexType& r) {
    const Row& row = rows_[r];
    for (RowIterator row_it = row.begin(); row_it != row.end(); ++row_it) {
      const IndexType& c = row_it->first;
      cols_[c].erase(r);
    }
    rows_.erase(r);
  }

  // Delete a column indexed by "c".
  void delete_col(const IndexType& c) {
    const Column& col = cols_[c];
    for (ColumnIterator col_it = col.begin(); col_it != col.end(); ++col_it) {
      const IndexType& r = col_it->first;
      rows_[r].erase(c);
    }
    cols_.erase(c);
  }

  // Swap this instance with another in constant time.
  void swap(SparseMatrix& other) {
    rows_.swap(other.rows_);
    cols_.swap(other.cols_);
  }

 private:
  Rows rows_;
  Columns cols_;
};

}

#endif // _SPARSEMATRIX_H_

