/// @addtogroup homologicalAlgebra
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file ReducibleFreeChainComplex.h
///
/// @author Przemyslaw Witek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2011 by the CAPD Group.
//
// This file constitutes a part of the CAPD library,
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details.

#if !defined(_EXPERIMENTAL_REDUCIBLEFREECHAINCOMPLEX_H_)
#define _EXPERIMENTAL_REDUCIBLEFREECHAINCOMPLEX_H_

#include <iostream>
#include <vector>
#include <set>
#include <map>

#include "capd/homAlgebra/ChainT.h"
#include "capd/redAlg/data_structures/sparse_matrix.h"

namespace experimental {

template<typename freeModuleType, typename P_GeneratorCode>
class ReducibleFreeChainComplex;

template<typename freeModuleType, typename P_GeneratorCode>
void swap(ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>& rfcc1, ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>& rfcc2);

template<typename freeModuleType, typename P_GeneratorCode>
std::ostream& operator<<(std::ostream& out, const ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>& rfcc);

template<typename freeModuleType, typename P_GeneratorCode=typename freeModuleType::GeneratorType>
class ReducibleFreeChainComplex{
 public:
  typedef int IntType;
  typedef P_GeneratorCode GeneratorCode;
  typedef typename freeModuleType::GeneratorType GeneratorType;
  typedef typename freeModuleType::MatrixType MatrixType;
  typedef typename freeModuleType::MatrixType::ScalarType ScalarType;
  typedef ChainT<std::map<GeneratorCode, ScalarType> > Chain;
  typedef Chain CoChain;
  typedef int (ReducibleFreeChainComplex::*ReducibleFreeChainComplex_void_Ptr)();

  struct ReducedPair;
  struct ReduciblePair;
  class ReduciblePairComparer;

  ReducibleFreeChainComplex() {}

  template<typename freeModuleType2>
  ReducibleFreeChainComplex(const FreeChainComplex<freeModuleType2>& fcc);

  template<typename GeneratorType2>
  ReducibleFreeChainComplex(const std::vector<GeneratorType2>& generators);

  template<typename GeneratorType2>
  ReducibleFreeChainComplex(const std::set<GeneratorType2>& generators);

  template<typename P_Set>
  ReducibleFreeChainComplex(const P_Set& generators, int store_reduced_pairs = false);

  bool removeSimpleHomologyGenerators(std::vector<std::vector<GeneratorCode> >& simpleHomologyGenerators);

  operator FreeChainComplex<freeModuleType>();

  int reduce();
  int reduceNaively();
  int reduceViaSort();
  int reduceViaLocalSort();
  int reduceWithPriorityQueue();

  void reduce(Chain& chain);
  void restore(Chain& chain);

  friend void swap<freeModuleType, P_GeneratorCode>(
    ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>& rfcc1,
    ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>& rfcc2);

  friend std::ostream& operator<< <freeModuleType, P_GeneratorCode>(
    std::ostream& out,
    const ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>& rfcc);

 private:
  bool findAnyReduciblePair(int q, GeneratorCode* a, GeneratorCode* b) const;
  void reducePair(int q, const GeneratorCode& a, const GeneratorCode& b);
  void reducePair(int q, const GeneratorCode& a, const GeneratorCode& b, std::set<ReduciblePair, ReduciblePairComparer>& queue);
  Chain make_chain(const typename data_structures::SparseMatrix<GeneratorCode, ScalarType>::Column& boundary) const;
  CoChain make_cochain(const typename data_structures::SparseMatrix<GeneratorCode, ScalarType>::Row& coboundary) const;

  int top_dim_;
  vector<data_structures::SparseMatrix<GeneratorCode, ScalarType> > bd_;
  bool store_reduced_pairs_;
  vector<ReducedPair> reduced_pairs_;
};

template<typename freeModuleType, typename P_GeneratorCode>
struct ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::ReducedPair{
  int q_;
  GeneratorCode a_;
  GeneratorCode b_;
  Chain b_boundary_;      // coface boundary is stored multiplied by the reduction coefficient lambda
  CoChain a_coboundary_;    // face coboundary is stored multiplied by the reduction coefficient lambda

  ReducedPair(
    int q,
    GeneratorCode b,
    GeneratorCode a,
    const Chain& b_boundary,
    const CoChain& a_coboundary,
    ScalarType coeff
  ) : q_(q), a_(a), b_(b), b_boundary_(b_boundary), a_coboundary_(a_coboundary) {
     b_boundary_ *= coeff; // multiply here by lambda for efficiency reasons
     a_coboundary_ *= coeff; // multiply here by lambda for efficiency reasons
  }

  void reduce(Chain& chain){
    // -- fcout << " RFCC Reducing through pair (" << a << "," << b << ")\n";  //--
    int chainOwnDim = chain.ownDim();
    if (chainOwnDim == q_ - 1) {
      // -- fcout << " RFCC Reduced by (" << chain.at(a_) * b_boundary << "\n";  //--
      chain -= chain.at(a_) * b_boundary_;  // lambda already stored in boundary
      // chain.erase(a_);  // not needed, guaranteed by the above instruction
    } else if (chainOwnDim == q_) {
      // -- fcout << " RFCC Reduced by (" << b_ << "\n";  //--
      chain.erase(b_);
    }
  }

  void restore(Chain& chain) {
    int chainOwnDim = chain.ownDim();
    if (chainOwnDim == q_) {
      ScalarType s = chain * a_coboundary_;
      chain[b_] -= s;  // lambda already stored in coboundary
    }
  }
};

template<typename freeModuleType, typename P_GeneratorCode>
struct ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::ReduciblePair {
  ReduciblePair(const GeneratorCode& _a, const GeneratorCode& _b) : a(_a), b(_b), cost(0) {}
  ReduciblePair(const GeneratorCode& _a, const GeneratorCode& _b, int _cost) : a(_a), b(_b), cost(_cost) {}
  bool operator==(const ReduciblePair& other) const {
    return a == other.a && b == other.b && cost == other.cost;
  }
  bool operator!=(const ReduciblePair& other) const {
    return !operator==(other);
  }
  bool operator<(const ReduciblePair& other) const {
    return cost < other.cost || (cost == other.cost && (b < other.b || (b == other.b && a < other.a)));
  }
  GeneratorCode a;
  GeneratorCode b;
  int cost;
};

template<typename freeModuleType, typename P_GeneratorCode>
class ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::ReduciblePairComparer {
 public:
  ReduciblePairComparer(const data_structures::SparseMatrix<GeneratorCode, ScalarType>& matrix) : matrix_(matrix) {}
  bool operator()(const ReduciblePair& pair1, const ReduciblePair& pair2) const {
    int cost1 = cost(pair1);
    int cost2 = cost(pair2);
    return cost1 < cost2 || (cost1 == cost2 && (pair1.b < pair2.b || (pair1.b == pair2.b && pair1.a < pair2.a)));
  }
 private:
  int cost(const ReduciblePair& pair) const {
    typename data_structures::SparseMatrix<GeneratorCode, ScalarType>::ColumnsIterator cols_it = matrix_.cols().find(pair.b);
    if (cols_it == matrix_.cols().end()) return -1;
    typename data_structures::SparseMatrix<GeneratorCode, ScalarType>::RowsIterator rows_it = matrix_.rows().find(pair.a);
    if (rows_it == matrix_.rows().end()) return -1;
    return (cols_it->second.size() - 1) * (rows_it->second.size() - 1);
  }
  const data_structures::SparseMatrix<GeneratorCode, ScalarType>& matrix_;
};

}

#endif // _EXPERIMENTAL_REDUCIBLEFREECHAINCOMPLEX_H_
/// @}
