/// @addtogroup homologicalAlgebra
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file ReducibleFreeChainComplex.hpp
///
/// @author Przemyslaw Witek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2011 by the CAPD Group.
//
// This file constitutes a part of the CAPD library,
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details.

#if !defined(_EXPERIMENTAL_REDUCIBLEFREECHAINCOMPLEX_HPP_)
#define _EXPERIMENTAL_REDUCIBLEFREECHAINCOMPLEX_HPP_

#include <functional>
#include "capd/redAlg/ReducibleFreeChainComplex.h"
#include "capd/auxil/Stopwatch.h"
#include "capd/redAlg/data_structures/sparse_matrix.h"

// useful iteration macros
#define FOR(i, a, b) for (int i = (a); i <= (b); ++i)
#define FORD(i, a, b) for (int i = (a); i >= (b); --i)
#define FOREACH(it, coll) for (typeof((coll).begin()) it = (coll).begin(); it != (coll).end(); ++it)

using data_structures::SparseMatrix;
using std::set;
using std::vector;

namespace experimental {

template<typename freeModuleType, typename P_GeneratorCode>
template<typename freeModuleType2>
ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::ReducibleFreeChainComplex(
    const FreeChainComplex<freeModuleType2>& fcc) : store_reduced_pairs_(false) {
  std::cout << "--- ReducibleFreeChainComplex(fcc)" << std::endl;
  try {
    top_dim_ = fcc.topDim();
    bd_.resize(top_dim_ + 2);
    FOR(q, 0, top_dim_) {
      const MatrixType& bdMtrx = fcc.boundaryHomomorphism(q).coefMatrix();
      int m = bdMtrx.numberOfRows();
      int n = fcc.chainModule(q).numberOfGenerators();
      bd_[q].resize(m, n);
      for (int j = 0; j < n; ++j) {
        if (q > 0) {
          for (int i = 0; i < m; ++i) {
            bd_[q].set_non_existing(i, j, bdMtrx[i][j]);
          }
        } else {
          bd_[q].col(j);
        }
      }
    }
    bd_[top_dim_+1].resize(fcc.chainModule(top_dim_).numberOfGenerators(), 0);
  } catch (std::bad_alloc& e) {
    throw std::runtime_error("Memory request for ReducibleFreeChainComplex failed");
  }
}

template<typename freeModuleType, typename P_GeneratorCode>
template<typename GeneratorType2>
ReducibleFreeChainComplex<freeModuleType,P_GeneratorCode>::ReducibleFreeChainComplex(
    const std::vector<GeneratorType2>& generators) : store_reduced_pairs_(false) {
  std::cout << "--- ReducibleFreeChainComplex(vector)" << std::endl;
  try {
    std::cout << "generators.size() = " << generators.size() << std::endl;
    if (generators.empty())
      return;
    top_dim_ = generators.begin()->embeddingDimension();
    bd_.resize(top_dim_ + 2);
    vector<map<GeneratorType2, GeneratorCode> > codes(top_dim_ + 1);
    FOREACH(b_it, generators) {
      const GeneratorType2& b = *b_it;
      int b_dim = b.dimension();
      codes[b_dim].insert(make_pair(b, codes[b_dim].size()));
    }
    bd_[top_dim_+1].resize(codes[top_dim_].size(), 0);
    FORD(q, top_dim_, 1) {
      std::cout << "q = " << q << std::endl;
      vector<pair<pair<GeneratorCode, GeneratorCode>, ScalarType> > triples;
      FOREACH(b_it, codes[q]) {
        const GeneratorType2& b = b_it->first;
        const GeneratorCode& b_code = b_it->second;
        map<GeneratorType2, ScalarType> b_faces;
        b.boundary(b_faces);
        FOREACH(a_it, b_faces) {
          const GeneratorType2& a(a_it->first);
          int a_dim = a.dimension();
          GeneratorCode a_code = codes[a_dim].insert(make_pair(a, codes[a_dim].size())).first->second;
//           bd_[q].set(a_code, b_code, a_it->second);
          triples.push_back(make_pair(make_pair(a_code, b_code), a_it->second));
        }
      }
      bd_[q].resize(codes[q-1].size(), codes[q].size());
      FOREACH(triple, triples)
        bd_[q].set_non_existing(triple->first.first, triple->first.second, triple->second);
    }
    bd_[0].resize(0, codes[0].size());
    // Add generators with dimension 0.
    FOREACH(b_it, codes[0])
      bd_[0].col(b_it->second);
  } catch (std::bad_alloc& e) {
    throw std::runtime_error("Memory request for ReducibleFreeChainComplex from vector<GeneratorType2> failed");
  }
}

template<typename freeModuleType, typename P_GeneratorCode>
template<typename GeneratorType2>
ReducibleFreeChainComplex<freeModuleType,P_GeneratorCode>::ReducibleFreeChainComplex(
    const std::set<GeneratorType2>& generators) : store_reduced_pairs_(false) {
  std::cout << "--- ReducibleFreeChainComplex(set)" << std::endl;
  try {
    std::cout << "generators.size() = " << generators.size() << std::endl;
    if (generators.empty())
      return;
    top_dim_ = generators.begin()->embDim();
    bd_.resize(top_dim_ + 2);
    vector<map<GeneratorType2, GeneratorCode> > codes(top_dim_ + 1);
    FOREACH(b_it, generators) {
      const GeneratorType2& b = *b_it;
      int b_dim = b.ownDim();
      codes[b_dim].insert(make_pair(b, codes[b_dim].size()));
    }
    bd_[top_dim_+1].resize(codes[top_dim_].size(), 0);
    FORD(q, top_dim_, 1) {
      std::cout << "q = " << q << std::endl;
      bd_[q].resize(codes[q-1].size(), codes[q].size());
      FOREACH(b_it, codes[q]) {
        const GeneratorType2& b = b_it->first;
        const GeneratorCode& b_code = b_it->second;
        ChainT<map<GeneratorType2, ScalarType> > b_faces;
        b.boundary(b_faces);
        FOREACH(a_it, b_faces) {
          const GeneratorType2& a(a_it->first);
          int a_dim = a.ownDim();
          typename map<GeneratorType2, GeneratorCode>::const_iterator it = codes[a_dim].find(a);
          if (it != codes[a_dim].end())
             bd_[q].set_non_existing(it->second, b_code, a_it->second);
        }
      }
    }
    bd_[0].resize(0, codes[0].size());
    // Add generators with dimension 0.
    FOREACH(b_it, codes[0])
      bd_[0].col(b_it->second);
  } catch (std::bad_alloc& e) {
    throw std::runtime_error("Memory request for ReducibleFreeChainComplex from set<GeneratorType2> failed");
  }
}

template<typename freeModuleType, typename P_GeneratorCode>
template<typename P_Set>
ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::ReducibleFreeChainComplex(
    const P_Set& generators, int store_reduced_pairs) : store_reduced_pairs_(store_reduced_pairs) {
  std::cout << "--- ReducibleFreeChainComplex(P_Set)" << std::endl;
  try {
    std::cout << "generators.size() = " << generators.cardinality() << std::endl;
    if (generators.cardinality() == 0)
      return;
    top_dim_ = generators.embDim();
    bd_.resize(top_dim_ + 2);
    vector<set<GeneratorType> > codes(top_dim_ + 1);
    for (typename P_Set::PointCoordIterator b_it = generators.begin(); b_it != generators.end(); ++b_it) {
      GeneratorType b(b_it);
      codes[b.ownDim()].insert(b);
    }
    bd_[top_dim_+1].resize(codes[top_dim_].size(), 0);
    FORD(q, top_dim_, 1) {
      std::cout << "q = " << q << std::endl;
      bd_[q].resize(codes[q-1].size(), codes[q].size());
      FOREACH(b_it, codes[q]) {
        const GeneratorType& b = *b_it;
        ChainT<map<GeneratorCode, ScalarType> > b_faces;
        b.boundary(b_faces);
        FOREACH(a_it, b_faces) {
          const GeneratorType& a = a_it->first;
          if (codes[q-1].count(a))
             bd_[q].set_non_existing(a, b, a_it->second);
        }
      }
    }
    bd_[0].resize(0, codes[0].size());
    // Add generators with dimension 0.
    FOREACH(b_it, codes[0])
      bd_[0].col(*b_it);
  } catch (std::bad_alloc& e) {
    throw std::runtime_error("Memory request for ReducibleFreeChainComplex from set<GeneratorType2> failed");
  }
}

template<typename freeModuleType, typename P_GeneratorCode>
bool ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::removeSimpleHomologyGenerators(
    std::vector<std::vector<P_GeneratorCode> >& simpleHomologyGenerators) {
  std::cout << "--- removeSimpleHomologyGenerators" << std::endl;
  Stopwatch sw;
  bool all_removed = true;
  simpleHomologyGenerators.clear();
  simpleHomologyGenerators.resize(top_dim_ + 1);
  FORD(q, top_dim_, 0) {
    FOREACH(b_it, bd_[q].cols()) {
      const GeneratorCode& b = b_it->first;
      if (!bd_[q].col(b).size() && !bd_[q+1].row(b).size()) {
        simpleHomologyGenerators[q].push_back(b);
      } else {
        all_removed = false;
      }
    }
    FOREACH(b_it, simpleHomologyGenerators[q]) {
      const GeneratorCode& b = *b_it;
      bd_[q].delete_col(b);
      bd_[q+1].delete_row(b);
    }
  }
  std::cout << "removeSimpleHomologyGenerators: processTime = " << sw.processTime() << ", worldTime = " << sw.worldTime() << std::endl;
  return all_removed;
}

template<typename freeModuleType, typename P_GeneratorCode>
ReducibleFreeChainComplex<freeModuleType,P_GeneratorCode>::operator FreeChainComplex<freeModuleType>() {
  std::cout << "--- Converter" << std::endl;
  Stopwatch sw;

  FreeChainComplex<freeModuleType> result;
  if (top_dim_ == -1)
    return result;
  result.m_topDim = top_dim_;
  // First we reserve space and call implicit constructor for all chain modules
  result.m_chainModule.resize(result.m_topDim + 1);
  result.m_boundaryHomomorphism.resize(result.m_topDim + 1);
  // Generate the chain modules.
  FOR(q, 0, result.m_topDim) {
    map<GeneratorCode, int> generators;
    FOREACH(b_it, bd_[q].cols()) {
      const GeneratorCode& b = b_it->first;
      generators[b] = 0;
    }
    FreeModule<GeneratorCode, MatrixType, IntType> fm(generators);
    swap(result.m_chainModule[q], fm);
  }
  // Compute the boundary map.
  FORD(q, result.m_topDim, 1) {
    // Generate the boundary operator for q-dimensional module.
    MatrixType boundaryOperator(
      result.m_chainModule[q-1].numberOfGenerators(),
      result.m_chainModule[q].numberOfGenerators()
    );
    FOREACH(b, result.m_chainModule[q])
      FOREACH(a, result.m_chainModule[q-1])
        boundaryOperator[a->second][b->second] = bd_[q].get(a->first, b->first);
    // Create a homomorphism based on the boundary operator.
    FreeModuleHomomorphism<GeneratorCode, MatrixType, IntType> fmh(
      &result.m_chainModule[q],
      &result.m_chainModule[q-1],
      boundaryOperator
    );
    swap(result.m_boundaryHomomorphism[q], fmh);
  }
  std::cout << "converter: processTime = " << sw.processTime() << ", worldTime = " << sw.worldTime() << std::endl;
  return result;
}

template<typename freeModuleType, typename P_GeneratorCode>
int ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::reduce() {
  return reduceViaSort();
}

template<typename freeModuleType, typename P_GeneratorCode>
int ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::reduceNaively() {
  std::cout << "--- reduceNaively()" << std::endl;
  int pairs_reduced = 0;
  int reduction_cost = 0;
  long double reducePairProcessorTime = 0.0;
  long double reducePairWorldTime = 0.0;

  FORD(q, top_dim_, 1) {
    std::cout << "q = " << q << std::endl;
    GeneratorCode a, b;
    while (findAnyReduciblePair(q, &a, &b)) {
      pairs_reduced++;
      reduction_cost += (bd_[q].col(b).size() - 1) * (bd_[q].row(a).size() - 1);
//      Stopwatch sw;
      reducePair(q, a, b);
//      reducePairProcessorTime += sw.processTime();
//      reducePairWorldTime += sw.worldTime();
    }
    std::cout << "pairsReduced = " << pairs_reduced << ", reductionCost = " << reduction_cost << ", reduceProcessorTime = " << reducePairProcessorTime << ", reduceWorldTime = " << reducePairWorldTime << std::endl;
  }
  return pairs_reduced;
}

template<typename freeModuleType, typename P_GeneratorCode>
int ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::reduceViaSort() {
  std::cout << "--- reduceViaSort" << std::endl;
  int pairs_reduced = 0;
  int reduction_cost = 0;
  long double reducePairProcessorTime = 0.0;
  long double reducePairWorldTime = 0.0;

  FORD(q, top_dim_, 1) {
    std::cout << "q = " << q << std::endl;
    while (true) {
      std::vector<ReduciblePair> reducible_pairs;
      FOREACH(cols_it, bd_[q].cols())
        FOREACH(col_it, cols_it->second)
          if (isInvertible(col_it->second))
            reducible_pairs.push_back(ReduciblePair(col_it->first, cols_it->first, (cols_it->second.size()-1)*(bd_[q].row(col_it->first).size()-1)));
      if (reducible_pairs.empty())
        break;
      std::cout << "reducible_pairs.size() = " << reducible_pairs.size() << std::endl;
      std::sort(reducible_pairs.begin(), reducible_pairs.end());
      FOREACH(pair, reducible_pairs) {
        if (isInvertible(bd_[q].get(pair->a, pair->b))) {
          pairs_reduced++;
//          reduction_cost += (bd_[q].col(pair->b).size() - 1) * (bd_[q].row(pair->a).size() - 1);
//          Stopwatch sw;
          reducePair(q, pair->a, pair->b);
//          reducePairProcessorTime += sw.processTime();
//          reducePairWorldTime += sw.worldTime();
        }
      }
    }
    std::cout << "pairsReduced = " << pairs_reduced << ", reductionCost = " << reduction_cost << ", reduceProcessorTime = " << reducePairProcessorTime << ", reduceWorldTime = " << reducePairWorldTime << std::endl;
  }
  return pairs_reduced;
}

template<typename freeModuleType, typename P_GeneratorCode>
int ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::reduceViaLocalSort() {
  std::cout << "--- reduceViaLocalSort";
  return 0;
}

template<typename freeModuleType, typename P_GeneratorCode>
int ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::reduceWithPriorityQueue() {
  std::cout << "--- reduceWithPriorityQueue" << std::endl;
  int pairs_reduced = 0;
  int reduction_cost = 0;
  long double reducePairProcessorTime = 0.0;
  long double reducePairWorldTime = 0.0;

  FORD(q, top_dim_, 1) {
    std::cout << "q = " << q << std::endl;
    // create priority queue for choosing the best (at the moment) reducible pair
    ReduciblePairComparer comparer(bd_[q]);
    std::set<ReduciblePair, ReduciblePairComparer> queue(comparer);
    // fill queue with reducible pairs
    FOREACH(cols_it, bd_[q].cols())
      FOREACH(col_it, cols_it->second)
        if (isInvertible(col_it->second))
          queue.insert(ReduciblePair(col_it->first, cols_it->first));
    std::cout << "queue.size() = " << queue.size() << std::endl;
    int iteration = 0;
    while (!queue.empty()) {
      ReduciblePair pair = *queue.begin();
      queue.erase(queue.begin());
      // a pair is not reducible anymore
      if (!isInvertible(bd_[q].get(pair.a, pair.b)))
        continue;
      // construct the set of pairs affected by reduction
      vector<ReduciblePair> affectedPairs;
      FOREACH(it, bd_[q].col(pair.b))
        FOREACH(it2, bd_[q].row(it->first)) {
//          queue.erase(ReduciblePair(it->first, it2->first));
          affectedPairs.push_back(ReduciblePair(it->first, it2->first));
        }
      FOREACH(it, bd_[q].row(pair.a))
        FOREACH(it2, bd_[q].col(it->first)) {
//          queue.erase(ReduciblePair(it2->first, it->first));
          affectedPairs.push_back(ReduciblePair(it2->first, it->first));
        }
      // reduce pair
      pairs_reduced++;
      reduction_cost += (bd_[q].col(pair.b).size() - 1) * (bd_[q].row(pair.a).size() - 1);
      Stopwatch sw;
      reducePair(q, pair.a, pair.b, queue);
      reducePairProcessorTime += sw.processTime();
      reducePairWorldTime += sw.worldTime();
      // insert affected pairs which are still reducible back into the queue
      FOREACH(affectedPair, affectedPairs)
        if (isInvertible(bd_[q].get(affectedPair->a, affectedPair->b)))
          queue.insert(*affectedPair);

      iteration++;
      if (iteration % 10000 == 0)
        std::cout << "iteration = " << iteration << ", queue.size() = " << queue.size() << std::endl;
    }
    std::cout << "pairsReduced = " << pairs_reduced << ", reductionCost = " << reduction_cost << ", reduceProcessorTime = " << reducePairProcessorTime << ", reduceWorldTime = " << reducePairWorldTime << std::endl;
  }
  return pairs_reduced;
}

template<typename freeModuleType, typename P_GeneratorCode>
bool ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::findAnyReduciblePair(int q, GeneratorCode* a, GeneratorCode* b) const {
  FOREACH(cols_it, bd_[q].cols())
    FOREACH(col_it, cols_it->second)
      if (isInvertible(col_it->second)) {
        *a = col_it->first;
        *b = cols_it->first;
        return true;
      }
  return false;
}

template<typename freeModuleType, typename P_GeneratorCode>
typename ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::Chain ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::make_chain(const typename SparseMatrix<GeneratorCode, ScalarType>::Column& boundary) const {
  Chain chain;
  FOREACH(it, boundary)
    chain.insert(*it);
  return chain;
}

template<typename freeModuleType, typename P_GeneratorCode>
typename ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::CoChain ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::make_cochain(const typename SparseMatrix<GeneratorCode, ScalarType>::Row& coboundary) const {
  CoChain cochain;
  FOREACH(it, coboundary)
    cochain.insert(*it);
  return cochain;
}

template<typename freeModuleType, typename P_GeneratorCode>
void ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::reducePair(int q, const GeneratorCode& a, const GeneratorCode& b) {
  ScalarType coeff_ab = bd_[q].get(a, b);
  const typename SparseMatrix<GeneratorCode, ScalarType>::Column& b_boundary = bd_[q].col(b);
  const typename SparseMatrix<GeneratorCode, ScalarType>::Row& a_coboundary = bd_[q].row(a);
  if (store_reduced_pairs_)
    reduced_pairs_.push_back(ReducedPair(q, b, a, make_chain(b_boundary), make_cochain(a_coboundary), inverse(coeff_ab)));
  FOREACH(c_it, a_coboundary) {
    const GeneratorCode& c = c_it->first;
    if (!(c == b)) {
      const ScalarType& coeff_ab_ac = coeff_ab * c_it->second;
      // bd(c) -= <b, a> * <c, a> * bd(b)
      FOREACH(r_it, b_boundary) {
        const GeneratorCode& r = r_it->first;
        if (!(r == a)) {
          bd_[q].set(r, c, bd_[q].get(r, c) - coeff_ab_ac * r_it->second);
        }
      }
    }
  }
  bd_[q+1].delete_row(b);
  bd_[q].delete_col(b);
  bd_[q].delete_row(a);
  bd_[q-1].delete_col(a);
}

template<typename freeModuleType, typename P_GeneratorCode>
void ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::reducePair(int q, const GeneratorCode& a, const GeneratorCode& b, set<ReduciblePair, ReduciblePairComparer>& queue) {
//  std::cout << "--- reduce(" << q << ", " << a << ", " << b << ")" << std::endl;
  ScalarType coeff_ab = bd_[q].get(a, b);
  const typename SparseMatrix<GeneratorCode, ScalarType>::Column& b_boundary = bd_[q].col(b);
  const typename SparseMatrix<GeneratorCode, ScalarType>::Row& a_coboundary = bd_[q].row(a);
  if (store_reduced_pairs_)
    reduced_pairs_.push_back(ReducedPair(q, b, a, make_chain(b_boundary), make_cochain(a_coboundary), inverse(coeff_ab)));
  // for each generator c in a's coboundary
  FOREACH(c_it, a_coboundary) {
    const GeneratorCode& c = c_it->first;
    if (!(c == b)) {
      const ScalarType& coeff_ac = c_it->second;
      // for each generator r in b's boundary
      FOREACH(r_it, b_boundary) {
        const GeneratorCode& r = r_it->first;
        if (!(r == a)) {
          const ScalarType& coeff_rb = r_it->second;
          // compute the new value for (r, c)
          ScalarType value = bd_[q].get(r, c) - coeff_ab * coeff_ac * coeff_rb;
          bd_[q].set(r, c, value);
          // a new reducible pair found, add it to the list
          if (isInvertible(value))
            queue.insert(ReduciblePair(r, c));
        }
      }
    }
  }
  bd_[q+1].delete_row(b);
  bd_[q].delete_col(b);
  bd_[q].delete_row(a);
  bd_[q-1].delete_col(a);
}

template<typename freeModuleType, typename P_GeneratorCode>
inline void ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::reduce(Chain& chain) {
  std::cout << "--- reducechain" << std::endl;
  FOR(i, 0, reduced_pairs_.size() - 1)
    reduced_pairs_[i].reduce(chain);
}

template<typename freeModuleType, typename P_GeneratorCode>
inline void ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>::restore(Chain& chain) {
  std::cout << "--- restore" << std::endl;
  FORD(i, reduced_pairs_.size() - 1, 0)
    reduced_pairs_[i].restore(chain);
}

template<typename freeModuleType, typename P_GeneratorCode>
std::ostream& operator<<(std::ostream& out, const ReducibleFreeChainComplex<freeModuleType, P_GeneratorCode>& rfcc) {
  out << "--- cout" << std::endl;
  return out;
}

}

#endif // _EXPERIMENTAL_REDUCIBLEFREECHAINCOMPLEX_HPP_

/// @}
