/////////////////////////////////////////////////////////////////////////////
/// @file Mathematica
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-10-01
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////


#include <WolframLibrary.h>
#include <map>
#include <capd/apiRedHom/Complex.h>
#include <boost/assert.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/make_shared.hpp>
#include <capd/auxil/Logger.h>
#include <capd/apiRedHom/Algorithms.h>
#include <capd/auxil/ignoreUnused.h>

// http://reference.wolfram.com/language/LibraryLink/tutorial/Examples.html

namespace capd
{

}


using namespace capd;


namespace
{


  template<typename Object>
  class MemoryManagment
  {
    typedef std::map<mint, boost::shared_ptr<Object> > Memory;

  public:
    size_t size() const { return _memory.size(); }

    void reserve(mint id)
    {
      _memory[id].reset(static_cast<Object*>(NULL));
    }

    void assign(mint id, boost::shared_ptr<Object> obj)
    {
      _memory[id] = obj;
    }

    void erase(mint id)
    {
      _memory.erase(id);
    }

    Object& operator[](mint id)
    {
      typename Memory::iterator it = _memory.find(id);
      BOOST_ASSERT(_memory.end() != it);
      return *(it->second);
    }

  private:
    Memory _memory;
  };


  typedef MemoryManagment<apiRedHom::Complex> Complexes;
  static Complexes complexes;
  static MemoryManagment<std::string> homologySignatures;
}



EXTERN_C DLLEXPORT void manage_complexes(WolframLibraryData libData, mbool mode, mint id)
{
  capd::auxil::ignoreUnused(libData);

  if (mode == 0) {
    CAPD_DEBUG("Create complex id: " << id);
    complexes.reserve(id);
    homologySignatures.reserve(id);
  } else {
    CAPD_DEBUG("Erase complex id: " << id);
    complexes.erase(id);
    homologySignatures.erase(id);
  }
}

EXTERN_C DLLEXPORT mint WolframLibrary_getVersion( ) {
    return WolframLibraryVersion;
}

EXTERN_C DLLEXPORT int WolframLibrary_initialize( WolframLibraryData libData) {
  capd::auxil::ignoreUnused(libData);
  INIT_CAPD_LOGGER;
  return (*libData->registerLibraryExpressionManager)("RedHomCOMPLEXES", manage_complexes);
}

EXTERN_C DLLEXPORT void WolframLibrary_uninitialize( WolframLibraryData libData)
{
  capd::auxil::ignoreUnused(libData);
  (*libData->unregisterLibraryExpressionManager)("RedHomCOMPLEXES");
}


EXTERN_C DLLEXPORT int RedHom_Mathematica_GetComplexesCount(WolframLibraryData libData,
                                               mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);
  mint len = complexes.size();
  CAPD_DEBUG("Complexes count: " << len);
  MArgument_setInteger(Res, len);

  return LIBRARY_NO_ERROR;
}

EXTERN_C DLLEXPORT int RedHom_Mathematica_ComplexCardinality(WolframLibraryData libData,
                                                             mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);
  mint id = MArgument_getInteger(Args[0]);

  MArgument_setInteger(Res, complexes[id].cardinality());
  return LIBRARY_NO_ERROR;
}

EXTERN_C DLLEXPORT int RedHom_Mathematica_ComplexDim(WolframLibraryData libData,
                                                     mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);
  mint id = MArgument_getInteger(Args[0]);
  MArgument_setInteger(Res, complexes[id].dim());

  return LIBRARY_NO_ERROR;
}

EXTERN_C DLLEXPORT int RedHom_Mathematica_CreateSimplicialComplex(WolframLibraryData libData,
                                                      mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);
  mint id = MArgument_getInteger(Args[0]);
  CAPD_DEBUG("Create simplicial complex id: " << id);
  complexes.assign(id, boost::make_shared<apiRedHom::SimplicialComplex>());
  CAPD_DEBUG("...created");
  return LIBRARY_NO_ERROR;
}

EXTERN_C DLLEXPORT int RedHom_Mathematica_CreateCubicalComplex(WolframLibraryData libData,
                                                               mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);
  mint id = MArgument_getInteger(Args[0]);
  MTensor msize = MArgument_getMTensor(Args[1]);
  const mint mdim = libData->MTensor_getFlattenedLength(msize);
  mint* mdata = libData->MTensor_getIntegerData(msize);
  std::vector<size_t> size(mdim);

  CAPD_DEBUG("Create cubical complex id: " << id << " dim: " << mdim);

  for (size_t idx = 0; idx < size.size(); ++idx) {
    size[idx] = mdata[idx];
  }

  complexes.assign(id, boost::make_shared<apiRedHom::CubicalComplex>(size));
  CAPD_DEBUG("...created");
  return LIBRARY_NO_ERROR;
}

EXTERN_C DLLEXPORT int RedHom_Mathematica_InsertSimplex(WolframLibraryData libData,
                                                        mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);
  mint id = MArgument_getInteger(Args[0]);
  MTensor simplices = MArgument_getMTensor(Args[1]);
  mint len = libData->MTensor_getFlattenedLength(simplices);
  mint* data = libData->MTensor_getIntegerData(simplices);

  CAPD_DEBUG("Inserting simplex with len: " << len);

  apiRedHom::SimplicialComplex& complex = dynamic_cast<apiRedHom::SimplicialComplex&>(complexes[id]);
  apiRedHom::SimplicialComplex::CellCode cellCode(len);

  for (size_t idx = 0; idx < cellCode.size(); ++idx) {
    cellCode[idx] = data[idx];
  }

  std::sort(cellCode.begin(), cellCode.end());
  complex.insert(cellCode);

  return LIBRARY_NO_ERROR;
}


EXTERN_C DLLEXPORT int RedHom_Mathematica_InsertCube(WolframLibraryData libData,
                                                        mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);
  mint id = MArgument_getInteger(Args[0]);
  MTensor mcube = MArgument_getMTensor(Args[1]);
  mint const* mdims = libData->MTensor_getDimensions(mcube);
  const mint mdim = mdims[0];

  CAPD_DEBUG("Inserting cube with dim: " << mdim);

  apiRedHom::CubicalComplex& complex = dynamic_cast<apiRedHom::CubicalComplex&>(complexes[id]);
  apiRedHom::CubicalComplex::CellCode cellCode(mdim);

  for (size_t idx = 0; idx < cellCode.size(); ++idx) {
    const mint midx = static_cast<mint>(idx);
    mint pos[2][2] = {{midx + 1, 1}, {midx + 1, 2}};
    mint interval[2];
    libData->MTensor_getInteger(mcube, pos[0], &interval[0]);
    libData->MTensor_getInteger(mcube, pos[1], &interval[1]);

    cellCode[idx].first = interval[0];
    cellCode[idx].second = interval[1] - interval[0];
  }

  complex.insert(cellCode);

  return LIBRARY_NO_ERROR;
}

EXTERN_C DLLEXPORT int RedHom_Mathematica_FillWithBoundaries(WolframLibraryData libData,
                                                        mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);
  const mint id = MArgument_getInteger(Args[0]);
  CAPD_DEBUG("Filling with boundaries " << id);

  apiRedHom::CubicalComplex& complex = dynamic_cast<apiRedHom::CubicalComplex&>(complexes[id]);
  complex.fillWithBoundaries();

  return LIBRARY_NO_ERROR;
}


namespace
{
  int ConvertBettiVector(WolframLibraryData libData, const std::vector<int>& betti, MArgument Res)
  {
    MTensor result;
    mint dims[1] = {static_cast<mint>(betti.size())};
    int err = (*libData->MTensor_new)(MType_Integer, 1, dims, &result);
    if (err) return err;

    mint* data = libData->MTensor_getIntegerData(result);
    for (size_t idx = 0; idx < betti.size(); ++idx) {
      data[idx] = betti[idx];
    }

    MArgument_setMTensor(Res, result);

    return err;
  }

}

EXTERN_C DLLEXPORT int RedHom_Mathematica_BettiNumbersOverZ(WolframLibraryData libData,
                                                       mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);
  mint id = MArgument_getInteger(Args[0]);

  apiRedHom::BettiNumbersOverZ alg(complexes[id]);
  const std::vector<int> betti = alg();
  int err = ConvertBettiVector(libData, betti, Res);

  return err;
}

EXTERN_C DLLEXPORT int RedHom_Mathematica_BettiNumbersOverZ2(WolframLibraryData libData,
                                                       mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);
  mint id = MArgument_getInteger(Args[0]);

  apiRedHom::BettiNumbersOverZ2 alg(complexes[id]);
  const std::vector<int> betti = alg();
  int err = ConvertBettiVector(libData, betti, Res);

  return err;
}

EXTERN_C DLLEXPORT int RedHom_Mathematica_BettiNumbersOverZp(WolframLibraryData libData,
                                                       mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);
  mint id = MArgument_getInteger(Args[0]);
  const mint p = MArgument_getInteger(Args[1]);

  apiRedHom::BettiNumbersOverZp alg(complexes[id], p);
  const std::vector<int> betti = alg();
  int err = ConvertBettiVector(libData, betti, Res);

  return err;
}

EXTERN_C DLLEXPORT int RedHom_Mathematica_HomologyOverZ(WolframLibraryData libData,
                                                   mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);
  const mint id = MArgument_getInteger(Args[0]);

  apiRedHom::HomologyOverZ alg(complexes[id]);
  const std::string signature = alg().str;
  homologySignatures.assign(id, boost::make_shared<std::string>(signature));
  MArgument_setUTF8String(Res,  const_cast<char *>(homologySignatures[id].c_str()));

  return LIBRARY_NO_ERROR;
}

EXTERN_C DLLEXPORT int RedHom_Mathematica_HomologyOverZ2(WolframLibraryData libData,
                                                   mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);
  const mint id = MArgument_getInteger(Args[0]);

  apiRedHom::HomologyOverZ2 alg(complexes[id]);
  const std::string signature = alg().str;
  homologySignatures.assign(id, boost::make_shared<std::string>(signature));
  MArgument_setUTF8String(Res,  const_cast<char *>(homologySignatures[id].c_str()));

  return LIBRARY_NO_ERROR;
}

EXTERN_C DLLEXPORT int RedHom_Mathematica_HomologyOverZp(WolframLibraryData libData,
                                                   mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);
  const mint id = MArgument_getInteger(Args[0]);
  const mint p = MArgument_getInteger(Args[1]);

  apiRedHom::HomologyOverZp alg(complexes[id], p);
  const std::string signature = alg().str;
  homologySignatures.assign(id, boost::make_shared<std::string>(signature));
  MArgument_setUTF8String(Res,  const_cast<char *>(homologySignatures[id].c_str()));

  return LIBRARY_NO_ERROR;
}


#include <capd/persistence/AbstractPointCloud.h>
#include <capd/persistence/PointCloudTypes.h>
#include <capd/persistence/PointCloudUtils.h>
#include <capd/persistence/PointCloudSimplicialCellSComplex.h>
#include <capd/persistence/ComputePersistentHomology.h>
#include <boost/foreach.hpp>

EXTERN_C DLLEXPORT int RedHom_Mathematica_PointCloudPersistentHomology(WolframLibraryData libData,
                                                                       mint Argc, MArgument *Args, MArgument Res)
{
  capd::auxil::ignoreUnused(libData);
  capd::auxil::ignoreUnused(Argc);
  capd::auxil::ignoreUnused(Args);
  capd::auxil::ignoreUnused(Res);

  MTensor mpoints = MArgument_getMTensor(Args[0]);
  mint maxComplexDim = MArgument_getInteger(Args[1]);
  mreal maxDist = MArgument_getReal(Args[2]);

  mint mlen = libData->MTensor_getFlattenedLength(mpoints);
  mint mrank = libData->MTensor_getRank(mpoints);
  const mint* mdim = libData->MTensor_getDimensions(mpoints);
  mreal* mdata = libData->MTensor_getRealData(mpoints);

  CAPD_DEBUG("Point cloud data len: " << mlen);
  CAPD_DEBUG("Point cloud data rank: " << mrank);
  CAPD_DEBUG("Point cloud data dim: " << mdim[0] << " " << mdim[1]);

  BOOST_ASSERT(mrank == 2);

  if (maxComplexDim == 0) {
    CAPD_ERROR("Why do you compute persistence for max complex dimension == " << maxComplexDim);
    MTensor result;
    mint dims[2] = {0, 2};
    int err = (*libData->MTensor_new)(MType_Real, 2, dims, &result);
    MArgument_setMTensor(Res, result);
    return err;
  }

  const size_t dim = mdim[1];
  std::vector<capd::vectalg::Vector<double, 0> > points(mlen/dim);

  for (size_t idx = 0; idx < points.size(); ++idx) {
    points[idx] = capd::vectalg::Vector<double, 0>(dim);
    for (size_t d = 0; d < dim; ++d) {
      points[idx][d] = mdata[idx*dim + d];
    }
  }

  BOOST_ASSERT(points.size() > 0 && points[0].dimension() == dim);

  CAPD_DEBUG("Computing persistence homology");


  using namespace capd::persistence;
  typedef EuclideanPointCloudType<0> Traits;
  typedef Traits::PointCloudType PointCloud;
  boost::scoped_ptr<PointCloud> pc;


  pc.reset(makeSizeLimitedPointCloud<Traits>(points.begin(), points.end(),
                                             0, maxDist));

  CAPD_INFO("Point Cloud with: " << pc->size() << " points and " << pc->edges() << " edges constructed");
  CAPD_INFO("Building simplicial complex with maximu dimension: " << maxComplexDim);

  typedef PointCloudSimplicialCellSComplex PCComplex;
  PCComplex pccomplex(*pc, maxComplexDim);
  pccomplex();

  CAPD_INFO("Complex constructed, its size: " << pccomplex.size());
  CAPD_INFO("Computing persistence homology");

  typedef capd::persistence::ComputePersistentHomology<PointCloudSimplicialCellSComplex> ComputePersistentHomology;
  ComputePersistentHomology persistence(pccomplex);
  persistence();

  CAPD_INFO("TOTAL Persistent pairs: " << persistence.pairs().size() << ", cycles: " << persistence.cycles().size());

  const int persistenceDim = maxComplexDim - 1;

  ComputePersistentHomology::Diagram diagram = persistence.diagram(persistenceDim);
  ComputePersistentHomology::Pairs pairs = persistence.pairs(persistenceDim);
  ComputePersistentHomology::Cycles cycles = persistence.cycles(persistenceDim);

  CAPD_INFO("Diagram size in dim: " << (int)persistenceDim << " diagram: " << diagram.size() << ", pairs: " << pairs.size() << ", cycles: " << cycles.size());


  MTensor result;
  mint dims[2] = {static_cast<mint>(diagram.size()), 2};
  int err = (*libData->MTensor_new)(MType_Real, 2, dims, &result);
  if (err) return err;

  mreal* resultData = libData->MTensor_getRealData(result);
  for (size_t idx = 0; idx < diagram.size(); ++idx) {
    resultData[2*idx] = diagram[idx].first;
    resultData[2*idx + 1] = diagram[idx].second;
  }

  MArgument_setMTensor(Res, result);

  return err;
}
