/////////////////////////////////////////////////////////////////////////////
/// @file helpers.cpp
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2013-12-10
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2013 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.


#include "helpers.h"
#include <capd/apiRedHom/c_interface.hpp>

#include <capd/apiRedHom/Complex.h>

#include <boost/lambda/lambda.hpp>

#include <assert.h>
#include <vector>
#include <algorithm>
#include <deque>
#include <cstdio>
#include <algorithm>
#include <capd/apiRedHom/ComplexHelper.h>


using namespace capd::apiRedHom;

namespace
{

  struct DoConvertPureCubicalComplex
  {
    RedHom_ComplexInfo* complex;
    Obj pureCubicalComplex;

    Obj binaryArray;
    int dimension;
    std::vector<int> size;
    std::vector<int> coord;

    DoConvertPureCubicalComplex(RedHom_ComplexInfo* complex, Obj pureCubicalComplex):
      complex(complex),
      pureCubicalComplex(pureCubicalComplex)
    {
      Obj _dimension = Call_1_Wrapper(Dimension, pureCubicalComplex);
      Obj _size = Call_2_Wrapper(EvaluateProperty, pureCubicalComplex, GAPString("arraySize"));

      binaryArray = ElmPRec(pureCubicalComplex, RNamName("binaryArray"));
      dimension = INT_INTOBJ(_dimension);

      size.resize(dimension);
      coord.resize(dimension);

      for (int d = 0; d < dimension; ++d) {
	size[d] = INT_INTOBJ(ELM_PLIST(_size, d + 1));
      }
    }

    virtual void emit() = 0;

    void go(const int dim)
    {
      if (dim == dimension) {
        Obj value = Call_2_Wrapper(ArrayValue, binaryArray, GAPIntList(&coord[0], coord.size()));

        if (INT_INTOBJ(value) == 1) {
          emit();
        }

        return;
      }

      assert(dim <= dimension);

      const size_t length = size[dim];

      for (size_t pos = 1; pos <= length; ++pos) {
        coord[dim] = pos;
        //coord[dim] = pos - 1;
        go(dim + 1);
      }
    }

    virtual void operator()() = 0;

  };


  struct DoConvertPureCubicalComplexToCubSComplex: public DoConvertPureCubicalComplex
  {
    std::vector<int> allOnes;

    DoConvertPureCubicalComplexToCubSComplex(RedHom_ComplexInfo* complex, Obj pureCubicalComplex):
      DoConvertPureCubicalComplex(complex, pureCubicalComplex)
    {
      allOnes.resize(dimension);
      std::fill(allOnes.begin(), allOnes.end(), 1);

      RedHom_CreateCubicalComplex(complex, dimension, &size[0]);
    }

    void emit()
    {
      std::vector<int> coord(this->coord.begin(), this->coord.end());
      std::for_each(coord.begin(), coord.end(), boost::lambda::_1 -= 1);
      RedHom_AddCubeToComplex(complex, &coord[0], &allOnes[0]);
    }

    void operator()()
    {
      go(0);
      RedHom_CubicalComplexFillWithBoundaries(complex);
    }

  };

  struct DoConvertPureCubicalComplexToSComplex: public DoConvertPureCubicalComplex
  {
    std::deque<std::vector<size_t> > cubes;

    DoConvertPureCubicalComplexToSComplex(RedHom_ComplexInfo* complex, Obj pureCubicalComplex):
      DoConvertPureCubicalComplex(complex, pureCubicalComplex)
    {
    }

    void emit()
    {
      cubes.push_back(std::vector<size_t>(coord.begin(), coord.end()));
      // std::cerr << "CUBE";
      // for (int i = 0; i < coord.size(); ++i)
      // 	{
      // 	  std::cerr << " " << coord[i];
      // 	}
      // std::cerr << std::endl;
   }

    void  operator()()
    {
      go(0);
      SetComplex(complex, CubicalHelper::CreateGeneralizedSComplexFromFullCubes(dimension,
                                                                                cubes));
    }

  };

}

void ConvertPureCubicalComplex(Obj pureCubicalComplex, RedHom_ComplexInfo* complex)
{
  DoConvertPureCubicalComplexToCubSComplex doit(complex, pureCubicalComplex);
  doit();
}

void ConvertPureCubicalComplexToSComplex(Obj pureCubicalComplex, RedHom_ComplexInfo* complex)
{
  DoConvertPureCubicalComplexToSComplex doit(complex, pureCubicalComplex);
  doit();
}

int ConvertGAPIntList(Obj list, int** data)
{
  const int size = LEN_PLIST(list);
  *data = (int*) malloc(sizeof(int) * size);

  for (int i = 0; i < size; ++i) {
    (*data)[i] = INT_INTOBJ(ELM_PLIST(list, i + 1));
  }

  return size;
}

Obj GAPString(const char* str)
{
  const int len = strlen(str);
  Obj gapStr = NEW_STRING(len);
  C_NEW_STRING(gapStr, len, str);
  return gapStr;
}


Obj GAPBettiList(const int* betti, const int dim)
{
  const size_t len = dim > 0 ? dim : 1;
  Obj gapBetti = NEW_PLIST(T_PLIST_HOM, len);
  SET_LEN_PLIST(gapBetti, len);

  if (dim == 0) {
    SET_ELM_PLIST(gapBetti, 1, INTOBJ_INT(0));
    CHANGED_BAG(gapBetti);
  }

  for (int i = 0; i < dim; ++i) {
    SET_ELM_PLIST(gapBetti, i + 1, INTOBJ_INT(betti[i]));
    CHANGED_BAG(gapBetti);
  }

  return gapBetti;
}

Obj GAPIntList(const int* array, const unsigned long size)
{
  Obj gapList = NEW_PLIST(T_PLIST_HOM, size);
  SET_LEN_PLIST(gapList, size);

  for (size_t i = 0; i < size; ++i) {
    SET_ELM_PLIST(gapList, i + 1, INTOBJ_INT(array[i]));
    CHANGED_BAG(gapList);
  }

  return gapList;
}
