#include <iostream>
#include <fstream>

#include <boost/algorithm/string/join.hpp>

#include <boost/mpl/push_back.hpp>
#include <boost/mpl/assert.hpp>
#include <boost/mpl/fold.hpp>
#include <boost/mpl/int.hpp>
#include <boost/mpl/apply.hpp>
#include <boost/mpl/copy.hpp>
#include <boost/mpl/back_inserter.hpp>
#include <boost/mpl/range_c.hpp>
#include <boost/mpl/vector.hpp>
#include <boost/mpl/joint_view.hpp>
#include <boost/mpl/for_each.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/ref.hpp>
#include <boost/tuple/tuple_io.hpp>


#define USE_PRECOMPILED_CAPDREDHOM_COMPLEX_CUBSCOMPLEX_HPP

#include <capd/complex/CubSComplex.h>
#include <capd/complex/SimplexSComplex.hpp>
#include <capd/complex/SimplexIO.hpp>
#include <capd/complex/SComplex.h>
#include <capd/complex/SComplexDefaultTraits.h>
#include <capd/complex/ComputeHomologyOrchestrator.h>

using namespace capd::complex;

#include <capd/complex/CubSComplexReader.hpp>
#include <redHom/util/RedHomOptions.h>
#include <capd/complex/SComplexReader.hpp>
#include "SComplexBuilder.hpp"
#include "SComplexBuilderFromSimplices.hpp"

#include <capd/rings/Zp.h>
#include <capd/multiPrec/MpInt.h>
#include <capd/auxil/Logger.h>
#include <capd/homAlgebra/HomologySignature.h>
#include <capd/complex/EulerCharacteristic.h>
#include <capd/rings/Z2.h>
#include <capd/cubSet/BmpCubCelSetBuilder.hpp>
#include <capd/cubSet/readCubicalSet.hpp>




#ifndef MAX_CUBICAL_COMPLEX_DIM
#define MAX_CUBICAL_COMPLEX_DIM 10
#endif

enum ComplexMajorId {SComplexMajorId = 10,
                     SComplexSimplicesMajorId,
                     SComplexCubesMajorId,
                     CubSComplexMajorId,
                     SimplexSComplexMajorId};

typedef boost::tuple<ComplexMajorId, int, boost::any> ComplexTuple;

template<typename Complex, ComplexMajorId MAJOR_ID, int MINOR_ID>
struct ComplexDescriptor {
  typedef Complex type;
  static const ComplexMajorId majorId = MAJOR_ID;
  static const int minorId = MINOR_ID;
};


struct CubComplexDescriptor {
  template<typename DIM>
  struct apply {
    typedef ComplexDescriptor<CubSComplex<DIM::value>, CubSComplexMajorId, DIM::value> type;
  };
};

typedef boost::mpl::fold<boost::mpl::range_c<int,2, MAX_CUBICAL_COMPLEX_DIM + 1>, boost::mpl::vector<>,
                         boost::mpl::push_back<boost::mpl::_1, CubComplexDescriptor::apply<boost::mpl::_2> > >::type CubComplexDescriptors;

BOOST_MPL_ASSERT_RELATION( boost::mpl::size<CubComplexDescriptors>::value, ==, MAX_CUBICAL_COMPLEX_DIM - 1 );



typedef boost::mpl::joint_view<CubComplexDescriptors,
                               boost::mpl::vector<
                                 ComplexDescriptor<SComplex<SComplexDefaultTraits>, SComplexMajorId, 0>,
                                 ComplexDescriptor<SComplex<SComplexDefaultTraits>, SComplexCubesMajorId, 0>,
                                 ComplexDescriptor<SComplex<SComplexDefaultTraits>, SComplexSimplicesMajorId, 0>,
                                 ComplexDescriptor<SimplexSComplex, SimplexSComplexMajorId, 0>
                               >
                              > ComplexDescriptors;

using namespace redHom::util;

class MultiDimCubSComplexReader {

  struct ComplexConstructor {
    const int dim;
    std::ifstream& file;
    ComplexTuple& complex;

    ComplexConstructor(int _dim,  std::ifstream& _file, ComplexTuple& _complex):
      dim(_dim), file(_file), complex(_complex) {
      BOOST_ASSERT(boost::get<2>(complex).empty());
    }

    template<typename Complex, ComplexMajorId MajorId, int MinorId>
    void operator()(ComplexDescriptor<Complex, MajorId, MinorId>) {
      if (MajorId == CubSComplexMajorId && dim == MinorId) {
        BOOST_ASSERT(boost::get<2>(complex).empty());

        typedef typename Complex::BCubCellSet CubCellSet;
        CRef<CubCellSet> cubCellSetCR(new CubCellSet());
        BmpCubCelSetBuilder<CubCellSet> csb(cubCellSetCR);
        readCubicalSet(file, csb);

        complex = ComplexTuple(MajorId, MinorId,
                               boost::any(boost::shared_ptr<Complex>(new Complex(cubCellSetCR))));
      }
    }

  };

public:
  static ComplexTuple read(ifstream& file, const std::string& fileName, RedHomOptions::FileType fileType) {
    if (fileType != RedHomOptions::FT_CUB && fileType != RedHomOptions::FT_PCUB && fileType != RedHomOptions::FT_FCUB && fileType != RedHomOptions::FT_BMD) {
      throw std::runtime_error("Wrong file type for cubical complex");
    }

    ComplexTuple complex;

    CubSetBuilder csb;
    readCubicalSet(file,csb,true); //true means only get dim and type
    file.seekg(0);

    unsigned int embDim=csb.getDim();

    ComplexConstructor constructor(embDim, file, complex);
    boost::mpl::for_each<CubComplexDescriptors>(constructor);

    return complex;
  }

};

class SComplexReaderFromCubes {

public:
  static ComplexTuple read(ifstream& file, const std::string& fileName, RedHomOptions::FileType fileType) {
    if (fileType != RedHomOptions::FT_CUB && fileType != RedHomOptions::FT_PCUB && fileType != RedHomOptions::FT_FCUB) {
      throw std::runtime_error("Wrong file type for general cubical complex");
    }

    SComplexReader<SComplexDefaultTraits> reader;

    return boost::make_tuple(SComplexCubesMajorId, 0, reader(file, 3, 1));
  }

};

class SComplexReaderFromSimplices {

public:
  static ComplexTuple read(ifstream& file, const std::string& fileName, RedHomOptions::FileType fileType) {
    std::vector<std::set<int> > simplices;
    if (fileType == RedHomOptions::FT_DAT) {
      simplices = parseDat(file);
    } else if (fileType == RedHomOptions::FT_OBJ) {
      simplices = parseObj(file);
    } else {
      throw std::runtime_error("Wrong file type for general simplicial complex");
    }

    SComplexBuilderFromSimplices<long, SComplexDefaultTraits> builder(3);
    boost::shared_ptr<SComplex<SComplexDefaultTraits> > complex = builder(simplices, 3, 1);

    return boost::make_tuple(SComplexSimplicesMajorId, 0, complex);
  }

};

class SComplexReaderGeneral {

public:
  static ComplexTuple read(ifstream& file, const std::string& fileName, RedHomOptions::FileType fileType) {
    boost::shared_ptr<SComplex<SComplexDefaultTraits> > complex;

    if (fileType == RedHomOptions::FT_GEN) {
      typedef SComplex<SComplexDefaultTraits> Complex;
      Complex::Size complexSize = 0, kappaSize = 0;
      file >> complexSize >> kappaSize;

      Complex::KappaMap kappaMap(kappaSize);
      Complex::Dims dims(complexSize);

      file >> boost::tuples::set_open('[') >> boost::tuples::set_close(']') >> boost::tuples::set_delimiter(':');

      for (Complex::Size cellNum = 0; cellNum < complexSize; ++cellNum) {
        boost::tuple<Complex::Id, Complex::Dim, Complex::Color> cell;
        file >> cell;
        if (!(boost::get<0>(cell) >= 0 && boost::get<0>(cell) < complexSize)) {
          throw std::runtime_error("Only ids from range 0..complexSize supported for reading");
        }

        if (boost::get<2>(cell) != 1) {
          throw std::runtime_error("Only color #1 supported for reading");
        }

        dims[boost::get<0>(cell)] = boost::get<1>(cell);
      }

      file >> boost::tuples::set_open('(') >> boost::tuples::set_close(')') >> boost::tuples::set_delimiter(',');
      for (Complex::Size k = 0; k < kappaSize; ++k) {
        file >> kappaMap[k];
      }

      std::cout << "Creating general SComplex,  size " << dims.size() << ", kappa size " << kappaMap.size() << std::endl;
      complex.reset(new Complex(3, dims, kappaMap, 1));
    } else {
      throw std::runtime_error("Wrong file type for general simplicial complex");
    }

    return boost::make_tuple(SComplexSimplicesMajorId, 0, complex);
  }

};

class SimplexSComplexReader {

public:
  static ComplexTuple read(ifstream& file, const std::string& fileName, RedHomOptions::FileType fileType) {
    boost::shared_ptr<SimplexSComplex> complex(new SimplexSComplex());
    if (fileType == RedHomOptions::FT_DAT) {
      parseDat(file, *complex);
    } else     if (fileType == RedHomOptions::FT_OBJ) {
      parseObj(file, *complex);
    } else {
      throw std::runtime_error("Wrong file type for simplicial complex");
    }

    return boost::make_tuple(SimplexSComplexMajorId, 0, complex);
  }

};


template<typename ComplexReader, int MAJOR_ID>
struct ComplexReaderDescriptor {
  typedef ComplexReader type;
  static const int majorId = MAJOR_ID;
};

typedef boost::mpl::vector<ComplexReaderDescriptor<MultiDimCubSComplexReader, CubSComplexMajorId>,
                           ComplexReaderDescriptor<SComplexReaderFromCubes, SComplexCubesMajorId>,
                           ComplexReaderDescriptor<SComplexReaderFromSimplices, SComplexSimplicesMajorId>,
                           ComplexReaderDescriptor<SComplexReaderGeneral, SComplexMajorId>,
                           ComplexReaderDescriptor<SimplexSComplexReader, SimplexSComplexMajorId>
                           > ComplexReaderDescriptors;


struct ComplexReader {
  const ComplexMajorId majorId;
  ComplexTuple& complex;
  const std::string fileName;
  RedHomOptions::FileType fileType;

  ComplexReader(ComplexMajorId _majorId, ComplexTuple& _complex, const std::string _fileName, RedHomOptions::FileType _fileType):
    majorId(_majorId), complex(_complex), fileName(_fileName), fileType(_fileType) {
    BOOST_ASSERT(boost::get<2>(complex).empty());
  }

  template<typename Reader, int MajorId>
  void operator()(ComplexReaderDescriptor<Reader, MajorId>) {
    if (MajorId == majorId) {
      BOOST_ASSERT(boost::get<2>(complex).empty());
      ifstream file;

      file.open(fileName.c_str());
      if (!file) {
        throw std::logic_error("File not found: " + fileName);
      }

      complex = Reader::read(file, fileName, fileType);

      file.close();
    }
  }

};


template<typename ComplexT, typename ScalarT, typename IntT>
class RedHomOrchestrator
{
  typedef ComplexT Complex;
  typedef ScalarT Scalar;

public:
  RedHomOrchestrator():
    extractedSignature(HomologySignature<Scalar, IntT>())
  {}

  RedHomOrchestrator(const HomologySignature<Scalar, IntT>& extractedSignature):
    extractedSignature(extractedSignature)
  {}

  void operator()(const RedHomOptions& options, Complex& complex)
  {
    Stopwatch stopwatch;
    const long eulerCharacteristic = EulerCharacteristic<ComplexT>(complex)();
    fcout << "Euler characteristic is: \n\n" << eulerCharacteristic << std::endl;

    ComputeHomologyOrchestrator<ComplexT, ScalarT, IntT> computeHomology(complex,
                                                                         options.isShave(),
                                                                         options.isCoreduce(),
                                                                         options.isMorse(), options.isKMS());

    std::vector<IntT> betti = computeHomology();
    HomologySignature<Scalar, IntT> result = computeHomology.signature();
    const long reducedEulerCharacteristic = EulerCharacteristic<ComplexT>(complex)();

    fcout << "Signature calculated in: " << stopwatch << std::endl;
    fcout << "-----------------\n"  << std::endl;
    fcout << "Computed homology is: \n\n" << result  << std::endl;
    fcout << "Reduced Euler characteristic is: \n\n" << reducedEulerCharacteristic << std::endl;
  }

private:
  HomologySignature<Scalar, IntT> extractedSignature;
};

namespace {

  void validateOptions(const RedHomOptions& options) throw(std::logic_error) {
    if (options.getInputFilesCount() != 1) {
      throw std::logic_error("Support exactly one input file");
    }

    if (options.getModulo() != 0 && options.isBigInt()) {
      throw std::logic_error("Do not support big integers and modulo calculations");
    }
  }

  ComplexMajorId determineComplexMajorId(const RedHomOptions& options) throw(std::logic_error) {
    if (options.isCubicalFileType() && ! options.isGeneralSComplex()) {
      fcout << "Using CubSComplex" << std::endl;
      return CubSComplexMajorId;
    } else if (options.isCubicalFileType() && options.isGeneralSComplex()) {
      fcout << "Using SComplex for cubes" << std::endl;
      return SComplexCubesMajorId;
    } else if (options.isSimplicialFileType() && !options.isGeneralSComplex()) {
      fcout << "Using SimplexSComplex" << std::endl;
      return SimplexSComplexMajorId;
    } else if (options.isSimplicialFileType() && options.isGeneralSComplex()) {
      fcout << "Using SComplex for simplices" << std::endl;
      return SComplexSimplicesMajorId;
    } else if (options.isGeneralFileType()) {
      std::cout << "Using general SComplex" << std::endl;
      return SComplexMajorId;
    }

    throw logic_error("Unsupported file reader type");
  }

}


template<typename Derive>
class UniversalAlgorithmExecutor {
  std::string name;

protected:
  ComplexTuple& complex;

  UniversalAlgorithmExecutor(ComplexTuple& _complex, const std::string _name):
    name(_name), complex(_complex) {
  }

public:

  template<typename Complex, ComplexMajorId MajorId, int MinorId>
  void operator()(ComplexDescriptor<Complex, MajorId, MinorId>) {
    if (boost::get<0>(complex) == MajorId && boost::get<1>(complex) == MinorId) {
      Stopwatch stopwatch;
      boost::shared_ptr<Complex>& complexPtr = boost::any_cast<boost::shared_ptr<Complex>&>(boost::get<2>(complex));

      fcout << "Starting: " << name << std::endl;
      ((Derive*)this)->run(*complexPtr);
      fcout << "Finished: " << name << " in: " << stopwatch << std::endl;
    }
  }

};

template<typename ScalarT, typename IntT>
struct RedHomOrchestratorExecutor: public UniversalAlgorithmExecutor<RedHomOrchestratorExecutor<ScalarT, IntT> >
{
  const RedHomOptions& options;

  RedHomOrchestratorExecutor(ComplexTuple& complex, const RedHomOptions& options) :
    UniversalAlgorithmExecutor<RedHomOrchestratorExecutor<ScalarT, IntT> >(complex, "RedHomOrchestrator"),
    options(options)
  {}

  template<typename Complex>
  void run(Complex& complex)
  {
    RedHomOrchestrator<Complex, ScalarT, IntT> orchestrator;
    orchestrator(this->options, complex);
  }

};


int main(int argc, const char** argv) {
  INIT_CAPD_LOGGER;

  try {
    fcout.open("capd-report.txt");
    fcout.turnOn();

    std::string executableName("redHom");
    std::string version(" 0.9 ");
    std::string copyright("(C) CAPD::RedHom Group, 2010");
    std::string logo(executableName+version+copyright);
    std::string webpage("http://redhom.ii.uj.edu.pl");

    fcout << logo  << std::endl;
    fcout << webpage  << std::endl;
    fcout << "-----------------"  << std::endl;

    Stopwatch totalStopwatch;
    RedHomOptions options(argc, argv);

    if (options.isHelp()) {
      fcout << options.getHelp() << std::endl;
      return 0;
    }

    try {
      validateOptions(options);
    } catch(...) {
      fcout << options.getHelp() << std::endl;
      throw;
    }

    ComplexTuple complex;
    ComplexMajorId complexMajorId = determineComplexMajorId(options);
    ComplexReader complexReader(complexMajorId, complex, options.getInputFile(0), options.getFileType());

    Stopwatch readStopwatch;
    boost::mpl::for_each<ComplexReaderDescriptors>(complexReader);
    fcout << "Complex read in: " << readStopwatch << std::endl;

    if (options.isBigInt()) {
#ifdef __HAVE_MPFR__
      RedHomOrchestratorExecutor<capd::multiPrec::MpInt, capd::multiPrec::MpInt> orchestrator(complex, options);
      boost::mpl::for_each<ComplexDescriptors>(orchestrator);
#else
      throw std::runtime_error("Please install libmpfr, reconfigure and recompile.");
#endif
    } else if (options.getModulo() == 0) {
      RedHomOrchestratorExecutor<int, int> orchestrator(complex, options);
      boost::mpl::for_each<ComplexDescriptors>(orchestrator);
    } else if (options.getModulo() == 2) {
      RedHomOrchestratorExecutor<Z2, int> orchestrator(complex, options);
      CAPD_INFO("Modulo: " << 2);
      boost::mpl::for_each<ComplexDescriptors>(orchestrator);
    } else if (options.getModulo() != 0) {
      Zp::setMod(options.getModulo());
      RedHomOrchestratorExecutor<Zp, int> orchestrator(complex, options);
      CAPD_INFO("Modulo: " << Zp::getMod());
      boost::mpl::for_each<ComplexDescriptors>(orchestrator);
    }

    fcout << "Total execution: " << totalStopwatch << std::endl;
  } catch (std::exception& ex) {
    std::cerr << "Error: " << ex.what() << std::endl;
    return 1;
  } catch (const char* ex) {
    std::cerr << "Error: " << ex << std::endl;
    return 1;
  } catch (const std::string& ex) {
    std::cerr << "Error: " << ex << std::endl;
    return 1;
  } catch(...) {
    std::cerr << "Unknown exception" << std::endl;
    return 1;
  }

  return 0;
}
