#include "redHom/util/RedHomOptions.h"

using redHom::util::RedHomOptions;

RedHomOptions::RedHomOptions(int ac, const char** av, int inputFilesCount):
  desc("Allowed options"),
  inputFiles(inputFilesCount),
  generalSComplex(false),
  homBasis(false),
  rfcAtStart(false),
  experimentalRfcAtStart(false),
  rfcAtEnd(false),
  experimentalRfcAtEnd(false),
  noDefault(false),
  coreduce(false),
  shave(false),
  morse(false),
  kms(false),
  bmatrices(false),
  saveFiltration(false),
  fileType(FT_UNKNOWN),
  modulo(0),
  bigInt(false)
{
  using namespace boost::program_options;

  desc.add_options()
    ("help,h", "produce help message")
    ;

    desc.add_options()
      ("bmatrices", bool_switch(&bmatrices), "save boundary matrices to a file (sparse format accepted by LinBox)")
      ("filtration", bool_switch(&saveFiltration), "save complex filtration to a file (name = input file name + .filtration)")
      ;

    options_description fileDesc;
    fileDesc.add_options()
      ("fileType,f", value<FileType>(&fileType),
       "input file type: cub | pcub | fcub | bmd | obj | dat");

    positional_options_description positionalFileDesc;

    for (int i = 0; i < inputFilesCount; ++i) {
      std::string key = getInputFileKey(i);
      positionalFileDesc.add(key.c_str(), 1);
      fileDesc.add_options()(key.c_str(), value<std::string>(&inputFiles[i]), "input file");
    }

    desc.add(fileDesc);

    options_description complexDesc;
    complexDesc.add_options()
      ("general-scomplex,g", bool_switch(&generalSComplex), "use general SComplex");

    options_description algorithmDesc;
    algorithmDesc.add_options()
      ("RFC-at-start", bool_switch(&rfcAtStart), "execute RFC at start")
      ("ex-RFC-at-start", bool_switch(&experimentalRfcAtStart), "execute experimental RFC at start")
      ("coreduce", bool_switch(&coreduce), "execute coreduction algorithm")
      ("shave", bool_switch(&shave), "execute shave")
      ("morse", bool_switch(&morse), "use Discrete Morse approach")
      ("kms", bool_switch(&kms), "use KMS (algebraic reductions)")
      ("noDefault", bool_switch(&noDefault), "do not assume default algorithms flow")
      ("RFC-at-end", bool_switch(&rfcAtEnd), "execute RFC at end")
      ("ex-RFC-at-end", bool_switch(&experimentalRfcAtEnd), "execute experimental RFC at end")
      ("homBasis", bool_switch(&homBasis), "compute hom basis");

    desc.add(complexDesc);
    desc.add(algorithmDesc);
    desc.add_options()("statisticsFile", value<std::string>(&statisticsFile), "output statistics file, '-' - stdout");
    desc.add_options()("modulo", value<int>(&modulo), "modulus parameter for Z_p calculations. If p == 0 then integer calculations");
    desc.add_options()("bigInt", value<bool>(&bigInt), "use GMP big integers");
    store(command_line_parser(ac, av).options(desc).positional(positionalFileDesc).run(), vm);
    notify(vm);

}

bool RedHomOptions::isHelp() const {
  return vm.count("help");
}

std::string RedHomOptions::getHelp() const {
  std::ostringstream helpStream;
  helpStream << desc;

  return helpStream.str();
}

bool RedHomOptions::useDefaultAlgorithms() const {
  return !noDefault && (!morse && !coreduce && !shave && !kms);
}

RedHomOptions::FileType RedHomOptions::getFileType() const {
  if (fileType == FT_UNKNOWN) {
    std::string fileName = getInputFile(0);
    if (boost::iends_with(fileName, ".dat")) {
      return FT_DAT;
    } else if (boost::iends_with(fileName, ".obj")) {
      return FT_OBJ;
    } else if (boost::iends_with(fileName, ".cub")) {
      return FT_CUB;
    } else if (boost::iends_with(fileName, ".pcub")) {
      return FT_PCUB;
    } else if (boost::iends_with(fileName, ".fcub")) {
      return FT_FCUB;
    } else if (boost::iends_with(fileName, ".bmd")) {
      return FT_BMD;
    } else if (boost::iends_with(fileName, ".gen")) {
      return FT_GEN;
    }

  }
  return fileType;
}

bool RedHomOptions::isCubicalFileType() const {
  FileType ft = getFileType();
  return ft == FT_CUB || ft == FT_PCUB || ft == FT_FCUB || ft == FT_BMD;
}

bool RedHomOptions::isSimplicialFileType() const {
  FileType ft = getFileType();
  return ft == FT_OBJ || ft == FT_DAT;
}

bool RedHomOptions::isGeneralFileType() const {
  FileType ft = getFileType();
  return ft == FT_GEN;
}

const std::string& RedHomOptions::getInputFile(int i) const {
  static const std::string empty;
  return i < inputFiles.size() ? inputFiles[i] : empty;
}

size_t RedHomOptions::getInputFilesCount() const {
  size_t size = 0;
  for (size_t i = 0; i < inputFiles.size(); ++i) {
    size += vm.count(getInputFileKey(i));
  }
  return size;
}

const std::vector<std::string> RedHomOptions::getInputFiles() const {
  return inputFiles;
}

bool RedHomOptions::isGeneralSComplex() const {
  return generalSComplex;
}

bool RedHomOptions::isHomBasis() const {
  return homBasis;
}

bool RedHomOptions::isRfcAtStart() const {
  return rfcAtStart;
}

bool RedHomOptions::isExperimentalRfcAtStart() const {
  return experimentalRfcAtStart;
}

bool RedHomOptions::isRfcAtEnd() const {
  return useDefaultAlgorithms() || rfcAtEnd;
}

bool RedHomOptions::isExperimentalRfcAtEnd() const {
  return experimentalRfcAtEnd;
}

bool RedHomOptions::isCoreduce() const {
  return useDefaultAlgorithms() || coreduce;
}

bool RedHomOptions::isShave() const {
  return useDefaultAlgorithms() || shave;
}

bool RedHomOptions::isMorse() const {
  return useDefaultAlgorithms() || morse;
}

bool RedHomOptions::isKMS() const {
  return useDefaultAlgorithms() || kms;
}

bool RedHomOptions::isMatchingReduce() const
{
  return true;
}

bool RedHomOptions::isBMatrices() const {
  return bmatrices;
}

bool RedHomOptions::isSaveFiltration() const {
  return saveFiltration;
}

bool RedHomOptions::isStatisticsFile() const
{
  return statisticsFile.size() > 0;
}

const std::string& RedHomOptions::getStatisticsFile() const
{
  return statisticsFile;
}

std::string RedHomOptions::getInputFileKey(int i) const {
  return "input-file" + boost::lexical_cast<std::string>(i);
}

int RedHomOptions::getModulo() const
{
  return modulo;
}

bool RedHomOptions::isBigInt() const
{
  return bigInt;
}
