/////////////////////////////////////////////////////////////////////////////
/// @file Algorithms.cpp
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2013-12-03
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2013 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.


#include <capd/apiRedHom/Algorithms.h>
#include <capd/apiRedHom/Complex.h>
#include <capd/complex/ComputeHomologyOrchestrator.h>

#include <AlgorithmExecutor.h>
#include <ComplexDescriptors.h>
#include <capd/persistence/ImagePersistentHomology.h>
#include <capd/persistence/DistanceMatrixPersistentHomology.h>
#include <capd/complex/EulerCharacteristic.h>
#include <capd/complex/ZigZag.h>
#include <capd/complex/BuildGreedyDiscreteVectorField.h>


using namespace capd;
using namespace capd::apiRedHom;

using namespace capd::apiRedHom;

RedHomOrchestratorOverZ::RedHomOrchestratorOverZ(Complex& complex):
  _complex(complex),
  _reductions(true),
  _coreductions(true),
  _morse(true)
{}


namespace
{

  template<typename Scalar, typename Int>
  struct RunRedHomOrchestratorOverZ
  {
    RunRedHomOrchestratorOverZ(bool reductions, bool coreductions, bool morse):
      _reductions(reductions),
      _coreductions(coreductions),
      _morse(morse)
    {}

    template<typename TypeDescriptor>
    std::vector<Scalar> operator()(TypeDescriptor, AnyWrapper& anyComplex)
    {
      typedef typename TypeDescriptor::type Complex;
      Complex& complex = *anyComplex.get<boost::shared_ptr<Complex> >();
      capd::complex::ComputeHomologyOrchestrator<Complex, Scalar, Int> compute(complex, _reductions, _coreductions, _morse, true);
      return compute();
    }

  private:
    bool _reductions;
    bool _coreductions;
    bool _morse;
  };

}

std::vector<RedHomOrchestratorOverZ::Scalar> RedHomOrchestratorOverZ::operator()()
{
  std::vector<RedHomOrchestratorOverZ::Scalar> result;
  RunAlgorithm<capd::complex::Complexes>(_complex.anyComplex(), RunRedHomOrchestratorOverZ<Scalar, Scalar>(_reductions, _coreductions, _morse), boost::lambda::var(result) = boost::lambda::_1);
  return result;
}


const ImagePersistentHomology::Diagram& ImagePersistentHomology::operator()()
{
  const size_t rows = _image.size();
  BOOST_ASSERT(rows > 0);
  const size_t cols = _image[0].size();
  BOOST_ASSERT(cols > 0);

  persistence::ImagePersistentHomology::Image image(rows, cols);

  for (size_t row = 0; row < rows; ++row) {
    BOOST_ASSERT(_image[row].size() == cols);

    for (size_t col = 0; col < cols; ++col) {
      image[row][col] = _image[row][col];
    }
  }

  persistence::ImagePersistentHomology impl(image);
  impl();

  _signature.reset(new persistence::PersistentSignature(impl.signature()));

  return _signature->diagram();
}

const ImagePersistentHomology::Diagram& ImagePersistentHomology::diagram(size_t dim) const
{
  return _signature->diagram(dim);
}

const ImagePersistentHomology::Diagram& ImagePersistentHomology::diagram() const
{
  return _signature->diagram();
}

void DistanceMatrixPersistentHomology::operator()()
{
  const size_t rows = _distance.size();
  BOOST_ASSERT(rows > 0);
  const size_t cols = _distance[0].size();
  BOOST_ASSERT(cols > 0);

  persistence::DistanceMatrixPersistentHomology::DistanceMatrix distance(rows, cols);

  for (size_t row = 0; row < rows; ++row) {
    BOOST_ASSERT(_distance[row].size() == cols);

    for (size_t col = 0; col < cols; ++col) {
      distance[row][col] = _distance[row][col];
    }
  }

  delete _impl;
  _impl = new persistence::DistanceMatrixPersistentHomology(distance, _dim, _degreeLimit);
  (*_impl)();
}

DistanceMatrixPersistentHomology::~DistanceMatrixPersistentHomology()
{
  delete _impl;
  _impl = 0;
}

DistanceMatrixPersistentHomology::Pairs DistanceMatrixPersistentHomology::pairs(int dim) const
{
  return _impl->pairs(dim);
}

DistanceMatrixPersistentHomology::Cycles DistanceMatrixPersistentHomology::cycles(int dim) const
{
  return _impl->cycles(dim);
}

DistanceMatrixPersistentHomology::Diagram DistanceMatrixPersistentHomology::diagram(int dim) const
{
  return _impl->diagram(dim);
}


namespace
{
  struct RunEulerCharacteristic
  {

    template<typename TypeDescriptor>
    long operator()(TypeDescriptor, AnyWrapper& anyComplex)
    {
      typedef typename TypeDescriptor::type Complex;
      Complex& complex = *anyComplex.get<boost::shared_ptr<Complex> >();
      capd::complex::EulerCharacteristic<Complex> compute(complex);
      return compute();
    }
  };

}


long EulerCharacteristic::operator()()
{
  long result;
  RunAlgorithm<capd::complex::Complexes>(_complex.anyComplex(), RunEulerCharacteristic(), boost::lambda::var(result) = boost::lambda::_1);
  return result;
}


namespace
{
  struct RunZigZag
  {
    size_t _dim;

    template<typename TypeDescriptor>
    void operator()(TypeDescriptor, AnyWrapper& anyComplex)
    {
      typedef typename TypeDescriptor::type Complex;
      Complex& complex = *anyComplex.get<boost::shared_ptr<Complex> >();
      DoZigZag(complex, _dim);
    }
  };

}


void ZigZag::operator()()
{
  RunZigZag zigzag = {_dim};
  RunAlgorithm<capd::complex::Complexes>(_complex.anyComplex(), zigzag);
}


namespace
{
  struct RunCountCriticalCells
  {
    int method;

    template<typename TypeDescriptor>
    size_t operator()(TypeDescriptor, AnyWrapper& anyComplex)
    {
      typedef typename TypeDescriptor::type Complex;
      Complex& complex = *anyComplex.get<boost::shared_ptr<Complex> >();
      if (method == 0) {
        return capd::complex::CountCriticalCellsViaCoreduction<int>(complex);
      } else {
        return capd::complex::CountCriticalCellsViaReduction<int>(complex);
      }
    }
  };

}


size_t CountCriticalCells::operator()()
{
  RunCountCriticalCells algo = {_method};
  size_t result = 0;
  RunAlgorithm<capd::complex::Complexes>(_complex.anyComplex(), algo, boost::lambda::var(result) = boost::lambda::_1);

  return result;
}
