/////////////////////////////////////////////////////////////////////////////
/// @file Algorithms_RelativeBettiNumbers
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-07-23
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////


#include <capd/apiRedHom/Algorithms.h>
#include <capd/apiRedHom/Complex.h>
#include <capd/complex/RelativeHomology.h>

#include <AlgorithmExecutor.h>
#include <ComplexDescriptors.h>
#include <capd/rings/Zp.h>
#include <capd/rings/Z2.h>


using namespace capd;
using namespace capd::apiRedHom;
namespace b = boost;
namespace bl = boost::lambda;


namespace
{
  template<typename Scalar, typename Int>
  struct RunRelativeBettiNumbers
  {
    Complex& _subspace;
    bool _differenceClosed;
    bool _reductions;
    bool _coreductions;
    bool _morse;

    RunRelativeBettiNumbers(Complex& subspace, bool differenceClosed,
                            bool reductions, bool coreductions, bool morse):
      _subspace(subspace), _differenceClosed(differenceClosed),
      _reductions(reductions),
      _coreductions(coreductions),
      _morse(morse)
    {}


    template<typename TypeDescriptor>
    std::vector<Int> operator()(TypeDescriptor, AnyWrapper& anyComplex)
    {
      typedef typename TypeDescriptor::type Complex;
      Complex& complex = *anyComplex.get<boost::shared_ptr<Complex> >();
      Complex& subcomplex = _subspace.rawComplex<Complex>();

      capd::complex::RelativeHomology<Complex, Scalar, Int> alg(complex, subcomplex, _differenceClosed);
      alg.reductions(_reductions);
      alg.coreductions(_coreductions);
      alg.morse(_morse);

      return alg();
    }
  };
}

std::vector<int> apiRedHom::RelativeBettiNumbersOverZ::operator()()
{
  std::vector<int> result;
  RunAlgorithm<capd::complex::Complexes>(_complex.anyComplex(),
                                         RunRelativeBettiNumbers<int, int>(_subcomplex, _differenceClosed,
                                                                           _reductions, _coreductions, _morse),
                                         bl::var(result) = bl::_1);
  return result;
}

std::vector<int> apiRedHom::RelativeBettiNumbersOverZp::operator()()
{
  const int oldP = Zp::getMod();
  Zp::setMod(p);

  std::vector<int> result;
  RunAlgorithm<capd::complex::Complexes>(_complex.anyComplex(),
                                         RunRelativeBettiNumbers<Zp, int>(_subcomplex, _differenceClosed,
                                                                           _reductions, _coreductions, _morse),
                                         bl::var(result) = bl::_1);

  Zp::setMod(oldP);
  return result;
}

std::vector<int> apiRedHom::RelativeBettiNumbersOverZ2::operator()()
{
  std::vector<int> result;
  RunAlgorithm<capd::complex::Complexes>(_complex.anyComplex(),
                                         RunRelativeBettiNumbers<Z2, int>(_subcomplex, _differenceClosed,
                                                                           _reductions, _coreductions, _morse),
                                         bl::var(result) = bl::_1);
  return result;
}


std::vector<int> capd::apiRedHom::ComputeRelativeBettiNumbersOverZ(Complex& complex, Complex& subComplex, bool differenceClosed)
{
  RelativeBettiNumbersOverZ algo(complex, subComplex, differenceClosed);
  return algo();
}

std::vector<int> capd::apiRedHom::ComputeRelativeBettiNumbersOverZp(Complex& complex, Complex& subComplex, bool differenceClosed, int p)
{
  RelativeBettiNumbersOverZp algo(complex, subComplex, differenceClosed, p);
  return algo();
}

std::vector<int> capd::apiRedHom::ComputeRelativeBettiNumbersOverZ2(Complex& complex, Complex& subComplex, bool differenceClosed)
{
  RelativeBettiNumbersOverZ2 algo(complex, subComplex, differenceClosed);
  return algo();
}
