/////////////////////////////////////////////////////////////////////////////
/// @file ComplexHelper
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-01-18
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#include <capd/apiRedHom/ComplexHelper.h>
#include <capd/apiRedHom/Algorithms.h>
#include <capd/complex/SimplicesReader.h>
#include <capd/apiRedHom/Complex.h>
#include <capd/complex/SComplexKappaBuilderFromCubes.h>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/make_shared.hpp>
#include <capd/auxil/Logger.h>

namespace bl=boost::lambda;

namespace capd
{
  namespace apiRedHom
  {

    namespace
    {
      template<typename T, typename U>
      std::vector<T> ConvertVector(const std::vector<U>& v)
      {
        return std::vector<T>(v.begin(), v.end());
      }
    }

    std::vector<size_t> CubicalHelper::CubicalComplexSize(const CubicalHelper::FullCubeRange& fullCubeRange)
    {
      std::vector<size_t> size;
      for (auto cube : fullCubeRange) {
        if (cube.size() > size.size()) {
          size.resize(cube.size());
        }
        for (size_t idx = 0; idx < cube.size(); ++idx) {
          if (size[idx] < cube[idx] + 1) {
            size[idx] = cube[idx] + 1;
          }
        }
      }
      return size;
    }

    boost::shared_ptr<GeneralizedSComplex> CubicalHelper::CreateGeneralizedSComplexFromFullCubes(int dim, const CubicalHelper::FullCubeRange& fullCubeRange)
    {
      typedef complex::SComplexKappaBuilderFromCubes<GeneralizedSComplex::Id, GeneralizedSComplex::Dim> Builder;
      Builder builder(dim);

      std::for_each(fullCubeRange.begin(), fullCubeRange.end(), bl::bind(&Builder::addCell, &builder,
                                                                         bl::bind(&ConvertVector<int, size_t>, bl::_1)));
      builder();

      return boost::make_shared<GeneralizedSComplex>(builder.dims(),
                                                     GeneralizedSComplex::KappaMapEntries(builder.kappaMap()));
    }

    boost::shared_ptr<CubicalComplex> CubicalHelper::CreateFromFullCubes(const FullCubeRange& fullCubeRange, int type)
    {
      boost::shared_ptr<CubicalComplex> complex = boost::make_shared<CubicalComplex>(CubicalComplexSize(fullCubeRange), CubicalComplex::CellCodeRange(), type);
      std::for_each(fullCubeRange.begin(), fullCubeRange.end(),
                    bl::bind(static_cast<void (CubicalComplex::*)(const CubicalComplex::FullCube&)>(&CubicalComplex::insert),
                             complex.get(), bl::_1));
      complex->fillWithBoundaries();
      return complex;
    }

    boost::shared_ptr<Complex> ComplexHelper::create(std::istream& stream, InputType inputType)
    {
      BOOST_ASSERT(inputType == INPUT_TYPE_OBJ || inputType == INPUT_TYPE_DAT);

      complex::SimplicesReader reader(stream);
      boost::shared_ptr<SimplicialComplex> complex = boost::make_shared<SimplicialComplex>();
      reader(bl::bind<void>(&SimplicialComplex::insert, bl::var(*complex), bl::_1));

      return complex;
    }


    std::vector<int> CubicalHelper::BettiNumbers(const FullCubeRange& fullCubeRange, int modulo)
    {
      boost::shared_ptr<CubicalComplex> complex = CubicalHelper::CreateFromFullCubes(fullCubeRange);
      return ComputeBettiNumbers(*complex, modulo);
    }

  }
}
