/////////////////////////////////////////////////////////////////////////////
/// @file c_interface.cpp
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2013-12-07
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2013 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.

#include "capd/apiRedHom/c_interface.h"
#include "capd/apiRedHom/c_interface.hpp"
#include "capd/apiRedHom/Complex.h"
#include "capd/apiRedHom/Algorithms.h"

#include <boost/iterator/zip_iterator.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/make_shared.hpp>
#include <capd/auxil/Logger.h>


using namespace capd::apiRedHom;

void RedHom_DeleteComplex(RedHom_ComplexInfo* complexInfo)
{
  delete static_cast<boost::shared_ptr<Complex>*>(complexInfo->complexPtr);
  complexInfo->complexPtr = NULL;
}

void RedHom_CreateSimplicialComplex(RedHom_ComplexInfo* complexInfo)
{
  SetComplex(complexInfo, boost::make_shared<SimplicialComplex>());
}

void RedHom_CreateCubicalComplex(RedHom_ComplexInfo* complexInfo, int dim, const int* size)
{
  //const int type = 4; //Old
  const int type = 0; // Dense

  SetComplex(complexInfo, boost::make_shared<CubicalComplex>(std::vector<size_t>(size, size + dim), CubicalComplex::CellCodeRange(), type));
}

void RedHom_CreateGeneralizedSComplex(RedHom_ComplexInfo* complexInfo, unsigned int size, int* dims, unsigned int kappaEntries, unsigned int* cofaces, unsigned int* faces, int* coincidenceIndices)
{
  std::vector<GeneralizedSComplex::KappaMapEntry> entries(boost::make_zip_iterator(boost::make_tuple(cofaces, faces, coincidenceIndices)),
					       boost::make_zip_iterator(boost::make_tuple(cofaces + kappaEntries, faces + kappaEntries, coincidenceIndices + kappaEntries)));

  SetComplex(complexInfo, boost::make_shared<GeneralizedSComplex>(GeneralizedSComplex::Dims(dims, dims + size), entries));
}

int RedHom_ComplexCardinality(RedHom_ComplexInfo* complexInfo)
{
  return GetComplex<Complex>(complexInfo).cardinality();
}

int RedHom_ComplexDim(RedHom_ComplexInfo* complexInfo)
{
  return GetComplex<Complex>(complexInfo).dim();
}

void RedHom_AddSimplexToComplex(RedHom_ComplexInfo* complexInfo, const int* vertices, int size)
{
  GetComplex<SimplicialComplex>(complexInfo).insert(SimplicialComplex::CellCode(vertices, vertices + size));
}

void RedHom_CubicalComplexFillWithBoundaries(RedHom_ComplexInfo* complexInfo)
{
  GetComplex<CubicalComplex>(complexInfo).fillWithBoundaries();
}

void RedHom_AddCubeToComplex(RedHom_ComplexInfo* complexInfo, const int* coords, const int* unitInterval)
{
  const int dim = RedHom_ComplexDim(complexInfo);
  CubicalComplex::CellCode cube(dim);

  for (int i = 0; i < dim; ++i) {
    cube[i] = std::make_pair(coords[i], unitInterval[i]);
  }

  GetComplex<CubicalComplex>(complexInfo).insert(cube);
}

int RedHom_BettiNumbersOverZOfComplex(RedHom_ComplexInfo* complexInfo, int* bettiVector)
{
  std::fill(bettiVector, bettiVector + RedHom_ComplexDim(complexInfo) + 1, 0);

  BettiNumbersOverZ bettiAlg(GetComplex<Complex>(complexInfo));
  std::vector<int> betti = bettiAlg();
  std::copy(betti.begin(), betti.end(), bettiVector);
  return betti.size();
}


int RedHom_Coreduction(RedHom_ComplexInfo* complexInfo)
{
  Coreduction algo = {GetComplex<Complex>(complexInfo)};
  int reduced = algo();

  return reduced;
}

int RedHom_RedHomOrchestratorOverZ(RedHom_ComplexInfo* complexInfo, int* bettiVector, int reductions, int coreductions, int morse)
{
  std::fill(bettiVector, bettiVector + RedHom_ComplexDim(complexInfo) + 1, 0);

  RedHomOrchestratorOverZ alg(GetComplex<Complex>(complexInfo));
  alg.reductions(reductions);
  alg.coreductions(coreductions);
  alg.morse(morse);

  std::vector<int> betti = alg();
  std::copy(betti.begin(), betti.end(), bettiVector);
  return betti.size();
}

int RedHom_RelativeBettiNumbersOverZOfComplex(RedHom_ComplexInfo* complexInfo, RedHom_ComplexInfo* subcomplexInfo, int* bettiVector, int keepClosed, int reductions, int coreductions, int morse)
{
  std::fill(bettiVector, bettiVector + RedHom_ComplexDim(complexInfo) + 1, 0);

  RelativeBettiNumbersOverZ alg(GetComplex<Complex>(complexInfo), GetComplex<Complex>(subcomplexInfo), keepClosed);
  alg.reductions(reductions);
  alg.coreductions(coreductions);
  alg.morse(morse);

  std::vector<int> betti = alg();
  std::copy(betti.begin(), betti.end(), bettiVector);
  return betti.size();
}

// int RedHom_FundGroup(RedHom_ComplexInfo* complexInfo, int** buffer)
// {
//   std::vector<int> result;
//   FundGroup algo(GetComplex<Complex>(complexInfo));
//   result = algo();

//   *buffer = static_cast<int*>(malloc(sizeof(int) * result.size()));
//   std::copy(result.begin(), result.end(), *buffer);

//   return static_cast<int>(result.size());
// }

int RedHom_FundamentalGroup(RedHom_ComplexInfo* complexInfo, int** buffer, int vectorFieldType)
{
  std::vector<int> result;
  FundamentalGroup algo(GetComplex<Complex>(complexInfo), vectorFieldType);
  result = algo();

  *buffer = static_cast<int*>(malloc(sizeof(int) * result.size()));
  std::copy(result.begin(), result.end(), *buffer);

  return static_cast<int>(result.size());
}

void RedHom_ZigZag(RedHom_ComplexInfo* complexInfo, int dim)
{
  ZigZag algo(GetComplex<Complex>(complexInfo), dim);
  algo();
}

unsigned long RedHom_CountCriticalCells(RedHom_ComplexInfo* complexInfo, int method)
{
  CountCriticalCells algo(GetComplex<Complex>(complexInfo), method);
  return algo();
}

void RedHom_Init()
{
  INIT_CAPD_LOGGER;
}
