/////////////////////////////////////////////////////////////////////////////
/// @file PHATInterface.cpp
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-06-24
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2014 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.

#include <capd/persistence/PHATInterface.h>
#include <capd/config-capdRedHom.h>

#include <algorithm>
#include <stdexcept>

using capd::persistence::PHATInterface;

#ifdef HAVE_PHAT

#include <phat/boundary_matrix.h>
#include <phat/algorithms/twist_reduction.h>
#include <phat/persistence_pairs.h>
#include <phat/compute_persistence_pairs.h>

const bool PHATInterface::_enabled = true;

namespace
{
  template<typename Representation >
  void ComputeCycles(const phat::boundary_matrix< Representation >& boundary_matrix, PHATInterface::Cycles& cycles) {
    const size_t cols = boundary_matrix.get_num_cols();
    std::vector<bool> lowInversed(cols, false);

    for(size_t idx = 0; idx < cols; idx++) {
      if( !boundary_matrix.is_empty( idx ) ) {
        const size_t low = boundary_matrix.get_max_index(idx);
        lowInversed[low] = true;
      }
    }

    cycles.clear();

    for(size_t idx = 0; idx < cols; idx++) {
      if(boundary_matrix.is_empty(idx)) {
        if (! lowInversed[idx]) {
          cycles.push_back(idx);
        }
      }
    }
  }

}


void PHATInterface::operator()(const Filtration& filtration, size_t size, Pairs& pairs, Cycles& cycles)
{
  phat::boundary_matrix<> boundary;
  boundary.set_num_cols(size);


  int64_t idx = 0;
  for (Filtration::const_iterator it = filtration.begin(), end = filtration.end();
       it != end; ++it, ++idx) {
    Column tmp(it->second);

    std::sort(tmp.begin(), tmp.end());
    boundary.set_dim(idx, it->first);
    boundary.set_col(idx, tmp);
  }

  phat::persistence_pairs phatPairs;
  phat::compute_persistence_pairs<phat::twist_reduction>(phatPairs, boundary);

  pairs.resize(phatPairs.get_num_pairs());

  for (size_t idx = 0, end = phatPairs.get_num_pairs(); idx != end; ++idx) {
    pairs[idx] = phatPairs.get_pair(idx);
  }

  ComputeCycles(boundary, cycles);
}



#else

const bool PHATInterface::_enabled = false;

void PHATInterface::operator()(const Filtration& filtration, size_t size, Pairs& pairs, Cycles& cycles)
{
  throw std::logic_error("Not implemented, please compile with PHAT support.");
}

#endif
