/////////////////////////////////////////////////////////////////////////////
/// @file AlgorithmsTest.cpp
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2013-12-03
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2013 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.

#include "capd/apiRedHom/Algorithms.h"
#include "capd/apiRedHom/Complex.h"

#include "Fixtures.h"

#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/assign.hpp>
#include <boost/array.hpp>


using namespace boost::assign;
using namespace boost::lambda;
using namespace capd::apiRedHom;
using namespace capd::test;

BOOST_AUTO_TEST_SUITE(AlgorithmsSuite)


BOOST_AUTO_TEST_CASE_TEMPLATE(betti, Fixture, ComplexFixtures)
{
  {
    Fixture fixture;
    BettiNumbersOverZ bettiAlg(*fixture.complex);
    std::vector<int> betti = bettiAlg();
    BOOST_CHECK_EQUAL_COLLECTIONS(fixture.betti.begin(), fixture.betti.end(), betti.begin(), betti.end());
  }

  {
    Fixture fixture;
    std::vector<int> betti = ComputeBettiNumbersOverZ(*fixture.complex);
    BOOST_CHECK_EQUAL_COLLECTIONS(fixture.betti.begin(), fixture.betti.end(), betti.begin(), betti.end());
  }

  {
    Fixture fixture;
    BettiNumbersOverZp bettiAlg(*fixture.complex, 1009);
    std::vector<int> betti = bettiAlg();
    BOOST_CHECK_EQUAL_COLLECTIONS(fixture.betti.begin(), fixture.betti.end(), betti.begin(), betti.end());
  }

  {
    Fixture fixture;
    std::vector<int> betti = ComputeBettiNumbersOverZp(*fixture.complex, 1009);
    BOOST_CHECK_EQUAL_COLLECTIONS(fixture.betti.begin(), fixture.betti.end(), betti.begin(), betti.end());
  }

  {
    Fixture fixture;
    BettiNumbersOverZ2 bettiAlg(*fixture.complex);
    std::vector<int> betti = bettiAlg();
    BOOST_CHECK_EQUAL_COLLECTIONS(fixture.bettiMod(2).begin(), fixture.bettiMod(2).end(), betti.begin(), betti.end());
  }

  {
    Fixture fixture;
    std::vector<int> betti = ComputeBettiNumbersOverZ2(*fixture.complex);
    BOOST_CHECK_EQUAL_COLLECTIONS(fixture.bettiMod(2).begin(), fixture.bettiMod(2).end(), betti.begin(), betti.end());
  }

}

BOOST_AUTO_TEST_CASE_TEMPLATE(bettiViaSmith, Fixture, ComplexFixtures)
{
  Fixture fixture;

  BettiNumbersViaSmithFormOverZ bettiAlg(*fixture.complex);
  std::vector<int> betti = bettiAlg();
  BOOST_CHECK_EQUAL_COLLECTIONS(fixture.betti.begin(), fixture.betti.end(), betti.begin(), betti.end());
}


BOOST_AUTO_TEST_CASE_TEMPLATE(coreduction, Fixture, ComplexFixtures)
{
  Fixture fixture;

  Coreduction algo = {*fixture.complex};
  size_t reduced = algo();
  BOOST_CHECK_EQUAL(fixture.coreductions, reduced);
  // TODO
  //  BOOST_CHECK_EQUAL(fixture.cardinality - reduced, fixture.complex->cardinality());
}


BOOST_AUTO_TEST_CASE_TEMPLATE(orchestrator_all, Fixture, ComplexFixtures)
{
  Fixture fixture;

  RedHomOrchestratorOverZ alg(*fixture.complex);

  std::vector<int> betti = alg();
  BOOST_CHECK_EQUAL_COLLECTIONS(fixture.betti.begin(), fixture.betti.end(), betti.begin(), betti.end());
}


// BOOST_AUTO_TEST_CASE_TEMPLATE(fundGroup, Fixture, ComplexFixtures)
// {
//   Fixture fixture;

//   FundGroup alg(*fixture.complex);

//   // TODO how to test it? Now only run it
//   alg();

//   //  BOOST_CHECK_EQUAL_COLLECTIONS(fixture.betti.begin(), fixture.betti.end(), betti.begin(), betti.end());
// }



BOOST_AUTO_TEST_CASE(relative_homology)
{

  {
    std::vector<int> size(3, 10);

    CubicalSphere<> sphereFixture;
    CubicalAntipodalPoints<> pointsFixture;

    RelativeBettiNumbersOverZ relative(*sphereFixture.complex, *pointsFixture.complex, false);
    std::vector<int> expectedBetti = list_of(0)(1)(1).convert_to_container<std::vector<int> >();
    std::vector<int> betti = relative();

    BOOST_CHECK_EQUAL_COLLECTIONS(expectedBetti.begin(), expectedBetti.end(), betti.begin(), betti.end());
  }

  {
    std::vector<int> size(3, 10);

    CubicalSphere<> sphereFixture;
    CubicalAntipodalPoints<> pointsFixture;

    std::vector<int> betti = ComputeRelativeBettiNumbersOverZ(*sphereFixture.complex, *pointsFixture.complex, false);
    std::vector<int> expectedBetti = list_of(0)(1)(1).convert_to_container<std::vector<int> >();

    BOOST_CHECK_EQUAL_COLLECTIONS(expectedBetti.begin(), expectedBetti.end(), betti.begin(), betti.end());
  }

  {
    std::vector<int> size(3, 10);

    CubicalSphere<> sphereFixture;
    CubicalAntipodalPoints<> pointsFixture;

    RelativeBettiNumbersOverZp relative(*sphereFixture.complex, *pointsFixture.complex, false, 1009);
    std::vector<int> expectedBetti = list_of(0)(1)(1).convert_to_container<std::vector<int> >();
    std::vector<int> betti = relative();

    BOOST_CHECK_EQUAL_COLLECTIONS(expectedBetti.begin(), expectedBetti.end(), betti.begin(), betti.end());
  }

  {
    std::vector<int> size(3, 10);

    CubicalSphere<> sphereFixture;
    CubicalAntipodalPoints<> pointsFixture;

    std::vector<int> betti = ComputeRelativeBettiNumbersOverZp(*sphereFixture.complex, *pointsFixture.complex, false, 1009);
    std::vector<int> expectedBetti = list_of(0)(1)(1).convert_to_container<std::vector<int> >();

    BOOST_CHECK_EQUAL_COLLECTIONS(expectedBetti.begin(), expectedBetti.end(), betti.begin(), betti.end());
  }

  {
    std::vector<int> size(3, 10);

    CubicalSphere<> sphereFixture;
    CubicalAntipodalPoints<> pointsFixture;

    RelativeBettiNumbersOverZ2 relative(*sphereFixture.complex, *pointsFixture.complex, false);
    std::vector<int> expectedBetti = list_of(0)(1)(1).convert_to_container<std::vector<int> >();
    std::vector<int> betti = relative();

    BOOST_CHECK_EQUAL_COLLECTIONS(expectedBetti.begin(), expectedBetti.end(), betti.begin(), betti.end());
  }

  {
    std::vector<int> size(3, 10);

    CubicalSphere<> sphereFixture;
    CubicalAntipodalPoints<> pointsFixture;

    std::vector<int> betti = ComputeRelativeBettiNumbersOverZ2(*sphereFixture.complex, *pointsFixture.complex, false);
    std::vector<int> expectedBetti = list_of(0)(1)(1).convert_to_container<std::vector<int> >();

    BOOST_CHECK_EQUAL_COLLECTIONS(expectedBetti.begin(), expectedBetti.end(), betti.begin(), betti.end());
  }

}

BOOST_AUTO_TEST_SUITE_END()
