/////////////////////////////////////////////////////////////////////////////
/// @file CubicalComplexTest.cpp
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2013-12-03
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2013 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.


#include "capd/apiRedHom/Complex.h"
#include "capd/apiRedHom/Algorithms.h"
#include "CommonTools.h"
#include "Fixtures.h"

#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/assign.hpp>
#include <boost/foreach.hpp>
#include <boost/range/iterator_range.hpp>

using namespace boost::assign;
using namespace boost::lambda;

using namespace capd::apiRedHom;
using namespace capd::test;


namespace std
{
  template<typename F, typename S>
  ostream& operator<<(ostream& s, const pair<F, S>& p) {
    s << '<' << p.first << ',' << p.second << '>';
    return s;
  }
}

BOOST_AUTO_TEST_SUITE(CubicalComplexSuite)


BOOST_AUTO_TEST_CASE(insert)
{
  std::vector<size_t> size(2);
  CubicalComplex::CellCode cube = list_of<CubicalComplex::Interval>(0, true)(0, false);
  std::vector<CubicalComplex::CellCode> cubes = list_of(cube);

  size[0] = size[1] = 10;

  CubicalComplex complex(size, cubes);
  complex.insert(cube);

  BOOST_CHECK_EQUAL(2, complex.dim());
  BOOST_CHECK_EQUAL(1, complex.cardinality());

  CubicalComplex::CellCode cube2 = list_of<CubicalComplex::Interval>(1, true)(1, false);
  complex.insert(cube2);

  BOOST_CHECK_EQUAL(2, complex.cardinality());
}


BOOST_AUTO_TEST_CASE(euler_1)
{
  std::vector<size_t> size(2);
  size[0] = size[1] = 10;

  CubicalComplex complex(size);

  CubicalComplex::CellCode cube = list_of<CubicalComplex::Interval>(1, true)(1, false);
  complex.insert(cube);
  EulerCharacteristic eCh(complex);

  BOOST_CHECK_EQUAL(-1, eCh());
}


BOOST_AUTO_TEST_CASE(euler_2)
{
  std::vector<size_t> size(2);
  size[0] = size[1] = 10;

  std::vector<CubicalComplex::CellCode> cubes = list_of
    (list_of<CubicalComplex::Interval>(1, true)(1, false).convert_to_container<CubicalComplex::CellCode>())
    (list_of<CubicalComplex::Interval>(1, false)(1, true).convert_to_container<CubicalComplex::CellCode>())
    (list_of<CubicalComplex::Interval>(1, false)(1, false).convert_to_container<CubicalComplex::CellCode>())
    ;

  CubicalComplex complex(size, cubes);
  EulerCharacteristic eCh(complex);

  BOOST_CHECK_EQUAL(-1, eCh());
}

BOOST_AUTO_TEST_CASE(euler_3)
{
  std::vector<size_t> size(2);
  size[0] = size[1] = 10;

  std::vector<CubicalComplex::CellCode> cubes = list_of
    (list_of<CubicalComplex::Interval>(1, true)(1, false).convert_to_container<CubicalComplex::CellCode>())
    (list_of<CubicalComplex::Interval>(1, false)(1, true).convert_to_container<CubicalComplex::CellCode>())
    (list_of<CubicalComplex::Interval>(1, false)(1, false).convert_to_container<CubicalComplex::CellCode>())

    (list_of<CubicalComplex::Interval>(1, false)(2, false).convert_to_container<CubicalComplex::CellCode>())
    (list_of<CubicalComplex::Interval>(2, false)(1, false).convert_to_container<CubicalComplex::CellCode>())
    ;

  CubicalComplex complex(size, cubes);
  EulerCharacteristic eCh(complex);

  BOOST_CHECK_EQUAL(1, eCh());
}


BOOST_AUTO_TEST_CASE_TEMPLATE(iterator_1, Fixture, CubicalFixtures)
{
  Fixture fixture;

  BOOST_CHECK_EQUAL(fixture.cardinality, fixture.complex->cardinality());
  BOOST_CHECK_EQUAL(fixture.dim, fixture.complex->dim());

  EulerCharacteristic eCh(*fixture.complex);

  BOOST_CHECK_EQUAL(fixture.euler, eCh());
}

BOOST_AUTO_TEST_CASE(circle)
{
  std::vector<size_t> size(2);
  size[0] = size[1] = 10;

  std::vector<CubicalComplex::CellCode> cubes = list_of
    (list_of<CubicalComplex::Interval>(1, true)(1, false).convert_to_container<CubicalComplex::CellCode>())
    (list_of<CubicalComplex::Interval>(1, false)(1, true).convert_to_container<CubicalComplex::CellCode>())
    (list_of<CubicalComplex::Interval>(2, false)(1, true).convert_to_container<CubicalComplex::CellCode>())
    (list_of<CubicalComplex::Interval>(1, true)(2, false).convert_to_container<CubicalComplex::CellCode>())

    ;

  CubicalComplex complex(size, cubes);
  complex.fillWithBoundaries();

  BOOST_CHECK_EQUAL(8, complex.cardinality());

  EulerCharacteristic eCh(complex);

  BOOST_CHECK_EQUAL(0, eCh());
}

BOOST_AUTO_TEST_CASE(rectangle)
{
  std::vector<size_t> size(2);
  size[0] = size[1] = 10;

  std::vector<CubicalComplex::CellCode> cubes = list_of
    (list_of<CubicalComplex::Interval>(1, true)(1, true).convert_to_container<CubicalComplex::CellCode>())
    ;

  CubicalComplex complex(size, cubes);
  complex.fillWithBoundaries();

  BOOST_CHECK_EQUAL(9, complex.cardinality());

  EulerCharacteristic eCh(complex);

  BOOST_CHECK_EQUAL(1, eCh());

  BOOST_CHECK(complex.find(list_of<CubicalComplex::Interval>(1, true)(1, false).convert_to_container<CubicalComplex::CellCode>()));
  BOOST_CHECK( ! complex.find(list_of<CubicalComplex::Interval>(2, true)(2, true).convert_to_container<CubicalComplex::CellCode>()));

  BOOST_CHECK_EQUAL(1, complex.find(list_of<CubicalComplex::Interval>(1, true)(1, false).convert_to_container<CubicalComplex::CellCode>()).get().dim());
  BOOST_CHECK_EQUAL(1, complex[list_of<CubicalComplex::Interval>(1, true)(1, false).convert_to_container<CubicalComplex::CellCode>()].dim());

}


BOOST_AUTO_TEST_SUITE_END()
