/////////////////////////////////////////////////////////////////////////////
/// @file BdNeighboursTest
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-03-09
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////


#include <TestTools.h>
#include <capd/complex/BdNeighbours.h>
#include <../apiRedHom/Fixtures.h>


#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/mpl/list.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/assign.hpp>
#include <boost/foreach.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/typeof/typeof.hpp>

#include <map>
#include <set>
#include <vector>
#include <algorithm>
#include <utility>
#include <sstream>
#include <boost/range/irange.hpp>
#include <capd/complex/CubicalComplexFromArray.h>
#include <boost/range/adaptors.hpp>

namespace bl = boost::lambda;
namespace b = boost;
namespace ba = boost::assign;

using namespace capd;
using namespace capd::complex;
using namespace capd::test;

BOOST_AUTO_TEST_SUITE(BdNeighboursTestSuite)

BOOST_AUTO_TEST_CASE_TEMPLATE(exists, Fixture, ComplexFixtures)
{
  Fixture fixture;
  typedef typename Fixture::ComplexType ComplexType;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();
  if (complex.size() == 0 || complex.getDim() < 1) {
    return;
  }

  for (auto cell : complex.iterators().dimCells(1)) {
    BdNeighbours<ComplexType, 1> neighbours(complex, cell);
    const size_t neighoursCount = std::distance(neighbours.begin(), neighbours.end());
    BOOST_REQUIRE(neighoursCount > 0);
    BOOST_CHECK_EQUAL(1, neighbours.begin()->getDim());
  }

  for (auto cell : complex.iterators().dimCells(1)) {
    BdNeighbours<ComplexType, 0> neighbours(complex, cell, 1);
    const size_t neighoursCount = std::distance(neighbours.begin(), neighbours.end());
    BOOST_REQUIRE(neighoursCount > 0);
    BOOST_CHECK_EQUAL(1, neighbours.begin()->getDim());
  }
}

BOOST_AUTO_TEST_CASE_TEMPLATE(allDims, Fixture, ComplexFixtures)
{
  Fixture fixture;
  typedef typename Fixture::ComplexType ComplexType;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();
  if (complex.size() == 0 || complex.getDim() <= 1) {
    return;
  }

  for (size_t dim : boost::irange((size_t)1, (size_t)complex.getDim())) {
    for (auto cell : complex.iterators().dimCells(dim)) {
      BdNeighbours<ComplexType, 0> neighbours(complex, cell, dim);
      for (auto x : neighbours) {
        BOOST_CHECK_EQUAL(dim, x.getDim());
        //        BOOST_CHECK_EQUAL(cell.getColor(), x.getColor());
      }
    }
  }
}

BOOST_AUTO_TEST_CASE(simplicialCircle)
{
  typedef SimplicialCircle Fixture;
  Fixture fixture;
  typedef typename Fixture::ComplexType ComplexType;
  ComplexType& complex = fixture.complex->rawComplex<ComplexType>();

  for (auto cell : complex.iterators().dimCells(1)) {
    {
      BdNeighbours<ComplexType, 1> neighbours(complex, cell);
      std::set<typename ComplexType::Cell> uniqNeighbours(neighbours.begin(), neighbours.end());

      for (auto x : neighbours) {
        BOOST_CHECK_EQUAL(cell.getDim(), x.getDim());
      }

      BOOST_REQUIRE_EQUAL(2, uniqNeighbours.size());
    }

    {
      BdNeighbours<ComplexType, 0> neighbours(complex, cell, 1);
      std::set<typename ComplexType::Cell> uniqNeighbours(neighbours.begin(), neighbours.end());

      for (auto x : neighbours) {
        BOOST_CHECK_EQUAL(cell.getDim(), x.getDim());
      }
      BOOST_REQUIRE_EQUAL(2, uniqNeighbours.size());
    }

  }
}


BOOST_AUTO_TEST_CASE(colorDoesNotMatter)
{
  std::vector<size_t> shape = {{3, 3}};
  boost::multi_array<size_t, 2> array(shape);

  std::fill(array.origin(), array.origin() + array.num_elements(), 1);
  array[1][1] = 2;
  auto _complex = BuildCubicalComplexFromArray(array);
  auto& complex = *_complex;

  {
    auto bdNeighbours = makeBdNeighbours<2>(complex, complex[complex(0, 0)]);
    auto uniq = Uniq(bdNeighbours.begin(), bdNeighbours.end());
    BOOST_CHECK_EQUAL(3, boost::distance(uniq));
  }

  {
    auto bdNeighbours = makeBdNeighbours<2>(complex, complex[complex(0, 1)]);
    auto uniq = Uniq(bdNeighbours.begin(), bdNeighbours.end());
    BOOST_CHECK_EQUAL(5, boost::distance(uniq));
  }

}

BOOST_AUTO_TEST_SUITE_END()
