/////////////////////////////////////////////////////////////////////////////
/// @file BuildGreedyDiscreteVectorFieldTest
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-04-02
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////

#include <TestTools.h>
#include <capd/complex/BuildGreedyDiscreteVectorField.h>
#include <capd/auxil/Logger.h>
#include <../apiRedHom/Fixtures.h>

#include <capd/complex/SComplex.h>
#include <capd/complex/DefaultReduceStrategy.hpp>
#include <capd/complex/AKQStrategy.hpp>
#include <capd/complex/Coreduction.h>

#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/mpl/list.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/assign.hpp>
#include <boost/foreach.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/typeof/typeof.hpp>

#include <map>
#include <vector>
#include <algorithm>
#include <utility>
#include <sstream>
#include <capd/complex/SComplexDefaultTraits.h>
#include <boost/range/adaptors.hpp>
#include <boost/range/algorithm.hpp>
#include <boost/range/irange.hpp>
#include <capd/complex/ComplexFunctors.h>
#include <capd/complex/ComplexTraits.h>

namespace capd
{
  namespace test
  {

  }
}

namespace bl = boost::lambda;
namespace b = boost;
namespace ba = boost::assign;

using namespace capd;
using namespace capd::test;
using namespace capd::complex;

BOOST_AUTO_TEST_SUITE(BuildGreedyDiscreteVectorFieldTestSuite)

BOOST_AUTO_TEST_CASE( init )
{

}


template<typename Fixture>
size_t LegacyCriticalCells()
{
  Fixture fixture;
  typedef typename Fixture::ComplexType ComplexType;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();

  typedef capd::complex::SComplex<capd::complex::SComplexDefaultTraits> OutputComplex;
  typedef AKQReduceStrategy<ComplexType, OutputComplex, int> AKQ;
  capd::complex::Coreduction<AKQ, int, int> algorithm(new AKQ(complex));
  algorithm();

  return algorithm.getStrategy()->getOutputComplex()->size();
}

BOOST_AUTO_TEST_CASE_TEMPLATE(compareWithLegacy, Fixture, ComplexFixtures)
{
  Fixture fixture;
  typedef typename Fixture::ComplexType ComplexType;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();

  size_t cardinality = complex.cardinality();
  size_t a = CountCriticalCellsViaCoreduction<int>(complex);
  size_t b = LegacyCriticalCells<Fixture>();

  if (cardinality > 0) {
    BOOST_CHECK(a > 0);
  }

  BOOST_CHECK_EQUAL(b, a);
}

template<typename VectorField>
void CheckVectorFieldGraphNeighbours(const VectorField& vf)
{
  const typename VectorField::Graph& graph = vf.graph();

  for (auto vertex : graph.vertices()) {
    for (auto neighbour : graph.neighbours(vertex)) {
      if (vf.isCritical(vertex)) {
        BOOST_CHECK(!b::get<2>(neighbour));
      }

      if (vf.isKing(vertex)) {
        BOOST_CHECK(vf.complex()[b::get<0>(neighbour)].getDim() < vf.complex()[vertex].getDim());
        BOOST_CHECK(!b::get<2>(neighbour));
      }

      if (vf.isQueen(vertex)) {
        if (vf.complex()[b::get<0>(neighbour)].getDim() > vf.complex()[vertex].getDim()) {
          BOOST_CHECK(b::get<2>(neighbour));
        } else {
          BOOST_CHECK(!b::get<2>(neighbour));
        }
      }
    }
  }
}

template<typename VectorField>
void CheckVectorFieldGraph(const VectorField& vf)
{
  auto isInField = boost::adaptors::filtered(bl::bind(&VectorField::isInField, &vf, bl::_1));
  auto cellId = boost::adaptors::transformed(MakeCellId<typename VectorField::Complex>());
  const typename VectorField::Graph& graph = vf.graph();

  if (IsColorable<typename VectorField::Complex>::value) {
    BOOST_CHECK_EQUAL(boost::distance(graph.orderedVertices()),
                      boost::distance(vf.complex().iterators().allCells()
                                      | cellId | isInField));
  }

  for (auto vertex : graph.vertices()) {
    BOOST_CHECK(vf.isInField(vertex));
    if (vf.isCritical(vertex)) {
      BOOST_CHECK_EQUAL(boost::distance(graph.neighbours(vertex)),
                        boost::distance(vf.complex().iterators().bdCells(vf.complex()[vertex])
                                        | cellId | isInField));
    }

    if (vf.isKing(vertex)) {
      BOOST_CHECK_EQUAL(boost::distance(graph.neighbours(vertex)) + 1,
                        boost::distance(vf.complex().iterators().bdCells(vf.complex()[vertex])
                                        | cellId | isInField));
    }

    if (vf.isQueen(vertex)) {
      BOOST_CHECK_EQUAL(boost::distance(graph.neighbours(vertex)),
                        boost::distance(vf.complex().iterators().bdCells(vf.complex()[vertex])
                                        | cellId | isInField) + 1);
    }
  }

  auto orderedVertices = graph.orderedVertices();
  if (orderedVertices.size() > 0) {
    BOOST_CHECK_EQUAL(0, boost::distance(graph.neighbours(*orderedVertices.begin()))); // vertices in topological order
  }

  BOOST_CHECK(graph.size() >= boost::distance(graph.vertices()));
  BOOST_CHECK_EQUAL(graph.vertices(), boost::sort(orderedVertices));

  CheckVectorFieldGraphNeighbours(vf);
}

template<typename VectorField>
size_t CheckVectorField(const VectorField& vf)
{
  typedef typename VectorField::Complex Complex;;
  const typename VectorField::Graph& graph = vf.graph();

  auto allIds = graph.orderedVertices();

  auto isInField = boost::adaptors::filtered(bl::bind(&VectorField::isInField, &vf, bl::_1));
  auto isQueen = boost::adaptors::filtered(bl::bind(&VectorField::isQueen, &vf, bl::_1));
  auto isKing = boost::adaptors::filtered(bl::bind(&VectorField::isKing, &vf, bl::_1));
  auto isCritical = boost::adaptors::filtered(bl::bind(&VectorField::isCritical, &vf, bl::_1));


  for (auto id : allIds | isInField) {
    auto cell = vf.complex()[id];

    if (vf.isCritical(cell.getId())) {
      continue;
    }

    auto matched = vf.matched(cell.getId());
    BOOST_CHECK_EQUAL(cell.getId(), vf.matched(matched));

    if (vf.isKing(cell.getId())) {
      BOOST_CHECK(vf.isQueen(matched));
      auto queens = vf.complex().iterators().bdCells(cell)
        | boost::adaptors::transformed(MakeCellId<Complex>())
        | isQueen;
      BOOST_CHECK_EQUAL(1, boost::count_if(queens, bl::_1 == matched));
    }

    if (vf.isQueen(cell.getId())) {
      BOOST_CHECK(vf.isKing(matched));
      auto kings = vf.complex().iterators().cbdCells(cell)
        | boost::adaptors::transformed(MakeCellId<Complex>())
        | isKing;
      BOOST_CHECK_EQUAL(1, boost::count_if(kings, bl::_1 == matched));
    }

  }

  size_t inField =  (boost::distance(allIds | isCritical)
                     +
                     boost::distance(allIds | isKing)
                     +
                     boost::distance(allIds | isQueen));

  BOOST_CHECK_EQUAL(inField, (boost::distance(allIds | isInField)));
  return inField;
}

BOOST_AUTO_TEST_CASE_TEMPLATE(checkVectorFieldCoreduction, Fixture, ComplexFixtures)
{
  Fixture fixture;
  typedef typename Fixture::ComplexType ComplexType;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();
  const size_t cardinality = complex.cardinality();
  typename BuildGreedyDiscreteVectorFieldViaCoreductions<ComplexType, int>::type algo(complex);

  auto& vf = algo();
  size_t inField = CheckVectorField(vf);
  BOOST_CHECK_EQUAL(cardinality, inField);
  CheckVectorFieldGraph(vf);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(checkVectorFieldReduction, Fixture, ComplexFixtures)
{
  Fixture fixture;
  typedef typename Fixture::ComplexType ComplexType;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();
  const size_t cardinality = complex.cardinality();
  typename BuildGreedyDiscreteVectorFieldViaReductions<ComplexType, int>::type algo(complex);

  auto& vf = algo();
  size_t inField = CheckVectorField(vf);
  BOOST_CHECK_EQUAL(cardinality, inField);
  CheckVectorFieldGraph(vf);
}


BOOST_AUTO_TEST_SUITE_END()
