/////////////////////////////////////////////////////////////////////////////
/// @file BuildMorseComplexTest
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-04-15
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////


#include <TestTools.h>
#include <capd/complex/BuildMorseComplex.h>
#include <capd/auxil/Logger.h>
#include <../apiRedHom/Fixtures.h>
#include <capd/complex/BuildGreedyDiscreteVectorField.h>
#include <capd/complex/ComputeHomologySignature.hpp>

#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/mpl/list.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/assign.hpp>
#include <boost/foreach.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/typeof/typeof.hpp>

#include <map>
#include <vector>
#include <algorithm>
#include <utility>
#include <sstream>


namespace capd
{
  namespace test
  {

  }
}

namespace bl = boost::lambda;
namespace b = boost;
namespace ba = boost::assign;

using namespace capd;
using namespace capd::test;
using namespace capd::complex;

BOOST_AUTO_TEST_SUITE(BuildMorseComplexTestSuite)

template<typename Complex, typename Fixture>
void CheckStrongMorseInequalities(const Complex& complex, const Fixture& fixture)
{
  std::vector<size_t> countCWDim(fixture.dim + 1);

  for (auto cell : complex.iterators(1).allCells()) {
    countCWDim.at(cell.getDim()) += 1;
  }

  std::vector<size_t> betti(fixture.betti);
  betti.resize(fixture.dim + 1);

  CAPD_DEBUG(countCWDim << " " << betti);

  for (size_t p = 0; p < fixture.dim + 1; ++p) {
    int coeef = (p % 2 == 0 ? 1 : -1);
    int left = 0;
    int right = 0;
    for (size_t i = 0; i <= p; ++i) {
      left += coeef * countCWDim[i];
      right += coeef * betti[i];
      coeef *= -1;
    }

    CAPD_DEBUG("p = " << p << " " << left << " " << right);
    BOOST_CHECK(left >= right);
  }
}

template<typename Complex>
std::vector<size_t> BettiNumbers(Complex& complex)
{
  ComputeHomologySignature<Complex, int, size_t> algo(complex, false);
  return algo().ranks();
}

template<typename VectorField, typename Fixture>
void CheckMorseComplex(const VectorField& field, const Fixture& fixture)
{
  auto complex = ConvertToMorseComplex<size_t>(field);
  CAPD_DEBUG(complex.size() << " " << complex.cardinality());

  CheckStrongMorseInequalities(complex, fixture);

  auto betti = BettiNumbers(complex);
  BOOST_CHECK_EQUAL(fixture.betti, betti);
}


BOOST_AUTO_TEST_CASE_TEMPLATE(checkCoreduction, Fixture, ComplexFixtures)
{
  Fixture fixture;
  typedef typename Fixture::ComplexType ComplexType;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();
  typename BuildGreedyDiscreteVectorFieldViaCoreductions<ComplexType, int>::type algo(complex);

  CheckMorseComplex(algo(), fixture);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(checkReduction, Fixture, ComplexFixtures)
{
  Fixture fixture;
  typedef typename Fixture::ComplexType ComplexType;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();
  typename BuildGreedyDiscreteVectorFieldViaReductions<ComplexType, int>::type algo(complex);

  CheckMorseComplex(algo(), fixture);
}

BOOST_AUTO_TEST_SUITE_END()
