/////////////////////////////////////////////////////////////////////////////
/// @file CellSComplexTest
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-06-27
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.

#include <capd/complex/CellSComplex.h>
#include <capd/auxil/Logger.h>

#include <TestTools.h>

#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/mpl/list.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/assign.hpp>
#include <boost/foreach.hpp>

#include <map>
#include <vector>
#include <algorithm>
#include <utility>
#include <sstream>


namespace bl = boost::lambda;
namespace b = boost;
namespace ba = boost::assign;

using namespace capd;
using namespace capd::complex;

BOOST_AUTO_TEST_SUITE(CellSComplexTestSuite)

BOOST_AUTO_TEST_CASE(simplicialFromIterators)
{
  typedef SimplicialCellSComplex<SimplicialCellSComplexTraitsT<int> > Complex;

  typedef Complex::CellCode CellCode;

  std::vector<CellCode> simplices = ba::list_of
    (ba::list_of(0).operator CellCode())
    (ba::list_of(1).operator CellCode())
    (ba::list_of(2).operator CellCode())
    (ba::list_of(0)(1).operator CellCode())
    (ba::list_of(0)(2).operator CellCode())
    (ba::list_of(1)(2).operator CellCode())
    (ba::list_of(0)(1)(2).operator CellCode())
    ;

  Complex complex(simplices.begin(), simplices.end(), simplices.size());

  BOOST_CHECK_EQUAL(simplices.size(), complex.size());

  std::vector<int> dims(3);

  for (Complex::Id idx = 0; idx < complex.size(); ++idx) {
    Complex::Cell cell = complex[idx];

    BOOST_ASSERT(cell.dim() <= 2);
    BOOST_ASSERT(cell.dim() >= 0);
    dims[cell.dim()]++;
  }

  BOOST_CHECK_EQUAL(3, dims[0]);
  BOOST_CHECK_EQUAL(3, dims[1]);
  BOOST_CHECK_EQUAL(1, dims[2]);

  BOOST_CHECK_EQUAL(2, complex[ba::list_of(0)(1)(2).operator CellCode()].dim() );
}


BOOST_AUTO_TEST_CASE(simplicialFromStorage)
{
  typedef SimplicialCellSComplex<SimplicialCellSComplexTraitsT<int> > Complex;

  typedef Complex::CellCode CellCode;

  std::vector<CellCode> simplices = ba::list_of
    (ba::list_of(0).operator CellCode())
    (ba::list_of(1).operator CellCode())
    (ba::list_of(2).operator CellCode())
    (ba::list_of(0)(1).operator CellCode())
    (ba::list_of(0)(2).operator CellCode())
    (ba::list_of(1)(2).operator CellCode())
    (ba::list_of(0)(1)(2).operator CellCode())
    ;

  Complex complex(simplices);

  BOOST_CHECK_EQUAL(2, complex[ba::list_of(0)(1)(2).operator CellCode()].dim() );
}


BOOST_AUTO_TEST_CASE(notOrderedSimplices)
{
  typedef SimplicialCellSComplex<SimplicialCellSComplexTraitsT<int> > Complex;

  typedef Complex::CellCode CellCode;


  std::vector<CellCode> simplices = ba::list_of
    (ba::list_of(2)(1)(0).operator CellCode())
    (ba::list_of(1)(0).operator CellCode())
    (ba::list_of(2)(0).operator CellCode())
    (ba::list_of(2)(1).operator CellCode())
    (ba::list_of(0).operator CellCode())
    (ba::list_of(1).operator CellCode())
    (ba::list_of(2).operator CellCode())
    ;

  Complex complex(simplices);
  std::vector<Complex::Dim> dims(complex.size());

  for (Complex::Id idx = 0, size = complex.size(); idx < size; ++idx) {
    Complex::Cell cell = complex[idx];
    dims[cell.id()] = cell.dim();
    BOOST_FOREACH(Complex::Boundary::value_type bdCell, complex.boundary(cell)) {
      BOOST_ASSERT(bdCell < cell.id());
    }
  }

  BOOST_ASSERT(std::is_sorted(dims.begin(), dims.end()));
}


BOOST_AUTO_TEST_CASE(duplicatedSimplices)
{
  typedef SimplicialCellSComplex<SimplicialCellSComplexTraitsT<int> > Complex;

  typedef Complex::CellCode CellCode;


  std::vector<CellCode> simplices = ba::list_of
    (ba::list_of(2)(1)(0).operator CellCode())
    (ba::list_of(1)(0).operator CellCode())
    (ba::list_of(2)(0).operator CellCode())
    (ba::list_of(2)(1).operator CellCode())
    (ba::list_of(0).operator CellCode())
    (ba::list_of(1).operator CellCode())
    (ba::list_of(2).operator CellCode())
    (ba::list_of(0).operator CellCode())
    (ba::list_of(1).operator CellCode())
    (ba::list_of(2).operator CellCode())
    (ba::list_of(0)(1).operator CellCode())
    (ba::list_of(0)(2).operator CellCode())
    (ba::list_of(1)(2).operator CellCode())
    (ba::list_of(0)(1)(2).operator CellCode())
    ;

  Complex complex(simplices);
  std::vector<Complex::Dim> dims(complex.size());

  for (Complex::Id idx = 0, size = complex.size(); idx < size; ++idx) {
    Complex::Cell cell = complex[idx];
    dims[cell.id()] = cell.dim();
    BOOST_FOREACH(Complex::Boundary::value_type bdCell, complex.boundary(cell)) {
      BOOST_ASSERT(bdCell < cell.id());
    }
  }
}


BOOST_AUTO_TEST_CASE(checkBoundary)
{
  typedef SimplicialCellSComplex<SimplicialCellSComplexTraitsT<int> > Complex;

  typedef Complex::CellCode CellCode;

  std::vector<CellCode> simplices = ba::list_of
    (ba::list_of(0).operator CellCode())
    (ba::list_of(1).operator CellCode())
    (ba::list_of(2).operator CellCode())
    (ba::list_of(0)(1).operator CellCode())
    (ba::list_of(0)(2).operator CellCode())
    (ba::list_of(1)(2).operator CellCode())
    (ba::list_of(0)(1)(2).operator CellCode())
    ;

  Complex complex(simplices);

  for (Complex::Id idx = 0, size = complex.size(); idx < size; ++idx) {
    Complex::Cell cell = complex[idx];

    if (cell.dim() == 0) {
      BOOST_CHECK_EQUAL(0, complex.boundary(cell).size());
    } else {
      BOOST_CHECK_EQUAL(cell.dim() + 1, complex.boundary(cell).size());
    }
  }
}


BOOST_AUTO_TEST_SUITE_END()
