/////////////////////////////////////////////////////////////////////////////
/// @file ChainMapTest.cpp
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2013-12-17
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2013 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.

#include "capd/complex/ChainMap.hpp"
#include "capd/complex/HomologyModel.h"
#include "ChainMapFixtures.h"
#include <TestTools.h>

#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/mpl/list.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/assign.hpp>
#include <boost/typeof/typeof.hpp>

namespace bl = boost::lambda;
namespace b = boost;

using namespace boost::assign;
using namespace boost::lambda;

using namespace capd::test;
using namespace capd::complex;


BOOST_AUTO_TEST_SUITE(ChainMapSuite)

template<typename Fixture, typename Function>
void checkIdentity(Function function)
{
  typedef typename Fixture::DomHomologyModel::Chain DomChain;
  typedef typename Fixture::ImgHomologyModel::Chain ImgChain;
  Fixture fixture;

  BOOST_AUTO(chainMap, fixture.makeChainMap(function));

  if (fixture.domComplexOrg().cardinality() <= 0) {
    return;
  }

  {
    DomChain domChain(*fixture.domComplexOrg().iterators(1).allCells().begin());
    ImgChain imgChain = chainMap(domChain);

    BOOST_CHECK_EQUAL_COLLECTIONS(domChain.begin(), domChain.end(),
                                  imgChain.begin(), imgChain.end());
  }

  {
    BOOST_AUTO(allCells, fixture.domComplexOrg().iterators(1).dimCells(0));

    DomChain domChain(allCells.begin(), allCells.end());
    ImgChain imgChain = chainMap(domChain);

    BOOST_CHECK(domChain.size() > 0);

    BOOST_CHECK_EQUAL_COLLECTIONS(domChain.begin(), domChain.end(),
                                  imgChain.begin(), imgChain.end());
  }
}

BOOST_AUTO_TEST_CASE_TEMPLATE(identity_bind, Fixture, AuthomorphismChainMapFixtures)
{
  checkIdentity<Fixture>(bl::bind(bl::_3, bl::_1, bl::_2));
}

namespace
{
  template<typename Model>
  struct IdentityMap
  {
    template<typename Func>
    void operator()(const typename Model::Complex::CellCode& code, const typename Model::Scalar& scalar, Func func)
    {
      func(code, scalar);
    }
  };
}

BOOST_AUTO_TEST_CASE_TEMPLATE(identity_functor, Fixture, AuthomorphismChainMapFixtures)
{
  checkIdentity<Fixture>(IdentityMap<typename Fixture::DomHomologyModel>());
}


namespace
{

  template<typename Model>
  struct TimesThreeMap
  {
    template<typename Func>
    void operator()(const typename Model::Complex::CellCode& code, const typename Model::Scalar& scalar, Func func)
    {
      func(code, typename Model::Scalar(2) * scalar);
      func(code, scalar);
    }
  };
}

BOOST_AUTO_TEST_CASE_TEMPLATE(times_three_template_func, Fixture, AuthomorphismChainMapFixtures)
{
  typedef typename Fixture::DomHomologyModel::Chain DomChain;
  typedef typename Fixture::ImgHomologyModel::Chain ImgChain;
  Fixture fixture;

  BOOST_AUTO(chainMap, fixture.makeChainMap(TimesThreeMap<typename Fixture::DomHomologyModel>()));

  if (fixture.domComplexOrg().cardinality() <= 0) {
    return;
  }

  {
    DomChain domChain(*fixture.domComplexOrg().iterators(1).allCells().begin());
    ImgChain imgChain = chainMap(domChain);

    domChain *= 3;

    BOOST_CHECK_EQUAL_COLLECTIONS(domChain.begin(), domChain.end(),
                                  imgChain.begin(), imgChain.end());
  }
}

BOOST_AUTO_TEST_CASE(simplicial_circle_to_cubical_point)
{
  typedef SimplicialCircleToCubicalPointChainMap Fixture;

  typedef Fixture::DomHomologyModel::Chain DomChain;
  typedef Fixture::ImgHomologyModel::Chain ImgChain;
  Fixture fixture;

  BOOST_AUTO(constCell, fixture.imgComplexOrg().iterators(1).allCells().begin()->code());
  BOOST_AUTO(chainMap, fixture.makeChainMap(bl::bind(bl::_3, bl::var(constCell), bl::_2)));

  {
    BOOST_AUTO(allCells, fixture.domComplexOrg().iterators(1).dimCells(1));

    DomChain domChain(allCells.begin(), allCells.end());
    ImgChain imgChain = chainMap(domChain);

    BOOST_CHECK_EQUAL(3, domChain.size());
    BOOST_CHECK_EQUAL(1, imgChain.size());
    BOOST_CHECK_EQUAL(3, imgChain.begin()->second);
  }


}

BOOST_AUTO_TEST_SUITE_END()
