/////////////////////////////////////////////////////////////////////////////
/// @file CubCellComplexTest
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-09-16
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////


#include <capd/complex/CubCellComplex.h>
#include <capd/auxil/Logger.h>

#include <TestTools.h>

#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/mpl/list.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/assign.hpp>
#include <boost/foreach.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/typeof/typeof.hpp>
#include <boost/iterator/iterator_facade.hpp>
#include <boost/range/iterator_range.hpp>

#include <map>
#include <vector>
#include <algorithm>
#include <utility>
#include <sstream>


namespace bl = boost::lambda;
namespace b = boost;
namespace ba = boost::assign;

using namespace capd;
using namespace capd::complex;

BOOST_AUTO_TEST_SUITE(CubCellComplexTestSuite)


typedef boost::mpl::list<CubCellComplex<DenseCubMemoryModel<2, bool> >,
                         CubCellComplex<DenseCubMemoryModel<0, bool> >,
                         CubCellComplex<FilteredDenseCubMemoryModel<2> >,
                         CubCellComplex<FilteredDenseCubMemoryModel<0> >
                         > CubCellComplexes;


BOOST_AUTO_TEST_CASE_TEMPLATE(allIterator, CubCellComplexT, CubCellComplexes)
{
  std::vector<size_t> lengths = ba::list_of(1)(1);

  CubCellComplexT complex(lengths);
  typedef typename CubCellComplexT::CellCode CellCode;
  typedef typename CubCellComplexT::CellCode::value_type CellCodeValue;

  BOOST_CHECK_EQUAL(0, std::distance(complex.iterators().allCells().begin(),
                                     complex.iterators().allCells().end()));

  BOOST_CHECK_EQUAL(0, std::distance(complex.iterators(0).allCells().begin(),
                                     complex.iterators(0).allCells().end()));

  {
    complex.insert(complex(0, 0));
  }

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(0, false)(0, false);
    complex.insert(cellCode);
  }


  BOOST_CHECK_EQUAL(2, std::distance(complex.iterators().allCells().begin(),
                                     complex.iterators().allCells().end()));

  BOOST_CHECK_EQUAL(2, std::distance(complex.iterators(1).allCells().begin(),
                                     complex.iterators(1).allCells().end()));

  BOOST_CHECK_EQUAL(0, std::distance(complex.iterators(0).allCells().begin(),
                                     complex.iterators(0).allCells().end()));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(bdIterator, CubCellComplexT, CubCellComplexes)
{
  std::vector<size_t> lengths = ba::list_of(2)(2);

  CubCellComplexT complex(lengths);
  typedef typename CubCellComplexT::CellCode CellCode;
  typedef typename CubCellComplexT::CellCode::value_type CellCodeValue;


  CellCode fullCellCode = ba::list_of<CellCodeValue>(0, true)(0, true);
  complex.insert(fullCellCode);

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(0, false)(0, true);
    complex.insert(cellCode);
  }

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(0, true)(0, false);
    complex.insert(cellCode);
  }

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(1, false)(0, true);
    complex.insert(cellCode);
  }

  {
    typename CubCellComplexT::Iterators::BdCells bd = complex.iterators().bdCells(complex[fullCellCode]);
    BOOST_CHECK_EQUAL(3, std::distance(bd.begin(), bd.end()));
  }

  {
    typename CubCellComplexT::ColoredIterators::BdCells bd = complex.iterators(0).bdCells(complex[fullCellCode]);
    BOOST_CHECK_EQUAL(0, std::distance(bd.begin(), bd.end()));
  }

  {
    typename CubCellComplexT::ColoredIterators::BdCells bd = complex.iterators(1).bdCells(complex[fullCellCode]);
    BOOST_CHECK_EQUAL(3, std::distance(bd.begin(), bd.end()));
  }

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(0, true)(1, false);
    complex.insert(cellCode);
  }

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(1, true)(0, false);
    complex.insert(cellCode);
  }

  {
    typename CubCellComplexT::Iterators::BdCells bd = complex.iterators().bdCells(complex[fullCellCode]);

    BOOST_CHECK_EQUAL(4, std::distance(bd.begin(), bd.end()));
  }

  {
    typename CubCellComplexT::ColoredIterators::BdCells bd = complex.iterators(0).bdCells(complex[fullCellCode]);
    BOOST_CHECK_EQUAL(0, std::distance(bd.begin(), bd.end()));
  }

  {
    typename CubCellComplexT::ColoredIterators::BdCells bd = complex.iterators(1).bdCells(complex[fullCellCode]);
    BOOST_CHECK_EQUAL(4, std::distance(bd.begin(), bd.end()));
  }
}



BOOST_AUTO_TEST_CASE_TEMPLATE(allColoredIterator, CubCellComplexT, CubCellComplexes)
{
  std::vector<size_t> lengths = ba::list_of(1)(1);

  CubCellComplexT complex(lengths);
  typedef typename CubCellComplexT::CellCode CellCode;
  typedef typename CubCellComplexT::CellCode::value_type CellCodeValue;

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(0, true)(0, true);
    complex.insert(cellCode);
    complex.color(complex[cellCode], 0);
  }

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(0, false)(0, false);
    complex.insert(cellCode);
    complex.color(complex[cellCode], 1);
  }


  BOOST_CHECK_EQUAL(1, std::distance(complex.iterators(0).allCells().begin(),
                                     complex.iterators(0).allCells().end()));

  BOOST_CHECK_EQUAL(1, std::distance(complex.iterators(1).allCells().begin(),
                                     complex.iterators(1).allCells().end()));

}


BOOST_AUTO_TEST_CASE_TEMPLATE(bdColoredIterator, CubCellComplexT, CubCellComplexes)
{
  std::vector<size_t> lengths = ba::list_of(20)(20);

  CubCellComplexT complex(lengths);
  typedef typename CubCellComplexT::CellCode CellCode;
  typedef typename CubCellComplexT::CellCode::value_type CellCodeValue;


  CellCode fullCellCode = ba::list_of<CellCodeValue>(0, true)(0, true);
  complex.insert(fullCellCode);
  complex.color(complex[fullCellCode], 1);

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(0, false)(0, true);
    complex.insert(cellCode);
    complex.color(complex[cellCode], 0);
  }

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(0, true)(0, false);
    complex.insert(cellCode);
    complex.color(complex[cellCode], 1);
  }

  {
    BOOST_CHECK_EQUAL(1, std::distance(complex.iterators(0).bdCells(complex[fullCellCode]).begin(),
                                       complex.iterators(0).bdCells(complex[fullCellCode]).end()));

    BOOST_CHECK_EQUAL(1, std::distance(complex.iterators(1).bdCells(complex[fullCellCode]).begin(),
                                       complex.iterators(1).bdCells(complex[fullCellCode]).end()));
  }

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(0, true)(1, false);
    complex.insert(cellCode);
    complex.color(complex[cellCode], 1);
  }


  {
    BOOST_CHECK_EQUAL(1, std::distance(complex.iterators(0).bdCells(complex[fullCellCode]).begin(),
                                       complex.iterators(0).bdCells(complex[fullCellCode]).end()));

    BOOST_CHECK_EQUAL(2, std::distance(complex.iterators(1).bdCells(complex[fullCellCode]).begin(),
                                       complex.iterators(1).bdCells(complex[fullCellCode]).end()));
  }

}


BOOST_AUTO_TEST_CASE_TEMPLATE(fillWithBoundary, CubCellComplexT, CubCellComplexes)
{
  std::vector<size_t> lengths = ba::list_of(1)(1);

  CubCellComplexT complex(lengths);
  typedef typename CubCellComplexT::CellCode CellCode;
  typedef typename CubCellComplexT::CellCode::value_type CellCodeValue;

  {
    complex.fillWithBoundary(complex.insert(complex(0, 0)));
  }


  BOOST_CHECK_EQUAL(9, std::distance(complex.iterators().allCells().begin(),
                                     complex.iterators().allCells().end()));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(fillWithBoundaries, CubCellComplexT, CubCellComplexes)
{
  std::vector<size_t> lengths = ba::list_of(1)(1);

  CubCellComplexT complex(lengths);
  typedef typename CubCellComplexT::CellCode CellCode;
  typedef typename CubCellComplexT::CellCode::value_type CellCodeValue;

  complex.fillWithBoundaries();
  complex.insert(complex(0, 0));
  BOOST_CHECK_EQUAL(1, std::distance(complex.iterators().allCells().begin(),
                                     complex.iterators().allCells().end()));

  complex.fillWithBoundaries();
  BOOST_CHECK_EQUAL(9, std::distance(complex.iterators().allCells().begin(),
                                     complex.iterators().allCells().end()));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(fillWithBoundariesColor, CubCellComplexT, CubCellComplexes)
{
  std::vector<size_t> lengths = ba::list_of(1)(1);

  CubCellComplexT complex(lengths);
  typedef typename CubCellComplexT::CellCode CellCode;
  typedef typename CubCellComplexT::CellCode::value_type CellCodeValue;

  complex.fillWithBoundaries();
  typename CubCellComplexT::Cell cell = complex.insert(complex(0, 0));
  cell.setColor(1);

  BOOST_CHECK_EQUAL(1, std::distance(complex.iterators(1).allCells().begin(),
                                     complex.iterators(1).allCells().end()));

  complex.fillWithBoundaries();
  BOOST_CHECK_EQUAL(9, std::distance(complex.iterators(1).allCells().begin(),
                                     complex.iterators(1).allCells().end()));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(fillWithBoundaryColor, CubCellComplexT, CubCellComplexes)
{
  std::vector<size_t> lengths = ba::list_of(2)(2);

  CubCellComplexT complex(lengths);
  typedef typename CubCellComplexT::CellCode CellCode;
  typedef typename CubCellComplexT::CellCode::value_type CellCodeValue;

  {
    complex.fillWithBoundary(complex.insert(complex(1, 1)));
    complex.fillWithBoundary(complex.insert(complex(0, 0)), 1, false);
  }

  BOOST_CHECK_EQUAL(0, complex[complex(1, 1)].color());
  BOOST_CHECK_EQUAL(1, complex[complex(0, 0)].color());

  BOOST_CHECK_EQUAL(17, std::distance(complex.iterators().allCells().begin(),
                                      complex.iterators().allCells().end()));

  BOOST_CHECK_EQUAL(9, std::distance(complex.iterators(1).allCells().begin(),
                                     complex.iterators(1).allCells().end()));

  BOOST_CHECK_EQUAL(8, std::distance(complex.iterators(0).allCells().begin(),
                                     complex.iterators(0).allCells().end()));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(dim, CubCellComplexT, CubCellComplexes)
{
  std::vector<size_t> lengths = ba::list_of(2)(1);

  CubCellComplexT complex(lengths);
  typedef typename CubCellComplexT::CellCode CellCode;
  typedef typename CubCellComplexT::CellCode::value_type CellCodeValue;

  {
    complex.fillWithBoundary(complex.insert(complex(0, 0)));
    complex.fillWithBoundary(complex.insert(complex(1, 0)));
  }

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(2, false)(1, false);
    BOOST_CHECK_EQUAL(0, complex[cellCode].dim());
  }

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(2, false)(0, true);
    BOOST_CHECK_EQUAL(1, complex[cellCode].dim());
  }

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(1, true)(0, true);
    BOOST_CHECK_EQUAL(2, complex[cellCode].dim());
  }
}


BOOST_AUTO_TEST_CASE_TEMPLATE(cbdIterator, CubCellComplexT, CubCellComplexes)
{
  std::vector<size_t> lengths = ba::list_of(2)(2);

  CubCellComplexT complex(lengths);
  typedef typename CubCellComplexT::CellCode CellCode;
  typedef typename CubCellComplexT::CellCode::value_type CellCodeValue;


  CellCode pointCellCode = ba::list_of<CellCodeValue>(1, false)(1, false);
  complex.insert(pointCellCode);

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(1, false)(1, true);
    complex.insert(cellCode);
  }

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(1, true)(1, false);
    complex.insert(cellCode);
  }

  {
    typename CubCellComplexT::Iterators::CbdCells cbd = complex.iterators().cbdCells(complex[pointCellCode]);

    BOOST_CHECK_EQUAL(2, std::distance(cbd.begin(),
                                       cbd.end()));
  }

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(0, true)(1, false);
    complex.insert(cellCode);
  }

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(1, false)(0, true);
    complex.insert(cellCode);
  }

  {
    typename CubCellComplexT::Iterators::CbdCells cbd = complex.iterators().cbdCells(complex[pointCellCode]);

    BOOST_CHECK_EQUAL(4, std::distance(cbd.begin(), cbd.end()));
  }

}

BOOST_AUTO_TEST_CASE_TEMPLATE(compilationOfCopyIterators, CubCellComplexT, CubCellComplexes)
{
  std::vector<size_t> lengths = ba::list_of(1)(1);
  CubCellComplexT complex(lengths);
  {
    auto it1 = complex.iterators().allCells().begin();
    auto it2 = complex.iterators().allCells().end();
    it1 = it2;
  }
}

BOOST_AUTO_TEST_SUITE_END()
