/////////////////////////////////////////////////////////////////////////////
/// @file CubicTessellationTest
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-09-11
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////


#include <capd/complex/CubicTessellation.h>
#include <capd/auxil/Logger.h>

#include <TestTools.h>

#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/mpl/list.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/assign.hpp>
#include <boost/foreach.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/typeof/typeof.hpp>

#include <map>
#include <vector>
#include <algorithm>
#include <utility>
#include <sstream>
#include <iosfwd>
#include <iterator>


namespace bl = boost::lambda;
namespace b = boost;
namespace ba = boost::assign;

using namespace capd;
using namespace capd::complex;

namespace boost {

}

BOOST_AUTO_TEST_SUITE(CubicTessellationTestSuite)

BOOST_AUTO_TEST_CASE(empty)
{
  std::vector<size_t> lengths;
  CubicTessellation<0> qubes(lengths);
  BOOST_CHECK_EQUAL(0, qubes.size());
}

BOOST_AUTO_TEST_CASE(index_grid2D)
{
  std::vector<size_t> lengths = ba::list_of(4)(2);
  CubicTessellation<0> qubes(lengths);

  BOOST_REQUIRE_EQUAL(45, qubes.size());

  typedef CubicTessellation<0>::CellCode CellCode;

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false)(0, false);
    BOOST_CHECK_EQUAL(0, qubes[cellCode]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false)(1, false);
    BOOST_CHECK_EQUAL(2, qubes[cellCode]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(4, false)(0, false);
    BOOST_CHECK_EQUAL(40, qubes[cellCode]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false)(0, true);
    BOOST_CHECK_EQUAL(1, qubes[cellCode]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false)(1, true);
    BOOST_CHECK_EQUAL(3, qubes[cellCode]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(1, false)(0, false);
    BOOST_CHECK_EQUAL(10, qubes[cellCode]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(4, false)(2, false);
    BOOST_CHECK_EQUAL(44, qubes[cellCode]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, true)(0, true);
    BOOST_CHECK_EQUAL(6, qubes[cellCode]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, true)(1, true);
    BOOST_CHECK_EQUAL(8, qubes[cellCode]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(1, true)(0, true);
    BOOST_CHECK_EQUAL(16, qubes[cellCode]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(3, true)(1, true);
    BOOST_CHECK_EQUAL(38, qubes[cellCode]);
  }

}

BOOST_AUTO_TEST_CASE(index_grid3D_1)
{
  std::vector<size_t> lengths = ba::list_of(1)(1)(1);
  CubicTessellation<0> qubes(lengths);

  BOOST_REQUIRE_EQUAL(27, qubes.size());

  typedef CubicTessellation<0>::CellCode CellCode;

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false)(0, false)(0, false);
    BOOST_CHECK_EQUAL(0, qubes[cellCode]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(1, false)(1, false)(1, false);
    BOOST_CHECK_EQUAL(qubes.size() - 1, qubes[cellCode]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, true)(0, true)(0, true);
    BOOST_CHECK_EQUAL(13, qubes[cellCode]);
  }

}

BOOST_AUTO_TEST_CASE(index_grid3D_2)
{
  std::vector<size_t> lengths = ba::list_of(2)(2)(2);
  CubicTessellation<0> qubes(lengths);

  BOOST_REQUIRE_EQUAL(125, qubes.size());

  typedef CubicTessellation<0>::CellCode CellCode;

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false)(0, false)(0, false);
    BOOST_CHECK_EQUAL(0, qubes[cellCode]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(2, false)(2, false)(2, false);
    BOOST_CHECK_EQUAL(qubes.size() - 1, qubes[cellCode]);
  }

}


BOOST_AUTO_TEST_CASE(index_id_grid1D)
{
  std::vector<size_t> lengths = ba::list_of(1);
  CubicTessellation<0> qubes(lengths);

  BOOST_REQUIRE_EQUAL(3, qubes.size());

  typedef CubicTessellation<0>::CellCode CellCode;

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false);
    BOOST_CHECK_EQUAL(cellCode, qubes[0]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(1, false);
    BOOST_CHECK_EQUAL(cellCode, qubes[2]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, true);
    BOOST_CHECK_EQUAL(cellCode, qubes[1]);
  }

}

BOOST_AUTO_TEST_CASE(index_id_grid2D)
{
  std::vector<size_t> lengths = ba::list_of(4)(2);
  CubicTessellation<0> qubes(lengths);

  BOOST_REQUIRE_EQUAL(45, qubes.size());

  typedef CubicTessellation<0>::CellCode CellCode;

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false)(0, false);
    BOOST_CHECK_EQUAL(cellCode, qubes[0]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false)(1, false);
    BOOST_CHECK_EQUAL(cellCode, qubes[2]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(4, false)(0, false);
    BOOST_CHECK_EQUAL(cellCode, qubes[40]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(4, false)(2, false);
    BOOST_CHECK_EQUAL(cellCode, qubes[44]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(3, true)(1, true);
    BOOST_CHECK_EQUAL(cellCode, qubes[38]);
  }

}

BOOST_AUTO_TEST_CASE(index_id_grid3D)
{

  const int DIM=3;
  std::vector<size_t> lengths = ba::list_of(1)(1)(1);
  CubicTessellation<DIM> qubes(lengths);

  BOOST_REQUIRE_EQUAL(27, qubes.size());

  typedef CubicTessellation<DIM>::CellCode CellCode;

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false)(0, false)(0, false);
    BOOST_CHECK_EQUAL(cellCode, qubes[0]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(1, false)(1, false)(1, false);
    BOOST_CHECK_EQUAL(cellCode, qubes[qubes.size() - 1]);
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, true)(0, true)(0, true);
    BOOST_CHECK_EQUAL(cellCode, qubes[13]);
  }

}


BOOST_AUTO_TEST_CASE(dim_grid3D)
{

  const int DIM=3;
  std::vector<size_t> lengths = ba::list_of(2)(3)(1);
  CubicTessellation<DIM> qubes(lengths);

  BOOST_REQUIRE_EQUAL(105, qubes.size());

  typedef CubicTessellation<DIM>::CellCode CellCode;

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false)(0, false)(0, false);
    BOOST_CHECK_EQUAL(0, qubes.dim(qubes[cellCode]));
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(1, false)(1, false)(1, false);
    BOOST_CHECK_EQUAL(0, qubes.dim(qubes[cellCode]));
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, true)(0, true)(0, true);
    BOOST_CHECK_EQUAL(3, qubes.dim(qubes[cellCode]));
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(1, true)(0, true)(1, false);
    BOOST_CHECK_EQUAL(2, qubes.dim(qubes[cellCode]));
  }

}


BOOST_AUTO_TEST_CASE(iterators_length)
{
  const int DIM=2;
  typedef CubicTessellation<DIM>::CellCode CellCode;
  std::vector<size_t> lengths = ba::list_of(2)(2);
  CubicTessellation<DIM> qubes(lengths);

  {
    CubicTessellation<DIM>::All all = qubes.all();
    BOOST_REQUIRE_EQUAL(qubes.size(), std::distance(all.begin(), all.end()));
  }

  {
    CubicTessellation<DIM>::All all = ((const CubicTessellation<DIM>&) qubes).all();
    BOOST_REQUIRE_EQUAL(qubes.size(), std::distance(all.begin(), all.end()));
  }


  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false)(0, false);
    CubicTessellation<DIM>::Bd bd = qubes.bd(qubes[cellCode]);
    BOOST_REQUIRE_EQUAL(0, std::distance(bd.begin(), bd.end()));
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false)(0, true);
    CubicTessellation<DIM>::Bd bd = qubes.bd(qubes[cellCode]);
    BOOST_REQUIRE_EQUAL(2, std::distance(bd.begin(), bd.end()));
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, true)(0, true);
    CubicTessellation<DIM>::Bd bd = qubes.bd(qubes[cellCode]);
    BOOST_REQUIRE_EQUAL(4, std::distance(bd.begin(), bd.end()));
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false)(0, false);
    CubicTessellation<DIM>::Cbd cbd = qubes.cbd(qubes[cellCode]);
    BOOST_REQUIRE_EQUAL(2, std::distance(cbd.begin(), cbd.end()));
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(1, false)(1, false);
    CubicTessellation<DIM>::Cbd cbd = qubes.cbd(qubes[cellCode]);
    BOOST_REQUIRE_EQUAL(4, std::distance(cbd.begin(), cbd.end()));
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(2, false)(2, false);
    CubicTessellation<DIM>::Cbd cbd = qubes.cbd(qubes[cellCode]);
    BOOST_REQUIRE_EQUAL(2, std::distance(cbd.begin(), cbd.end()));
  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, true)(0, true);
    CubicTessellation<DIM>::Cbd cbd = qubes.cbd(qubes[cellCode]);
    BOOST_REQUIRE_EQUAL(0, std::distance(cbd.begin(), cbd.end()));
  }

}

BOOST_AUTO_TEST_CASE(check_boundary)
{
  const int DIM=3;
  typedef CubicTessellation<DIM>::CellCode CellCode;
  typedef CubicTessellation<DIM>::Bd Bd;
  typedef Bd::iterator BdIt;

  std::vector<size_t> lengths = ba::list_of(2)(2)(2);
  CubicTessellation<DIM> qubes(lengths);

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, false)(0, false)(0, true);
    Bd bd = qubes.bd(qubes[cellCode]);
    BdIt bdIt = bd.begin();

    {
      CellCode bdcellCode = ba::list_of<CellCode::value_type>(0, false)(0, false)(0, false);
      BOOST_REQUIRE_EQUAL(bdcellCode, qubes[*bdIt]);
    }

    ++bdIt;
    {
      CellCode bdcellCode = ba::list_of<CellCode::value_type>(0, false)(0, false)(1, false);
      BOOST_REQUIRE_EQUAL(bdcellCode, qubes[*bdIt]);
    }

  }


  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, true)(0, true)(0, true);
    Bd bd = qubes.bd(qubes[cellCode]);
    BdIt bdIt = bd.begin();

    {
      CellCode bdcellCode = ba::list_of<CellCode::value_type>(0, false)(0, true)(0, true);
      BOOST_REQUIRE_EQUAL(bdcellCode, qubes[*bdIt]);
    }

    ++bdIt;

    {
      CellCode bdcellCode = ba::list_of<CellCode::value_type>(1, false)(0, true)(0, true);
      BOOST_REQUIRE_EQUAL(bdcellCode, qubes[*bdIt]);
    }

    ++bdIt;

    {
      CellCode bdcellCode = ba::list_of<CellCode::value_type>(0, true)(0, false)(0, true);
      BOOST_REQUIRE_EQUAL(bdcellCode, qubes[*bdIt]);
    }

    ++bdIt;

    {
      CellCode bdcellCode = ba::list_of<CellCode::value_type>(0, true)(1, false)(0, true);
      BOOST_REQUIRE_EQUAL(bdcellCode, qubes[*bdIt]);
    }

    ++bdIt;

    {
      CellCode bdcellCode = ba::list_of<CellCode::value_type>(0, true)(0, true)(0, false);
      BOOST_REQUIRE_EQUAL(bdcellCode, qubes[*bdIt]);
    }

    ++bdIt;

    {
      CellCode bdcellCode = ba::list_of<CellCode::value_type>(0, true)(0, true)(1, false);
      BOOST_REQUIRE_EQUAL(bdcellCode, qubes[*bdIt]);
    }

    ++bdIt;

    BOOST_CHECK(bd.end() == bdIt);
  }


}

BOOST_AUTO_TEST_CASE(check_coboundary)
{
  const int DIM=3;
  typedef CubicTessellation<DIM>::CellCode CellCode;
  typedef CubicTessellation<DIM>::Cbd Cbd;
  typedef Cbd::iterator CbdIt;

  std::vector<size_t> lengths = ba::list_of(2)(2)(2);
  CubicTessellation<DIM> qubes(lengths);

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, true)(0, false)(0, true);
    Cbd cbd = qubes.cbd(qubes[cellCode]);
    CbdIt cbdIt = cbd.begin();

    {
      CellCode cbdcellCode = ba::list_of<CellCode::value_type>(0, true)(0, true)(0, true);
      BOOST_REQUIRE_EQUAL(cbdcellCode, qubes[*cbdIt]);
    }

    ++cbdIt;
    BOOST_CHECK(cbd.end() == cbdIt);

  }

  {
    CellCode cellCode = ba::list_of<CellCode::value_type>(0, true)(1, false)(0, true);
    Cbd cbd = qubes.cbd(qubes[cellCode]);
    CbdIt cbdIt = cbd.begin();

    {
      CellCode cbdcellCode = ba::list_of<CellCode::value_type>(0, true)(1, true)(0, true);
      BOOST_REQUIRE_EQUAL(cbdcellCode, qubes[*cbdIt]);
    }

    ++cbdIt;
    {
      CellCode cbdcellCode = ba::list_of<CellCode::value_type>(0, true)(0, true)(0, true);
      BOOST_REQUIRE_EQUAL(cbdcellCode, qubes[*cbdIt]);
    }

    ++cbdIt;
    BOOST_CHECK(cbd.end() == cbdIt);

  }


}

BOOST_AUTO_TEST_SUITE_END()
