/////////////////////////////////////////////////////////////////////////////
/// @file IteratedCoboundaryTest
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-03-11
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////


#include <capd/complex/IteratedCoboundary.h>
#include <capd/auxil/Logger.h>
#include <../apiRedHom/Fixtures.h>
#include <TestTools.h>

#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/mpl/list.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/assign.hpp>
#include <boost/foreach.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/typeof/typeof.hpp>

#include <map>
#include <vector>
#include <algorithm>
#include <utility>
#include <sstream>
#include <capd/complex/CubicalComplexFromArray.h>
#include <boost/type_traits/remove_const.hpp>

namespace capd
{
  namespace test
  {

  }
}

namespace bl = boost::lambda;
namespace b = boost;
namespace ba = boost::assign;

using namespace capd;
using namespace capd::test;
using namespace capd::complex;

BOOST_AUTO_TEST_SUITE(IteratedCoboundaryTestSuite)


template<typename Range, typename Cell>
void check(Range& gcbd, Cell& cell, size_t cbdCellsSize, size_t depth)
{
  std::set<typename boost::remove_const<Cell>::type> gcbdUniq(gcbd.begin(), gcbd.end());
  BOOST_CHECK_EQUAL(cbdCellsSize, gcbdUniq.size());

  for(auto cbdCell : gcbdUniq) {
    BOOST_CHECK_EQUAL(cell.getDim() + depth, cbdCell.getDim());
  }
}

template<size_t DEPTH, typename ComplexType>
void testStatic(ComplexType& complex, const typename ComplexType::Cell& cell, size_t cbdCellsSize)
{
  {
    auto boundary = makeIteratedCoboundary<DEPTH>(complex);
    auto gcbd = boundary(cell);
    check(gcbd, cell, cbdCellsSize, DEPTH);
  }

}

template<typename ComplexType>
void testDynamic(ComplexType& complex, const typename ComplexType::Cell& cell, size_t cbdCellsSize, size_t depth)
{
  {
    auto boundary = makeIteratedCoboundary(complex, depth);
    auto gcbd = boundary(cell);
    check(gcbd, cell, cbdCellsSize, depth);
  }
}


BOOST_AUTO_TEST_CASE_TEMPLATE(directCoboundary, Fixture, ComplexFixtures)
{
  Fixture fixture;
  typedef typename Fixture::ComplexType ComplexType;
  typedef typename ComplexType::Cell Cell;;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();

  for (Cell cell : complex.iterators().allCells()) {
    auto cbdCells = complex.iterators().cbdCells(cell);
    size_t cbdCellsSize = std::distance(cbdCells.begin(), cbdCells.end());

    testStatic<1>(complex, cell, cbdCellsSize);
    testDynamic(complex, cell, cbdCellsSize, 1);
  }
}

BOOST_AUTO_TEST_CASE(gotoFullCells)
{
  typedef SimplicialTriangle Fixture;
  Fixture fixture;
  typedef typename Fixture::ComplexType ComplexType;
  typedef typename ComplexType::Cell Cell;;
  ComplexType& complex = fixture.complex->rawComplex<ComplexType>();
  auto cell = complex[fixture.b.rawCell<Cell>().code()];
  testStatic<2>(complex, cell, 1);
  testDynamic(complex, cell, 1, 2);

}

BOOST_AUTO_TEST_CASE(colorDoesNotMatter)
{
  std::vector<size_t> shape = {{3, 3}};
  boost::multi_array<size_t, 2> array(shape);

  std::fill(array.origin(), array.origin() + array.num_elements(), 1);
  array[1][1] = 2;
  auto _complex = BuildCubicalComplexFromArray(array);
  auto& complex = *_complex;
  auto cellCode = complex(1, 1);
  cellCode[0].second = cellCode[1].second = false;

  testStatic<2>(complex, complex[cellCode], 4);
  testDynamic(complex, complex[cellCode], 4, 2);
}

BOOST_AUTO_TEST_SUITE_END()
