/////////////////////////////////////////////////////////////////////////////
/// @file RelativeHomologyTest.cpp
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-05-26
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2014 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.


#include <TestTools.h>
#include "../apiRedHom/Fixtures.h"

#include <capd/complex/RelativeHomology.h>

#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/mpl/list.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/assign.hpp>

#include <map>
#include <vector>
#include <algorithm>
#include <utility>

namespace bl = boost::lambda;
namespace b = boost;

using namespace boost::assign;
using namespace boost::lambda;

using namespace capd::test;
using namespace capd::complex;

BOOST_AUTO_TEST_SUITE(RelativeHomologyTestSuite)

BOOST_AUTO_TEST_CASE_TEMPLATE(remove_all, Fixture, ComplexFixtures)
{
  typedef typename Fixture::ComplexType ComplexType;

  Fixture fixture;
  Fixture subspace;
  ComplexType& complex = fixture.rawComplex();
  ComplexType& subcomplex = subspace.rawComplex();

  typedef RelativeHomology<ComplexType, int, int> Alg;
  Alg alg(complex, subcomplex, false);
  std::vector<typename Alg::Scalar> betti = alg();

  BOOST_CHECK_EQUAL(0, betti.size());
}


BOOST_AUTO_TEST_CASE(ball_simplcial)
{
  typedef SimplicialComplexFixture::ComplexType ComplexType;
  SimplicialTriangle space;
  SimplicialCircle subspace;

  ComplexType& complex = space.rawComplex();
  ComplexType& subcomplex = subspace.rawComplex();

  typedef RelativeHomology<ComplexType, int, int> Alg;
  Alg alg(complex, subcomplex, false);
  std::vector<Alg::Scalar> betti = alg();
  std::vector<Alg::Scalar> expectedBetti = list_of(0)(0)(1);

  BOOST_CHECK_EQUAL_COLLECTIONS(expectedBetti.begin(), expectedBetti.end(), betti.begin(), betti.end());
}

BOOST_AUTO_TEST_CASE(ball_cubical)
{
  typedef CubicalComplexFixture<2>::ComplexType ComplexType;
  CubicalRectangle<> space;
  CubicalCircle<> subspace;

  ComplexType& complex = space.rawComplex();
  ComplexType& subcomplex = subspace.rawComplex();

  typedef RelativeHomology<ComplexType, int, int> Alg;
  Alg alg(complex, subcomplex, false);
  std::vector<Alg::Scalar> betti = alg();
  std::vector<Alg::Scalar> expectedBetti = list_of(0)(0)(1);

  BOOST_CHECK_EQUAL_COLLECTIONS(expectedBetti.begin(), expectedBetti.end(), betti.begin(), betti.end());
}

BOOST_AUTO_TEST_CASE(ball_generalized)
{
  typedef GeneralizedComplexFixture::ComplexType ComplexType;
  GeneralizedRectangle space;
  GeneralizedCircle subspace;

  ComplexType& complex = space.rawComplex();
  ComplexType& subcomplex = subspace.rawComplex();

  typedef RelativeHomology<ComplexType, int, int> Alg;
  Alg alg(complex, subcomplex, false);
  std::vector<Alg::Scalar> betti = alg();
  std::vector<Alg::Scalar> expectedBetti = list_of(0)(0)(1);

  BOOST_CHECK_EQUAL_COLLECTIONS(expectedBetti.begin(), expectedBetti.end(), betti.begin(), betti.end());
}

BOOST_AUTO_TEST_CASE(open_circle_simplicial)
{
  // This test checks if we do not try to extract points in Coreduction Algorithm. We cannot do it after Excision.
  typedef SimplicialComplexFixture::ComplexType ComplexType;
  SimplicialCircle space;
  SimplicialPoint subspace;

  ComplexType& complex = space.rawComplex();
  ComplexType& subcomplex = subspace.rawComplex();

  typedef RelativeHomology<ComplexType, int, int> Alg;
  Alg alg(complex, subcomplex, false);
  std::vector<Alg::Scalar> betti = alg();
  std::vector<Alg::Scalar> expectedBetti = list_of(0)(1);

  BOOST_CHECK_EQUAL_COLLECTIONS(expectedBetti.begin(), expectedBetti.end(), betti.begin(), betti.end());
}


BOOST_AUTO_TEST_SUITE_END()
