/////////////////////////////////////////////////////////////////////////////
/// @file SqueezedComplexTest
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2015-01-20
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2015 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////


#include <capd/complex/SqueezedComplex.h>
#include <capd/complex/Coreduction.h>
#include <capd/complex/Shave.h>

#include <capd/auxil/Logger.h>

#include <../apiRedHom/Fixtures.h>
#include <TestTools.h>

#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/mpl/list.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/assign.hpp>
#include <boost/foreach.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/typeof/typeof.hpp>

#include <map>
#include <vector>
#include <algorithm>
#include <utility>
#include <sstream>
#include <boost/typeof/typeof.hpp>
#include <capd/complex/ComputeHomologyOrchestrator.hpp>

namespace bl = boost::lambda;
namespace b = boost;
namespace ba = boost::assign;

using namespace capd;
using namespace capd::complex;
using namespace capd::test;

BOOST_AUTO_TEST_SUITE(SqueezedComplexTestSuite)

template<typename Complex, typename Iterators>
size_t CheckIterators(Complex& complex, Iterators cells, bool up)
{
  // This function is to check compilation!
  size_t count = 0;
  for (typename Iterators::iterator it = cells.begin(), end = cells.end(); it != end; ++it) {
    it->getColor();
    it->getDim();
    ++count;
    if (up) {
      CheckIterators(complex, complex.iterators().cbdCells(*it), up);
      CheckIterators(complex, complex.iterators(1).cbdCells(*it), up);
    } else {
      CheckIterators(complex, complex.iterators().bdCells(*it), up);
      CheckIterators(complex, complex.iterators(1).bdCells(*it), up);
    }
  }

  return count;
}

template<typename Complex>
void CheckIteratorFamily(Complex& complex)
{
  CheckIterators(complex, complex.iterators().allCells(), true);
  BOOST_CHECK(CheckIterators(complex, complex.iterators().dimCells(0), true) > 0 || complex.size() == 0);
  BOOST_CHECK(CheckIterators(complex, complex.iterators(1).dimCells(0), true) > 0 || complex.size() == 0);

  CheckIterators(complex, complex.iterators().allCells(), false);
  CheckIterators(complex, complex.iterators().dimCells(1), false);
}

template<typename Complex, typename Squeezed>
void CheckIdentity(Complex& complex, Squeezed& squeezed)
{
  BOOST_CHECK_EQUAL(complex.cardinality(), squeezed.size());
  for (size_t idx = 0; idx < squeezed.size(); ++idx) {
    BOOST_CHECK_EQUAL(idx, squeezed[idx].getId());
    BOOST_CHECK_EQUAL(squeezed[idx].getOrg().getDim(), squeezed[idx].getDim());
  }

  BOOST_FOREACH(const typename Squeezed::iterator::value_type& cell, squeezed) {
    BOOST_CHECK_EQUAL(cell.getOrg().getDim(), cell.getDim());
  }

  CheckIteratorFamily(squeezed);

  if (squeezed.size() > 0) {
    BOOST_CHECK(squeezed.iterators().dimCells(0).begin() != squeezed.iterators().dimCells(0).end());
    BOOST_CHECK(squeezed.iterators(1).dimCells(0).begin() != squeezed.iterators(1).dimCells(0).end());
  }
}

BOOST_AUTO_TEST_CASE_TEMPLATE(identity, Fixture, ComplexFixtures)
{
  typedef typename Fixture::ComplexType ComplexType;
  Fixture fixture;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();
  typedef SqueezedComplex<ComplexType> Squeezed;;
  Squeezed squeezed(complex);
  CheckIdentity(complex, squeezed);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(wrapwrap, Fixture, ComplexFixtures)
{
  typedef typename Fixture::ComplexType ComplexType;
  Fixture fixture;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();
  typedef SqueezedComplex<ComplexType> Squeezed;;
  Squeezed squeezed(complex);
  typedef SqueezedComplex<Squeezed> Squeezed2;;
  Squeezed2 squeezed2(squeezed);

  CheckIdentity(complex, squeezed2);
  CheckIdentity(squeezed, squeezed2);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(wrapAndCoreduce, Fixture, ComplexFixtures)
{
  typedef typename Fixture::ComplexType ComplexType;
  Fixture fixture;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();
  typedef SqueezedComplex<ComplexType> Squeezed;;
  Squeezed squeezed(complex);

  boost::shared_ptr<typename CoreductionFactory<Squeezed, int, int>::DefaultAlgorithm> algo =
    CoreductionFactory<Squeezed, int, int>::createDefault(squeezed);

  size_t reduced = (*algo)();

  BOOST_CHECK_EQUAL(fixture.coreductions, reduced);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(wrapWrapAndCoreduce, Fixture, ComplexFixtures)
{
  typedef typename Fixture::ComplexType ComplexType;
  Fixture fixture;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();
  typedef SqueezedComplex<ComplexType> Squeezed;;
  Squeezed squeezed(complex);

  typedef SqueezedComplex<Squeezed> Squeezed2;;
  Squeezed2 squeezed2(squeezed);
  boost::shared_ptr<typename CoreductionFactory<Squeezed2, int, int>::DefaultAlgorithm> algo =
    CoreductionFactory<Squeezed2, int, int>::createDefault(squeezed2);

  size_t reduced = (*algo)();

  BOOST_CHECK_EQUAL(fixture.coreductions, reduced);
}


BOOST_AUTO_TEST_CASE_TEMPLATE(wrapAndComputeHomology, Fixture, ComplexFixtures)
{
  typedef typename Fixture::ComplexType ComplexType;
  Fixture fixture;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();
  typedef SqueezedComplex<ComplexType> Squeezed;;
  Squeezed squeezed(complex);

  typedef ComputeHomologyOrchestrator<Squeezed, int, int> Alg;
  Alg alg(squeezed, true, true, true, true);

  std::vector<typename Alg::Int> betti = alg();
  BOOST_CHECK_EQUAL_COLLECTIONS(fixture.betti.begin(), fixture.betti.end(), betti.begin(), betti.end());
}


BOOST_AUTO_TEST_CASE_TEMPLATE(shaveWrapAndComputeHomology, Fixture, ComplexFixtures)
{
  typedef typename Fixture::ComplexType ComplexType;
  Fixture fixture;
  ComplexType& complex = fixture.complex->template rawComplex<ComplexType>();

  {
    boost::shared_ptr<typename ShaveAlgorithmFactory<ComplexType>::DefaultAlgorithm> algo =
      ShaveAlgorithmFactory<ComplexType>::createDefault(complex);

    (*algo)();
  }

  typedef SqueezedComplex<ComplexType> Squeezed;;
  Squeezed squeezed(complex);

  typedef ComputeHomologyOrchestrator<Squeezed, int, int> Alg;
  Alg alg(squeezed, true, true, true, true);

  std::vector<typename Alg::Int> betti = alg();
  BOOST_CHECK_EQUAL_COLLECTIONS(fixture.betti.begin(), fixture.betti.end(), betti.begin(), betti.end());
}


BOOST_AUTO_TEST_SUITE_END()
