/////////////////////////////////////////////////////////////////////////////
/// @file ComputePersistentHomologyTest
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-06-30
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////


#include <capd/persistence/ComputePersistentHomology.h>
#include <capd/auxil/Logger.h>

#include <TestTools.h>
#include "PointCloudFixtures.h"
#include <capd/persistence/PointCloudUtils.h>
#include <capd/persistence/PointCloudSimplicialCellSComplex.h>

#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/mpl/list.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/assign.hpp>
#include <boost/foreach.hpp>

#include <map>
#include <vector>
#include <algorithm>
#include <utility>
#include <sstream>
#include <capd/persistence/ComputePersistentHomology.h>
#include <capd/persistence/PointCloudTypes.h>
#include <boost/scoped_ptr.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/typeof/typeof.hpp>
#include <boost/math/special_functions/fpclassify.hpp>





// TODO remove it!!!!
#include <capd/complex/CellComplex.h>
#include <capd/complex/CubCellComplex.h>
#include <capd/persistence/DistanceMatrixPointCloud.h>
#include <capd/vectalg/Matrix.hpp>


namespace bl = boost::lambda;
namespace b = boost;
namespace ba = boost::assign;

using namespace capd;
using namespace capd::persistence;

BOOST_AUTO_TEST_SUITE(ComputePersistentHomologyTestSuite)

BOOST_AUTO_TEST_CASE(HouseExample)
{
  typedef capd::persistence::ComputePersistentHomology<PointCloudSimplicialCellSComplex> ComputePersistentHomology;

  if (!ComputePersistentHomology::enabled()) {
    return;
  }

  typedef EuclideanPointCloudType<0>::PointCloudType PointCloudType;
  typedef PointCloudType::Point Point;

  std::vector<Point> points = ba::list_of
    (Point((double[]){ -1.0, 0.0 }))
    (Point((double[]){ 1.0, 0.0 }))
    (Point((double[]){ 1.0, 2.0 }))
    (Point((double[]){ -1.0, 2.0 }))
    (Point((double[]){ 0.0, 3.0 }));
  ;

  boost::scoped_ptr<PointCloudType> pc(makeSizeLimitedPointCloud<EuclideanPointCloudType<0> >(points.begin(), points.end(), 0));

  typedef PointCloudSimplicialCellSComplex PCComplex;
  PCComplex pccomplex(*pc, 0);
  pccomplex();

  ComputePersistentHomology persistence(pccomplex);
  persistence();

  BOOST_CHECK_EQUAL(5, persistence.pairs().size());
  BOOST_CHECK_EQUAL(1, persistence.cycles().size());

  BOOST_CHECK_EQUAL(4, persistence.pairs(0).size());
  BOOST_CHECK_EQUAL(1, persistence.cycles(0).size());

  BOOST_CHECK_EQUAL(1, persistence.pairs(1).size());
  BOOST_CHECK_EQUAL(0, persistence.cycles(1).size());

  BOOST_CHECK_EQUAL(0, persistence.pairs(2).size());
  BOOST_CHECK_EQUAL(0, persistence.cycles(2).size());

  BOOST_CHECK_EQUAL(0, persistence.pairs(3).size());
  BOOST_CHECK_EQUAL(0, persistence.cycles(3).size());
}


BOOST_AUTO_TEST_CASE(DistanceMatrixTriangle)
{
  typedef capd::persistence::ComputePersistentHomology<PointCloudSimplicialCellSComplex> ComputePersistentHomology;

  if (!ComputePersistentHomology::enabled()) {
    return;
  }

  double d[3][3] = { {0, 1, 1}, {1, 0, 1}, {1, 1, 0} };
  const DistanceMatrixPointCloud::Metric::DistanceMatrix dm(d);
  DistanceMatrixPointCloud distance(dm);
  DegreeLimitedNeighbourhood<DistanceMatrixPointCloud> neighbourhood(distance, 0, 0);
  neighbourhood();

  typedef PointCloudSimplicialCellSComplex PCComplex;
  PCComplex pccomplex(distance, 0);
  pccomplex();

  ComputePersistentHomology persistence(pccomplex);
  persistence();

  BOOST_CHECK_EQUAL(2, persistence.pairs().size());
  BOOST_CHECK_EQUAL(1, persistence.cycles().size());

  BOOST_CHECK_EQUAL(2, persistence.pairs(0).size());
  BOOST_CHECK_EQUAL(1, persistence.cycles(0).size());

  BOOST_CHECK_EQUAL(0, persistence.pairs(1).size());
  BOOST_CHECK_EQUAL(0, persistence.cycles(1).size());

  //  BOOST_CHECK_EQUAL(0, persistence.pairs(2).size());
  //BOOST_CHECK_EQUAL(0, persistence.cycles(2).size());
}



BOOST_AUTO_TEST_CASE_TEMPLATE(all_euclidean, Fixture, capd::persistence::test::PointCloudFixtures)
{
  typedef capd::persistence::ComputePersistentHomology<PointCloudSimplicialCellSComplex> ComputePersistentHomology;

  if (!ComputePersistentHomology::enabled()) {
    return;
  }

  typedef EuclideanPointCloudType<0>::PointCloudType PointCloudType;
  Fixture fixture;
  const typename Fixture::Points& points = fixture.points();
  boost::scoped_ptr<PointCloudType> pc(makeSizeLimitedPointCloud<EuclideanPointCloudType<0> >(points.begin(), points.end(), 0));

  typedef PointCloudSimplicialCellSComplex PCComplex;
  PCComplex pccomplex(*pc, fixture.maxDim());
  pccomplex();

  ComputePersistentHomology persistence(pccomplex);
  persistence();

  BOOST_CHECK(0 < persistence.pairs().size());
  BOOST_CHECK(0 < persistence.cycles().size());

  BOOST_FOREACH(const ComputePersistentHomology::Pairs::value_type& pair, persistence.pairs()) {
    BOOST_CHECK(pccomplex[pair.first].color() <= pccomplex[pair.second].color());
  }

  ComputePersistentHomology::Cycles cycles = persistence.cycles();
  BOOST_AUTO(cellIdFiltration, bl::bind(&PCComplex::Cell::color,
                                        bl::bind(static_cast<PCComplex::Cell(PCComplex::*)(PCComplex::Id)>(&PCComplex::operator[]), bl::var(pccomplex), bl::_1)));

  std::vector<double> cycleColors(boost::make_transform_iterator(cycles.begin(), cellIdFiltration),
                                   boost::make_transform_iterator(cycles.end(), cellIdFiltration));
  BOOST_CHECK(std::is_sorted(cycleColors.begin(), cycleColors.end()));
  ComputePersistentHomology::Diagram diagram = persistence.diagram();
  BOOST_CHECK(std::is_sorted(diagram.begin(), diagram.end(),
                                          ComputePersistentHomology::DiagramIntervalComparator()));
  BOOST_CHECK((boost::math::isinf)(diagram.back().second));
}


BOOST_AUTO_TEST_CASE(cubicalCircle)
{
  typedef capd::complex::CubCellComplex<capd::complex::FilteredDenseCubMemoryModel<2> > CubCellComplex;

  typedef capd::persistence::ComputePersistentHomology<CubCellComplex> ComputePersistentHomology;

  if (!ComputePersistentHomology::enabled()) {
    return;
  }

  std::vector<size_t> lengths = ba::list_of(1)(1);

  CubCellComplex complex(lengths);
  typedef CubCellComplex::CellCode CellCode;
  typedef CubCellComplex::CellCode::value_type CellCodeValue;

  {
    CellCode cellCode = ba::list_of<CellCodeValue>(0, true)(0, true);
    complex.fillWithBoundary(complex.insert(cellCode));
    complex.remove(cellCode);
  }

  ComputePersistentHomology persistence(complex);
  persistence();

  BOOST_CHECK_EQUAL(2, persistence.cycles().size());
  BOOST_CHECK_EQUAL(0, persistence.pairs().size());
}

BOOST_AUTO_TEST_SUITE_END()
