/////////////////////////////////////////////////////////////////////////////
/// @file PHATInterfaceTest.cpp
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-06-24
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2014 by the CAPD::RedHom Group.
//
// This file constitutes a part of the CAPD::RedHom library,
// distributed under the terms of the GNU General Public License.
// Consult  http://redhom.ii.edu.pl/ for details.

#include <capd/persistence/PHATInterface.h>
#include <capd/auxil/Logger.h>

#include <TestTools.h>

#define BOOST_TEST_DYN_LINK

#include "../apiRedHom/Fixtures.h"

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/mpl/list.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/assign.hpp>
#include <boost/foreach.hpp>

#include <map>
#include <vector>
#include <algorithm>
#include <utility>
#include <sstream>
#include <numeric>
#include <boost/foreach.hpp>

namespace bl = boost::lambda;
namespace b = boost;

using namespace boost::assign;
using namespace boost::lambda;

using namespace capd;
using namespace capd::persistence;


BOOST_AUTO_TEST_SUITE(PHATInterfaceTestSuite)

BOOST_AUTO_TEST_CASE(triangle)
{
  if (! PHATInterface::enabled()) {
    return;
  }

  PHATInterface phat;
  typedef PHATInterface::Column Column;
  std::vector<PHATInterface::FiltrationElement> filtration = pair_list_of
    (0, Column())(0, Column())(0, Column())
    (1, list_of(0)(1).operator Column())(1, list_of(0)(2).operator Column())(1, list_of(1)(2).operator Column())
    (2, list_of(3)(4)(5).operator Column());

  const PHATInterface::Pairs expectedPairs = pair_list_of(1, 3)(2, 4)(5, 6);
  PHATInterface::Pairs pairs;
  PHATInterface::Cycles cycles;

  phat(filtration, filtration.size(), pairs, cycles);

  BOOST_CHECK_EQUAL_COLLECTIONS(expectedPairs.begin(), expectedPairs.end(),
                                pairs.begin(), pairs.end());
  BOOST_REQUIRE_EQUAL(1, cycles.size());
  BOOST_CHECK_EQUAL(0, cycles[0]); // the one connected component
}


BOOST_AUTO_TEST_CASE(emptyTriangle)
{
  if (! PHATInterface::enabled()) {
    return;
  }

  PHATInterface phat;
  typedef PHATInterface::Column Column;
  std::vector<PHATInterface::FiltrationElement> filtration = pair_list_of
    (0, Column())(0, Column())(0, Column())
    (1, list_of(0)(1).operator Column())(1, list_of(0)(2).operator Column())(1, list_of(1)(2).operator Column());

  const PHATInterface::Pairs expectedPairs = pair_list_of(1, 3)(2, 4);
  PHATInterface::Pairs pairs;
  PHATInterface::Cycles cycles;

  phat(filtration, filtration.size(), pairs, cycles);

  BOOST_CHECK_EQUAL_COLLECTIONS(expectedPairs.begin(), expectedPairs.end(),
                                pairs.begin(), pairs.end());
  BOOST_REQUIRE_EQUAL(2, cycles.size());
  BOOST_CHECK_EQUAL(0, cycles[0]); // the one connected component
  BOOST_CHECK_EQUAL(5, cycles[1]); // the one connected component
}


namespace
{

  template<typename Complex>
  std::vector<PHATInterface::FiltrationElement> CreateFiltration(Complex& complex)
  {
    typedef PHATInterface::Column Column;
    const size_t cardinality = complex.cardinality();
    std::vector<PHATInterface::FiltrationElement> filtration;
    std::vector<size_t> idToPosition(cardinality);
    filtration.reserve(cardinality);

    for (size_t dim = 0; dim <= complex.getDim(); ++dim) {
      typedef typename Complex::Iterators::DimCells DimCells;
      typedef typename Complex::Iterators::BdCells BdCells;

      DimCells dimCells = complex.iterators().dimCells(dim);

      for(typename DimCells::iterator dimIt = dimCells.begin(), dimEnd = dimCells.end();
          dimIt != dimEnd; ++dimIt) {

        idToPosition[dimIt->getId()] = filtration.size();
        filtration.push_back(PHATInterface::FiltrationElement());

        filtration.back().first = dimIt->getDim();

        BdCells bdCells = complex.iterators().bdCells(*dimIt);
        for(typename BdCells::iterator bdIt = bdCells.begin(), bdEnd = bdCells.end();
            bdIt != bdEnd; ++bdIt) {

          filtration.back().second.push_back(idToPosition[bdIt->getId()]);
        }
      }

    }

    return filtration;
  }
}

BOOST_AUTO_TEST_CASE_TEMPLATE(cycles_all, Fixture, capd::test::ComplexFixtures)
{
  if (! PHATInterface::enabled()) {
    return;
  }

  return;
  // TODO MJ CreteFiltration gives random segfault on CubicalPoint fixture.
  // Only on some machines! Memory reated problem. Maybe wrong iterators in cubical setting?
  Fixture fixture;

  std::vector<PHATInterface::FiltrationElement> filtration = CreateFiltration(fixture.rawComplex());

  PHATInterface phat;
  PHATInterface::Pairs pairs;
  PHATInterface::Cycles cycles;

  phat(filtration, filtration.size(), pairs, cycles);
  size_t bettiSum = std::accumulate(fixture.betti.begin(), fixture.betti.end(), 0);

  BOOST_CHECK_EQUAL(bettiSum, cycles.size());
}

BOOST_AUTO_TEST_SUITE_END()
