/////////////////////////////////////////////////////////////////////////////
/// @file PointCloudSimplicialCellSComplexTest
///
/// @author Mateusz Juda <mateusz.juda@{ii.uj.edu.pl,gmail.com}>
///
/// @date 2014-06-27
/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2014 by the CAPD Group.
//
// This file constitutes a part of the CAPD library (capdRedHom),
// distributed under the terms of the GNU General Public License.
// Consult http://capd.ii.uj.edu.pl and  http://redhom.ii.edu.pl/ for details.
/////////////////////////////////////////////////////////////////////////////


#include <capd/persistence/PointCloudSimplicialCellSComplex.h>
#include <capd/auxil/Logger.h>

#include <TestTools.h>
#include "PointCloudFixtures.h"
#include <capd/persistence/PointCloudUtils.h>

#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/mpl/list.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/assign.hpp>
#include <boost/foreach.hpp>

#include <map>
#include <vector>
#include <algorithm>
#include <utility>
#include <sstream>
#include <capd/persistence/PointCloudTypes.h>
#include <capd/auxil/Logger.h>
#include <boost/scoped_ptr.hpp>
#include <capd/persistence/DistanceMatrixPointCloud.h>
#include <capd/vectalg/Matrix.hpp>
#include <capd/persistence/DegreeLimitedNeighbourhood.h>




namespace bl = boost::lambda;
namespace b = boost;
namespace ba = boost::assign;

using namespace capd;
using namespace capd::persistence;

BOOST_AUTO_TEST_SUITE(PointCloudSimplicialCellSComplexTestSuite)

BOOST_AUTO_TEST_CASE(EuclideanPointCloud)
{
  typedef EuclideanPointCloudType<0>::PointCloudType PointCloudType;
  typedef PointCloudType::Point Point;
  typedef PointCloudType::Scalar Scalar;

  std::vector<Point> points = ba::list_of
    (Point((Scalar[]){0, 0}))
    (Point((Scalar[]){10, 10}))
    (Point((Scalar[]){0, 1}))
    (Point((Scalar[]){2, 0}));

  {
    boost::scoped_ptr<PointCloudType> pc(makeSizeLimitedPointCloud<EuclideanPointCloudType<0> >(points.begin(), points.end(), 3));
    PointCloudSimplicialCellSComplex complex(*pc, 0);
    complex();

    BOOST_CHECK_EQUAL(8, complex.size());
  }

  {
    boost::scoped_ptr<PointCloudType> pc(makeDegreeLimitedPointCloud<EuclideanPointCloudType<0> >(points.begin(), points.end(), 0));
    PointCloudSimplicialCellSComplex complex(*pc, 0);
    complex();

    BOOST_CHECK_EQUAL(15, complex.size());
  }

  {
    boost::scoped_ptr<PointCloudType> pc(makeSizeLimitedPointCloud<EuclideanPointCloudType<0> >(points.begin(), points.end(), 0));
    PointCloudSimplicialCellSComplex complex(*pc, 0);
    complex();

    BOOST_CHECK_EQUAL(15, complex.size());
  }

}


BOOST_AUTO_TEST_CASE(DistanceMatrix)
{
  typedef DistanceMatrixPointCloud PointCloudType;
  double d[3][3] = { {0, 1, 1}, {1, 0, 1}, {1, 1, 0} };
  PointCloudType::Metric::DistanceMatrix distance(d);


  PointCloudType pc(distance);
  DegreeLimitedNeighbourhood<PointCloudType> neighbourhood(pc, 0, 0);
  neighbourhood();
  BOOST_CHECK_EQUAL(3, pc.edges());
  BOOST_CHECK_EQUAL(3, pc.size());

  PointCloudSimplicialCellSComplex complex(pc, 0);

  complex();

  BOOST_CHECK_EQUAL(2, complex.dim());
  BOOST_CHECK_EQUAL(7, complex.size());
}


BOOST_AUTO_TEST_CASE_TEMPLATE(all_euclidean, Fixture, capd::persistence::test::PointCloudFixtures)
{
  typedef EuclideanPointCloudType<0>::PointCloudType PointCloudType;
  Fixture fixture;
  const typename Fixture::Points& points = fixture.points();
  boost::scoped_ptr<PointCloudType> pc(makeSizeLimitedPointCloud<EuclideanPointCloudType<0> >(points.begin(), points.end(), 0));

  typedef PointCloudSimplicialCellSComplex PCComplex;
  PCComplex pccomplex(*pc, fixture.maxDim());

  pccomplex();

  std::vector<size_t> count(fixture.maxDim() + 1);

  for (PCComplex::Complex::Id idx = 0, size = pccomplex.size(); idx < size; ++idx) {
    count[pccomplex[idx].dim()]++;
  }

  BOOST_CHECK_EQUAL(fixture.size(), count[0]);
  BOOST_CHECK_EQUAL(fixture.edges(), count[1]);

  for (PCComplex::Complex::Id idx = 0, size = pccomplex.size(); idx < size; ++idx) {
    typedef PCComplex::Iterators::BdCells Boundary;
    Boundary cellBd = pccomplex.iterators().bdCells(pccomplex[idx]);

    if (pccomplex[idx].dim() == 0) {
      BOOST_CHECK_EQUAL(0,  pccomplex[idx].color());
      BOOST_CHECK_EQUAL(0, cellBd.size());
    } else {
      BOOST_CHECK_EQUAL(pccomplex[idx].dim() + 1, cellBd.size());
    }

    BOOST_FOREACH(Boundary::value_type bd, cellBd) {
      BOOST_CHECK(pccomplex[bd.id()].color() <= pccomplex[idx].color());

      if (pccomplex[idx].dim() == 1) {
        BOOST_CHECK(pccomplex[idx].color() > 0);
      }

    }
  }
}

BOOST_AUTO_TEST_SUITE_END()
