struct ExtendedProofData : public ProofData{
  ExtendedProofData(RosslerPoincareMap& PM)
    : ProofData(PM)
  {}
  
  void writeArray(DMatrix A, std::ostream& out);
  void writeVector(DVector A, std::ostream& out);
  void writeColumnVector(DVector A, std::ostream& out);
  void writeDiagMatrix(DVector A, std::ostream& out);

  void exportSet(HSet& h, std::ostream& out);
  void exportChainOfCoveringRelations(HSet& first, HSet& last, HSetChain& chain, const char* filename);
  void exportAllChainsOfCoveringRelations();

  void exportCoordinateSystemsAtPeriodicPoints();
  void exportDiametersAtPeriodicPoints();


  void exportAllChainCenters();
  void exportChainCenters(HSet& first, HSet& last, HSetChain& chain, const char* filename);
  
};

// ##############################################################

void ExtendedProofData::writeArray(DMatrix A, std::ostream& out){
  out << "\\left[\\begin{smallmatrix}\n";
  for(int i=1;i<=3;++i){
    out << "" << A(i,1) 
        << "&" << A(i,2) 
        << "&" << A(i,3) << "\\\\\n";
  }
  out << "\\end{smallmatrix}\\right]";
}

// ##############################################################

void ExtendedProofData::writeVector(DVector r, std::ostream& out){
  out << "\\left(" << r[0]  << "," << r[1] << "," << r[2] << "\\right)";
}

// ##############################################################

void ExtendedProofData::writeColumnVector(DVector r, std::ostream& out){
  out << "\\left[\\begin{smallmatrix}"
      << r[0] 
      << "\\\\" << r[1] 
      << "\\\\" << r[2]
      << "\\end{smallmatrix}\\right]";
}

// ##############################################################

void ExtendedProofData::writeDiagMatrix(DVector r, std::ostream& out){
  out << "\\mathrm{Diag}\\left(\\begin{smallmatrix}\n";
  out << "" << r[0] 
      << ",&" << r[1]
      << ",&" << r[2] << "\n";
  out << "\\end{smallmatrix}\\right)\\\\\n";
}

// ##############################################################

void ExtendedProofData::exportChainCenters(HSet& first, HSet& last, HSetChain& chain, const char* filename){
  std::ofstream out;
  out.open(filename);
  out.precision(17);
  out << first._x[0] << " " << first._x[1] << " "  << first._x[2] << std::endl;
  for(unsigned i=0;i<chain.size();++i){
    out << chain[i]->_x[0] << " " << chain[i]->_x[1] << " "  << chain[i]->_x[2] << std::endl;
  }
  out << last._x[0] << " " << last._x[1] << " "  << last._x[2] << std::endl;
  out.close();
}

// ##############################################################

void ExtendedProofData::exportAllChainCenters(){
  exportChainCenters(P1,P2,chainP1P2,"out/P1P2centers.dat");
  exportChainCenters(P2,P1,chainP2P1,"out/P2P1centers.dat");
  exportChainCenters(P1,P3,chainP1P3,"out/P1P3centers.dat");
  exportChainCenters(P3,P4,chainP3P4,"out/P3P4centers.dat");
  exportChainCenters(P4,P3,chainP4P3,"out/P4P3centers.dat");

  exportChainCenters(Q2,Q1,chainQ2Q1,"out/Q2Q1centers.dat");
  exportChainCenters(Q1,Q2,chainQ1Q2,"out/Q1Q2centers.dat");
  exportChainCenters(Q1,P4,chainQ1P4,"out/Q1P4centers.dat");
  exportChainCenters(Q1,P1,chainQ1P1,"out/Q1P1centers.dat");
}

// ##############################################################

void ExtendedProofData::exportCoordinateSystemsAtPeriodicPoints(){
  std::cout << "A_1^8&=&"; writeArray(P1._C,std::cout);
  std::cout << "A_2^8&=&"; writeArray(P2._C,std::cout);
  std::cout << "A_3^8&=&"; writeArray(P3._C,std::cout);
  std::cout << "A_4^{16}&=&"; writeArray(P4._C,std::cout);
  std::cout << "A_1^{12}&=&"; writeArray(Q1._C,std::cout);
  std::cout << "A_2^{14}&=&"; writeArray(Q2._C,std::cout);
}

// ##############################################################

void ExtendedProofData::exportDiametersAtPeriodicPoints(){
  std::cout << "D_1^8&=&"; writeDiagMatrix(P1._r,std::cout);
  std::cout << "D_2^8&=&"; writeDiagMatrix(P2._r,std::cout);
  std::cout << "D_3^8&=&"; writeDiagMatrix(P3._r,std::cout);
  std::cout << "D_4^{16}&=&"; writeDiagMatrix(P4._r,std::cout);
  std::cout << "D_1^{12}&=&"; writeDiagMatrix(Q1._r,std::cout);
  std::cout << "D_2^{14}&=&"; writeDiagMatrix(Q2._r,std::cout);
}

void ExtendedProofData::exportSet(HSet& h, std::ostream& out){
  out << h.N << "&";
  out.precision(17);
  writeColumnVector(h._x,out);
  out << "&";
  writeArray(h._C,out); 
  out << "&";
  out.precision(10);
  writeColumnVector(h._r,out);
  out << "\\\\\n";
}

void ExtendedProofData::exportChainOfCoveringRelations(HSet& first, HSet& last, HSetChain& chain, const char* filename)
{
  std::ofstream out;
  out.open(filename);
  out << "\\begin{array}{c|c|c|c} \\hline\\mathrm{p}& x & C & r\\\\\\hline";
//  exportSet(first,out);
//  out << "\\hline\n";
  for(auto i=chain.begin();i!=chain.end();++i){
    exportSet(**i,out);
    out << "\\hline\n";
  }
//  exportSet(last,out);
  out << "\\end{array}\n";
  out.close();
}

void ExtendedProofData::exportAllChainsOfCoveringRelations()
{
  exportChainOfCoveringRelations(P1,P2,chainP1P2,"supplementary/chainP1P2.tex");
  exportChainOfCoveringRelations(P2,P1,chainP2P1,"supplementary/chainP2P1.tex");
  exportChainOfCoveringRelations(P3,P4,chainP3P4,"supplementary/chainP3P4.tex");
  exportChainOfCoveringRelations(P4,P3,chainP4P3,"supplementary/chainP4P3.tex");

  exportChainOfCoveringRelations(P1,P3,chainP3P4,"supplementary/chainP1P3.tex");

  exportChainOfCoveringRelations(Q1,Q2,chainQ1Q2,"supplementary/chainQ1Q2.tex");
  exportChainOfCoveringRelations(Q2,Q1,chainQ2Q1,"supplementary/chainQ2Q1.tex");

  exportChainOfCoveringRelations(Q1,P1,chainQ1P1,"supplementary/chainQ1P1.tex");
  exportChainOfCoveringRelations(Q1,P4,chainQ1P4  ,"supplementary/chainQ1P4.tex");  
}

