/**
 * This file provide a data structure and algorithms for computation of forward invariant sets for maps.
 * Forward invariant set is represented as a union of boxes of fixed size.
 * The main algorithm takes an initial box and computes its image as a union of boxes. 
 * Then it repeats the same procedure for each new box added to the set. If there are no new boxes then the algorithms stops 
 * and the set contains an enclosure (union of boxes) of a forward invariant set.
 * 
 * Warning: The algortithm is designed for attractors, only! 
*/

#ifndef DANIEL_WILCZAK_INV_SET_ENCLOSURE_H
#define DANIEL_WILCZAK_INV_SET_ENCLOSURE_H

#include <set>
#include <iostream>
#include <fstream>
#include <stdexcept>
#include <omp.h>
#include "capd/capdlib.h"

typedef capd::ZVector Point;

/// Linear (lex) order for vectors Z^n used to encode boxes of fixed diameters in R^n.
/// This order is needed by the standard data structure 'set' from which the class InvSetEnclosure inherites.
template<class PointT>
struct less{

  bool operator() (const PointT& x, const PointT& y) const
  {
    for(unsigned i=0;i<x.dimension();++i)
    {
      if(x[i]>y[i])
        return false;
      if(x[i]<y[i])
        return true;
    }
    return false;
  }
};

typedef std::set<Point,less<Point> > BS;

/**
 * This is the main class implemented in this file. 
 * The template parameter stands for interval vector type used in rigorous computations.
*/
template< class VectorT >
class InvSetEnclosure : public BS
{
public:

  typedef Point Vertex;
  typedef BS BaseSet;
  typedef typename VectorT::ScalarType ScalarType;

  /// use iterators from std::set
  using BaseSet::iterator;
  using BaseSet::const_iterator;
  using BaseSet::begin;
  using BaseSet::end;
  using BaseSet::insert;
  using BaseSet::find;
  
  /// rebind an interval vector to double vector
  typedef typename VectorT::template rebind<double>::other DVector;

  /// The constructor requires resolution, i.e. a vector of integers. 
  /// The coefficients define size of boxes used to cover an invariant set. 
  /// The size of box is called denominator and it ir related to resolution by the formula
  /// denominator[i] = 2^resolution[i]
  InvSetEnclosure(const Vertex& _resolution)
    : resolution(_resolution), denominator(resolutionToDenominator(_resolution))
  {}

  const Vertex& getResolution() const   { return resolution;  }
  const Vertex& getDenominator() const  { return denominator; }

  /// resolution to denominator conversion
  /// denominator[i] = 2^resolution[i]
  static Vertex resolutionToDenominator(const Vertex& resolution)
  { 
    Vertex result(resolution.dimension());
    for(unsigned i=0;i<resolution.dimension();++i)
      result[i] = (1 << int(resolution[i]));
    return result;
  }
  
  /// Static member function that converts a vector of integers (node in the graph) to an interval vector. 
  /// Node in the graph is a vector of integers which encodes left corner of a box in R^n.
  /// Given size of box (second argument denominator) the function computes and returns interval vector that represents this box.
  static VectorT vertexToVector(const Vertex& vertex, const Vertex& denominator)
  {
    VectorT result(vertex.dimension());
    for(unsigned i=0;i<vertex.dimension();++i)
      result[i] = ScalarType(int(vertex[i]),int(vertex[i])+1)/int(denominator[i]);
    return result;
  }

  /// This algorithm covers the domain (interval vector) by set of boxes and inserts these boxes to the set of boxes 'G'.
  void computeCover(BaseSet& G, const VectorT& domain)
  {
    int dim = domain.dimension();
    Vertex left(dim), right(dim);
    int numberOfElements = 1;
    int i;
    capd::rounding::DoubleRounding::roundNearest();

    for(i=0;i<dim;++i)
    {
      double L = domain[i].leftBound();
      double R = domain[i].rightBound();
      left[i] = static_cast<int>(floor(L*int(denominator[i])));
      right[i] = static_cast<int>(ceil(R*int(denominator[i])));
      // these numbers can be the same! So we must assure that at least one element is added
      if(left[i]==right[i])
        right[i] = right[i]+1;
      numberOfElements *= right[i]-left[i];
    }

    Vertex coeffs = left;
    for(i=0;i<numberOfElements;++i)
    {
      if(G.find(coeffs)== G.end())
        G.insert(coeffs);
      int k = dim;
      do
      {
        --k;
        coeffs[k] = int(coeffs[k])+1;
        if(int(coeffs[k])==int(right[k]))
          coeffs[k] = int(left[k]);
        else
          k=0;
      }while(k);
    }
  }

  /**
   * This algorithm realizes one step of the main algorithm that computes forward invariant set.
   * Given 
   * - a set of boxes 'vec'
   * - a map 'f'
   * it computes set of boxes that covers f(union(vec)) and inserts it to the set 'uncheckedBoxes'.
   * The algorithm is parallelized by the OpenMp library.
  */ 
  template<class MapT>
  void computeImageOfBoxes(MapT&f, std::vector<Vertex>& vec, BaseSet& uncheckedBoxes)
  {
    int counter,i;
    int unSize = vec.size();
    int threads=1;
    while(threads<omp_get_max_threads()) threads*=2;
    
    // temporary sets for each thread. Synchronization is to slow!
    BaseSet* newBoxes = new BaseSet[threads];
    
    #pragma omp parallel for schedule(dynamic)
    for(counter=0;counter<unSize;++counter)
    {
      // take an unchecked box from 'vec' and convert it to interval vector 'x'
      VectorT x = vertexToVector(vec[counter],denominator);

      try{
        // here we evaluate the abstract map 'f'
        x = f(x);
      }catch(std::exception& e)
      {
         std::cout << "InvSetEnclosure::computeImageOfBoxes\nFatal and unexpected error occured when compute value of map!!!" << std::endl;
         std::cout << e.what() << std::endl;
         std::cout << "x=" << x << std::endl;
         exit(0);
      }
      
      // After successfull evaluation of the map at 'x'
      // the enclosure of f(x) is assumed to be in 'enc'.
      // We compute its cover and add the result to the set specific for this thread.
      computeCover(newBoxes[omp_get_thread_num()],x);
    }

    // This step removes from each set vertices that are already in the main data structure and which has been already checked. 
    // Again the algorithm is parallel. This step is very important because unnecessary inserting to a huge data structure extremely are very slow!
    #pragma omp parallel for
    for(i=0;i<omp_get_max_threads();++i)
    {
      for(auto b = newBoxes[i].begin();b!=newBoxes[i].end();)
      {
        auto p = b;
        ++b;
        if( this->find(*p) != this->end() )
          newBoxes[i].erase(p);
      }
    }
    
    // After removing already checked boxes we merge the remaining.
    // This is a tricky parallel reduction. After it finishes, 
    // the result of merge is written to newBoxes[0]
    int k = threads/2;
    while(k){
      #pragma omp parallel for
      for(i=0;i<k;++i)
      {
        for(auto it=newBoxes[i+k].begin();it!=newBoxes[i+k].end();++it){
          if(newBoxes[i].find(*it)==newBoxes[i].end())
            newBoxes[i].insert(*it);
        }
        newBoxes[i+k].clear();
      }
      k/=2;
    }

    std::cout << "New boxes: " << newBoxes[0].size() << std::endl;

    // copying of data from temporary newBoxes[0] to 'uncheckedBoxes'
    this->insert(newBoxes[0].begin(),newBoxes[0].end());
    uncheckedBoxes.insert(newBoxes[0].begin(),newBoxes[0].end());
    delete[] newBoxes;
  }

  /**
   * This is the main loop of the algorithm that encloses forward invariant set.
   * Given an initial point 'initVector' it encloses its forward trajectory.
   * The algorithm stops when there are no boxes for which we did not compute the value of the map.
  */
  template<class MapT>
  void encloseForwardInvariantSet(MapT& f, VectorT initVector)
  {
    // clear the set (no boxes in the cover)
    this->clear();
    
    // cover the initVector by the boxes
    computeCover(*this,initVector);

    // this is for performance reasons, only. Operations on large data structrures might be slowand therefore we limit them as much as possible.
    // The chunk size can be arbitrarty positive number and it stands for maximal number of boxes processed in parallel by the previous algorithm 'computeImageOfBoxes'.
    // This particular value comes from experiments as well working.
    const int MAX_CHUNK_SIZE = 32*1024;
    BaseSet uncheckedBoxes = static_cast<BaseSet&>(*this);
    std::vector<Vertex> vec;
    vec.reserve(MAX_CHUNK_SIZE);

    do
    {
      int i=0;
      // here we take a chunk of boxes
      vec.clear();
      auto b = uncheckedBoxes.begin();
      while(b!=uncheckedBoxes.end() and i<MAX_CHUNK_SIZE){
        vec.push_back(std::move(*b));
        ++i;
        ++b;
      }
      uncheckedBoxes.erase(uncheckedBoxes.begin(),b);
      
      // and process it 
      this->computeImageOfBoxes(f,vec,uncheckedBoxes);
      
      std::cout << "====> Unchecked boxes: " << uncheckedBoxes.size() << std::endl;
      std::cout << "====> Current size of cover: " << this->size() << std::endl;
    }while(uncheckedBoxes.size()!=0);
  }

  /**
   * This function stores set of boxes to a file
  */
  void save(const char* fileName) const
  {
    std::ofstream f;
    f.open(fileName);
    f << resolution;
    typename BaseSet::const_iterator b=this->begin(), e=this->end();
    while(b!=e)
    {
      f << std::endl << *b;
      ++b;
    }
    f.close();
  }

  /**
   * This function reads set of boxes from a file. 
   * The second argument says if we have to clear the set before reading.
  */
  void load(const char* fileName, bool clearGraph=true)
  {
    if(clearGraph)
      this->clear();
    std::ifstream f;
    f.open(fileName);
    Vertex temp;
    f >> resolution;
    denominator = resolutionToDenominator(resolution);

    do
    {
      f >> temp;
      this->insert(temp);
      f.get();
    }while(!f.eof() and f.peek());
    f.close();
  }

private:
  Vertex resolution;
  Vertex denominator;
};

#endif
