/**
 * This file contains algorithms used to prove the existence of a periodic orbit. 
 * Moreover, its stability type is computed (eigenvalues of derivaterive of Poincare map at periodic point).
*/ 

#ifndef DANIEL_WILCZAK_MULTIPLE_SHOOTING
#define DANIEL_WILCZAK_MULTIPLE_SHOOTING

#include <complex>
#include "capd/capdlib.h"
#include "eigenvalues.h"
#include "RosslerPoincareMap.h"

/**
 * We will use two algorithms (interval Newton and interval Krawczyk operator)
 * for validation that a periodic orbit exists. They differ by the way they compute corrector term.
 * 
 * Thus, for having the same interface, we add extra (unused) arguments to interval Newton corrector.
*/ 
// ##################################################################

capd::IVector NewtonCorrector(
  const capd::IVector& imCenter, 
  const capd::IMatrix& D, 
  const capd::IVector& /*Y*/,
  const capd::IVector& /*center*/) 
{
  return capd::matrixAlgorithms::krawczykInverse(D) * imCenter;
}

// ##################################################################

capd::IVector KrawczykCorrector(
  const capd::IVector& imCenter, 
  const capd::IMatrix& D, 
  const capd::IVector& Y,
  const capd::IVector& center) 
{
  using namespace capd;
  DMatrix D1 = vectalg::convertObject<DMatrix>(D);
  DMatrix C1 = matrixAlgorithms::gaussInverseMatrix(D1);
  IMatrix C = vectalg::convertObject<IMatrix>(C1);
  // check if C is invertible
  matrixAlgorithms::gaussInverseMatrix(C);
  
  IMatrix K = C*D;
  for(IMatrix::size_type i=1;i<=D.numberOfRows();++i)
    K(i,i) -= 1.;
  return C* imCenter + K*(Y-center);
}

// ##################################################################

/**
  * The following function computes the Interval Newton/Krawczyk Operator for the map
  *
  * F: (x0,x1,x2,...x_{n-1}) -> (x0-P(x_{n-1}),x1-P(x0), .... , x_{n-1} - P(x_{n-2}))
  *
  * or the Interval Krawczyk Operator
  *   K = x - C*F(x) + (Id - C*dF(X))(X-x0)
  * It verifies the existence and uniqueness of a periodic point in a given set X.
  *
  * Parameters are:
  * @param pm - an instance of PoincareMap for the Rossler system
  * @param X[] - array of intervals that contains our candidate for periodic orbits.
  * @param N - total dimension, i.e. period*dim
  *
  * @returns result - true if the orbit has been validated
*/
template<class Method, class Map>
void MultipleShooting(Map& f, Method correctorMethod, capd::interval X[], int N, int dim, int order1, int order2, bool& result)
{
  try{
    using namespace capd;
    IVector imCenter(N), Y(N,X);
    IMatrix D = IMatrix::Identity(N);
    IVector center = midVector(Y);
    
    for(int i=0;i<N;i+=dim){
      int prev = (N+i-dim)%N;
      f.setOrder(order1);
      IVector x = f(IVector(dim,center.begin()+prev));
      for(int j=0;j<dim;++j) imCenter[i+j] = center[i+j] - x[j];
   
      // Then we compute derivative at the set X
      f.setOrder(order2);
      IMatrix DP = f.derivative(IVector(dim,X+prev));
      for(int j=0;j<dim;++j)
        for(int c=0;c<dim;++c)
          D[i+j][prev+c] -= DP[j][c];
    }
    
    // compute corrector of interval Newton/Krawczyk operator
    IVector corrector = correctorMethod(imCenter,D,Y,center);
    IVector Operator = center - corrector;
    result = result or subsetInterior(Operator, Y);
    if(result){
      capd::vectalg::intersection(Operator.begin(),Y.begin(),X,X+N);
    }
  }catch(std::exception& e){
  }
}

// ##############################################################
/**
  * This function validates (if possible) the existence of a periodic orbit near an approximate point 'x'.
  * This is done by subsequent calls to interval Newton and interval Krawczyk algorithms.
  * 
  * Eventually, of the orbit is validated, the function prints to the standard output 
  * spectrum of derivative of Poincare map at the periodic point.
  *   
  * Parameters are:
  * @param PM - an instance of PoincareMap for the Rossler system
  * @param x - candidate for a periodic orbit.
  * @param period - period of this orbit
  * @param N - total dimension, i.e. period*dim
  *
  * @returns - true if the orbit has been validated
 */ 
bool provePeriodicOrbit(RosslerPoincareMap& PM, capd::DVector x, int period, int dim, int order1=35, int order2=25, double tol=1e-8){
  using namespace capd;
  std::cout << "\n######################################\n";
  DVector p = x;
  DMatrix Q =  DMatrix::Identity(3);
  DMatrix temp(3,3);
  int N = period*dim;
  interval X[N];
  interval s =  tol*interval(-1,1);
  for(int i=0;i<N;i+=dim){
    for(int j=0;j<dim;++j)
      X[i+j] = x[j]+s;
    x = PM(x,temp);
    Q = temp*Q;
  }
  DVector rV(dim), iV(dim);
  DMatrix rVec(dim,dim), iVec(dim,dim);
  alglib::computeEigenvaluesAndEigenvectors(Q,rV,iV,rVec,iVec);  

  bool result = false;
  MultipleShooting(PM,KrawczykCorrector,X,N,dim,order1,order2,result);
  MultipleShooting(PM,NewtonCorrector,X,N,dim,order1,order2,result); 
  MultipleShooting(PM,KrawczykCorrector,X,N,dim,order1,order2,result);
  MultipleShooting(PM,NewtonCorrector,X,N,dim,order1,order2,result); 
  MultipleShooting(PM,KrawczykCorrector,X,N,dim,order1,order2,result);
  MultipleShooting(PM,NewtonCorrector,X,N,dim,order1,order2,result); 
  MultipleShooting(PM,KrawczykCorrector,X,N,dim,order1,order2,result);
  MultipleShooting(PM,NewtonCorrector,X,N,dim,order1,order2,result); 
  IMaxNorm norm;
  capd::interval error = norm(IVector(dim,X) - IVector(p));
  std::cout << "Radius=" << error << std::endl;
  PM.setOrder(order2);
  if(result){
    // check the inclusion and the estimation on periodic orbit
    interval returnTime, rt;
    if(iV[0]!=0.) 
      rVec.column(1) = iVec.column(1);
    IMatrix init =  vectalg::convertObject<IMatrix>(rVec);
    IMatrix D = PM.derivative(IVector(dim,X),period,returnTime,init);
    D = matrixAlgorithms::krawczykInverse(init)*D;
    
    auto lambda = computeComplexEigenvalues(D);
    std::cout << "eigenvalues:\n"
      << " & " << lambda[0] 
      << " & " << lambda[1]   
      << " & " << lambda[2]  
      << " & " << returnTime << "\\\\ \\hline \n"; 
  }
  return result;
}

#endif
