/**
 * This file contains nonrigorous routines that prepare data for rigorous computations.
 * It also contains routines that call the main algorithms 
 * - enclosing invariant set
 * - validation of covering relations
 * - validation of cone condition
 * with the initial data from this object.
*/ 

#ifndef DANIEL_WILCZAK_PROOF_DATA_H
#define DANIEL_WILCZAK_PROOF_DATA_H

// ##############################################################
/**
 * This is an auxiliary routine. 
 * @param p - an approximate periodic point
 * @param period - period of this point
 * @param s1, s2, s3 positive reals
 * 
 * @returns in constructs an h-set centered at p with the coordinate system given by eigenvectors of DP^{period}(p) and with size (s1,s2,s3)
*/ 
HSet getHSetAtFixedPoint(RosslerPoincareMap& PM, DVector p, int period, double s1, double s2, double s3=1e-7){
  DMatrix ev(3,3);
  computeCoordSystem(PM,p,period,ev);
  return HSet(p,ev,DVector({s1,s2,s3}),period);
}

// ##############################################################
/**
 * This routine construct a sequence of h-sets linking h-sets centered at P1 and P2, repsectively. 
 * Data from hand-made simulation.
*/ 
HSetChain createP1P2HeteroclinicChain(RosslerPoincareMap& PM, HSet h, HSet last, int length, double t){
  HSetChain result(length+1);
  // take an approiximate heteroclinic point
  DVector p = h._x + h._C.column(0)*t;
  DMatrix D(3,3), temp(3,3), C(3,3);
  
  //diameters of the sets along heteroclinic chain
  DVector r[] = { 
    DVector({7e-4,4e-4,1e-7}), // 0
    DVector({1e-3,2e-4,1e-7}), // 1
    DVector({1e-3,8e-5,1e-7}), // 2
    DVector({1e-3,8e-5,1e-7}), // 3
    DVector({1e-3,1e-4,1e-7}), // 4
    DVector({1e-3,8e-5,1e-7}), // 5
    DVector({2e-3,8e-5,1e-7}), // 6
    DVector({2e-3,8e-5,1e-7}), // 7
    DVector({2e-3,8e-5,1e-7})  // 8
  };
  
  // some heurestic routine that computes coordinate systems along this sequence
  result[0] = new HSet(p,h._C,r[0]);
  for(int i=1;i<=length;++i){
    D = result[i-1]->_C;
    for(int k=0;k<8;++k) { p = PM(p,temp); D = temp*D; }
    if(i<=4)
      C = h._C;
    else 
      C = last._C;
    C.column(0) = D.column(0);
    normalizeColumns(C);
    // insert new h-set into the sequence
    result[i] = new HSet(p,C,r[i]);
  }
  return result;
}

// ##############################################################
/**
 * This routine construct a sequence of h-sets linking h-sets centered at P2 and P1, repsectively. 
 * Data from hand-made simulation.
*/ 

HSetChain createP2P1HeteroclinicChain(RosslerPoincareMap& PM, HSet h, HSet last, int length, double t){
  HSetChain result(length+1);
  DVector p = h._x + h._C.column(0)*t;
  DMatrix D(3,3), temp(3,3), C(3,3);
  DVector r[] = { 
    DVector({1e-3,4e-4,1e-7}),   // 0
    DVector({3e-4,3e-4,1e-7}),   // 1
    DVector({3e-4,2.5e-4,1e-7}), // 2
    DVector({7e-4,4e-4,1e-7}),   // 3
    DVector({2e-3,1.5e-3,1e-7}), // 4
    DVector({7e-4,3e-4,1e-7}),   // 5
    DVector({1e-3,1e-4,1e-7})    // 6
  };
  result[0] = new HSet(p,h._C,r[0]);
  for(int i=1;i<=length;++i){
    D = result[i-1]->_C;
    for(int k=0;k<8;++k) { p = PM(p,temp); D = temp*D; }
    if(i<=4){
      C = h._C;
      C.column(1) = D.column(1);
    }
    else 
      C = last._C;
    C.column(0) = D.column(0);
    normalizeColumns(C);
    result[i] = new HSet(p,C,r[i]);
  }
  return result;
}

// ##############################################################

/**
 * This routine construct a sequence of h-sets linking h-sets centered at P1 and P3, repsectively. 
 * Data from hand-made simulation.
*/ 
HSetChain createP1P3HeteroclinicChain(RosslerPoincareMap& PM, HSet h, HSet last, int length, double t){
  HSetChain result(length+1);
  DVector p = h._x + h._C.column(0)*t;
  DMatrix D(3,3), temp(3,3), C(3,3);
  DVector r[] = { 
    DVector({7e-4,1e-4,1e-7}), // 0
    DVector({1e-3,4e-5,1e-7}), // 1
    DVector({1e-3,4e-5,1e-7}), // 2
    DVector({1e-3,4e-5,1e-7}), // 3
    DVector({1e-3,4e-5,1e-7}), // 4
    DVector({5e-4,6e-5,1e-7}), // 5
    DVector({4e-4,1e-4,1e-7}), // 6
    DVector({7e-4,2e-4,1e-7}), // 7
    DVector({7e-4,3e-4,1e-7})
  };
  result[0] = new HSet(p,h._C,r[0]);
  for(int i=1;i<=length;++i){
    D = result[i-1]->_C;
    for(int k=0;k<8;++k) { p = PM(p,temp); D = temp*D; }
    if(i<=6)
      C = h._C;
    else 
      C = last._C;
    C.column(0) = D.column(0);
    normalizeColumns(C);
    result[i] = new HSet(p,C,r[i]);
  }
  return result;
}

// ##############################################################

/**
 * This routine construct a sequence of h-sets linking h-sets centered at P3 and P4, repsectively. 
 * Data from hand-made simulation.
*/ 
HSetChain createP3P4HeteroclinicChain(RosslerPoincareMap& PM, HSet h, HSet last, int length, double t){
  HSetChain result(length+1);
  DVector p = h._x + h._C.column(0)*t;
  DMatrix D(3,3), temp(3,3), C(3,3);
  DVector r[] = { 
    DVector({1.e-4,3e-4,1e-7}),   // 0
    DVector({1.5e-4,1.5e-4,1e-7}),// 1
    DVector({2e-4,8e-5,1e-7}),    // 2
    DVector({3e-4,5e-5,1e-7}),    // 3
    DVector({1e-4,4e-5,1e-7}),    // 4
    DVector({4e-4,4e-5,1e-7}),    // 5
    DVector({5e-5,4e-5,1e-7}),    // 6
    DVector({6e-4,4e-5,1e-7}),    // 7
    DVector({2e-4,4e-5,1e-7}),    // 8
    DVector({6e-4,4e-5,1e-7}),    // 9
    DVector({2e-4,4e-5,1e-7})     // 10
  };
  result[0] = new HSet(p,h._C,r[0]);
  int i;
  for(i=1;i<=length;++i){
    D = result[i-1]->_C;
    for(int k=0;k<8;++k) { p = PM(p,temp); D = temp*D; }
    if(i<=5)
      C = h._C;
    else 
      C = last._C;
    C.column(0) = D.column(0);
    normalizeColumns(C);
    result[i] = new HSet(p,C,r[i]);
  }
  return result;
}

// ##############################################################

/**
 * This routine construct a sequence of h-sets linking h-sets centered at P4 and P3, repsectively. 
 * Data from hand-made simulation.
*/ 
HSetChain createP4P3HeteroclinicChain(RosslerPoincareMap& PM, HSet h, HSet last, int length, double t){
  HSetChain result(length+1);
  DVector p = h._x + h._C.column(0)*t;
  DMatrix D(3,3), temp(3,3), C(3,3);
  DVector r[] = { 
    DVector({7e-4,3e-4,1e-7}),  // 0
    DVector({2e-4,2e-4,1e-7}),  // 1
    DVector({7e-4,1e-4,1e-7}),  // 2
    DVector({2e-4,5e-5,1e-7}),  // 3
    DVector({1e-3,1e-4,1e-7}),  // 4
    DVector({2e-4,5e-5,1e-7}),  // 5
    DVector({1e-3,1e-4,1e-7}),  // 6
    DVector({4e-4,1.5e-4,1e-7}),// 7
    DVector({2e-4,1e-4,1e-7}),  // 8
    DVector({2e-4,1e-4,1e-7}),  // 9
    DVector({1e-4,1e-4,1e-7})   // 10
  };
  result[0] = new HSet(p,h._C,r[0]);
  int i;
  for(i=1;i<=length;++i){
    D = result[i-1]->_C;
    for(int k=0;k<8;++k) { p = PM(p,temp); D = temp*D; }
    if(i<=9)
      C = h._C;
    else 
      C = last._C;
    C.column(0) = D.column(0);
    normalizeColumns(C);
    result[i] = new HSet(p,C,r[i]);
  }
  return result;
}


// ##############################################################

/**
 * This routine construct a sequence of h-sets linking h-sets centered at Q1 and Q2, repsectively. 
 * Data from hand-made simulation.
*/ 
HSetChain createQ1Q2HeteroclinicChain(RosslerPoincareMap& PM, HSet& h, HSet& last, int length, DVector t){
  HSetChain result(length+1);
  DVector p = h._x + h._C*t;
  DMatrix D(3,3), temp(3,3), C(3,3);
  DVector r[] = { 
    DVector({1e-5,5e-5,1e-6}), // 0
    DVector({1e-5,5e-4,1e-6}), // 1
    DVector({1e-5,1e-3,1e-6}), // 2
    DVector({3e-4,1e-4,1e-6}), // 3
    DVector({5e-4,1e-3,1e-6}), // 4
    DVector({7e-4,3e-3,1e-6}), // 5
    DVector({1e-3,3e-3,1e-6}), // 6
    DVector({2e-2,2e-2,4e-6})  // 7
  };
  result[0] = new HSet(p,h._C,r[0],8);
  for(int i=1;i<=length;++i){
    D = result[i-1]->_C;
    for(int k=0;k<8;++k) { p = PM(p,temp); D = temp*D; }
    if(i<=6)
      C = h._C;
    else 
      C = last._C;
    C.column(0) = D.column(0);
    C.column(1) = D.column(1);
    normalizeColumns(C);
    result[i] = new HSet(p,C,r[i],8);
  }
  return result;
}

// ##############################################################
/**
 * This routine construct a sequence of h-sets linking h-sets sentered at Q2 and Q1, repsectively. 
 * Data from hand-made simulation.
*/ 
HSetChain createQ2Q1HeteroclinicChain(RosslerPoincareMap& PM, HSet& h, HSet& last, int length, DVector t){
  HSetChain result(length+1);
  DVector p = h._x + h._C*t;
  DVector r[] = { 
    DVector({3e-4,1.5e-4,1e-6}), // 0
    DVector({4e-4,3e-4,1e-6}), // 1
    DVector({2e-3,4e-4,1e-6})
  };
  int n = 14;
  result[0] = new HSet(p,h._C,r[0],n);
  DMatrix temp(3,3);
  for(int i=1;i<=length;++i){
    DMatrix D = result[i-1]->_C;
    for(int k=0;k<n;++k) {p = PM(p,temp); D=temp*D; }
    DMatrix C = D;
    C.column(2) = h._C.column(2);
    normalizeColumns(C);
    result[i] = new HSet(p,C,r[i],n);
  }
  return result;
}

// ##############################################################
/**
 * This routine construct a sequence of h-sets linking h-sets sentered at Q1 and P4, repsectively. 
 * Data from hand-made simulation.
*/ 
HSetChain createQ1P4HeteroclinicChain(RosslerPoincareMap& PM, HSet h, HSet last, int length, double t){
  HSetChain result(length+1);
  DVector p = h._x + h._C*DVector(t,0.,0.);
  DMatrix D(3,3), temp(3,3), C(3,3);
  DVector r[] = { 
    DVector({6e-5,1e-5,1e-6}), // 0
    DVector({7e-5,1e-5,1e-6}), // 1
    DVector({8e-5,1e-5,1e-6}), // 2
    DVector({1e-4,1e-5,1e-6})  // 3
  };
  int N = 12;
  C.column(0) = h._C.column(1);
  C.column(1) = h._C.column(0);
  C.column(2) = h._C.column(2);
  result[0] = new HSet(p,C,r[0],N);
  for(int i=1;i<=length;++i){
    D = result[i-1]->_C;
    for(int k=0;k<N;++k) { p = PM(p,temp); D = temp*D; }
    C.column(0) = D.column(0);
    C.column(1) = D.column(1);
    normalizeColumns(C);
    if(i==3) N = 16;
    result[i] = new HSet(p,C,r[i],N);
  }
  return result;
}

// ##############################################################
/**
 * This routine construct a sequence of h-sets linking h-sets sentered at Q1 and P1, repsectively. 
 * Data from hand-made simulation.
*/ 
HSetChain createQ1P1HeteroclinicChain(RosslerPoincareMap& PM, HSet h, HSet last, int length,double t ){
  HSetChain result(length+1);
  DVector p = h._x + h._C*DVector(t,0.,0.);
  DMatrix D(3,3), temp(3,3), C(3,3);
  DVector r[] = { 
    DVector({1e-4,1e-5,1e-6}), // 0
    DVector({1e-4,1e-5,1e-6}), // 1
    DVector({1e-4,1e-5,1e-6}), // 2
    DVector({4e-4,1e-5,1e-6}), // 3
    DVector({5e-5,1e-5,1e-6})  // 4
  };
  int N = 12;
  C = h._C;
  result[0] = new HSet(p,C,r[0],N);
  int i;
  for(i=1;i<=length;++i){
    D = result[i-1]->_C;
    for(int k=0;k<N;++k) { p = PM(p,temp); D = temp*D; }
    C.column(0) = D.column(0);
    C.column(1) = D.column(1);
    normalizeColumns(C);
    if(i==3) {N = 16; }
    result[i] = new HSet(p,C,r[i],N);
  }
  return result;
}

// ##############################################################
/**
 * Main data structure that stores information about all h-sets
*/ 
struct ProofData{
  ProofData(RosslerPoincareMap& PM) 
    : PM(PM), 
      enclosureFilename("out/attractor.dat") ,
      needToTabulateBinomialForParallelComputation(binomial(40,20))
  {}
  
  void initHSetsAtPeriodicPoints();
  void initHeterloclinicChains();

  void encloseAttractor();

  void checkCoveringRelationsAtPeriodicPoints();
  void checkWeakChaoticCoveringRelations();
  void checkHyperChaoticCoveringRelations();
  void checkLinkingCoveringRelations();
  void checkAllCoveringRelations();
  
  void computeEigenvaluesAtPeriodicPoints();
  void checkConeConditions();
  
  // members
  RosslerPoincareMap& PM;
  HSet P1, P2, P3, P4, Q1, Q2;
  
  // weak-chaotic chains
  HSetChain chainP1P2,chainP2P1,chainP1P3, chainP3P4, chainP4P3;
  
  // hyperchaotic chains
  HSetChain chainQ1Q2, chainQ2Q1;

  // linking between hyperchaos and chaos
  HSetChain chainQ1P4,chainQ1P1;

  // data structure that stores positive invariant set - see InSetEnclosure.h
  typedef InvSetEnclosure<RosslerPoincareMap::VectorType> InvSet;
  typedef InvSet::Vertex Vertex;
  std::string enclosureFilename;
  
  volatile long long needToTabulateBinomialForParallelComputation;
};

// ##############################################################
/**
 * this routine construct 6 h-sets at numerically found periodic points.
*/ 
void ProofData::initHSetsAtPeriodicPoints(){

  // -------------- WEAK CHAOS ----------------------------
  // period 8 orbits for weak chaos, they are out of attractor
  DVector p1({-104.32937253702462,0.028756669726685443,44.645081351998819});
  DVector p2({-104.26664163365506,0.028773972266421831,44.640115482927115});

  // period 8 orbits for weak chaos on attractor
  DVector p3({-104.42324539012806,0.028730815749171541,44.678254866134068});
  
  // period 16 orbits for weak chaos on attractor
  DVector p4({-104.39575243552828,0.028738382959034744,44.666264617071981});

  P1 = getHSetAtFixedPoint(PM,p1,8,1e-3,1e-4);
  P2 = getHSetAtFixedPoint(PM,p2,8,1e-3,6e-4);
  P3 = getHSetAtFixedPoint(PM,p3,8,1.5e-4,6e-4);
  P4 = getHSetAtFixedPoint(PM,p4,16,1e-3,1e-3);

  // -------------- HYPERCHAOS ----------------------------
      
  // period 12 orbit for hyperchaos (out of attractor)
  DVector q1({-103.69667754570543,0.028932144798038389,44.407870627484129});

  // period 14 orbit for hyperchaos (out of attractor, complex ev.)
  DVector q2({-103.37098255164607,0.029023312473829044,44.284349486019579});
   
  Q1 = getHSetAtFixedPoint(PM,q1,12,1e-5,5e-5,1e-6);
  Q2 = getHSetAtFixedPoint(PM,q2,14,3.5e-4,2e-4,1e-6);
}

// ##############################################################
/**
 * This routine creates heteroclinic chains of h-sets between periodic orbits
*/ 
void ProofData::initHeterloclinicChains(){
  // the last argument is the distance from the periodic point in some unstable direction
  // This has beeen found numerically as an orbit close to heteroclinic.
  // See (5.2) in the article.
  // weak chaos
  chainP1P2 = createP1P2HeteroclinicChain(PM,P1,P2,8,0.00158202);
  chainP2P1 = createP2P1HeteroclinicChain(PM,P2,P1,6,0.00128584);
  chainP3P4 = createP3P4HeteroclinicChain(PM,P3,P4,10,-0.000134682);
  chainP4P3 = createP4P3HeteroclinicChain(PM,P4,P3,10,0.00039344);
  chainP1P3 = createP1P3HeteroclinicChain(PM,P1,P3,8,-0.000884928);

  // hyperchaos
  chainQ1Q2 = createQ1Q2HeteroclinicChain(PM,Q1,Q2,7,DVector({-8.1098e-05,0.,0.}));
  chainQ2Q1 = createQ2Q1HeteroclinicChain(PM,Q2,Q1,2,DVector({-1.9639e-007,0.0010592239,0.}));

  // linking between hyperchaos and chaos
  // See (5.10) in the article.
  chainQ1P4 = createQ1P4HeteroclinicChain(PM,Q1,P4,3,-6.0427e-005);
  chainQ1P1 = createQ1P1HeteroclinicChain(PM,Q1,P1,4,-7.0488e-005);
}

// ##############################################################
/**
 * This function just calls the general algorithm for enclosing forward invariant set.
 * Size of single box in the covering is set to (2^{-10},2^{-10},2^{-12})
*/ 
void ProofData::encloseAttractor(){
  auto start = tic("enclosing attractor");
  
  // fix a resolution
  Vertex resolution = Vertex({10,10,12});
  InvSet enc(resolution);
  // we start enclosing the atractor from a single box that conatins q1 periodic orbit
  IVector initPoint(this->Q1._x); 
  
  // then we call the algorithm (see InvSetEnclosure.h)
  enc.encloseForwardInvariantSet(PM,initPoint);
  
  // and save computed attractor to file
  enc.save(this->enclosureFilename.c_str());
  
  tac(start);
}

// ##############################################################
/**
 * This function calls general algorithm for verification of chain of covering relations with one unstable direction.
 * We subsequently check sequences (5,3-5.7) from the article.
 * See covrel.h for the algorithms.
*/
void ProofData::checkWeakChaoticCoveringRelations(){
  checkHeteroclinicChain(PM,P1,P2,chainP1P2, " Heteroclinic chain P1 => ... => P2", "grids/P1P2.grid");
  checkHeteroclinicChain(PM,P2,P1,chainP2P1," Heteroclinic chain P2 => ... => P1", "grids/P2P1.grid");

  checkHeteroclinicChain(PM,P3,P4,chainP3P4, " Heteroclinic chain P3 => ... => P4","grids/P3P4.grid");
  checkHeteroclinicChain(PM,P4,P3,chainP4P3," Heteroclinic chain P4 => ... => P3","grids/P4P3.grid");

  checkHeteroclinicChain(PM,P1,P3,chainP1P3," Heteroclinic chain P1 => ... => P3","grids/P1P3.grid");
}

// ##############################################################
/**
 * This function calls general algorithm for verification of chain of covering relations.
 * We check that each h-sets centered at periodic orbit covers itself.
 * See covrel.h for the algorithms.
*/
void ProofData::checkCoveringRelationsAtPeriodicPoints(){
  //covering relations for periodic orbits
  checkCovering(PM,P1,P1,"P1=>P1","grids/perPoints.grid");
  checkCovering(PM,P2,P2,"P2=>P2","grids/perPoints.grid");
  checkCovering(PM,P3,P3,"P3=>P3","grids/perPoints.grid");
  checkCovering(PM,P4,P4,"P4=>P4","grids/perPoints.grid");

  //covering relations for periodic orbits
  check2DCovering(PM,Q1,Q1,"Q1=>Q1","grids/perPoints.grid");
  check2DCovering(PM,Q2,Q2,"Q2=>Q2","grids/perPoints.grid");
}

// ##############################################################

/**
 * This function calls general algorithm for verification of chain of covering relations.
 * We check two heteroclinic chains for hyperchaos.
 * See covrel.h for the algorithms.
*/
void ProofData::checkHyperChaoticCoveringRelations(){
  check2DHeteroclinicChain(PM,Q2,Q1,chainQ2Q1, " Heteroclinic chain Q2 => ... => Q1","grids/Q2Q1.grid");
  check2DHeteroclinicChain(PM,Q1,Q2,chainQ1Q2, " Heteroclinic chain Q1 => ... => Q2","grids/Q1Q2.grid");
}

// ##############################################################

/**
 * This function calls general algorithm for verification of chain of covering relations.
 * We check two heteroclinic chains between hyperchaos and chaos.
 * See covrel.h for the algorithms.
*/
void ProofData::checkLinkingCoveringRelations(){
  checkLinkingChain(PM,Q1,P4,chainQ1P4, " Heteroclinic chain Q1 => ... => P4","grids/Q1P4.grid");
  checkLinkingChain(PM,Q1,P1,chainQ1P1, " Heteroclinic chain Q1 => ... => P1","grids/Q1P1.grid");
}

// ##############################################################
/**
 * This function calls subsequently all previous function for covering relations.
*/

void ProofData::checkAllCoveringRelations(){
  auto start = tic("check coverinig relations");

  this->checkCoveringRelationsAtPeriodicPoints();
  this->checkWeakChaoticCoveringRelations();
  this->checkHyperChaoticCoveringRelations();
  this->checkLinkingCoveringRelations();

  tac(start);
}

// ##############################################################
/**
 * This function calls general algorithm for proving the existence of periodic orbit.
 * If suceed, it computes eigenvalues of derivative at periodic point.
 * See eigenvalues.h for algorithms.
 */ 
void ProofData::computeEigenvaluesAtPeriodicPoints(){
  PM.setTolerance(1e-18);
  int d=1;
  std::cout << "validate p4: " << provePeriodicOrbit(PM,P4._x,d*16,3) << std::endl;
  PM.setTolerance(1e-18);
  PM.setCrossingDirection(capd::poincare::Both);
  d=2;
  std::cout << "validate p1: " << provePeriodicOrbit(PM,P1._x,d*8,3) << std::endl;
  std::cout << "validate p2: " << provePeriodicOrbit(PM,P2._x,d*8,3) << std::endl;
  std::cout << "validate p3: " << provePeriodicOrbit(PM,P3._x,d*8,3) << std::endl;
  std::cout << "validate q1: " << provePeriodicOrbit(PM,Q1._x,d*12,3) << std::endl;
  std::cout << "validate q2: " << provePeriodicOrbit(PM,Q2._x,d*14,3) << std::endl;
}


// ##############################################################
/**
 * This function calls general algorithm for verification of
 * - the existence of bidirectional sequence of covering relations at periodic point
 * - check the cone condition at periodic point
 * This is done fore each periodic orbit p1,p2,p3,p4.q1.q2
*/ 
void ProofData::checkConeConditions(){
  auto start = tic("checking cone conditions");

  checkConeCondition(PM,P1,"P1",checkHeteroclinicChain); 
  checkConeCondition(PM,P2,"P2",checkHeteroclinicChain);
  checkConeCondition(PM,P3,"P3",checkHeteroclinicChain);
  checkConeCondition(PM,P4,"P4",checkHeteroclinicChain);

  checkConeCondition(PM,Q1,"Q1",check2DHeteroclinicChain);
  checkConeCondition(PM,Q2,"Q2",check2DHeteroclinicChain);

  tac(start);
}

// ##############################################################

#endif

