#ifndef DANIEL_WILCZAK_ROSSLER_POINCARE_MAP_H
#define DANIEL_WILCZAK_ROSSLER_POINCARE_MAP_H

#include <list>
#include <omp.h>
#include "capd/capdlib.h"

// ----------------------------- RosslerPoincareMap ---------------------
/**
 * This class represents Poincare map for the 4D Rossler system.
 * It provides methods for computation of Poincare map and its derivative.
*/

class RosslerPoincareMap
{
public:
  typedef capd::IVector VectorType;
  typedef capd::IMatrix MatrixType;

  // This is contructor of the class.
  // Here we declare vector fields, solvers and Poincare map.
  RosslerPoincareMap(capd::interval a, capd::interval b, capd::interval c, capd::interval d,int order=20)
    :   factor(1)
  {
    // for each thread on the machine we define own vector field object.
    // both rigorous and nonrigorous
    iVectorField = new capd::IMap* [omp_get_max_threads()];
    dVectorField = new capd::DMap* [omp_get_max_threads()];

    iSection = new capd::ICoordinateSection* [omp_get_max_threads()];
    dSection = new capd::DCoordinateSection* [omp_get_max_threads()];

    // for each thread on the machine we define own solver object.
    // both rigorous and nonrigorous
    iSolver = new capd::IOdeSolver* [omp_get_max_threads()];
    dSolver = new capd::DOdeSolver* [omp_get_max_threads()];

    // for each thread on the machine we define own Poincare map object.
    // both rigorous and nonrigorous
    iPoincareMap = new capd::IPoincareMap* [omp_get_max_threads()];
    dPoincareMap = new capd::DPoincareMap* [omp_get_max_threads()];

    // Arrays for objects are allocated. Define objects in arrays.
    for (int i=0; i<omp_get_max_threads();++i)
    {
      // define vector fields
      iVectorField[i] = new capd::IMap("par:a,b,c,d;var:x,z,w,y;fun:-(y+z),b+x*z,d*w+c*z,x+a*y+w;");
      dVectorField[i] = new capd::DMap("par:a,b,c,d;var:x,z,w,y;fun:-(y+z),b+x*z,d*w+c*z,x+a*y+w;");
      
      iSection[i] = new capd::ICoordinateSection(4,3);
      dSection[i] = new capd::DCoordinateSection(4,3);

      // set parameters of vector fields (rigorous)
      iVectorField[i]->setParameter("a",a);
      iVectorField[i]->setParameter("b",b);
      iVectorField[i]->setParameter("c",c);
      iVectorField[i]->setParameter("d",d);

      // set parameters of vector fields (long double version)
      dVectorField[i]->setParameter("a",a.mid().leftBound());
      dVectorField[i]->setParameter("b",b.mid().leftBound());
      dVectorField[i]->setParameter("c",c.mid().leftBound());
      dVectorField[i]->setParameter("d",d.mid().leftBound());

      // define solvers with a given vector field, order of Taylor method and StepControl
      iSolver[i] = new capd::IOdeSolver(*(iVectorField[i]),order);
      iSolver[i]->setAbsoluteTolerance(1e-12);
      iSolver[i]->setRelativeTolerance(1e-12);
      iPoincareMap[i] = new capd::IPoincareMap(*iSolver[i],*iSection[i],capd::poincare::PlusMinus);

      dSolver[i] = new capd::DOdeSolver(*(dVectorField[i]),order);
      dSolver[i]->setAbsoluteTolerance(1e-21);
      dSolver[i]->setRelativeTolerance(1e-21);
      dPoincareMap[i] = new capd::DPoincareMap(*dSolver[i],*dSection[i],capd::poincare::PlusMinus);
    }
  }

  // destructor - releases the memory
  ~RosslerPoincareMap(){
    // Arrays for objects are allocated. Define objects in arrays.
    for (int i=0; i<omp_get_max_threads();++i)
    {
      delete iVectorField[i];
      delete dVectorField[i];
      delete iSection[i];
      delete dSection[i];
      delete iSolver[i];
      delete iPoincareMap[i];
      delete dSolver[i];
      delete dPoincareMap[i];
    }
    delete iVectorField;
    delete dVectorField;
    delete iSection;
    delete dSection;
    delete iSolver;
    delete iPoincareMap;
    delete dSolver;
    delete dPoincareMap;    
  }
  
  // This is a method that sets the order of the Taylor method to each solver for each thread.
  void setOrder(int newOrder)
  {
    for (int i=0; i<omp_get_max_threads();++i)
    {
      iSolver[i]->setOrder(newOrder);
      dSolver[i]->setOrder(newOrder);
    }
  }

  void setCrossingDirection(capd::poincare::CrossingDirection d)
  {
    for (int i=0; i<omp_get_max_threads();++i)
    {
      iPoincareMap[i]->setCrossingDirection(d);
      dPoincareMap[i]->setCrossingDirection(d);
    }
  }

  // This is a function that sets a factor - this is a very technical parameter 
  // of doubleton sets that specifies when to rearange representation.
  void setFactor(double _factor)
  {
    factor = _factor;
  }

  // This function sets new parameters of step control method for each solver.
  void setTolerance(double tol)
  {
    for(int i=0;i<omp_get_max_threads();++i){
      iSolver[i]->setAbsoluteTolerance(tol);
      iSolver[i]->setRelativeTolerance(tol);
    }
  }

  // The section is 3D while the phase space is 4D. 
  // The functions below are used to convert 3D <-> 4D vectors and matrices.
  capd::DVector embed(capd::DVector u){
    return capd::DVector({u[0],u[1],u[2],0.});
  }

  VectorType embed(VectorType u){
    return VectorType({u[0],u[1],u[2],capd::interval(0.)});
  }

  capd::DMatrix embed(capd::DMatrix A){
    capd::DMatrix B(4,4);
    for(int i=1;i<4;++i)
      for(int j=1;j<4;++j)
        B(i,j) = A(i,j);
    return B;
  }

  capd::IMatrix embed(capd::IMatrix A){
    capd::IMatrix B(4,4);
    for(int i=1;i<4;++i)
      for(int j=1;j<4;++j)
        B(i,j) = A(i,j);
    return B;
  }

  capd::DMatrix project(capd::DMatrix A){
    capd::DMatrix B(3,3);
    for(int i=1;i<4;++i)
      for(int j=1;j<4;++j)
        B(i,j) = A(i,j);
    return B;
  }

  MatrixType project(MatrixType A){
    MatrixType B(3,3);
    for(int i=1;i<4;++i)
      for(int j=1;j<4;++j)
        B(i,j) = A(i,j);
    return B;
  }

  capd::DVector project(capd::DVector u){
    return capd::DVector(3,u.begin());
  }

  VectorType project(VectorType u){
    return VectorType(3,u.begin());
  }

  // This is a nonrigorous routine. 
  // It computes Poincare map P(u).
  capd::DVector operator()(capd::DVector u)
  {
    capd::rounding::DoubleRounding::roundNearest();
    capd::DVector x = embed(u);
    int i = omp_get_thread_num() % omp_get_max_threads();
    x = dPoincareMap[i]->operator()(x);
    return project(x);
  }

  // This is a nonrigorous routine. 
  // It computes Poincare n-th iteration of P at 'u'.
  capd::DVector operator()(capd::DVector u, int n)
  {
    capd::rounding::DoubleRounding::roundNearest();
    capd::DVector x = embed(u);
    int i = omp_get_thread_num() % omp_get_max_threads();
    for(int j=0;j<n;++j)
      x = dPoincareMap[i]->operator()(x);
    return project(x);
  }

  // This is a nonrigorous routine. 
  // It computes simultanously Poincare map and its derivative.
  capd::DVector operator()(capd::DVector u, capd::DMatrix& result)
  {
    capd::rounding::DoubleRounding::roundNearest();
    capd::DVector x = embed(u);
    // put the point to the main coordinate system
    capd::DMatrix D = capd::DMatrix(4,4);

    int i = omp_get_thread_num() % omp_get_max_threads();
    x = dPoincareMap[i]->operator()(x,D);
    D = dPoincareMap[i]->computeDP(x,D);
    result = project(D);
    return project(x);
  }

  // This is rigorous routine. For an interval vector u
  // it computes an interval vector that encloses P^n(u).
  VectorType operator()(VectorType u, int n=1)
  {
    capd::IVector x = embed(u);
    
    // define doubleton representation of the above
    capd::C0HOTripletonSet s(x);
    s.setC0Factor(factor);
    // call routine from the CAPD library
    x = (*iPoincareMap[omp_get_thread_num() % omp_get_max_threads()])(s,n);
    return project(x);
  }

  // This is rigorous routine. For an interval vector u
  // it computes an interval matrix that encloses DP^n(u)*init.
  MatrixType derivative(VectorType u, int n, capd::interval& returnTime, MatrixType init)
  {
    capd::IVector x = embed(u);
    
    // define doubleton representation of the above
    capd::C1HORect2Set s(capd::C1HORect2Set::C0BaseSet(x),capd::C1HORect2Set::C1BaseSet(embed(init)));
    capd::IMatrix D(4,4);
    auto& pm = *iPoincareMap[omp_get_thread_num() % omp_get_max_threads()];
    // call routine from the CAPD library
    x = pm(s,D,returnTime,n);
    return project(pm.computeDP(x,D));
  }

  // This is rigorous routine. For an interval vector u
  // it computes an interval vector that encloses derivative DP^n(u).
  MatrixType derivative(VectorType u, int n=1)
  {
    capd::interval rt;
    return derivative(u,n,rt,MatrixType::Identity(u.dimension()));
  }

  // This is rigorous routine. For a set represented by x + C*r
  // it computes an interval matrix that encloses DP^n(u)*init.
  MatrixType derivative(VectorType x, MatrixType C, VectorType r, int n, MatrixType init)
  {
    // define doubleton representation of the above
    capd::C1HORect2Set s(capd::C1HORect2Set::C0BaseSet(x,C,r),capd::C1HORect2Set::C1BaseSet(init));
    capd::IMatrix D(4,4);
    capd::interval returnTime;
    auto& pm = *iPoincareMap[omp_get_thread_num() % omp_get_max_threads()];
    // call routine from the CAPD library
    x = pm(s,D,returnTime,n);
    return project(pm.computeDP(x,D));
  }

  double factor;
  
  // Members: vector fields, solvers and Poicnare maps.
  capd::DMap** dVectorField;
  capd::IMap** iVectorField;
  
  capd::DCoordinateSection** dSection;
  capd::ICoordinateSection** iSection;

  capd::IOdeSolver** iSolver;
  capd::DOdeSolver** dSolver;

  capd::IPoincareMap** iPoincareMap;
  capd::DPoincareMap** dPoincareMap;
};


#endif

