/**
 * This file contains algorithms that
 * - construct sequence of h-sets, each centered at an approximate periodic orbit, with geometrically decreasing size
 * - check covering relations in both directions along this sequence
 * - check the cone condition at the smallest set in the sequence
*/ 

#ifndef DANIEL_WILCZAK_CONES_H
#define DANIEL_WILCZAK_CONES_H

// ##############################################################
/**
 * @param h - h-set centered at an approximate periodic point
 * @param length - length of the sequence to be constructed
 * @param factor - a positive real number
 * 
 * @returns a sequence of h-sets, each centered at the center of 'h' of the length 'length' with geometrically decreasing size of subsequent h-set.
 *          Two expanding coordinates are decresed, only. 
*/ 
HSetChain getPeriodicChain(HSet& h, int length, double factor){
  HSetChain result(length);
  DVector r = h._r;
  for(int i=0;i<length;++i){
    r[0] /= factor;
    r[1] /= factor;
    result[i] = new HSet(h._x, h._C,r,h.N);
  }
  return result;
}

// ##############################################################
/**
 * Check positive definiteness of a 3x3 matrix by Sylvester criterion 
*/ 
bool checkSylvesterCriterion(const IMatrix& T)
{
  // we verify if matrix is positive definite by simple Sylvester criterion
  // determinants of minors 1x1 2x2 and 3x3 minors should be positive numbers
  interval det1 = T(1,1);
  interval det2 = T(2,2)*T(1,1) - sqr(T(2,1));
  interval det3 = T(1,1)*T(2,2)*T(3,3) + T(1,2)*T(2,3)*T(3,1) + T(1,3)*T(2,1)*T(3,2)
                - T(2,2)*sqr(T(1,3)) - sqr(T(1,2))*T(3,3) - T(1,1)*sqr(T(2,3));

  return (det1 >0. && det2>0. && det3>0.);
}

// ##############################################################
/**
 * Check positive definiteness of a 3x3 matrix by Gershogorin criterion 
*/ 
bool checkGershgorinCriterion(const IMatrix& M)
{
  for(unsigned i=1;i<=M.numberOfRows();++i){
    interval p=0;
    for(unsigned j=1;j<=M.numberOfColumns();++j) {
      if(j!=i)
        p += capd::abs(M(i,j));
    }
    if( !(M(i,i)>p) ) return false;
  }
  return true;
}

// ##############################################################
/**
 * There is a theorem by J.Rohn that says, that a symmetric interval matrix is positive definite in and only if its corners are positive definite.
 * Since we are in 3D only, we use this cirtetion to check if an interval matrix M is positrive definite.
*/ 
template<class MatrixT>
bool checkPosDefMatrix(MatrixT M)
{
  MatrixT D(M.numberOfRows(),M.numberOfColumns());
  int size = (1 << (M.numberOfRows()*(M.numberOfRows()+1))/2);

  // this is the loop over each corner
  for(int k=0;k<size;++k){
    int u=k;
    MatrixT Q(M.numberOfRows(),M.numberOfColumns());
    for(unsigned p=1;p<=M.numberOfRows();++p) {
      for(unsigned s=p;s<=M.numberOfColumns();++s) {
        if(u%2==0)
          Q(p,s) = Q(s,p) = M(p,s).leftBound();
        else
          Q(p,s) = Q(s,p) = M(p,s).rightBound();
        u/=2;
      }
    }

    // if Sylvester criterion applies to the corner then check the next corner
    if(checkSylvesterCriterion(Q))  continue;

    // if not, then try to diaginalize and applu Gershgorin criterion
    if (!capd::matrixAlgorithms::symMatrixDiagonalize(Q,D,1e-12))
    {
      std::cout << "Cannot diagonalize a matrix\n";
      return false;
    }
    
    if(checkGershgorinCriterion(D)) continue;
      
    return false;
  }

  return true;
}

// ##############################################################
/**
 * This is the main algorithm that checks cone condition at an h-set.
 * 
 * @param h - h-set
 * @param g0,g1 - positive integers that stand for the ncessary subdivisions of the two main cordinates of 'h'
 * @param L1, L2, L3 - coefficients of the diagonal quadratic form Q from the definition of cone condition.
*/ 
bool checkConeCondition(RosslerPoincareMap& PM, HSet& h, int g0, int g1, double L1, double L2, double L3)
{
  // take diagonal quadratic form Q
  IMatrix Q(3,3); Q(1,1) = L1, Q(2,2) = L2; Q(3,3) = L3;

  IVector x = h.x();
  IMatrix C = h.C(PM);
  IVector r = h.r();
  IMatrix invC = matrixAlgorithms::krawczykInverse(PM.project(C));
  IMatrix T = transpose(invC)*Q*invC;
  T = intersection(T,transpose(T));

  // we will subdivide 'h', so precompute grids
  interval gu = (r[0].right()-r[0].left())/g0;
  interval gs = (r[1].right()-r[1].left())/g1;
  IVector L = leftVector(r);
  volatile bool result = true;
  int k;

  #pragma omp parallel for schedule(dynamic) 
  for(k=0;k<g0*g1;k++){
    if(!result) continue;
    int i = k%g1;
    int j = k/g1;
    // take subvector _r\subset r
    IVector _r(4);
    _r[0] = L[0] + interval(j,j+1)*gu;
    _r[1] = L[1] + interval(i,i+1)*gs;
    _r[2] = r[2];
    
    // compute derivative of Poincare map
    IMatrix D = PM.derivative(x,C,_r,h.N,C);

    // compute interval matrix DF^T*Q*DF-Q expressed in proper coordinate sytems
    IMatrix W = PM.project(invC)*D;
    W = transpose(W)*Q*W-Q;
    W = intersection(W,transpose(W));

    D = transpose(D)*T*D-Q;
    D = intersection(D,transpose(D));

    D = intersection(W,D);

    // check if this matrix is positive definite. 
    bool res = checkPosDefMatrix(D);
    if(!res)
    #pragma omp critical
    {
      printf("(%d,%d) out of (%d,%d)\n",j+1,i+1,g0,g1);
      result = res;
    }
  }
  return result;
}

// ##############################################################

typedef void (*CheckHChainProcedure)(RosslerPoincareMap& PM, HSet& begin, HSet& end, HSetChain& chain, const char message[], const char gridFile[]);
/**
 * This is the main algorithm that for a given set 'h' proves that there is a bidirectional seuqence of covering relations centered at the same point as 'h' is 
 * and the cone condition is satisfied on the smallest set.
 * 
 * @param h - h-set
 * @param checkHChain - an algorithm for checking covering relations. \
                        Note, we have different algorithm for 1 or 2 expanding directions in the set covrel.h but they have the same signature CheckHChainProcedure.
*/ 
void checkConeCondition(RosslerPoincareMap& PM, HSet h, const char filename[], CheckHChainProcedure checkHChain){
  int length, g1, g2;
  double factor, L1, L2, L3;

  // read data from config file (length of the sequence, parameters of quadratic form, number of subdivisions)
  std::string dir = std::string("grids/") + filename;
  std::string grid = dir + "Cones.grid";
  std::ifstream in(grid.c_str());
  in >> length >> factor >> L1 >> L2 >> L3 >> g1 >> g2;
  in.close();
  
  // first we construct sequence of covering relations relations
  HSetChain omegaChain = getPeriodicChain(h,length+1,factor);

  grid = dir + "Omega.grid";
  std::string desc = "Omega limit chain for ";
  desc += filename;
  
  // here we check that for each subsequent pair covering relation holds true
  // we use a routine chechHChain (either checkCovering or check2DCovering)
  HSet* end = omegaChain[length];
  checkHChain(PM,h,*end,omegaChain,desc.c_str(),grid.c_str());
  
  // we need bidirectional covering relation, so reverse this chain.
  HSetChain alphaChain(length+1);
  for(int i=0;i<=length;++i)
    alphaChain[i] = omegaChain[length-i];

  // here we check that for each subsequent pair in reveresed chain covering relation holds true
  // we use a routine chechHChain (either checkCovering or check2DCovering)
  grid = dir + "Alpha.grid";
  desc = "Alpha limit chain for ";
  desc += filename;
  checkHChain(PM,*end,h,alphaChain,desc.c_str(),grid.c_str());
  
  // eventually we check the cone condition at the smallest set in the sequence
  std::cout << "Cone condition for " << filename << ": " << std::flush
            << checkConeCondition(PM,*end,g1,g2,L1,L2,L3) << std::endl;
}

#endif
