/**
 * This file contains algorithms used to prove covering relations with one or two expanding directions.
 * Also chains of covering relations are supported.
*/ 

#ifndef DANIEL_WILCZAK_COVREL_H
#define DANIEL_WILCZAK_COVREL_H

#include "RosslerPoincareMap.h"
#include <sstream>

/**
 * This is an auxiliary data structure that stores information aboot number of subdivisions
 * of a set/wall/face needed to check the certain condition. It also provides methods for 
 * reading such data from config *.grid files. 
*/
struct Grid
{
  std::string description;
  int unstable;
  int stable;
  int accrossUnstable, accrossStable;
  bool status;
  Grid(const char fileName[], std::string description) 
  {
    std::ifstream file;
    file.open(fileName);
    char bufor[255];
    while(file >> bufor)
    {
      if(bufor[0] == '@')
      {
        file >> bufor;
        if(description!=std::string(bufor))
          continue;
        do{
          file.getline(bufor,255);
        }while(file.peek()=='@');
        file >> bufor >> unstable >> stable >> accrossUnstable >> accrossStable;
        file.close();
        status = true;
        return;
      }
    };
    status = false;
  }
}; 

// ##############################################################
/**
 * This is an abstract algorithm. It takes a set of the form x+C*r 
 * and checks if a condition 'c' is satisfied on the set
 * M*( PM^period(x+C*r) - y)
 * where
 * @param M - matrix
 * @param y - vector
 * @param PM - Poincare map
 * @param x+C*r - a set
 * 
 * @returns true if condition is satisfied
*/ 
template<class Condition>
bool checkCondition(
  RosslerPoincareMap& PM, 
  const IVector& x, 
  const IMatrix& C, 
  const IVector& r, 
  const IMatrix& M, 
  const IVector& y, 
  const Condition& c, int period)
{
  // we represent initial condition as tripleton
  C0HOTripletonSet h(x,C,r);
  interval returnTime;
  try{
    auto& pm = *PM.iPoincareMap[omp_get_thread_num()];
    // compute M*( PM^period(x+C*r) - y)
    IVector w = pm(h,y,M,returnTime,period);
    // and check the condition
    if(!c(w,r)) {
      #pragma omp critical
      {
        std::cout << r << std::endl;
        std::cout << w << std::endl;
      }
      return false;
    }
    
    return true;
  }catch(std::exception& e){
    #pragma omp critical
    {
      std::cout << "Exception caught when computing Poincare map: " << e.what() << std::endl ;
    }
  } 
  return false;
}

// ##############################################################
/**
 * This is an abstract algorithm. It checks if a condition 'c' is satisfied on the point
 * M*( PM^period(u) - y)
 * where u belongs to x+C*r. 
 * During these computations 'index' coordinate of 'r' is subdivided into 'g' pieces.
 * @param M - matrix
 * @param y - vector
 * @param PM - Poincare map
 * @param x+C*r - a set
 * 
 * @returns true if condition is satisfied
*/ 
template<class Condition>
bool checkFace(
    RosslerPoincareMap& PM, 
    const IVector& x, 
    const IMatrix& C, 
    const IVector& r, 
    const IMatrix& M, 
    const IVector& y, 
    const Condition& c, 
    int period, int g, int index
    )
{
  // we subdivide 'index' coordinate - compute size of one element
  const interval gu = (r[index].right()-r[index].left())/g;
  const IVector L = leftVector(r);
  volatile bool result = true;
  int j;
  
  #pragma omp parallel for shared(result) schedule(dynamic)
  for(j=0;j<g;j++){
    if(!result) continue;
    IVector _r = r;
    // take subset _r of r
    _r[index] = L[index] + interval(j,j+1)*gu;
    // and call the previous algorithm
    bool res = checkCondition(PM,x,C,_r,M,y,c,period);
    if(!res)
    #pragma omp critical
    {
      printf("%d out of %d\n",j+1,g);
      result = res;
    } 
  }
  return result;
}

// ##############################################################
/**
 * This is an abstract algorithm. It checks if a condition 'c' is satisfied on the point
 * M*( PM^period(u) - y)
 * where u belongs to x+C*r. 
 * During these computations two coordinates of 'r' is subdivided into 'g0' and 'g1' pieces, respectively.
 * @param M - matrix
 * @param y - vector
 * @param PM - Poincare map
 * @param x+C*r - a set
 * 
 * @returns true if condition is satisfied
*/ 

template<class Condition>
bool checkSet(
    RosslerPoincareMap& PM, 
    const IVector& x, 
    const IMatrix& C, 
    const IVector& r, 
    const IMatrix& M, 
    const IVector& y, 
    Condition c, 
    int period, int g0, int g1
  ){
  interval gu = (r[0].right()-r[0].left())/g0;
  interval gs = (r[1].right()-r[1].left())/g1;
  IVector L = leftVector(r);
  volatile bool result = true;
  int k;
  #pragma omp parallel for  schedule(dynamic)
  for(k=0;k<g0*g1;k++){
    if(!result) continue;
    int i = k%g1;
    int j = k/g1;
    // take a vector _r \subset r
    IVector _r(4);
    _r[0] = L[0] + interval(j,j+1)*gu;
    _r[1] = L[1] + interval(i,i+1)*gs;
    _r[2] = r[2];
    // call abstract algorithm
    bool res = checkCondition(PM,x,C,_r,M,y,c,period);
    if(!res)
    #pragma omp critical
    {
      printf("(%d,%d) out of (%d,%d)\n",j+1,i+1,g0,g1);
      result = res;
    } 
  }
  return result;
}

// ##############################################################
// macro for logging to standard ouput
#define CHECK_CONDITION(message,f) {\
  bool result = f;\
  std::cout << message << result << std::endl;\
  if(!result) { std::cout << "Condition not satisfied. Stopping the program.\n\n"; exit(0); } \
 }

// ##############################################################
/**
 * This is the main algorithm that checks covering relation between h-sets s1 and s2 in the case, 
 * when we have one unstable direction only.
 * 
 * @param gridFile - is a data file (text format) that contains information about the iterations, subdivisions, etc. needed to check this covering relation
 * @param desc - gridFile usually contains information about many covering relations. 'desc' is used as a key in seraching this file for the proper data. 'desc' can be for instance N1=>N2. 
 * 
*/ 
void checkCovering(RosslerPoincareMap& PM, HSet& s1, HSet& s2, const char desc[], const char gridFile[]){
  // prepare data: center and cordinates of s1, inverse coordinates in s2
  IVector x = s1.x();
  IVector y = s2.x();
  IMatrix C = s1.C(PM);
  IMatrix M = matrixAlgorithms::krawczykInverse(s2.C(PM));
  Grid grid(gridFile,desc);
  if(!grid.status){
    std::cout << "Warning: description " << desc << " not found in data file \n";
    return;
  }
  // lambda functions - abstract condition from the previous algorhtms 'checkCondition'.
  auto mapAcross = [&s2](IVector w, IVector) { return abs(w[0])>s2._r[0] or (abs(w[1])<s2._r[1] and abs(w[2])<s2._r[2]); };
  auto mapLeft   = [&s2](IVector w, IVector) { return w[0] < -s2._r[0]; };
  auto mapRight  = [&s2](IVector w, IVector) { return w[0] > s2._r[0]; };

  // auxiliary computation. We check if during the covering relation we should change the sign, 
  // i.e. if the left side s1 is mapped to the left side or to the right side of s2. Note, we have only one unstable direction.
  C0HOTripletonSet h(x + C*IVector({s1._r[0],0.,0.,0.}));
  interval returnTime;
  IPoincareMap& pm = *PM.iPoincareMap[0];
  IVector w = pm(h,y,M,returnTime,s1.N);
  double direction = w[0]>0 ? 1. : -1;
  
  // now the main checking
  std::cout << "Checking " << desc << "...\n" << std::flush;
  IVector r = s1.r();

  // left face of s1 must be mapped to the left of s2*direction
  r[0] = -s1._r[0]*direction;
  CHECK_CONDITION( "\tmap left : ", checkFace(PM,x,C,r,M,y,mapLeft,s1.N,grid.stable,1) );

  // right face of s1 must be mapped to the right of s2*direction
  r[0] = s1._r[0]*direction;
  CHECK_CONDITION( "\tmap right: " , checkFace(PM,x,C,r,M,y,mapRight,s1.N,grid.stable,1) );

  // entire set s1 must be mapped across s2
  r = s1.r();
  CHECK_CONDITION( "\taccross  : " , checkSet(PM,x,C,r,M,y,mapAcross,s1.N,grid.accrossUnstable,grid.accrossStable) );
}

// ##############################################################
/**
 * This is the main algorithm that checks covering relation between h-sets s1 and s2 in the case, 
 * when we have two unstable direction.
 * 
 * @param gridFile - is a data file (text format) that contains information about the iterations, subdivisions, etc. needed to check this covering relation
 * @param desc - gridFile usually contains information about many covering relations. 'desc' is used as a key in seraching this file for the proper data. 'desc' can be for instance N1=>N2. 
 * 
*/ 
void check2DCovering(RosslerPoincareMap& PM, HSet& s1, HSet& s2, const char desc[], const char gridFile[])
{
  Grid grid(gridFile,desc);
  if(!grid.status){
    std::cout << "Warning: description " << desc << " not found in data file \n";
    return;
  }

  // this case is more subtle than the case with one unstable direction. 
  // We will concturct homotopy with linear expanding map on two variables and zero at the third (as in the definition of covering relation).
  // Our linear map is just an approximate derivative at the center of s1 expressed in coordinate systems of s1 and s2.
  DVector _x = s1._x;
  DMatrix A = computeDerivative<DMatrix>(PM,_x,s1.N);
  DMatrix _L = matrixAlgorithms::gaussInverseMatrix(s2._C)*A*s1._C;
  IMatrix L(4,4);
  for(int i=1;i<=2;++i)
    for(int j=1;j<=2;++j)
      L(i,j)=_L(i,j);  

  // now L is our linear mapping
  IVector x = s1.x();
  IVector y = s2.x();
  IMatrix C = s1.C(PM);
  IMatrix M = matrixAlgorithms::krawczykInverse(s2.C(PM));
  
  // two abstract algorithm that must be checked for 2D covering relation
  // entire set s1 must be mapped across s2
  auto mapAcross = [&s2](IVector w, IVector ) { return  abs(w[0])>s2._r[0] or abs(w[1])>s2._r[1] or abs(w[2])<s2._r[2]; };
  
  // interval hull of L(exist set of s1) and PM(exist set of s1) must be mapped out of s2 
  // on the two coordinates marked as expanding.
  auto mapAway   = [&s2,&L](IVector rig, IVector p) { 
    IVector hul = intervalHull(L*p,rig);  
    bool result =  abs(hul[0])>s2._r[0] or abs(hul[1])>s2._r[1]; 
    if(! result) {
      std::cout << "hul=" << hul << std::endl;
      std::cout << "L=" << L << std::endl;
    }
    return result;
  };

  // now we check the above conditions
  std::cout << "Checking " << desc << "...\n" << std::flush;
  
  // we check if 4 unstable faces of s1 are mapped away s2
  IVector r = s1.r();
  r[0] = -s1._r[0];
  CHECK_CONDITION( "\tmap away 1: ", checkFace(PM,x,C,r,M,y,mapAway,s1.N,grid.stable,1) );
  r[0] = s1._r[0];
  CHECK_CONDITION( "\tmap away 2: ", checkFace(PM,x,C,r,M,y,mapAway,s1.N,grid.stable,1) );

  r = s1.r();
  r[1] = -s1._r[1];
  CHECK_CONDITION( "\tmap away 3: ", checkFace(PM,x,C,r,M,y,mapAway,s1.N,grid.unstable,0) );
  r[1] = s1._r[1];
  CHECK_CONDITION( "\tmap away 4: ", checkFace(PM,x,C,r,M,y,mapAway,s1.N,grid.unstable,0) );

  // we check if two stable faces are mapped across. Since PM is homeo onto image, the Jordan's theorem guarantee that this holds for each point from s1.
  r = s1.r();
  r[2] = -s1._r[2];
  CHECK_CONDITION( "\taccross  1: " , checkSet(PM,x,C,r,M,y,mapAcross,s1.N,grid.accrossUnstable,grid.accrossStable) );
  r[2] = s1._r[2];
  CHECK_CONDITION( "\taccross  2: " , checkSet(PM,x,C,r,M,y,mapAcross,s1.N,grid.accrossUnstable,grid.accrossStable) );
}

// ##############################################################
/**
 * This is the main algorithm that checks the sequence of covering relations linking 'begin' and 'end'. 
 * 
 * @param begin - first set in the sequence
 * @param end - last set in the sequence
 * @param chain - an array of sets between 'begin' and 'end'
 * @param alg1 - algorithm that check covering relation at the begining of the sequnce
 * @param alg2 - algorithm that checks covering relation for the last pair 
 * @param message - log to be printed to the standard output
 * @param gridFile - a file that contains information about iterations, subdivisions, etc.
 * 
*/ 
template<class Alg1, class Alg2>
void checkHeteroclinicChain(Alg1 alg1, Alg2 alg2, RosslerPoincareMap& PM, HSet& begin, HSet& end, HSetChain& chain, const char message[], const char gridFile[])
{
  printf("##################### %s #######################\n",message);
  // check covering relation between 'begin' and first set in the sequence
  alg1(PM,begin,*chain[0],"BEGIN=>N0",gridFile);
  unsigned i;
  unsigned N=chain.size();
  // loop over each pair in the sequence, i.e. N_i => N_{i+1}
  for(i=0;i<N-1;++i){
    std::ostringstream out;
    out << "N" << i << "=>N" << (i+1);
    alg1(PM,*chain[i],*chain[i+1],out.str().c_str(),gridFile);
  }
  
  // check N_{last} => end
  std::ostringstream out;
  out << "N" << i << "=>END";
  alg2(PM,*chain[i],end,out.str().c_str(),gridFile);
}


// ##############################################################
/**
 * This is an auxiliary algorithm that checks the sequence of covering relations linking 'begin' and 'end'. 
 * We assume that each covering relation has one unstable direction.
 * 
 * @param begin - first set in the sequence
 * @param end - last set in the sequence
 * @param chain - an array of sets between 'begin' and 'end'
 * @param message - log to be printed to the standard output
 * @param gridFile - a file that contains information about iterations, subdivisions, etc.
 * 
*/ 
void checkHeteroclinicChain(RosslerPoincareMap& PM, HSet& begin, HSet& end, HSetChain& chain, const char message[], const char gridFile[])
{
  checkHeteroclinicChain(checkCovering,checkCovering,PM,begin,end,chain,message,gridFile);
}

// ##############################################################
/**
 * This is an auxiliary algorithm that checks the sequence of covering relations linking 'begin' and 'end'. 
 * We assume that each covering relation has two unstable directions.
 * 
 * @param begin - first set in the sequence
 * @param end - last set in the sequence
 * @param chain - an array of sets between 'begin' and 'end'
 * @param message - log to be printed to the standard output
 * @param gridFile - a file that contains information about iterations, subdivisions, etc.
 * 
*/ 
void check2DHeteroclinicChain(RosslerPoincareMap& PM, HSet& begin, HSet& end, HSetChain& chain, const char message[], const char gridFile[])
{
  checkHeteroclinicChain(check2DCovering,check2DCovering,PM,begin,end,chain,message,gridFile);
}

// ##############################################################
/**
 * This is an auxiliary algorithm that checks the sequence of covering relations linking 'begin' and 'end'. 
 * We assume that each covering relation has two unstable direction except the last pair, when we have one dimensional expansion.
 * 
 * @param begin - first set in the sequence
 * @param end - last set in the sequence
 * @param chain - an array of sets between 'begin' and 'end'
 * @param message - log to be printed to the standard output
 * @param gridFile - a file that contains information about iterations, subdivisions, etc.
 * 
*/ 

void checkLinkingChain(RosslerPoincareMap& PM, HSet& begin, HSet& end, HSetChain& chain, const char message[], const char gridFile[])
{
  checkHeteroclinicChain(check2DCovering,checkCovering,PM,begin,end,chain,message,gridFile);
}

#endif
