/**
 * This file contains rigorous routines for enclosing the spectrum of an interval real matrix.
 */ 

#ifndef DANIEL_WILCZAK_EIGENVALUES
#define DANIEL_WILCZAK_EIGENVALUES

/**
 * Auxiliary function. It computes upper bound for the radius 
 * of the Gershogorin disc for the i-th row of matrix P.
 */ 
template<class Matrix>
typename capd::TypeTraits<typename Matrix::ScalarType>::Real 
GershgorinRadius(const Matrix& P, unsigned i){
  using namespace capd;
  typename capd::TypeTraits<typename Matrix::ScalarType>::Real r=0;
  for(unsigned j=0;j<P.numberOfColumns();++j){
    if(j!=i)
      r += (capd::abs(P[i][j])).rightBound();
  }
  return r;
}

// ########################################################

/**
 * This is a rigorous routine. It computes bound for the spectrum of a real interval matrix.
 * 
 * The algorithm applies interval Newton and an interval Krawczyk operator to the equation for the eigenvalue problem, i.e.
 * We solve the equation F(lambda,v2,v3,...,vn) = (A-lambda*Id)*(1,v2,...,vn) = 0 by the Krawczyk and Newton methods.
 * 
 * If succeed, then (1,v2,...,vn) is an eigenvector for the eigenvalue lambda.
 * The function returns a complex interval vector.
*/
template<class Matrix>
typename Matrix::RowVectorType::template rebind< std::complex<typename Matrix::ScalarType> >::other
computeComplexEigenvalues(Matrix _A){
  typedef typename Matrix::ScalarType Interval;
  typedef std::complex<Interval> CInterval;
  typedef typename Matrix::RowVectorType Vector;
  typedef typename capd::TypeTraits<Interval>::Real Float;
  typedef typename Matrix::template rebind<Float>::other DMatrix;
  typedef typename Vector::template rebind<Float>::other DVector;
  typedef typename Matrix::template rebind< CInterval>::other CMatrix;
  typedef typename Vector::template rebind< CInterval >::other CVector;

  if(_A.numberOfRows()!=_A.numberOfColumns())
    throw std::runtime_error("computeEigenvaluesAndEigenvectors error: cannot compute eigenvalues of a nonsquare matrix");

  // First we compute an approximate spectrum of the center matrix mid(_A)
  // just by a nonrigorous routine.
  const unsigned dim = _A.numberOfRows();
  CMatrix A = capd::vectalg::convertObject<CMatrix>(_A);

  DMatrix DA = capd::vectalg::midObject<DMatrix>(_A);
  DVector rV(dim), iV(dim);
  DMatrix rVec(dim,dim), iVec(dim,dim);
  capd::alglib::computeEigenvaluesAndEigenvectors(DA,rV,iV,rVec,iVec);  
  
  CMatrix B(dim,dim);  
  unsigned i,j;
  for(i=1;i<=dim;++i)
    for(j=1;j<=dim;++j)
      B(i,j) = CInterval(rVec(i,j),iVec(i,j));
  capd::matrixAlgorithms::gauss(B,(A*B),A);
  
  // Now we have 
  // B - an aproximate Jordan basis
  // A - almost diagonal matrix that contains (complex) eigenvalues on diagonal 
  // These will be our candidates for rigorous routines.
  
  CVector result(dim);
  // We solve the equation F(lambda,v2,v3,...,vn) = (A-lambda*Id)*(1,v2,...,vn)=0 by Newton method
  for(j=0;j<dim;++j){
    Float r = GershgorinRadius(A,j);
    Float s = r + 1e-8; // to be nonzero
    s *= 2;
    CInterval S(Interval(-s,s),Interval(-s,s));
    CVector U(dim);
    U = S;
    // candidate for eigenvalue 
    CInterval lambda(rV[j],iV[j]);
    
    // here we compute the interval Krawyczk operator
    CMatrix F = A;
    for(i=0;i<dim;++i){
      if(i!=j){
        F[i][j] = S;
        F[i][i] -= (lambda+S);
      } else{
        F[j][j] = 1.;
      }
    }
    CMatrix C = capd::matrixAlgorithms::krawczykInverse(midMatrix(F));
    CVector z = A.column(j);
    z[j] -= lambda;
    // Intrval Krawczyk Correction
    CVector N = C*z + (C*F-CMatrix::Identity(dim))*U;

    Float norm = capd::TypeTraits<Float>::zero();
    for(unsigned c=0;c<dim;++c){
      norm = capd::max(norm,rightBound(capd::abs(N[c].real())));
      norm = capd::max(norm,rightBound(capd::abs(N[c].imag())));
    }
    if(norm<s){
      // try intersect with interval Newton, if F invertible
      try{
        N = capd::vectalg::intersection(N,capd::matrixAlgorithms::gauss(F,z));
      }catch(...) {}
      result[j] = lambda - N[j];
    } else {
      result[j] = A[j][j] + CInterval(Interval(-r,r),Interval(-r,r));
    }
  }
  
  // Eventually we check if the obtained bounds are distinct. 
  // If so, each bound contains unique eigenvalue. 
  // Otherwise, if two or more bounds intersect, we must take the hull of them (in this case they were not reduced by Krawczyk or Newton and they are just Gershgorin discs).
  for(j=0;j<dim-1;++j){
    for(unsigned i=j+1;i<dim;++i)
    {
      CInterval p;
      if(capd::intervals::intersection(result[j],result[i],p)){
        result[j] = capd::intervals::intervalHull(result[j],result[i]);
        result[i] = result[j];
      }
    }
  }
  
  return result;
}

#endif
