/** author: Daniel Wilczak, 2015.
    This is the main file of the program. 
*/

/** 
 * Headers from the CAPD library.
 * The library provides (in particular) algorithms for rigorous computation 
 * of Poincare maps and their derivatives, manipulation of interval vectors and matrices. 
*/
#include "capd/intervals/lib.h"
#include "capd/intervals/IComplex.h"
#include "capd/capdlib.h"
using namespace capd;

/** Auxiliary algorithms. */

// Clock functions
#include "tictac.h"

// Some useful auxiliary functions, like normalization of columns, conversion of sequence of doubles to sequence of doubles, etc.
#include "utils.h"

// Rigorous algorithm for eigenvalue problem. An algorithm based on interval Krawczyk operator and interval Newton operator is implemented.
#include "eigenvalues.h"

/** Program specific data structures and algorithms. */

// Algorithms for the computation of Poincare map and its derivative for the 4D Rossler system.
#include "RosslerPoincareMap.h"

// Data structure and algorithms for enclosing forward invariant set of an abstract map.
// They are used to compute the trapping region for the Poincare map for the Rossler system. 
#include "InvSetEnclosure.h"

// Data structures that represent mathematical object h-set and a sequence of h-sets: HSetChain
#include "HSet.h"

// Algorithms used to check covering relations with one or two expanding directions.
#include "covrel.h"

// Algorithms used to check the cone condition at h-sets centered at approximate periodic point.
#include "cones.h"

// An algorithm used to prove the existence of a hyperbolic periodic point.
#include "MultipleShooting.h"

// This is a data structure that stores all the data used in the computations. Browse this file for details. 
// It also contains routines that create heteroclinic chains of covering relations (not automatic).
#include "ProofData.h"

// This file contains routines thath export supplementary data of the paper to files.
// These routines are not used in the computations regarding the main theorems. 
#include "ExtendedProofData.h"

// ##############################################################

int main(int argc, char* argv[])
{
  std::cout.precision(17);
  std::cout << std::boolalpha;

  try{   
    const double a=.27857;
    const double b=3;
    const double c=-0.3;
    const double d=0.05;
    RosslerPoincareMap PM(a,b,c,d,20);

    // nonrigorous computation - contruct all h-sets, details in ProofData.h. 
    ExtendedProofData data(PM);
    data.initHSetsAtPeriodicPoints();
    data.initHeterloclinicChains();

    // this routine exports data for the supplementary material. It is not a part of the proof.
    // data.exportAllChainsOfCoveringRelations();

    // rigorous computations - details in ProofData.h
    data.encloseAttractor();
    data.checkAllCoveringRelations();
    data.checkConeConditions();
    data.computeEigenvaluesAtPeriodicPoints();

  }catch(std::exception& e){
    std::cout << "Exception: " << e.what() << std::endl;
  }
  return 0;
}
